procedure frbs(image)

# Script to locate and extract RBS data from raw frames.  These
# extractions are only good for quicklook data reduction and the
# users should not attempt to do any real analysis with their resulting
# spectra.  This script is meant to be used in conjunction with the
# script 'prbs.cl'.  Both scripts are called from the task 'rbs'
# in the package 'rgo.quickie' and it is recommended that you use the
# rbs task rather than to call these scripts individually. 
# Most of the beautiful programing style and GUI interface was stolen
# from Jim Lewis ISIS script, which in turn stole ideas from
# scripts written by Brian Boyle (aka Ryan O'Bible).
# Ryan O'Bible,  RGO, February 1995 
# 
# Updates:
#
#      May 1995 -- Added bias subtraction, flat fielding and a check
#                  for header information.  Also fixed a number of
#                  bugs.  JRL and BJB
#
#      July 1995 -- Modified IDS routine for RBS data, BJB and JRL
#      21 Feb 1996 -- Major overhaul (JRL)

string  image   {prompt="Input image(s)"}
int     nspec   {1,prompt="Number of spectra in image?"}
real    stw     {8500,prompt="Start wavelength"}
real    app     {-2.,prompt="Angstroms/pix"}
int     blkfact {1,min=1,prompt="Blocking factor in spectral domain"}
bool    fconv   {no,prompt="Convert from FITS?"}
bool    crproc  {yes,prompt="Remove cosmic rays for extraction?"}
bool    lcal    {yes,prompt="Wavelength calibrate?"}
#bool    head    {yes,prompt="Does header information exist?"}
bool    bsub    {yes,prompt="Perform bias subraction"}
bool    ffield  {no,prompt="Perform flat field"}
string  fframe  {"",prompt="Flat field name"}
#string  sfunc   {"",prompt="Name of sensitivity function?"}


struct  *inimglist
struct  *tlist

begin

        string  dsect           # data section
        string  imglist         # equals image
        string  imtype          # image type
        string  mstype          # multispec image type
        string  infile          # temporary list for files
        string  img             # image name from fscan with extension stripped
        string  ct              # type of chip
        string  refim           # reference image for spectral tracing
        string  outfil          # dummy string
        string  tempname        # dummy string
        string  dumstr          # dummy string
        string  log1            # name of log file
        string  ff              # name of input flat field
        string  sf              # name of input sensitivity function
        string  bsect           # bias strip image section
        string  ysamp           # sample of bias strip for possible fit
        real    defsizx,defsizy # default size in case header items are missing
        real    endp            # end pixel for wavelength calibration
        real    mpp             # mm/pix
        real    start           # starting pixel for wavelength calibration
        real    rd              # CCD readnoise
        real    gn              # CCD gain
        real    axs,axe         # X range of light sensitive part of chip
        real    dxs,dys         # Start of window
        real    dxe,dye         # End of window
        real    xb,yb           # binning factor
        real    dumreal         # dummy real
        real    pos             # spectrum position
        real    bxstart,bystart # start of bias strip
        real    bxend,byend     # end of bias strip
        int     mm              # dummy
        int     iblk            # blocking factor in spectral domain
        int     nsp             # number of spectra
        int     nfound          # number of spectra found
        bool    crp             # TRUE if removing cosmic rays
        bool    fcon            # TRUE if input file is a FITS file
        bool    bproc           # TRUE if subtracting bias
        bool    fproc           # TRUE if dividing by flat field
        bool    procccd         # TRUE if did CCD processing

# Get query parameter
        imglist = image

# Expand file lists into temporary files.
        infile =  mktemp("temp")
        sections (imglist,option="fullname",>infile)
        inimglist = infile

# Get a few parameters
        iblk = blkfact
        nsp = nspec
        crp = crproc
        fcon = fconv

# Set up logging
        tee.append = yes
        log1 = "rbs.log"

# Get file name extension
        imtype = "."//envget("imtype")
        mstype = ".ms"//imtype

# Strip extension from flat field image and sensitivity function
# (if either exist...)
        ff = fframe
        strpextn(ff) | scan(ff)
#        sf = sfunc
#        strpextn(sf) | scan(sf)

# Loop through images
        while (fscan(inimglist,img) != EOF) {

# Strip off any extensions.  This is to avoid conflicts with names which
# might arise with FITS files or if the user has used meta-characters in
# the input file list
           strpextn(img) | scan(img)

# Clobber old output files
           imdelete(img//mstype, ver-, >& "dev$null")

# If the input file is a fits image, then convert it to an iraf file...
# This will become unnecessary when v2.11 arrives with its FITS image kernel
           if (fcon) {
              print(" Reading FITS image") | tee(log1)
              rfits(img//".fit", "", img, make+, lo-, sh-,
              datatype="", blank=0.0, scale+, oldir-, offset=0)
           }

# Get the chip name in the rbs.
           q_gets(img,"DETHEAD","unknown") | scan(ct)

# Get information about windowing of the chip
         
           q_getr(img,"XBIN",1.0) | scan(xb)
           q_getr(img,"YBIN",1.0) | scan(yb)

           hselect(img,"i_naxis1,i_naxis2",yes) | scan(defsizx,defsizy)
           defsizx = defsizx*xb; defsizy = defsizy*yb
           q_getr(img,"AXSTART",51.0) | scan(axs)
           q_getr(img,"AXEND",1074.0) | scan(axe)
           q_getr(img,"DXSTART",1.0) | scan(dxs)
           q_getr(img,"DXEND",defsizx) | scan(dxe)
           q_getr(img,"DYSTART",1.0) | scan(dys)
           q_getr(img,"DYEND",defsizy) | scan(dye)

           start = max(real(nint((axs-dxs)/xb+1.0)),1.0)
           endp = min(real(nint((axe-dxs+1.0)/xb)),real(nint(defsizx/xb)))

# Get the chip readout noise and gain.  If it doesn't exist, then
# substitute plausible values.
           q_getr(img,"READNOIS",INDEF) | scan(rd)
           if (rd == INDEF) {
              rd = 5.0
              print("Readnoise not in header -- assumed: ",rd) | tee (log1)
           }
           q_getr(img,"GAIN",INDEF) | scan(gn)
           if (gn == INDEF) {
              gn = 1.0
              print("Gain not in header -- assumed: ",gn) | tee (log1)
           }
           print(" ") | tee(log1)
           print(" CCD frame: ",img,"   RBS ") | tee(log1)
           print(" Chip: ",ct,"   Readnoise = ",rd," Gain = ",gn,) | tee(log1)

# Do you want to do CCD processing?
           bproc = bsub
           if ((axs - 6.0) <= dxs) bproc = no
           fproc = ffield
           if (fproc && !access(ff//imtype)) {
              fproc = no
              print("Warning: can't access flat-field image") | tee(log1)
           }
           procccd = no
           if(fproc || bproc){
              if (!access("backup/"))
                 mkdir("backup/")
              ccdred.verbose = no
              ccdred.backup = "backup/"
              ccdred.logfile = "rbs.log"

# Create a bias section
              bxstart = (6 - dxs)/xb + 1.0
              if (bxstart <= 0.0) bxstart = 1.0
              bxend = (axs - 6 - dxs)/xb + 1.0
              bystart = (101 - dys)/yb
              if (bystart <= 0) bystart = 1.0
              byend = (dye - dys)/xb + bystart

              bsect = "["//nint(bxstart)//":"//nint(bxend)
              bsect = bsect//","//nint(bystart)//":"//nint(byend)//"]"
              ysamp = nint(bystart)//":"//nint(byend)
            
# Perform bias subtraction and flat-field if required
              if(bproc)print (" Performing overscan-correction") | tee(log1)
              if(fproc)print (" Performing flat-field division") | tee(log1)
              ccdproc (img,ccdtype=" ", max_cache=0, noproc=no, fixpix=no, 
                 overscan=bproc, trim=no, zerocor=no, darkcor=no, flatcor=fproc, 
                 illumcor=no, fringecor=no, readcor=no, scancor=no, 
                 readaxis="line", fixfile="", biassec=bsect, trimsec="",
                 zero="", dark="", flat=ff, illum="", fringe="", minreplace=1.,
                 scantype="shortscan", nscan=1, interactive=no, function="legendre",
                 order=1, sample=ysamp, naverage=1, niterate=1, low_reject=3., 
                 high_reject=3., grow=0.)
              procccd = yes
           }

# Set airmass and Julian date  - you never know when they might come in handy.....
# However, only do it if header information exists
#           if(head){
#              setjd (img,observatory="lapalma",date="date-obs",time="timstart",
#                 exposure="timexpos",ra="ra", dec="dec", epoch="equinox", 
#                 jd="jd", hjd="hjd", ljd="ljd",utdate=yes, uttime=yes,
#                 listonly=no, >& "dev$null")
#              setairmass (img,observatory="lapalma", intype="beginning", 
#                 outtype="effective",date="date-obs", exposure="timexpos", 
#                 airmass="airmass", utmiddle="utmiddle", show=yes, update=yes, 
#                 override=yes, >& "dev$null")
#           }


# Remove cosmic rays from the frame to facilitate locating the spectra. Note
# that this is not the frame from which the spectra will be extracted.
           refim=img
           if (crp) {
              refim=mktemp(img)
              print(" Removing Cosmic Rays") | tee(log1)
              imdelete("_fmed*"//imtype, ver-, >& "dev$null")
              imdelete("_cr*"//imtype, ver-, >& "dev$null")
              imdelete("_mask*"//imtype, ver-, >& "dev$null")
              xzap(img,outlist=refim, nsigma=2., xboxsz=15, yboxsz=1, 
                 nobjsigma=0., nrings=0, donegative-,nnegsigma=5.,
                 imgsec="[*,*]", deletemasks+, cleanup+, verbose-,
                 inimglist="", outimglist="", statlist="", >& "dev$null")
              imdelete("_fmed"//img, ver-,>&"dev$null")
           }

# Trace the spectrum...
           print(" Tracing spectrum") | tee(log1)

           apextract.dispaxis = 1
           apall (refim, output=refim, apertures="", format="multispec", references="",
              nfind=nsp, profiles="", interactive-, find+,
              recenter+, resize+, edit-, trace+, fittrace-,
              extract-, extras-, review-, line=INDEF, nsum=100, 
              lower=-5., upper=5., apidtable="", b_function="chebyshev", 
              b_order=1, b_sample="-50:-21,21:50", b_naverage=-10, 
              b_niterate=1, b_low_reject=3., b_high_rejec=3., b_grow=0., 
              width=10., radius=10., threshold=0., minsep=5., maxsep=1000., 
              order="increasing", aprecenter="", npeaks=nsp, shift+,
              llimit=INDEF, ulimit=INDEF, ylevel=0.1, peak+, bkg+,
              r_grow=0., avglimits-, t_nsum=50, t_step=50, t_nlost=3, 
              t_function="legendre", t_order=4, t_sample="*", t_naverage=1, 
              t_niterate=2, t_low_reject=3., t_high_rejec=3., t_grow=0., 
              background="fit", skybox=1, weights="variance", pfit="fit1d", 
              clean-, saturation=INDEF, readnoise=rd, gain=gn, lsigma=4., 
              usigma=4., nsubaps=1, > "dev$null")

# Extract the spectrum...
           print(" Extracting spectrum ") | tee(log1)

           apall (img,output=img, apertures="", format="multispec", references=refim,
              nfind=nsp, profiles="", interactive-, find-, recenter-, 
              resize+, edit-, trace-, fittrace-, extract+, extras-, review-, 
              line=INDEF, nsum=10, lower=-5., upper=5., 
              apidtable="", b_function="chebyshev", b_order=1, 
              b_sample="-50:-21,21:50", b_naverage=-10, b_niterate=1, 
              b_low_reject=3., b_high_rejec=3., b_grow=0., width=10., 
              radius=10., threshold=0., minsep=5.,maxsep=1000., 
              order="increasing", aprecenter="", npeaks=nsp, shift+,
              llimit=INDEF, ulimit=INDEF, ylevel=0.1, peak+, bkg+, 
              r_grow=0., avglimits-, t_nsum=50, t_step=50, t_nlost=3, 
              t_function="legendre", t_order=4, t_sample="*", t_naverage=1, 
              t_niterate=2, t_low_reject=3., t_high_rejec=3., t_grow=0., 
              background="fit", skybox=1, weights="variance",pfit="fit1d", 
              clean+, saturation=65000, readnoise=rd, gain=gn, lsigma=3., 
              usigma=3., nsubaps=1, > "dev$null")
 
# Send out a friendly message and clean up the database directory.  Remove
# the cosmic ray processed frame if it exists...
           print(" ") | tee(log1)
           tempname = mktemp("temp")
           type("database/aplast",map+,dev="terminal") | 
              match("begin","",stop-,print+,meta+,>tempname) 
           nfound = 0
           tlist = tempname
           while(fscan(tlist,dumstr,dumstr,dumstr,mm,dumreal,pos) != EOF) {
              print(" Spectrum # ",mm, "found in row ",pos) | tee(log1)
              nfound = nfound + 1
           }
           print(" ") | tee(log1)
           tlist = ""
           delete(tempname, ver-, >& "dev$null")
           delete("database/ap"//refim, ver-, >& "dev$null")
           if (crp) imdelete(refim, ver-, >& "dev$null")

# Create a data section string for use later
           if (nfound > 1) 
              dsect = "["//int(start)//":"//int(endp)//",*]"
           else
              dsect = "["//int(start)//":"//int(endp)//"]"

# If you did CCD processing on this frame, then get rid of the processed
# frame and bring back the original...
           if (procccd) {
              imdelete(img, ver-)
              imrename("backup/"//img//imtype,img,verbose-)
           }

# Chop off the bits you don't want...
           tempname = mktemp("temp")
           outfil = img//mstype
           imrename(outfil,tempname,verbose-, >& "dev$null")
           imcopy(tempname//dsect,outfil,verbose-,>& "dev$null")
           imdelete(tempname,verify-)

# If you want to do a wavelength calibration then decide where on the chip
# the spectrum lies...
           if(lcal){
              print(" Start wavelength ",stw," Angstroms/pix ",
                 app," Binning factor ",xb) | tee(log1)
              print(" ") | tee(log1)

# Edit the WCS to include the wavelength information
              wcsreset(outfil,wcs="physical",verbose-)
              sapertures(outfil,apertures="",apidtable="",wcsreset=no,
                 verbose-,beam=INDEF,dtype=0,w1=stw,dw=app,z=INDEF,
                 aplow=INDEF,aphigh=INDEF,title="INDEF")
              dispcor(outfil,outfil,linearize+,database="",table="",
                 w1=INDEF,w2=INDEF,dw=INDEF,nw=INDEF,log-,flux+,samedisp+,
                 global-,ignoreaps-,confirm-,listonly-,verbose-,logfile=log1)


# If not doing a wavelength calibration, then reset the WCS so that logical
# and physical coordinates match...
           } else {
              wcsreset(outfil,wcs="physical",verbose-)
           }

# Block average the result if you want...

           if (iblk > 1) 
              blkavg(outfil,outfil,iblk,1,1,1,1,1,1,option="average")           

# If this came originally from a FITS file, then delete the IRAF file...
           if (fcon) imdelete(img, ver-, >& "dev$null")
        }

# Tidy and exit
        delete (infile, ver-, >& "dev$null")
        inimglist = ""
end
