procedure preset(input,wantchip,wanttrace,wantdek,norder,nobjt,trace,
   nobjd,track,dekker)
#------------------------------------------------------------------------------
#
#   -----------
#   P R E S E T
#   -----------
#
#   Description
#   -----------
#   Set up files necessary to use the previewer
#
#   
#   Scope of program
#   ----------------
#   - Only 2d frames supported
#   - Data type irrlevant
#   - Error files supported
#   - Batch execution not supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   CL script
#
#
#   Parameters
#   ----------
#   input      This is  an  input frame which has been taken on the instrument
#              for which you want  to  do  this  setup.   It  should  have  the
#              spectrum  of  a  point  source  covering as much of the spectral
#              range as possible.  It preferably should  have  medium  to  high
#              signal to noise. (character)(prompted for)
#
#   wantchip   This is true if you want to create a new chip file. (bool)
#              (prompted for)
#
#   wanttrace  This is true if you want to create a new trace file. (bool)
#              (prompted for)
#
#   wantdek    This is true if you want to create a new dekker file. (bool)
#              (prompted for)
#
#   norder     This  is  the order of the polynomial to fit for the trace.  The
#              quantity being fit is the position of the centre of  the  seeing
#              disc  as  a function of wavelength.  For FOS frames, an order of
#              5 is good.  ISIS and IDS frames don't generally  need  a  trace.
#              But if they do, then an order of 1 is usually sufficient.
#              (int)(prompted for)
#
#   nobjt     It  is  possible  to  trace  more  than one spectrum on a frame.
#             This is the number of spectra which are to be traced. (int)
#             (prompted for)
#
#   trace      This is the name of the output file for the trace. (character)
#              (prompted for)
#
#   nobjd      This is the number of apertures to include in the  dekker  file.
#              An  aperture  should  be  included if there is a chance that the
#              user may want to extract a spectrum from  that  aperture  during
#              the course of the night. (int)(prompted for)
#
#   track      If  a  trace  hasn't been created on this run, then the user can
#              decide whether a previously defined trace will be  used  in  the
#              creation of the dekker file. (bool)(prompted for)
#
#   dekker     This  is  the  name  for  the output file which will contain the
#              dekker aperture information. (character)(prompted for)
#
#   strace     This is true if you want to use a standard trace for a FOS frame.
#              This is not relevant for any other type of data. (character)(hidden)
#
#   gumby      This is a hidden parameter which, if switched on, gives the  new
#              user  advice  on how to use the program.  If gumby is used, then
#              each prompt for information is preceded by a  small  explanation
#              of  what is required of the user.  To use gumby mode, one should
#              type: `preset gumby+' on the command line and not use epar.
#              (bool)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   Not relevant
#
#   
#   Method
#   ------
#   - The user enters an input file name.
#   - The header is read to get chip info.  The standard chip file
#     is then modified accordingly.
#   - If the user wants to do a trace or a dekker, then the file
#     is bias subtracted and displayed.
#   - For a trace, the user indicates the location of each spectrum and
#     the spectral range desired with the cursor.
#   - For a dekker file, the user indicates the spatial range of the 
#     slit and the spectral range for each spectrum with the cursor.
#
#
#   External applications called
#   ----------------------------
#   Package dataio:
#      rfits
#
#   Package images:
#      imdelete
#      imrename
#      imstack
#
#   Package sam:
#      dekker
#      trace
#
#   Package scp:
#      chip
#      proc
#
#   Package tables:
#      tcreate
#      tdelete
#      tdump
#      tmerge
#      trename
#
#   Package tv:
#      display
#
#
#
#   Possible future upgrades
#   ------------------------
#   - A cleverer way of doing the chip file which can take into account
#     subsampling and windowing in the spectral direction.
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   09-DEC-93     - Original program 
#   15-DEC-93     - Added ability to use standard trace files for FOS 
#                   frames. (JRL)
#
#
#------------------------------------------------------------------------------

# Query Parameters

string      input      {prompt="Name of input file"}
bool        wantchip   {yes,prompt="Do you want to create a new chip file"}
bool        wanttrace  {yes,prompt="Do you want to create a new trace file"}
bool        wantdek    {yes,prompt="Do you want to create a new dekker file"}
int         norder     {1,min=0,prompt="Order of polynomial for trace fit"}
int         nobjt      {1,min=1,prompt="Number of objects to trace"}
string      trace      {prompt="Name of trace file"}
int         nobjd      {1,min=1,prompt="Number of objects to mark for dekker"}
bool        track      {yes,prompt="Use a trace with dekker?"}
string      dekker     {prompt="Name of dekker file"}

# Hidden parameters

bool        strace     {yes,prompt="Do you want to use a standard FOS trace?"}
bool        gumby      {no,prompt="Do you need lots of extra help to run this?"}
struct      *stdchipfile
struct      *temp

begin

      # Local variables

      string infile,telescope,instrum,ccd,filename,isiscolour,
         ccdfile,comment,chipname,dsect,bsect,exphead,ftrace,dek,suffix,
         stdtable,blank,inst,temp1,temp2,temp3,temp4,str1,extension
      real readout,gain
      int nx,ny,i,idum,ist,ifn,len,nord,nobj,axis
      bool wantc,wantt,wantd,trac,clobber,gum,standtr
      struct data,curtime

      # See if the user is a gumby or not...(i.e. whether lots of extra
      # help is required)

      gum = gumby

      # If there person needs help, then give a welcoming message now

      if (gum) {
         print("#############################################################")
         print("#                          PRESET                           #")
         print("#  This program sets up a number of files which are needed  #")
         print("#  by the previewer.  I'll try to explain as much as I can  #")
         print("#  as we go along.  You should run this program for each    #")
         print("#  instrument you use.  Note that I consider the arms of    #")
         print("#  ISIS to be separate instruments.  If at some stage of    #")
         print("#  the night you change your set up (e.g. the way the CCD   #")
         print("#  has been windowed, the slit mask on FOS) you should run  #")
         print("#  this program again.                                      #")
         print("#                                                           #")
         print("#  Don't forget that IRAF and hence SAM are both case       #")
         print("#  sensitive.  So, don't use upper case unless you really   #")
         print("#  really mean it!!                                         #")
         print("#############################################################")
      } 
 
      # Get input file name and get rid of the file extension if it exists

      blank = """"""
      extension = "."//envget("imtype")
      if (gum) {
         print(" ")
         print("#############################################################")
         print("#  I need a file which has at least one spectrum from the   #")
         print("#  instrument you want to consider. Don't type the file     #")
         print("#  extension like .imh or .hhh!                             #")
         print("#############################################################")
      }
      infile = input
      len = strlen(infile)
      if (substr(infile,len-3,len) == extension)
         infile = substr(infile,1,len-4)

      # Find out what you want to do here...

      if (gum) {
         print(" ")
         print("#############################################################")
         print("#  A chip file is required in order to do debiassing and    #")
         print("#  trimming of the input file.  If you already have one     #")
         print("#  then you can skip this bit.  If this is the first time   #")
         print("#  you have run this or if you have recently changed the    #")
         print("#  way the chip is windowed, then you need a new chip file. #")
         print("#############################################################")
      }
      wantc = wantchip
      if (gum) {
         display(infile,1,fill+)
         print(" ")
         print("#############################################################")
         print("#  A trace file is required if spectra are badly distorted, #")
         print("#  such as FOS spectra.  If this is a FOS spectrum, you     #")
         print("#  must have a trace file.  If this is data from ISIS or    #")
         print("#  IDS, you will only need a trace if the spectra are       #")
         print("#  tilted across the CCD rows by more than a few pixels. If #")
         print("#  you already have a perfectly good trace for this CCD,    #")
         print("#  then don't bother doing another one.  Have a look at the #")
         print("#  image display and see if you think you need one.         #")
         print("#############################################################")
      }
      wantt = wanttrace
      if (gum) {
         print(" ")
         print("#############################################################")
         print("#  A dekker file defines both the spatial and spectral      #")
         print("#  limits of a spectrum.  You'll need to create one now     #")
         print("#  unless you already have one for this instrument or the   #")
         print("#  aperture/slit configuration has changed recently.        #")
         print("#############################################################")
      }
      wantd = wantdek

      # Get telescope identifier and find out how this was observed

      imgets(infile,"TELESCOP")
      telescope = imgets.value
      if (telescope == "WHT") {
         imgets(infile,"INSTRUME")
      } else if (telescope == "INT") {
         imgets(infile,"DETID")
      } else {
         error(0,"No telescope information in the header")
      }
      instrum = substr(imgets.value,1,3)
      if (instrum == "FOS") {
         if (telescope == "WHT") {
            inst = "FOS2" }
         else {
            inst = "FOS1"
         }
      }
      else if (instrum == "ISI") {
         isiscolour = substr(imgets.value,6,6)
         inst = "ISIS"//isiscolour }
      else  {
         instrum = "IDS"
         inst = "IDS"
      }

      # If this is a FOS frame, then see if the user wants to use standard trace
      # information...

      if (instrum == "FOS" && wantt) 
         standtr = strace
      else 
         standtr = no

      # If the standard table exists, then recover the old information from it

      stdtable = "sam_preview.tab"
      filename = blank
      ftrace = blank
      dek = blank
      if (access(stdtable)) {
         temp1 = mktemp("SAM")
         tdump(table=stdtable,cdfile="",pfile="",datafile=temp1,
            columns=inst,rows="-",pwidth=80) 
         temp = temp1

         # If there was information for this instrument, then recover
         # the file names 
         
         if (fscan(temp,str1) != EOF) filename = str1
         if (fscan(temp,str1) != EOF) ftrace = str1
         if (fscan(temp,str1) != EOF) dek = str1
         delete(files=temp1,go-,ver-,def+,all+,sub+)
         if (ftrace != blank) {
            len = strlen(ftrace)
            if (substr(ftrace,len-3,len) == extension)
               ftrace = substr(ftrace,1,len-4)
         }
         if (dek != blank) {
            len = strlen(dek)
            if (substr(dek,len-3,len) == extension)
               dek = substr(dek,1,len-4)
         }
      }
      
      ###############################################
      #     SECTION FOR CREATING SCP CHIP FILE      #
      ###############################################

      # If this was a FOS file, then get the chip file
    
      if (instrum == "FOS") {
         filename = "fos"//substr(inst,4,4)//".chp"

         # If you want a new chip file and there is one lurking around,
         # then get rid of it.  If you don't want one, and there isn't
         # one in the current directory, then send a warning message

         if (wantc) {
            if (access(filename)) 
               delete(files=filename,go+,ver-,def+,all+,sub+) 
         }
         else {
            if (!access(filename)) 
               print("Warning: You have no chip file available") 
         }

         # Create the chip file if you want it

         if (wantc) copy("scp$lpochips/fos"//substr(inst,4,4),filename,ver-) 
      }

      # If this isn't a FOS file, then we have to build the chip file
      # up from scratch (since it has probably been windowed).

      else {

         # Get the name of the chip

         if (telescope == "WHT") {
            imgets(infile,"CCDCHIP") 
            ccd = substr(imgets.value,1,4) }
         else {
            imgets(infile,"DETHEAD")
            ccd = substr(imgets.value,5,8)
         }

         # Make this name lower case

         print(ccd) | lcase | scan(ccdfile)

         # Get the file name and see if you have one already

         filename = ccdfile//".chp"
         if (wantc) {
            if (access(filename)) 
               delete(files=filename,go+,ver-,def+,all+,sub+) 
         }
         else {
            if (!access(filename)) 
               print("Warning: You have no chip file available")
         }

         # If you don't have one already, then make one up...

         if (wantc) {

            # Get the current time for comment line

            time | scan(curtime)
            comment = ccd//" "//curtime

            # Get chip name for next line

            chipname = ccd

            # Size of array from file header

            imgets(infile,"i_naxis1")
            nx = int(imgets.value)
            imgets(infile,"i_naxis2")
            ny = int(imgets.value)

            # Set up the data section.  The X range should be as it is
            # in the standard chip file  since that isn't windowed in general.
            # The Y will start at 1 and end at the last pixel value
            # available.  This will all be changed when the telescopes start
            # writing sensible info into the header.

            stdchipfile = "scp$lpochips/"//ccdfile
            for (i = 1; i <= 3 ; i = i + 1)
               idum = fscan(stdchipfile,data)
            idum = fscan(stdchipfile,data)
            ist = stridx("[",data)
            ifn = stridx(",",substr(data,ist,strlen(data))) + ist - 1
            ifn = stridx(":",substr(data,ifn,strlen(data))) + ifn - 1
            dsect = substr(data,ist,ifn)//ny//"]"

            # Set up bias section.  Same rules apply as above, except start
            # at y = 1

            idum = fscan(stdchipfile,data)
            ist = stridx("[",data)
            ifn = stridx(",",substr(data,ist,strlen(data))) + ist - 1
            bsect = substr(data,ist,ifn)//"1:"//ny//"]"
 
            # Read readout noise

            idum = fscan(stdchipfile,data)
            ist = stridx("!",data) + 1
            readout = real(substr(data,ist,strlen(data)))

            # Read gain

            idum = fscan(stdchipfile,data)
            ist = stridx("!",data) + 1
            gain = real(substr(data,ist,strlen(data)))
            stdchipfile = ""

            # Now do the chip file

            scp.chip(chipfile=filename,comment=comment,chipname=chipname,
            nx=nx,ny=ny,data_reg=dsect,bias_reg=bsect,readout=readout,
            gain=gain,bcfile="",brfile="",bpfile="")
         }
      }

      # If you want either a trace or dekker file, then process the frame
      # and display the processed frame

      if ((wantt && !standtr) || wantd) {
         exphead = "EXPOSED"
         if (telescope == "INT") exphead = "TIMEXPOS"
         
         # Get a random filename suffix

         suffix = mktemp("P")

         # Message for gumbies

         if (gum) {
            print(" ")
            print("#############################################################")
            print("#  Hang-on a sec.  I'm just doing a bias subtraction...     #")
            print("#############################################################")
         }

         # Now do a bias strip subtraction

         scp.proc(bias-,dark-,flat-,prog+,chipfile=filename,bias_frames="",
            dark_frames="",flat_frames="",prog_frames=infile,
            exphead=exphead,meanbias="",sub_bias+,sub_frame-,over-,
            meandark="",dark_cor-,meanflat="",flat_div-,image_flat+,
            nord=0,axis=1,trim+,want_err+,suffix=suffix,sixteen-,
            repval=0.1,>"dev$null")
         display(infile//suffix,1,fill+)
      }

      ###############################################
      #     SECTION FOR CREATING SAM TRACE FILE     #
      ###############################################

      # Do this if you want to set up a trace file

      if (wantt) {
   
         # Set up parameters for trace

         if (instrum == "FOS") {
            axis = 2 }
         else {
            axis = 1
         }

         # If the user wants to use a standard trace, then copy it over here

         if (instrum == "FOS" && standtr) {
            ftrace = "stdtr_fos"
            if (telescope == "INT")
               ftrace = ftrace//1
            else
               ftrace = ftrace//2
           if (access(ftrace//extension)) imdelete(images=ftrace,go+,ver-,def+)
           rfits(fits_file="sam$stdtrace/"//ftrace//".fits",
           iraf_file=ftrace//extension,make+,long-,short-,data="",blank=0,
           scale+,oldiraf-,off=0)

         # Otherwise create it yourself...

         } else {

            # Another gumby message...

            if (gum) {
               print(" ")
               print("#############################################################")
               print("#  Right...You have chosen to do a trace.  I need to know   #")
               print("#  a few things about your prospective trace.  First of all #")
               print("#  I'll need to know what order of polynomial to fit.  For  #")
               print("#  FOS spectra, 5 is a good number and for ISIS and IDS     #")
               print("#  you generally only need a first order fit (a line).      #")
               print("#  I'll need to know how many spectra you want to fit with  #")
               print("#  a trace.  If all spectra have the same curvature then    #")
               print("#  can get away with only fitting one spectrum.  Otherwise  #")
               print("#  you'll need to fit any spectrum which you may want to    #")
               print("#  extract at some time.  Finally I need a name for the     #")
               print("#  output trace file.  Try to make it unique, but if it     #")
               print("#  isn't I can fix that.                                    #")
               print("#############################################################")
            }

            nord = norder
            nobj = nobjt
            ftrace = trace

            # See if the output exists already.  If it does and it can
            # be clobbered, then clobber it.  If it exists and can't be
            # clobbered, create a temporary name which the user can
            # change.

            len = strlen(ftrace)
            if (substr(ftrace,len-3,len) == extension) 
               ftrace = substr(ftrace,1,len-4)
            if (access(ftrace//extension)) {
               clobber = (envget("clobber") == "yes")
               if (clobber) {
                  imdelete(images=ftrace,go+,ver-,def+) }
               else {
                  print("You already have a file called "//ftrace)
                  ftrace = mktemp(ftrace)
                  print("The output will now be called "//ftrace)
                  print("Please do NOT rename it!")
               }
            }
         
            # Now, do the trace

            if (gum) {
               print(" ")
               print("#############################################################")
               print("#  In a moment you'll be given a cursor on the image.       #")
               print("#  You'll need to do the following for each spectrum which  #")
               print("#  you want to trace.  Place the cursor on the first of     #")
               print("#  the spectra to be traced.  Hit any key on the keyboard   #")
               print("#  except `q'.  This is to give me an approximate idea of   #")
               print("#  location of the spectrum.  Next I'll need to know how    #")
               print("#  much of the spectrum you want to fit.  You can do this   #")
               print("#  by (1) marking the two limits on the image with the      #")
               print("#  cursor or (2) typing the letter `w' which means that you #")
               print("#  want to fit over the whole of the spectral range.  Note  #")
               print("#  if you type `q' at any stage the program will quit.      #")
               print("#  You'll need to do this exercise for each spectrum you    #")
               print("#  want to trace.                                           #")
               print("#############################################################")
            }
            trace(input=infile//suffix,axis=axis,nord=nord,nobj=nobj,
            output=ftrace)
         }
      }

      ###############################################
      #     SECTION FOR CREATING SAM DEKKER FILE    #
      ###############################################

      if (wantd) {

         # Another gumby message

         if (gum) {
            print(" ")
            print("#############################################################")
            print("#  You have opted to create a dekker file.  I'll need to    #")
            print("#  a few more things now.  The first thing is the number of #")
            print("#  spectra which you want to include in the dekker file.    #")
            print("#  This should include any spectra which you might want to  #")
            print("#  extract at some stage.                                   #")
            print("#############################################################")
         }

         # Set up parameters for dekker

         if (instrum == "FOS") {
            axis = 2 }
         else {
            axis = 1
         }
         nobj = nobjd

         # If you did a new trace this time around or if you are using
         # FOS then you must include a trace in dekker.  Otherwise, you
         # are given a choice whether to use one or not.

         if (wantt || instrum == "FOS") {
            trac = yes }
         else {
            if (gum) {
               print(" ")
               print("#############################################################")
               print("#  You didn't create a trace this time around.  Do you want #")
               print("#  to use one in creating this dekker file?  If you do,     #")
               print("#  the program will follow the curvature in the defined     #")
               print("#  trace file.                                              #")
               print("#############################################################")
            }
            trac = track
         }

         # If you haven't created a trace this time around and you want
         # to use one, then get the name of the trace file here.  Check that
         # it has the right number of spectra...

         if (!wantt && trac && !access(ftrace//extension)) {
            ftrace = trace
            len = strlen(ftrace)
            if (substr(ftrace,len-3,len) == extension) 
               ftrace = substr(ftrace,1,len-4)
            imgets(ftrace,"i_naxis2")
            ny = int(imgets.value)
            if (ny != nobj && ny != 1) {
               print("The trace file has a different number of spectra")
               print("than the proposed dekker file.  If this is the number")
               print("of spectra you want, then redo the trace file first.")
               error("Rerun preset and redefine your trace")
            }  
         }

         # Standard trace file contains the maximum number of spectra possible.
         # If the new trace is from a standard trace file, and you want to mark
         # fewer spectra than the max, then subset the file. This is to make sure
         # that the trace and dekker files have the same number of objects for 
         # FOS.  If the dekker file has only one object, then it is assumed that
         # the user is after the first order spectrum.
   
         if (wantt && standtr) {
            if (nobj == 1) {
               temp1 = mktemp("SAM")
               imstack(images=ftrace//"[*,1]",output=temp1,title="*",
                  pixtype="*")
               imdelete(images=ftrace,go+,ver-,def+)
               imrename(oldnames=temp1,newnames=ftrace,ver-)
            }
         }

         # See if the output exists already.  If it does and it can
         # be clobbered, then clobber it.  If it exists and can't be
         # clobbered, create a temporary name which the user can
         # change.

         if (gum) {
            print(" ")
            print("#############################################################")
            print("#  Right, the last thing I need now is an output file name  #")
            print("#  for the dekker file.  Try and make it unique, but it     #")
            print("#  isn't a problem if you don't.                            #")
            print("#############################################################")
         }         
         dek = dekker
         len = strlen(dek)
         if (substr(dek,len-3,len) == extension) 
            dek = substr(dek,1,len-4)
         if (access(dek//extension)) {
            clobber = (envget("clobber") == "yes")
            if (clobber) {
               imdelete(images=dek,go+,ver-,def+) }
            else {
               print("You already have a file called "//dek)
               dek = mktemp(dek)
               print("The output will now be called "//dek)
               print("Please do NOT rename it!")
            }
         }

         # Now do the dekker

         if (gum) {
            print(" ")
            print("#############################################################")
            print("#  You'll now be given a cursor on the image.  For each     #")
            print("#  spectrum which you want to extract, you'll first need to #")
            print("#  mark the spatial limits of the slit.  Make sure you have #")
            print("#  enough pixels to include some sky bands.  Mark these     #")
            print("#  points with the cursor and any key except `q'.  If this  #")
            print("#  is a FOS frame, then don't mark too close to the edge of #")
            print("#  of the slit, or you might confuse sam's autofinder. Next #")
            print("#  I need to know how much of the spectrum is to be         #")
            print("#  extracted. You can do this in several ways: (1) mark the #")
            print("#  spectral limits with two cursor points; (2) type the     #")
            print("#  letter `w' which means you'll use the whole spectral     #")
            print("#  range; or (3) type the letter `t' which means that the   #")
            print("#  spectral range will be the same as any trace you might   #")
            print("#  be using. Do this for each spectrum in turn.  Don't type #")
            print("#  `q' unless you really want to quit!                      #")
            print("#############################################################")
         }
         dekker(axis=axis,nobj=nobj,track=trac,trace=ftrace,dekker=dek)
      }

      # Clean up garbage

      if ((wantt && !standtr) || wantd ) {
         imdelete(infile//suffix//"*"//extension,go+,ver-,def+)
         delete(files="scp_log.*",go+,ver-,def+,all+,sub+)
      }

      # See if the standard table exists.  If it doesn't, then create
      # it now...

      if (! access(stdtable)) {
         temp1 = mktemp("SAM")
         temp2 = mktemp("SAM")
         print(inst//" chr*32",>temp1)
         print(filename,>temp2)
         print(ftrace//extension,>>temp2)
         print(dek//extension,>>temp2)
         tcreate(table=stdtable,cdfile=temp1,datafile=temp2,uparfile="",
            nskip=0,nlines=0,nrows=0,hist+,extrapar=5,tbltype="default",
            extracol=0)
         delete(files=temp1,go-,ver-,def+,all+,sub+)
         delete(files=temp2,go-,ver-,def+,all+,sub+)
      } 

      # If it does exist, then merge the previous info...

      else {
         temp1 = mktemp("SAM") 
         temp2 = mktemp("SAM")
         temp3 = mktemp("SAM")
         temp4 = mktemp("SAM")
         print(inst//" chr*32",>temp1)
         print(filename,>temp2)
         print(ftrace//extension,>>temp2)
         print(dek//extension,>>temp2)
         tcreate(table=temp3,cdfile=temp1,datafile=temp2,uparfile="",
            nskip=0,nlines=0,nrows=0,hist+,extrapar=5,tbltype="default",
            extracol=0)
         tmerge(intable=stdtable//","//temp3,outtable=temp4,option="merge",
            allcols+,tbltype="column",allrows=100,extracol=0)
         tdelete(table=stdtable,go+,ver-,def+)
         tdelete(table=temp3,go+,ver-,def+)
         trename(intable=temp4,outtable=stdtable,ver-)
         delete(files=temp1,go-,ver-,def+,all+,sub+)
         delete(files=temp2,go-,ver-,def+,all+,sub+)
      }
end
