procedure preview(input,autofind,optimal,track)
#------------------------------------------------------------------------------
#
#   -------------
#   P R E V I E W
#   -------------
#
#   Description
#   -----------
#   Do quick look sky subtraction and extraction of spectra on a frame
#
#   
#   Scope of program
#   ----------------
#   - Only 2d frames supported
#   - Data type irrlevant
#   - Error files supported
#   - Batch execution not supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   CL script
#
#
#   Parameters
#   ----------
#   input      This  is  the  name of the input file which contains the spectra
#              to be extracted. (character)(prompted for)
#
#   autofind   The routine can automatically locate the spectra and sky  bands
#              with  the  defined limits of the dekker.  This is quite reliable
#              for medium to high signal/noise data.  For very low s/n data  it
#              may be necessary to do this by interactively. (bool)
#              (prompted for)
#   
#   optimal    This  is  true  if  the  user wants to do an optimal extraction.
#              For low s/n spectra, this can be quite beneficial.  The  routine
#              will  run  more  slowly if optimal extraction is used, basically
#              because it has an extra step to perform.   For  medium  to  high
#              s/n  data, the benefits of optimal extraction are not that great
#              and hence it is probably not worth doing it. (bool)
#              (prompted for)
#
#   track      If a trace has  been  defined  for  IDS  or  ISIS  (which  don't
#              normally  need  one) then the user is given the option to ignore
#              it.  This option is not given for FOS  spectra  as  they  always
#              require a trace. (bool)(prompted for)
#
#   reverse    This  is a hidden parameter.  It is assumed in this program that
#              the left hand side of an ISIS or IDS spectrum and that  the  top
#              of  a  FOS  spectrum  is the blue end.  If this is not the case,
#              then reverse should be set to true. (bool)(hidden)
#
#   gumby      This is a hidden parameter which, if switched on, gives the  new
#              user  advice  on how to use the program.  If gumby is used, then
#              each prompt for information is preceded by a  small  explanation
#              of  what is required of the user.  To use gumby mode, one should
#              type: `preview gumby+' on the command line and not use epar.
#              (bool)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   - Output is to temporary 1d files.
#
#   
#   Method
#   ------
#   - The user specifies a frame with spectra to extract.
#   - The input file header is examined for telescope and instrument
#     information.  The chip, trace and dekker information for that
#     particular combination is read from the standard preview
#     table.
#   - Old temporary files are deleted
#   - The input frame is bias subtracted and displayed, then the
#     extraction is done.
#   - The spatial and spectral limits of the seeing disc, the sky bands and
#     the dekkers are plotted on the image display.
#   - A rough wavelength calibration is done. If the user so specifies, the
#     wavelength direction is reversed on the frame.
#   - The final spectra are plotted.
#
#
#   External applications called
#   ----------------------------
#   Package images:
#      hedit
#      imdelete
#
#   Package onedspec:
#      specplot
#
#   Package scp:
#      proc
#
#   Package sam:
#      proc
#      showme
#
#   Package tables:
#      tdump
#
#   Package tv:
#      display
#
#
#   Possible future upgrades
#   ------------------------
#   - Wavelength calibration for IDS spectra needed. 
#   - A better overall method for wavelength calibration is needed which 
#     removes the ambiguity about the blue and red ends of the spectrum.
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   09-DEC-93     - Original program 
#
#
#------------------------------------------------------------------------------

# Query Parameters

string       input     {prompt="Name of input file"}
bool         autofind  {yes,prompt="Locate spectra automatically?"}
bool         optimal   {no,prompt="Want optimal extraction?"}
bool         track     {yes,prompt="Want to use a trace for this file?"}

# Hidden Parameters

bool         reverse   {no,prompt="Is wavelength sense reversed on frame?"}
bool         gumby     {no,prompt="Do you need lots of extra help to run this?"}
struct       *temp

begin

      # Local variables

      string infile,output,telescope,instrum,inst,isiscolour,
         stdtable,chipfile,trace,dekker,exphead,suffix,tempfil,plfile,
         hout,xlabel,tempfil2
      int len,axis,idum,nobj,nw,iobj
      bool auto,opt,want_err,trac,wavecal,rev,gum
      real cenwave,disper,pixsize,dlam,dummy,pixstart,lamcs[2],dlams[2],lam0

      # See if the user is a gumby or not...(i.e. whether lots of extra
      # help is required)

      gum = gumby

      # Initialise a few things...

      suffix = mktemp("P")
      chipfile = ""
      trace = ""
      dekker = ""
      tempfil = mktemp("SAM")
      tempfil2 = mktemp("SAM")
      stdtable = "sam_preview"

      # Get input file name and get rid of the file extension if it exists

      if (gum) {
         print("############################################################")
         print("#                        PREVIEW                           #")
         print("#  This program does the actual quick-look extractions on  #")
         print("#  your spectra.  I only need a few pieces of information  #")
         print("#  so this should be reasonably straight-forward.  What    #")
         print("#  I need is the name of the file with the spectra to be   #")
         print("#  extracted.                                              #")
         print("############################################################")
      }
      infile = input
      len = strlen(infile)
      if (substr(infile,len-3,len) == "."//envget("imtype")) 
         infile = substr(infile,1,len-4)

      # Get telescope identifier and find out how this was observed

      imgets(infile,"TELESCOP")
      telescope = imgets.value
      if (telescope == "WHT") {
         imgets(infile,"INSTRUME")
      } else if (telescope == "INT") {
         imgets(infile,"DETID")
      } else {
         beep
         print("No telescope information in the header")
         goto cleanup
      }
      instrum = substr(imgets.value,1,3)
      if (instrum == "FOS") {
         if (telescope == "WHT") {
            inst = "FOS2" }
         else {
            inst = "FOS1"
         }
      }
      else if (instrum == "ISI") {
         isiscolour = substr(imgets.value,6,6)
         inst = "ISIS"//isiscolour }
      else {
         instrum = "IDS"
         inst = "IDS"
      }
      
      # Open the standard table and dump the information for this
      # instrument to the temporary file.

      if (!access(stdtable//".tab")) {
         beep
         print("Can't find standard table sam_preview -- run preset first")
         goto cleanup
      }
      tdump(table=stdtable,cdfile="",pfile="",datafile=tempfil2,
         columns=inst,rows="-",pwidth=80)

      # Read the chip file name from the temporary file...

      temp = tempfil2
      idum = fscan(temp,chipfile)
      if (idum == EOF) {
         beep
         print("There is no information for this instrument in ")
         print("the sam_preview table. Try running preset first.")
         goto cleanup
      }
      if (chipfile == "") {
         beep
         print("You have no chip file defined for this instrument")
         print("Try running preset first")
      }

      # Read the trace file name from the temporary file. Defer judgement
      # on the file name until you see whether one is required.

      idum = fscan(temp,trace)
      if (idum == EOF) {
         beep
         print("There is no trace file information for this instrument")
         print("Try running preset first.")
         goto cleanup
      }

      # Read the dekker file name from the temporary file

      idum = fscan(temp,dekker)
      if (dekker == "" || idum == EOF) {
         beep
         print("There is no dekker file information for this instrument")
         print("Try running preset first")
         goto cleanup
      }
      temp = ""

      # Get some useful parameters

      if (gum) {
         print(" ")
         print("############################################################")
         print("#  It is possible for this program to locate each of the   #")
         print("#  spectra on your frame and its sky bands automatically.  #")
         print("#  For medium and high s/n data this works quite reliably. #")
         print("#  For low s/n data you may want to locate the spectra by  #")
         print("#  hand.  Possibly the best thing to do is to use the      #")
         print("#  autofinder and if you're not happy with the result then #")
         print("#  rerun this with the autofinder switched off.            #")
         print("############################################################")
      }
      auto = autofind
      if (gum) {
         print(" ")
         print("############################################################")
         print("#  You will be given the option to do an `optimal          #")
         print("#  extraction' of your spectra.  This is a means of        #")
         print("#  maximising the output s/n for a given input frame. This #")
         print("#  is quite useful for low s/n spectra, but the gain is    #")
         print("#  fairly minimal for high s/n spectra.  Not using optimal #")
         print("#  extraction may save a lot of time!                      #")
         print("############################################################")
      }
      opt = optimal
      output = mktemp("SAM")
      rev = reverse

      # Get rid of any old output files which might be lying around

      temp = "sam_ofiles"
      if (access("sam_ofiles")) {
         while (fscan(temp,tempfil) != EOF) {
            len = strlen(tempfil)
            if (substr(tempfil,len-3,len) == "."//envget("imtype")) 
            tempfil = substr(tempfil,1,len-4)
            imdelete(tempfil//"*."//envget("imtype"),go+,ver-,def+)
         }
      }
      temp = ""

      # Set up a few parameters for SAM.  For FOS, the y axis is the spectral axis, and
      # you must use a trace.  For ISIS or IDS, the x axis is the spectral axis.
      # If you haven't defined a trace, then it's assumed that you don't want to use
      # one.  If you have defined a trace, then you are given the option to ignore it.

      if (instrum == "FOS") {
         trac = yes
         if (trace == "") {
            beep
            print("There is no trace file for this instrument")
            print("Try running preset first.")
            goto cleanup
         }
         axis = 2 }
      else {
         axis = 1
         if (trace == "" ) {
            trac = no }
         else {
            if (gum) {
               print(" ")
               print("############################################################")
               print("#  I'm about to ask you if you want to use a trace for     #")
               print("#  this frame.  I know you have defined one but I thought  #")
               print("#  I would give you the option to ignore it.  Answer `yes' #")
               print("#  if you still want to use it.                            #")
               print("############################################################")
            }
            trac = track
         }
      }

      # Set up a few variables for scp.proc

      exphead = "EXPOSED"
      if (telescope == "INT") exphead = "TIMEXPOS"

      # Now do bias strip subtraction and create error array if it is needed

      if (gum) {
         print(" ")
         print("############################################################")
         print("#  Hang on while I bias subtract and display your frame... #")
         print("############################################################")
      }     
      want_err = opt
      scp.proc(bias-,dark-,flat-,prog+,chipfile=chipfile,bias_frames="",
         dark_frames="",flat_frames="",prog_frames=infile,
         exphead=exphead,meanbias="",sub_bias+,sub_frame-,over-,
         meandark="",dark_cor-,meanflat="",flat_div-,image_flat+,
         nord=0,axis=1,trim+,want_err=want_err,suffix=suffix,sixteen-,
         repval=0.1,>"dev$null")

      # Display the bias subtracted file

      display(image=infile//suffix,frame=1,fill+)

      # Now run sam.proc

      if (gum && !auto) {
         print(" ")
         print("############################################################")
         print("#  I'm starting to do the extraction now.  In a moment you #")
         print("#  be presented with a line plot.  This is the information #")
         print("#  between the dekker limits of your first spectrum all    #")
         print("#  summed together. (e.g. squashed in the spectral axis)   #")
         print("#  I need you to indicate some points with the cursor. You #")
         print("#  can do this by hitting any key on the keyboard except   #")
         print("#  `q'.  First, indicate the location of the peak of the   #")
         print("#  spectral emission.  Next mark the spatial limits of the #")
         print("#  spectrum with 2 cursor points.  These should be the     #")
         print("#  points where the spectrum has fallen down to a minimal  #")
         print("#  level.  Next, mark a band of pixels to be used for sky  #")
         print("#  on the left-hand side of the spectrum (use two more     #")
         print("#  cursor points).  Finally, mark a similar band on the    #")
         print("#  right-hand side with 2 cursor points.  That should be a #")
         print("#  total of 7 cursor points.  Hitting `q' will cause the   #")
         print("#  program to quit.  Do this for each spectrum plotted.    #")
         print("############################################################")
      } else if (gum && auto) {
         print(" ")
         print("############################################################")
         print("#  Hang on while I do the extraction...                    #")
         print("############################################################")
      }               
      sam.proc(input=infile//suffix,output=output,optimal=opt,autofind=auto,
      axis=axis,track=trac,trace=trace,dekker=dekker)

      # Show where the extraction took place...

      if (gum) {
         print(" ")
         print("############################################################")
         print("#  What you are seeing on the image display is tramlines   #")
         print("#  being drawn of the spatial limits which have been used  #")
         print("#  in the extractions I've just done. In red are the       #")
         print("#  limits of the dekker. The sky bands are drawn in green. #")
         print("#  Finally the limits of the seeing disc are in blue.      #")
         print("############################################################")
      }
      showme(frame=1)

      # Work out rough wavelength calibration for ISIS from header info

      if (gum) {
         print(" ")
         print("############################################################")
         print("#  I'm just working out a rough wavelength calibration...  #")
         print("############################################################")
      }
      wavecal = no
      if (instrum == "ISI") {
         imgets(infile,"CENWAVE")
         cenwave = real(imgets.value)
         imgets(infile,"DISPERSI")
         disper = real(imgets.value)
         imgets(infile,"CCDXPIXE")
         pixsize = 1.0e-3*real(imgets.value)
         imgets(infile//suffix,"i_naxis1")
         nw = int(imgets.value)
         dlam = disper*pixsize
         if (rev) dlam = -dlam
         wavecal = yes
         imgets(infile//suffix,"nobj")
         nobj = int(imgets.value)
         for (iobj = 1; iobj <= nobj; iobj = iobj + 1) {
            listpix(dekker//"[3,"//iobj//"]",wcs="logical",formats="",
               ver-) | scan(dummy,pixstart)
            lam0 = cenwave - (0.5*real(nw) - pixstart)*dlam
            if (nobj == 1) {
               hout = output }
            else if (iobj < 10) {
               hout = output//"_00"//iobj }
            else if (iobj < 100) {
               hout = output//"_0"//iobj }
            else {
               hout = output//iobj
            }
            hedit(images=hout,fields="crval1",value=lam0,add+,del-,ver-,show-,
            update+)
            hedit(images=hout,fields="cdelt1",value=dlam,add+,del-,ver-,show-,
            update+)
         }

      # Mean coefs for FOS wavelength calibration (only 1st and 2nd order
      # currently available...)

      } else if (instrum == "FOS") {
         if (inst == "FOS1") {
            dlams[1] = -10.72
            dlams[2] = -5.35
            lamcs[1] = 8332.5
            lamcs[2] =  4169.17 }
         else {
            dlams[1] = -8.69
            dlams[2]  = -4.34
            lamcs[1] = 7262.61
            lamcs[2] = 4664.30
         }
         wavecal = yes
         imgets(infile//suffix,"nobj")
         nobj =  int(imgets.value)
         imgets(infile//suffix,"i_naxis2")
         nw = int(imgets.value)
         for (iobj = 1; iobj <= min(2,nobj); iobj = iobj + 1) {
            listpix(dekker//"[3,"//iobj//"]",wcs="logical",formats="",
               ver-) | scan(dummy,pixstart)
            dlam = dlams[iobj]
            if (rev) dlam = -dlam
            cenwave = lamcs[iobj]
            lam0 = cenwave - (0.5*real(nw) - pixstart)*dlam
            if (nobj == 1) {
               hout = output }
            else if (iobj < 10) {
               hout = output//"_00"//iobj }
            else if (iobj < 100) {
               hout = output//"_0"//iobj }
            else {
               hout = output//iobj
            }
            hedit(images=hout,fields="crval1",value=lam0,add+,del-,ver-,show-,
            update+)
            hedit(images=hout,fields="cdelt1",value=dlam,add+,del-,ver-,show-,
            update+)
         }
         if (nobj > 2) 
            print("Sorry, couldn't wavelength calibrate orders higher than 2")
      } else if (instrum == "IDS") {
         print("Sorry, I can't wavelength calibrate IDS spectra yet...")
      }         
            
      # Now plot the output spectra

      if (gum) {
         print(" ")
         print("############################################################")
         print("#  I am now plotting the extracted spectra on a rough      #")
         print("#  wavelength scale (if it was available).  Note that the  #")
         print("#  wavelength calibration is only approximate and that     #")
         print("#  these spectra are NOT a replacement for a proper        #")
         print("#  treatment of the data.  You may save them for future    #")
         print("#  reference by using `prekeep'. If the wavelength scale   #")
         print("#  doesn't look right to you, then try re-running this     #")
         print("#  program in the following way:                           #")
         print("#                                                          #")
         print("#     sa> preview rev+                                     #")
         print("#                                                          #")
         print("#  This will reverse the wavelength calibration, which may #")
         print("#  at times be necessary.                                  #")
         print("############################################################")
      }
      imgets(infile//suffix,"nobj")
      nobj = int(imgets.value)
      if (nobj == 1) {
         plfile = output//"."//envget("imtype") }
      else {
         plfile = output//"_???."//envget("imtype")
      }
      if (wavecal) {
         xlabel = "Wavelength (Angstroms)"
      } else {
         xlabel = "Pixels"
      }
      specplot(plfile,autolay+,autosc+,yscale+,xlabel=xlabel,
      ylabel="Counts",cursor="dev$null")

      # Make a note of the output file names

      if (access("sam_ofiles")) delete(files="sam_ofiles",go-,ver-,
         def+,all+,sub+)
      print(plfile,>"sam_ofiles")

      # Clean some things up...

cleanup:
      delete(files=tempfil2,go-,ver-,def+,all+,sub+)
      imdelete("PROF*."//envget("imtype"),go+,ver-,def+)
      imdelete(infile//suffix//"*."//envget("imtype"),go+,ver-,def+)
      delete("scp_log.*",go-,ver-,def+,all+,sub+)
end

