procedure proc(input,output,optimal,autofind)
#------------------------------------------------------------------------------
#
#   -------
#   P R O C
#   -------
#
#   Description
#   -----------
#   Do spectrum find, sky subtraction, profile fit and extraction all
#   in one go.
#
#   
#   Scope of program
#   ----------------
#   - Only 2d frames supported
#   - Data type irrlevant
#   - Error files supported
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   CL script
#
#
#   Parameters
#   ----------
#   input      List of images with spectra to be processed (character)
#              (prompted for)
#
#   output     This the list of output files.  If there are more than one
#              spectrum on the frame, then each spectrum will be stored in a
#              separate file with this parameter as its root name.  E.g.  if
#              the output file is called `fred' then the second spectrum
#              extracted will be called `fred_002'.  If there is just one
#              spectrum on the frame, the output file will simply be called
#              `fred'. (character)(prompted for)
#
#   optimal    This is TRUE if the user wants optimal extraction.  Note that
#              it is not prompted for and it defaults to TRUE. (bool)
#              (prompted for)
#
#   autofind   This is TRUE if the user wants the spectra and sky bands
#              located automatically. (bool)(prompted for)
#
#   axis       The axis to which spectra are parallel -- i.e. the dispersion
#              axis.  1 = X, 2 = Y. (int)(hidden)
#
#   track      TRUE if a trace file is being used. (bool)(hidden)
#
#   trace      Name of the trace file if required. (character)(hidden)
#
#   dekker     Name of the dekker file. (character)(hidden)
#
#   nom_width  This is the nominal width of the spectrum in the spatial domain
#              and is used as the FWHM of a smoothing kernel in autofind mode.
#              (int)(hidden)
#
#   ansig      This is the number of sigma out to which you want to include as
#              part of the seeing disc.  The limits of the seeing disc are
#              therefore defined as x0 +/- ansig*sigma, where x0 and sigma
#              have been found by the autofinder.  This parameter should 
#              generally be 2-3. (real)(hidden)
#
#   max_sep    In automatic mode when the sky pixels are being chosen, the
#              routine needs to know how far away from the spectral pixels the
#              sky pixels are allowed to be.  Having sky pixels which are
#              reasonably  far from the spectral pixels could affect the sky
#              subtraction because of large scale variations in the sky data.
#              This is the maximum distance away from the edge of the defined
#              object region which sky pixels are allowed to be.  For  example,
#              if max_sep = 5 then the most distant sky pixel will be at most
#              5 pixels away from the nearest object pixel on that side.
#              (int)(hidden)
#
#   sky_ord    This is the order of the polynomial which will be fitted to the
#              sky rows.  One should be extremely careful about using orders
#              higher than about 2 since the interpolation in the object
#              pixels may involve areas where the gradient of the polynomials
#              becomes very large. (int)(hidden)
#
#   nreject    The sky polynomial fits are done iteratively with the worst
#              fitting pixel from each fit being rejected in each iteration.
#              This is the number of pixels which will be rejected. Again one
#              should be careful to not reject too many pixels as it may
#              become impossible to get a good fit. (int)(hidden)
#
#   pord       This is the order of the polynomials to be fit for the optimal
#              extraction profile (int)(hidden)
#
#   pol_dist   If a trace file is being used, then the algorithm used will be
#              that of Marsh. In this case there needs to be defined a
#              distance pixels between the polynomials in order for the 
#              interpolation scheme to work. The value must be less than 1 and
#              preferably would be less than 0.5. (real)(hidden)
#
#   pro_thr    The profile fitting is done iteratively where pixels whose
#              residuals are this number of sigma above the predicted profile
#              are rejected. The sigma here is the error for the individual
#              pixel. Very  high  signal/noise  data  should  have  a  
#              correspondingly high value of this threshold. It often isn't
#              worth doing more than one iteration as the change can be
#              minimal for high s/n data. In such a case this threshold can
#              be set very high (say 1000). (real)(hidden)
#
#   block      The user may block the spectra in wavelength in order to
#              increase the signal to noise. The value tells how many pixels
#              will be added together to make one block. It is recommended
#              that this value be 1 except in cases of very bad signal/noise.
#              (int)(hidden)
#
#   ext_thr    This is a threshold over which pixels will be rejected from
#              the optimal extraction. This is a way of getting rid of cosmic
#              ray hits and other defects and is only used when optimal
#              extraction is being used. The units are in number of sigma of
#              the individual pixels.  Very high s/n data should obviously
#              have very high values of this parameter. (real)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   - Output is to one dimensional files of type REAL
#
#   
#   Method
#   ------
#   - This routine staples together the programs find, sky, profile
#     and extract.  See the prologs for each of these for a description
#     of the methods employed.
#
#
#   External applications called
#   ----------------------------
#   Package images:
#      hedit
#      imdelete
#      sections
#
#   Package sam:
#      extract
#      find
#      profile
#      sky
#
#
#   Possible future upgrades
#   ------------------------
#     
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   09-DEC-93     - Original program 
#
#
#------------------------------------------------------------------------------

# Query Parameters

string    input          {prompt="List of input frames"}
string    output         {prompt="Output file name"}
bool      optimal        {yes,prompt="Do you want an optimal extraction?"}
bool      autofind       {yes,prompt="Find spectra automatically?"}

# Hidden Parameters (general)

int       axis           {1,min=1,max=2,prompt="Dispersion axis"}
bool      track          {no,prompt="Use a trace file"}
string    trace          {prompt="Name of trace file"}
string    dekker         {prompt="Name of dekker file"}

# Hidden Parameters (find)

int       nom_width      {7,min=1,prompt="Nominal full width of seeing disc"}
real      ansig          {2.0,min=0.01,
                          prompt="Number of sigma to delimit seeing disc"}
real      max_sep        {5,min=1,prompt="Maximum separation between sky and object rows"}
gcur      graphcur       {"",prompt="Location of spectrum"}

# Hidden Parameters (sky)

int       sky_ord        {0,min=0,prompt="Order of polynomial for sky fit"}
int       nreject        {1,min=0,prompt="Number of bad sky pixels to reject"}

# Hidden Parameters (profile)

int       pord           {2,min=0,prompt="Order of polynomial for profile fits"}
real      pol_dist       {0.4,min=0.01,max=1.0,
                          prompt="Distance between polynomials"}
real      pro_thr        {20.0,min=0.0,
                          prompt="Threshold in sigma for rejection of blocks"}
int       block          {1,min=1,prompt="Blocking factor"}

# Hidden Parameters (extract)

real      ext_thr        {20.0,min=0.0,prompt="Threshold for bad pixel deletion"}

# Image lists

struct    *inlist
struct    *outlist

begin

      # Local variables

      string tmpinfiles,tmpoutfiles,iname,oname,tmpskysub,tmpprof
      bool auto,opt
      int idummy,n_in,n_out

      # Expand input and output image lists

      tmpinfiles = mktemp("SAM")
      sections(input,option="fullname",>tmpinfiles)
      n_in = sections.nimages
      tmpoutfiles = mktemp("SAM")
      sections(output,option="fullname",>tmpoutfiles)
      n_out = sections.nimages
      if (n_in != n_out) {
         print("Input and output lists have different sizes!")
         bye
      }
                   
      # Get the rest of the parameters

      opt = optimal
      auto = autofind

      # Now loop through the input file list...

      inlist = tmpinfiles
      outlist = tmpoutfiles
      while (fscan(inlist,iname) != EOF) {
         idummy = fscan(outlist,oname)

         # First "find" the spectra

         print("Locating the spectra...")
         find(iname,auto,axis=axis,nom_width=nom_width,
            ansig=ansig,max_sep=max_sep,track=track,trace=trace,
            dekker=dekker,graphcur=graphcur)

         # Now do sky subtraction. Put sky subtracted frame into
         # a dummy file which you'll later delete

         print(" ")
         print(" ")
         print("Doing sky subtraction...")
         tmpskysub = mktemp("SAM")
         sky(iname,tmpskysub,sky_ord=sky_ord,nreject=nreject)

         # If this is an optimal extraction, then do the profile
         # fit now...

         if (opt) {
            print(" ")
            print(" ")
            print("Doing optimal extraction profile fits...")
            tmpprof = mktemp("PROF")
            profile(tmpskysub,tmpprof,block=block,pol_dist=pol_dist,
               pord=pord,pro_thr=pro_thr)
            hedit(iname,"pr_file",value=tmpprof,add+,del-,ver-,
               show-,up+)
         }

         # Now do the extraction

         print(" ")
         print(" ")
         print("Doing extraction...")
         extract(tmpskysub,oname,optimal=opt,ext_thr=ext_thr)

         # Get rid of the temporary sky file

         imdelete(tmpskysub//"*."//envget("imtype"),go_ahead+,verify-,default+)

      }

      # Get rid of temporary list files

      delete(tmpinfiles,go+,ver-,def+,all+,sub+)
      delete(tmpoutfiles,go+,ver-,def+,all+,sub+)
end
