procedure sam_poly(x,y,weights,npts,nterms,sumx,sumy,array,a,chisqr)
#--------------------------------------------------------------------------------------
#      SAM_POLY -- Fit a weighted polynomial to data points
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) x:          (real[npts]) Independent variable
#      (>) y:          (real[npts]) Dependent variable
#      (>) weights:    (real[npts]) Array of weights
#      (>) npts:       (int) Number of data points
#      (>) nterms:     (int) Number of terms in fit (order+1)
#      (W) sumx:       (double[2*nterms-1]) Workspace
#      (W) sumy:       (double[nterms]) Workspace
#      (W) array:      (double[nterms,nterms]) Workspace
#      (<) a:          (real[nterms]) Output coefficients of fit
#      (<) chisqr:     (real) Reduced chi-squared of fit
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        22-FEB-1993 -- Based on pp 141-2 of Bevington
#      UPDATE:
#
#--------------------------------------------------------------------------------------

# Parameters

int npts,nterms
real x[npts],y[npts],weights[npts],a[nterms],chisqr
double sumx[2*nterms-1],sumy[nterms],array[nterms,nterms]

# Local variables

int nmax,n,j,i,k,l,free
real xi,yi,weight,xterm,yterm,delta
double chisq

# Functions

real sam_determ()

errchk sam_determ

begin

      # Zero some arrays
      
      nmax = 2*nterms - 1
      call amovkd(0.0d0,sumx,nmax)
      call amovkd(0.0d0,sumy,nterms)
      chisq = 0.0d0

      # Do the necessary sums...

      do i = 1,npts {
         xi = x[i]
         yi = y[i]
         weight = weights[i]
         xterm = weight
         do n = 1,nmax {
            sumx[n] = sumx[n] + double(xterm)
            xterm = xterm*xi
         }
         yterm = weight*yi
         do n = 1,nterms {
            sumy[n] = sumy[n] + double(yterm)
            yterm = yterm*xi
         }
         chisq = chisq + double(weight)*double(yi)**2
      }

      # Construct matrices and calculate coefficients

      do j = 1,nterms {
         do k = 1,nterms {
            n = j + k - 1
            array[j,k] = sumx[n]
         }
      }

      # If the determinant is zero then signal an error

      delta = sam_determ(array,nterms)
      if (delta == 0.0) {
         chisqr = 0.0
         do j = 1,nterms {
            a[j] = 0.0
         }
         call error(0,"SAM_POLY: fit failed -- determinant was zero.")
      }

      # Otherwise, carry on...

      do l = 1,nterms {
         do j = 1,nterms {
            do k = 1,nterms {
               n = j + k - 1
               array[j,k] = sumx[n]
            }
            array[j,l] = sumy[j]
         }
         a[l] = sam_determ(array,nterms)/delta
      }

      # Calculate chi-squared

      do j = 1,nterms {
         chisq = chisq - 2.0d0*double(a[j])*sumy[j]
         do k = 1,nterms {
            n = j + k - 1
            chisq = chisq + double(a[j])*double(a[k])*sumx[n]
         }
      }
      free = npts - nterms
      chisqr = real(chisq)/real(free)
end

         