include <math.h>

procedure sub_finda(iptr,dekker,trace,nord,spec_rng,nom_width,ansig,
   idistmax,nx,profile,mask,pos,width,sky1,sky2)
#--------------------------------------------------------------------------------------
#      SUB_FINDA -- Main subroutine for finding a spectrum automatically
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) iptr:       (pointer) Pointer to input file
#      (>) dekker:     (real[2]) Limits of dekker
#      (>) trace:      (real[nord+4]) Trace coefficients
#      (>) nord:       (int) Order of trace polynomial
#      (>) spec_rng:   (real[2]) Requested spectral range
#      (>) nom_width:  (int) Nominal width of seeing disc
#      (>) ansig:      (real) Number of sigma for limits of spectrum
#      (>) idistmax:   (int) Maximum distance a sky row can be from the object
#      (>) nx:         (int) Number of pixels between dekker limits
#      (W) profile:    (real[nx]) Workspace
#      (W) mask:       (short[nx]) Workspace
#      (<) pos:        (real) Position of each object
#      (<) width:      (real[2]) Limits of seeing disc for each object
#      (<) sky1:       (real[2]) First sky band for each object
#      (<) sky2:       (real[2]) Second sky band for each object
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        22-FEB-1993
#      UPDATE:      15-NOV-1993 -- Major revision to include new peak
#                                  and background finding algorithm (JRL)
#
#--------------------------------------------------------------------------------------

# Parameters

pointer iptr
short mask[nx]
int nord,nom_width,idistmax,nx,width[2],sky1[2],sky2[2]
real dekker[2],trace[nord+4],ansig,spec_rng[2],pos,profile[nx]

# Local variables

pointer prptr,kptr,cnvptr,xptr
int iext,nk,nk1,nnx,i,istop1,istop2,isep,i1,i2,id1,id2
real sig_gau,anorm,denom,x0,sigma

errchk sam_trspec,malloc,sam_peaks

begin

      # Now sum the pixels in the trace area into these work arrays

      call sam_trspec(iptr,spec_rng,nord,trace,nx,profile)

      # Set up some sizes for the convolution

      sig_gau = nom_width/2.355  # Sigma of gaussian is same as that for seeing disc
      iext = nom_width
      nk = 2*iext + 1
      nk1 = nk/2 + 1
      nnx = nx + 2*iext

      # Get some workspace for the convolution and peak finding

      call malloc(prptr,nnx,TY_REAL)
      call malloc(kptr,nk,TY_REAL)
      call calloc(cnvptr,nnx,TY_REAL)
      call malloc(xptr,nx,TY_REAL)

      # Set up Gaussian kernel

      anorm = 1.0/(sqrt(TWOPI)*sig_gau)
      denom = -0.5/sig_gau**2
      do i = 1,nk 
         Memr[kptr+i-1] = anorm*exp(denom*(real(i-nk1)**2))

      # Move profile to the workspace, extending it at either end.

      do i = 1,iext
         Memr[prptr+i-1] = profile[1]
      do i = nx+iext+1,nnx
         Memr[prptr+i-1] = profile[nx]
      call amovr(profile,Memr[prptr+iext],nx)

      # Convolve the profile with the Gaussian

      call acnvrr(Memr[prptr],Memr[cnvptr],nx,Memr[kptr],nk)

      # Now find the local maximum and its gaussian width

      do i = 1,nx
         Memr[xptr+i-1] = real(i) + dekker[1] - 1.0
      call sam_peaks(Memr[xptr],Memr[cnvptr],nx,x0,sigma)

      # Ditch the workspace you no longer need

      call mfree(prptr,TY_REAL)
      call mfree(kptr,TY_REAL)
      call mfree(cnvptr,TY_REAL)
      call mfree(xptr,TY_REAL)

      # Deconvolove the profile sigma...

      sigma = sigma**2 - sig_gau**2
      if (sigma > 0.0) {
         sigma = sqrt(sigma) }
      else {
         call error(0,"Width of profile negative?")
      }

      # How many pixels to include...

      pos = x0
      istop1 = nint(x0 - ansig*sigma)
      istop2 = nint(x0 + ansig*sigma)
      id1 = nint(dekker[1])
      id2 = nint(dekker[2])
      if (istop1 <= id1 || istop2 >= id2) 
         call error(0,"Profile to wide -- try reducing 'ansig' parameter") 
      width[1] = real(istop1)
      width[2] = real(istop2)

      # Orgainise LHS sky band...

      isep = 0
      do i = istop1-1,id1,-1 {
         isep = isep + 1
         if (isep == idistmax) break
      }
      i1 = max(id1,i)
      i2 = istop1 - 1
      if (isep == idistmax) i2 = i2 - 1
      sky1[1] = real(i1) 
      sky1[2] = real(i2) 

      # Orgainise RHS sky band...

      isep = 0
      do i = istop2+1,id2 {
         isep = isep + 1
         if (isep == idistmax) break
      }
      i1 = istop2 + 1
      i2 = min(id2,i)
      if (isep == idistmax) i1 = i1 + 1
      sky2[1] = real(i1) 
      sky2[2] = real(i2) 
end
