procedure sub_profh(iptr,eptr,nx,ny,pord,block_size,pro_thr,
   profile,tot,variance,prof,var,xfit,yfit,wfit,sumy,coefs,pro_ok,keep,output)
#--------------------------------------------------------------------------------------
#      SUB_PROFH -- Main subroutine for finding optimal extract weights using
#                   Keith Horne's algorithm. (straightish spectra)
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) iptr:       (pointer) Pointer to input file
#      (>) eptr:       (pointer) Pointer to input error file
#      (>) nx:         (int) X dimension of input data
#      (>) ny:         (int) Y dimension of input data
#      (>) pord:       (int) Order of polynomials to be used in profile fit
#      (>) block_size: (int) Blocking factor in the spectral direction
#      (>) pro_thr:    (real) Threshold used for deleting bad blocks from fit
#      (W) profile:    (real[nx]) Workspace
#      (W) tot:        (real[nx]) Workspace
#      (W) variance:   (real[nx]) Workspace
#      (W) prof:       (real[nx,ny]) Workspace
#      (W) var:        (real[nx,ny]) Workspace
#      (W) xfit:       (real[ny,nx]) Workspace
#      (W) yfit:       (real[ny,nx]) Workspace
#      (W) wfit:       (real[ny,nx]) Workspace
#      (W) sumy:       (real[ny]) Workspace
#      (W) coefs:      (real[pord+1,nx]) Workspace
#      (W) pro_ok:     (bool[nx,ny]) Workspace
#      (W) keep:       (bool[ny]) Workspace
#      (<) output:     (real[nx,ny]) Output profile
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        22-FEB-1993
#      UPDATE:
#
#--------------------------------------------------------------------------------------

# Parameters

pointer iptr,eptr
int nx,ny,pord,block_size
real pro_thr,profile[nx],tot[nx],variance[nx],prof[nx,ny],var[nx,ny],
   xfit[ny,nx],yfit[ny,nx],wfit[ny,nx],sumy[ny],coefs[pord+1,nx],output[nx,ny]
bool pro_ok[nx,ny],keep[ny]

# Local variables

pointer ptr1,ptr2,wpt1,wpt2,wpt3
int nchunks,j,iy1,iy2,jj,ii,i,ndel,iter
real yscale,sum,y1,y2,chisqr,val,resid

# Functions

pointer imgl2r()
real apolr()

errchk imgl2r

begin

      # Initialise a few things...

      nchunks = ny/block_size
      if (mod(ny,block_size) != 0) nchunks = nchunks + 1
      yscale = 1.0/real(ny)

      # Form standard profile

      do j = 1,nchunks {
         iy1 = block_size*(j-1) + 1
         iy2 = iy1 + block_size
         if (j == nchunks) iy2 = ny

         # Initialise summation workspaces...

         call aclrr(profile,nx)
         call aclrr(tot,nx)
         call aclrr(variance,nx)

         # Sum up the profile along the x direction

         do jj = iy1,iy2 {
            ptr1 = imgl2r(iptr,jj)
            ptr2 = imgl2r(eptr,jj)
            do ii = 1,nx {
               profile[ii] = profile[ii] + Memr[ptr1+ii-1]
               variance[ii] = variance[ii] + Memr[ptr2+ii-1]
               tot[ii] = tot[ii] + 1.0
            }
         }

         # Divide by total contributions

         sum = 0.0
         do ii = 1,nx {
            if (tot[ii] > 0.0) {
               profile[ii] = profile[ii]/tot[ii]
               variance[ii] = variance[ii]/(tot[ii]**2)
               pro_ok[ii,j] = true }
            else {
               profile[ii] = 0.0
               variance[ii] = 0.0
               pro_ok[ii,j] = false
            }
            sum = sum + profile[ii]
         }

         # Normalise all this

         if (sum > 0.0) {
            do ii = 1,nx {
               if (pro_ok[ii,j]) {
                  prof[ii,j] = profile[ii]/sum
                  var[ii,j] = variance[ii]/sum**2
               }
            }
            keep[j] = true }
         else {
            do ii = 1,nx {
               pro_ok[ii,j] = false
            }
            keep[j] = false
         }
      } # End standard profile loop

      # Set up fitting arrays...Note they are set up backwards (e.g. y,x) for speed

      do i = 1,nx {
         do j = 1,nchunks {
            y1 = (j-1)*block_size + 1
            y2 = y1 + block_size
            xfit[j,i] = 0.5*(y1 + y2)*yscale
            yfit[j,i] = prof[i,j]
            if (keep[j] && pro_ok[i,j]) {
               wfit[j,i] = 1.0/var[i,j] }
            else {
               wfit[j,i] = 0.0
            }
         }
      }

      # Get some workspace

      call malloc(wpt1,2*(pord+1)-1,TY_DOUBLE)
      call malloc(wpt2,pord+1,TY_DOUBLE)
      call malloc(wpt3,(pord+1)**2,TY_DOUBLE)

      # Now do the fits and iterate until you've deleted no more points

      ndel = 1
      iter = 0
      while (ndel != 0) {
         iter = iter + 1
         ndel = 0

         # Initialise summation

         call aclrr(sumy,nchunks)

         # Loop for each column

         do i = 1,nx {

            # Do the fit for that column and evaluate the fit over that 
            # column if the fit worked.

            ifnoerr (call sam_poly(xfit[1,i],yfit[1,i],wfit[1,i],nchunks,pord+1,
               Memd[wpt1],Memd[wpt2],Memd[wpt3],coefs[1,i],chisqr)) {
               do j = 1,nchunks {
                  val = apolr(xfit[j,i],coefs[1,i],pord+1)
                  val = max(0.0,val)
                  output[i,j] = val
                  sumy[j] = sumy[j] + val
               } }
            else {
               do j = 1,nchunks {
                  pro_ok[i,j] = false
                  wfit[j,i] = 0.0
                  ndel = ndel + 1
               }
            }
         }

         # Now renormalise the profile

         do j = 1,nchunks {
            sum = sumy[j]
            if (sum != 0.0) {
               do i = 1,nx {
                  output[i,j] = output[i,j]/sum
               } }
            else {
               do i = 1,nx {
                  pro_ok[i,j] = false
                  wfit[j,i] = 0.0
                  ndel = ndel + 1
               }
               keep[j] = false
            }

            # Find which points to reject

            do i = 1,nx {
               if (pro_ok[i,j]) {
                  resid = (prof[i,j] - output[i,j])/sqrt(var[i,j])
                  if (abs(resid) > pro_thr) {
                     ndel = ndel + 1
                     pro_ok[i,j] = false
                     wfit[j,i] = 0
                  }
               }
            }
         }

         # Friendly message...

         call printf("Iteration # %d, %d points deleted from fit\n")
            call pargi(iter)
            call pargi(ndel)

      }  # End of iteration loop

      # Now evaluate the profile over the whole image for output...

      do j = 1,ny {
         y1 = real(j)*yscale
         sum = 0.0
         do i = 1,nx {
            val = apolr(y1,coefs[1,i],pord+1)
            val = max(val,0.0)
            output[i,j] = val
            sum = sum + val
         }
         do i = 1,nx {
            if (sum != 0.0) {
               output[i,j] = output[i,j]/sum }
            else {
               output[i,j] = 0.0
            }
         }
      }

      # Ditch the workspace

      call mfree(wpt1,TY_DOUBLE)
      call mfree(wpt2,TY_DOUBLE)
      call mfree(wpt3,TY_DOUBLE)
end
