procedure sub_profm(iptr,eptr,nx,ny,ny1,ny2,nchunks,block_size,nord,trace,
   pol_dist,npoly,pord,pro_thr,xs,ypos,profile,variance,tot,prof,var,a_fit,
   b_fit,save,pro_ok,keep,yexp,output)   
#--------------------------------------------------------------------------------------
#      SUB_PROFM -- Main subroutine for finding optimal extract weights using
#                   Tom Marsh's algorithm. (curved spectra)
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) iptr:       (pointer) Pointer to input file
#      (>) eptr:       (pointer) Pointer to input error file
#      (>) nx:         (int) X dimension of input data
#      (>) ny:         (int) Y dimension of input data
#      (>) ny1:        (int) Lower limit of requested spectral range
#      (>) ny2:        (int) Upper limit of requested spectral range
#      (>) nchunks:    (int) Number of blocks of given a blocking factor
#      (>) block_size: (int) Blocking factor in the spectral direction
#      (>) nord:       (int) Order of trace polynomial
#      (>) trace:      (real[nord+4]) Trace coefficients
#      (>) pol_dist:   (real) Space between polynomials
#      (>) npoly:      (int) Number of polynomials which will be fit
#      (>) pord:       (int) Order of polynomials to be used in profile fit
#      (>) pro_thr:    (real) Threshold used for deleting bad blocks from fit
#      (W) xs:         (real[nchunks]) Workspace
#      (W) ypos:       (real[nchunks]) Workspace
#      (W) profile:    (real[nx]) Workspace
#      (W) variance:   (real[nx]) Workspace
#      (W) tot:        (real[nx]) Workspace
#      (W) prof:       (real[nx,nchunks]) Workspace
#      (W) var:        (real[nx,nchunks]) Workspace
#      (W) a_fit:      (real[npoly*(pord+1)]) Workspace
#      (W) b_fit:      (real[npoly*(pord+1)]) Workspace
#      (W) save:       (real[nchunks,npoly]) Workspace
#      (W) pro_ok:     (bool[nx,nchunks]) Workspace
#      (W) keep:       (bool[nchunks]) Workspace
#      (W) yexp:       (real[nchunks,2*pord+1]) Workspace
#      (<) output:     (real[nx,ny]) Output profile
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        22-FEB-1993
#      UPDATE:      19-OCT-1993 -- Modified for speed (JRL)
#
#--------------------------------------------------------------------------------------

# Parameters

pointer iptr,eptr
int ny1,ny2,block_size,nx,nord,npoly,ny,pord,nchunks
real trace[nord+4],pro_thr,ypos[nchunks],xs[nchunks],profile[nx],tot[nx],variance[nx],
   bfit[npoly*(pord+1)],save[nchunks,npoly],a_fit[npoly*(pord+1),npoly*(pord+1)],
   prof[nx,nchunks],var[nx,nchunks],pol_dist,yexp[nchunks,2*pord+1],
   output[nx,ny]
bool pro_ok[nx,nchunks],keep[nchunks]

# Local variables

pointer ptr1,ptr2,wpt1,wpt2,wpt3
int idim,ixl,ixr,ix1,ix2,irow,j,iy1,iy2,jj,ii,ndel,iter,iq,k,n,
   jx1,jx2,i,ip,l,m,krank,ip1,ip2,index
real right,left,yscale,zero,pro_thr2,tau,yy,xshift,w1,w2,sum,xpol,sum2,
   xp,xpol_l,xpol_l1,xpol_k,xpol_k2,xl,xk,rnorm,xleft,x1,x2,wt,
   sumpre,resid

# Functions

real apolr(),sam_qint()
pointer imgl2r()

errchk imgl2r

begin

      # Initialise a few things

      left = 1.0
      right = real(nx)
      yscale = trace[nord+4]
      zero = apolr(trace[nord+2]*yscale,trace,nord+1)
      pro_thr2 = pro_thr**2
      idim = (pord + 1)*npoly
      tau = 1.0e-2
      call amovkr(0.0,output,nx*ny)

      # Spectra in temp file will be set up so that the trace will have
      # have been subtracted out in integer pixel amounts.  Therefore the
      # leftmost pixel you ever want to consider is the first one and the
      # rightmost pixel is the last one in a row

      ixl = 1
      ixr = nx
      ix1 = 2
      ix2 = nx - 1

      # Loop for block in the spectral direction and get a rough estimate
      # of the profile...

      do j = 1,nchunks {

         # First get the position of the block in both coords...

         iy1 = ny1 + block_size*(j-1)
         iy2 = iy1 + block_size - 1
         if (j == nchunks) iy2 = ny2
         yy = 0.5*real(iy1 + iy2)*yscale
         ypos[j] = yy
         xshift = apolr(yy,trace,nord+1) - zero
         xshift = xshift - real(nint(xshift))
         xs[j] = xshift

         # Initialise profile arrays

         call amovkr(0.0,profile,nx)
         call amovkr(0.0,tot,nx)
         call amovkr(0.0,variance,nx)

         # Now sum along the spacial direction

         do jj = iy1,iy2 {
            ptr1 = imgl2r(iptr,jj-ny1+1)
            ptr2 = imgl2r(eptr,jj-ny1+1)
            do ii = ixl,ixr {
               profile[ii] = profile[ii] + Memr[ptr1+ii-1]
               variance[ii] = variance[ii] + Memr[ptr2+ii-1]
               tot[ii] = tot[ii] + 1.0
            }
         }

         # Divide by the total # of contributions

         do ii = ixl,ixr {
            if (tot[ii] > 0.0) {
               profile[ii] = profile[ii]/tot[ii]
               variance[ii] = variance[ii]/(tot[ii]**2)
               pro_ok[ii,j] = true }
            else {
               profile[ii] = 0.0
               variance[ii] = 0.0
               pro_ok[ii,j] = false
            }
         }

         # How much was there total?

         w1 = 0.5 - xshift 
         w2 = xshift + 0.5
         sum = profile[ix1]*w1 + profile[ix2]*w2
         do i = ix1+1,ix2-1 {
            sum = sum + profile[i]
         }

         # Normalise all this...

         if (sum > 0.0) {
            do ii = ixl,ixr {
               if (pro_ok[ii,j]) {
                  prof[ii,j] = profile[ii]/sum
                  var[ii,j] = variance[ii]/sum**2
               }
            }
            keep[j] = true }
         else {
            do ii = ixl,ixr {
               pro_ok[ii,j] = false
            }
            keep[j] = false
         }
      } # End of the row loop

      # Set up a few things relating to the iteration scheme

      ndel = 1
      iter = 0

      # Get some workspace for the fitting routine 

      call malloc(wpt1,idim,TY_REAL)
      call malloc(wpt2,idim,TY_REAL)
      call malloc(wpt3,idim,TY_INT)

      # Set a look-up array for later use...

      do i = 0,2*pord {
         do j = 1,nchunks {
            yexp[j,i+1] = ypos[j]**real(i)
         }
      }

      # Loop until you haven't deleted any new points

      do while (ndel != 0) {
         ndel = 0
         iter = iter + 1

         # Now set up the RHS of Tom's equation A4

         do iq = 1,idim {
            k = (iq - 1)/(pord + 1) + 1
            n = iq - (pord + 1)*(k - 1)
            xpol = left + pol_dist*(real(k-1) + 0.5)
            sum2 = 0.0
            if (n == 1) {
               do j = 1,nchunks {
                  sum = 0.0
                  if (keep[j]) {
                     xp = xpol + xs[j]
                     jx1 = nint(xp - pol_dist)
                     jx1 = max(ixl,jx1)
                     jx2 = nint(xp + pol_dist)
                     jx2 = min(ixr,jx2)
                     do i = jx1,jx2 {
                        if (pro_ok[i,j]) {
                           sum = sum + prof[i,j]*sam_qint(i,xp,pol_dist)/var[i,j]
                        }
                     }
                     yy = yexp[j,n]
                     sum2 = sum2 + sum*yy
                     save[j,1] = sum
                  }
               }
            } else {
               do j = 1,nchunks {
                  if (keep[j]) {
                     yy = yexp[j,n]
                     sum2 = sum2 + save[j,1]*yy
                  }
               }
            }
            b_fit[iq] = sum2
         }

             
         # Now take a stab at the lhs matrix of Tom's equation A4.  Only
         # evaluate half of it as it is symmetric

         do ip = 1,idim {
            l = (ip - 1)/(pord + 1) + 1
            m = ip - (pord + 1)*(l - 1)
            xpol_l = left + pol_dist*(real(l-1) + 0.5)
            xpol_l1 = xpol_l - pol_dist
            do iq = 1,ip {
               k = (iq - 1)/(pord + 1) + 1
               n = iq - (pord + 1)*(k - 1)
               xpol_k = left + pol_dist*(real(k-1) + 0.5)
               xpol_k2 = xpol_k + pol_dist

               # Contribution will be zero unless both polynomials can be
               # affected by the same pixel.

               sum2 = 0.0
               if ((xpol_l1 - xpol_k2) < 1.0) {

                  # Evaluate the contributions only once for each configuration
                  # of the polynomials

                  if (m == 1 && n == 1) {
                     do j = 1,nchunks {
                        if (keep[j]) {
                           xshift = xs[j]
                           jx1 = nint(xpol_l1 + xshift)
                           jx2 = nint(xpol_k2 + xshift)
                           jx1 = max(1,jx1)
                           jx2 = min(nx,jx2)

                           # Forget pixels where there is no overlap

                           if (jx2 >= jx1) {
                              xl = xpol_l + xshift
                              xk = xpol_k + xshift
                              sum = 0.0
                              do i = jx1,jx2 {
                                 if (pro_ok[i,j]) {
                                    sum = sum + sam_qint(i,xl,pol_dist)*
                                    sam_qint(i,xk,pol_dist)/var[i,j]
                                 }
                              } }
                           else {
                              sum = 0.0
                           }
                           yy = yexp[j,m+n-1]
                           sum2 = sum2 + sum*yy
                           save[j,k] = sum
                        }
                     }
                  }
     
                  # Otherwise use stored information

                  else {
                     do j = 1,nchunks {
                        if (keep[j]) {
                           yy = yexp[j,m+n-1]
                           sum2 = sum2 + save[j,k]*yy
                        }
                     }
                  }
               }
               a_fit[iq,ip] = sum2
               if (ip != iq) a_fit[ip,iq] = sum2
            }
         }

         # Now solve matrix equation 

         call hfti(a_fit,idim,idim,idim,b_fit,1,1,tau,krank,rnorm,
            Memr[wpt1],Memr[wpt2],Memi[wpt3])

         # Ditch pixels with worst residuals

         call amovkr(0.0,profile,nx)
         do j = 1,nchunks {
            xshift = xs[j]
            xleft = left + xshift

            # Evaluate the fit...

            sumpre = 0.0
            do i = ix1,ix2 {
               x1 = real(i) - xleft - 0.5
               x2 = x1 + 1.0
               ip1 = int(x1/pol_dist + 0.5)
               ip1 = max(1,ip1)
               ip2 = int(x2/pol_dist + 1.5)
               ip2 = min(npoly,ip2)
               sum2 = 0.0
               do k = ip1,ip2 {
                  sum = 0.0
                  xp = xleft + pol_dist*(real(k-1) + 0.5)
                  do n = 1,pord+1 {
                     index = (pord + 1)*(k - 1) + n
                     sum = sum + b_fit[index]*yexp[j,n]
                  }
                  sum2 = sum2 + sum*sam_qint(i,xp,pol_dist)
               }
               sum2 = max(0.0,sum2)
               profile[i] = sum2
               if (i == ix1) {
                  wt = 0.5 - xshift }
               else if (i == ix2) {
                  wt = xshift + 0.5 }
               else {
                  wt = 1.0
               }
               sumpre = sumpre + wt*sum2
            }

            # Enforce normalisation

            do i = ix1,ix2 {
               if (sumpre != 0.0) profile[i] = profile[i]/sumpre               
            }

            # Now find which pixels to reject...

            do i = ix1,ix2 {
               if (pro_ok[i,j] && sumpre != 0.0) {
                  resid = (profile[i] - prof[i,j])**2/var[i,j]
                  if (resid > pro_thr2) {
                     ndel = ndel + 1
                     pro_ok[i,j] = false
                  } }
               else if (sumpre == 0.0) {
                  ndel = ndel + 1
                  pro_ok[i,j] = false
               }
            }
         }

         # News report to the user...

         call printf("Iteration %d, %d points deleted from fit \n")
            call pargi(iter)
            call pargi(ndel)
      } # End of iteration loop

      # Now evaluate the profile over the image for output

      irow = 0
      do j = ny1,ny2 {
         irow = irow + 1
         yy = real(j)*yscale
         xshift = apolr(yy,trace,nord+1) - zero
         xshift = xshift - real(nint(xshift))
         xleft = left + xshift
         sumpre = 0.0
         do i = ix1,ix2 {
            x1 = real(i) - xleft - 0.5
            x2 = x1 + 1.0
            ip1 = int(x1/pol_dist + 0.5)
            ip1 = max(1,ip1)
            ip2 = int(x2/pol_dist + 1.5)
            ip2 = min(npoly,ip2)
            sum2 = 0.0
            do k = ip1,ip2 {
               xp = xleft + pol_dist*(real(k-1) + 0.5)
               sum = 0.0
               do n = 1,pord+1 {
                  index = (pord+1)*(k-1) + n
                  sum = sum + b_fit[index]*yy**real(n-1)
               }
               sum2 = sum2 + sum*sam_qint(i,xp,pol_dist)
            }
            sum2 = max(0.0,sum2)
            profile[i] = sum2
            sumpre = sumpre + sum2
         }

         # Enforce normalisation

         do i = ix1,ix2 {
            if (sumpre != 0.0) {
               output[i,irow] = profile[i]/sumpre }
            else {
               output[i,irow] = 0.0
            }
         }
      }

      # Ditch the workspace

      call mfree(wpt1,TY_REAL)
      call mfree(wpt2,TY_REAL)
      call mfree(wpt3,TY_INT)

end

