include <imhdr.h>

procedure sub_sky(er_ex,toff,sky_ord,nreject,axis,
   sky1,sky2,ny1,ny2,workx,worky,workw,nw,coefs,optr,oeptr)
#--------------------------------------------------------------------------------------
#      SUB_SKY -- Main routine to sky subtract a frame
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) er_ex:      (bool) TRUE if an error array exists
#      (>) toff:       (short) Offset from input to section coords
#      (>) sky_ord:    (int) Order of polynomial to be fit to sky rows
#      (>) nreject:    (int) Maximum number of pixels to reject in sky row fit
#      (>) axis:       (int) Spectral axis
#      (>) sky1:       (real[2]) First sky band
#      (>) sky2:       (real[2]) Second sky band
#      (>) ny1:        (int) Lower limit of requested spectral range
#      (>) ny2:        (int) Upper limit of requested spectral range
#      (W) workx:      (real[nw]) Workspace
#      (W) worky:      (real[nw]) Workspace
#      (W) workw:      (real[nw]) Workspace
#      (>) nw:         (int) Size of workspaces
#      (W) coefs:      (real[sky_ord+1]) Workspace for polynomial coefs
#      (>) optr:       (pointer) Pointer to input/output file
#      (>) oeptr:      (pointer) Pointer to input/output variance file
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        22-FEB-1993
#      UPDATE:      08-NOV-1993 -- Now subtracts sky estimate in region from
#                                  lower limit of lh sky band to upper limit
#                                  of rh sky band. (JRL)
#
#--------------------------------------------------------------------------------------

# Parameters

int sky_ord,nreject,ny1,ny2,nw,axis
short toff
pointer optr,oeptr
real sky1[2],sky2[2],worky[nw],workx[nw],workw[nw],coefs[sky_ord+1]
bool er_ex

# Local variables

int nsterm,irow,j,nfit,i,ireject,k,l,n,lrem,ist,ifn,ioff,nrow,
   nnx,sk11,sk12,sk21,sk22
pointer wpt1,wpt2,wpt3,ptr1,ptr2,ptr3,ptr4
real chisqr,rmax,resid,sumw,avevar,varfit,skyval,err
bool ok

# Functions

real apolr()
pointer imgl2r(),impl2r()

errchk malloc,imgl2r,impl2r

begin
      
      # Grab some workspace

      nsterm = sky_ord + 1
      call malloc(wpt1,2*nsterm-1,TY_DOUBLE)
      call malloc(wpt2,nsterm,TY_DOUBLE)
      call malloc(wpt3,nsterm*nsterm,TY_DOUBLE)

      # Set up some constants...

      nnx = IM_LEN(optr,1)
      sk11 = nint(sky1[1]) - toff
      sk12 = nint(sky1[2]) - toff
      sk21 = nint(sky2[1]) - toff
      sk22 = nint(sky2[2]) - toff
      ist = sk11
      ifn = sk22

      # Loop through each row/column

      irow = 0
      nrow = ny2 - ny1 + 1
      do j = 1,nrow {
         irow = irow + 1
         nfit = 0

         # First get the input data

         ptr1 = imgl2r(optr,irow)
         if (er_ex) ptr2 = imgl2r(oeptr,irow)

         # Go through first sky band...
         
         do i = sk11,sk12 {
            ioff = i - 1
            if (er_ex) {
               err = Memr[ptr2+ioff] }
            else {
               err = 1.0
            }
            if (err != 0.0) {
               nfit = nfit + 1
               workx[nfit] = real(i)
               worky[nfit] = Memr[ptr1+ioff]
               workw[nfit] = 1.0/err
            }
         }

         # Now the second sky band...

         do i = sk21,sk22 {
            ioff = i - 1
            if (er_ex) {
               err = Memr[ptr2+ioff] }
            else {
               err = 1.0
            }
            if (err != 0.0) {
               nfit = nfit + 1
               workx[nfit] = real(i)
               worky[nfit] = Memr[ptr1+ioff]
               workw[nfit] = 1.0/err
            }
         }

         # If you haven't enough pixels to fit, then send out a message and
         # eventually this row will be flagged as bad by virtue of a zero
         # variance...

         ok = true
         if (nfit < nsterm) {
            ok = false
            call eprintf("Could not fit %s %d\n")
               if (axis == 1) {
                  call pargstr("column") }
               else {
                  call pargstr("row")
               }
               call pargi(j) }
               
         # Otherwise fit then reject as many pixels as you want/can
         
         else {
            ireject = min(nreject,nfit-nsterm)
            do k = 0,ireject {
               if (ok) {

                  # Do the fit. If it worked and this isn't the last iteration, 
                  # then flag the worst fitting point by giving it a zero weight.

                  ifnoerr (call sam_poly(workx,worky,workw,nfit,sky_ord+1,Memd[wpt1],
                  Memd[wpt2],Memd[wpt3],coefs,chisqr)) {
                     if (k != ireject) {
                        rmax = -1.0
                        do l = 1,nfit {
                           if (workw[l] != 0.0) {
                              resid = abs(worky[l] - apolr(workx[l],
                              coefs,nsterm))
                              if (resid > rmax) {
                                 rmax = resid
                                 lrem = l
                              }
                           }
                        }
                        workw[lrem] = 0.0
                     } 
                  }

                  # If the fit was unsuccessful, then flag the row as bad
   
                  else {
                     ok = false
                     break
                  }
               }
            } 
         }

         #  Get a row/column from the output file

         ptr3 = impl2r(optr,irow)
         call amovr(Memr[ptr1],Memr[ptr3],nnx)
         if (er_ex) {
            ptr4 = impl2r(oeptr,irow)
            call amovr(Memr[ptr2],Memr[ptr4],nnx)
         }

         # Now, if after all that you got a proper fit, subtract off the
         # sky.  

         if (ok) {
            if (er_ex) {

               # Attempt to add in a contribution from the sky subtraction
               # to the overall variances of the individual points.  See
               # Bevington p. 188.

               sumw = 0.0
               n = 0
               do i = 1,nfit {
                  if (workw[i] != 0.0) {
                     sumw = sumw + workw[i]
                     n = n + 1
                  }
               }
               if (n != 0) {
                  avevar = sumw/real(n**2)
                  varfit = chisqr/avevar }
               else {
                  varfit = 0.0
               }
            }
            do i = ist,ifn {
               ioff = i - 1
               skyval = apolr(real(i),coefs,nsterm)
               Memr[ptr3+ioff] = Memr[ptr3+ioff] - skyval
               if (er_ex) Memr[ptr4+ioff] = Memr[ptr4+ioff] + varfit
            }
         }

         # If you didn't get a good fit, then make the variance array all zeros.

         else {
            if (er_ex) {
               do i = ist,ifn {
                  ioff = i - 1
                  Memr[ptr4+ioff] = 0.0
               }
            }
         }
      
         # Send a friendly message...

         if (mod(irow,100) == 0) { 
            call printf("%d %s done out of %d \n")
               call pargi(irow)
               if (axis == 1) {
                  call pargstr("columns") }
               else {
                  call pargstr("rows")
               }
               call pargi(nrow)
         }     
      } # End of row/column loop

      # Ditch the workspace

      call mfree(wpt1,TY_DOUBLE)  
      call mfree(wpt2,TY_DOUBLE)  
      call mfree(wpt3,TY_DOUBLE)
end
