include <imhdr.h>
include <error.h>

procedure t_arc()
#------------------------------------------------------------------------------
#
#   -----
#   A R C
#   -----
#
#   Description
#   -----------
#   Transfer header information from object frames to their associated arc
#   frames.  This allows arcs to be extracted in the same way as the objects
#
#   
#   Scope of program
#   ----------------
#   - Dimensionality irrelevant
#   - Data type irrelevant
#   - Variance arrays irrelevant
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   basis      List of files which have header information to be transferred
#              (string) (prompted for)
#
#   output     List of files to which header information is to be transferred
#              (string) (prompted for)
#
#
#   Propogation of data structure
#   -----------------------------
#   Not relevant
#
#   
#   Method
#   ------
#   - The user specifies a list of images which have been processed and which have
#     the requisite positional information in the header.
#   - The position of the object, dekker file, trace file, etc information is
#     copied from the headers of these files to the headers of the output files.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgstr
#
#   Library ETC:
#      erract
#      error
#
#   Library FMTIO:
#      eprintf
#      pargi
#      sprintf
#   
#   Library IMIO:
#      imaddb
#      imaddi
#      imaddr
#      imastr
#      imgetb
#      imgeti
#      imgetr
#      imgstr
#      immap
#      imtclose
#      imtgetim
#      imtlen
#      imtopen
#      imunmap
#
#   Library MEMIO:
#      salloc
#      sfree
#      smark
#
#   Internal subroutines called
#   ---------------------------
#   None
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <error.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   22-FEB-93     - Original program based on VAX program SAM_ARC
#
#
#------------------------------------------------------------------------------

# Local variables

pointer    bname     # Pointer for basis file names
pointer    bptr      # Pointer for basis file headers
pointer    cdummy    # Pointer for dummy character string
pointer    cp1       # Pointer for dummy character string
int        i         # Integer counter
int        idum      # Integer dummy variable
pointer    imtlist1  # Pointer for basis file list
pointer    imtlist2  # Pointer for output file list
int        j         # Integer counter
int        list1     # Basis file list number
int        list2     # Output file list number
int        nobj      # Number of objects
pointer    oname     # Pointer for output file names
pointer    optr      # Pointer for output file headers
pointer    sp        # Stack pointer
bool       track     # TRUE if a trace is being used

# Functions

pointer immap()
int imgeti(),imtlen(),imtopen(),imtgetim()
real imgetr()
bool imgetb()

begin

      # Mark stack and get some workspace

      call smark(sp)
      call salloc(oname,SZ_FNAME,TY_CHAR)
      call salloc(bname,SZ_FNAME,TY_CHAR)
      call salloc(cdummy,SZ_FNAME,TY_CHAR)
      call salloc(cp1,SZ_LINE,TY_CHAR)
      call salloc(imtlist1,SZ_LINE,TY_CHAR)
      call salloc(imtlist2,SZ_LINE,TY_CHAR)

      # Get basis and output lists and check that they have the same size

      call clgstr("basis",Memc[imtlist1],SZ_FNAME)
      call clgstr("output",Memc[imtlist2],SZ_FNAME)

      # Open the template lists

      list1 = imtopen(Memc[imtlist1])
      list2 = imtopen(Memc[imtlist2])
      if (imtlen(list1) != imtlen(list2))
         call error(0,"Image lists don't have same number of components")

      # Loop for the files in the lists...

      while (imtgetim(list1,Memc[bname],SZ_FNAME) != EOF) {

         # Open the files...

         iferr {
            idum = imtgetim(list2,Memc[oname],SZ_FNAME)
            bptr = immap(Memc[bname],READ_ONLY,0)
            optr = immap(Memc[oname],READ_WRITE,0)
         } then {
            call eprintf("Error opening input files\n")
            call erract(EA_WARN)
         }

         # Get dispersion axis data

         call imaddi(optr,"axis",imgeti(bptr,"axis"))

         # Get trace file info

         track = imgetb(bptr,"track")
         call imaddb(optr,"track",track)
         if (track) {
            call imgstr(bptr,"trace",Memc[cdummy],SZ_FNAME)
            call imastr(optr,"trace",Memc[cdummy])
         }

         # Get dekker file info

         call imgstr(bptr,"dekker",Memc[cdummy],SZ_FNAME)
         call imastr(optr,"dekker",Memc[cdummy])

         # Get the profile file info

         ifnoerr (call imgstr(bptr,"pr_file",Memc[cdummy],SZ_FNAME)) {
            call imastr(optr,"pr_file",Memc[cdummy])
         }

         # Get the number of objects

         nobj = imgeti(bptr,"nobj")
         call imaddi(optr,"nobj",nobj)

         # Now loop for each object

         do i = 1,nobj {

            # Get info about spectrum postion and width

            call sprintf(Memc[cp1],SZ_LINE,"pos%d")
               call pargi(i)
            call imaddr(optr,Memc[cp1],imgetr(bptr,Memc[cp1]))
            do j = 1,2 {
               call sprintf(Memc[cp1],SZ_LINE,"swid%d_%d")
                  call pargi(i)
                  call pargi(j)
               call imaddr(optr,Memc[cp1],imgetr(bptr,Memc[cp1]))
            }

            # Get info about first sky band

            do j = 1,2 {
               call sprintf(Memc[cp1],SZ_LINE,"skya%d_%d")
                  call pargi(i)
                  call pargi(j)
               call imaddr(optr,Memc[cp1],imgetr(bptr,Memc[cp1]))
            }

            # Get info about second sky band

            do j = 1,2 {
               call sprintf(Memc[cp1],SZ_LINE,"skyb%d_%d")
                  call pargi(i)
                  call pargi(j)
               call imaddr(optr,Memc[cp1],imgetr(bptr,Memc[cp1]))
            }
         }

         # Close up files

         call imunmap(bptr)
         call imunmap(optr)

      } # End of template file list loop

      # Close up template files

      call imtclose(list1)
      call imtclose(list2)

      # Release workspace

      call sfree(sp)
end

