include <imhdr.h>
include <error.h>

define HELPFIL     "rgo$sam/curfile/dekhlp1.key"
define HELPFIL2    "rgo$sam/curfile/dekhlp2.key"
define PROMPT      "dekker: options for aperture definition"
define PROMPT2     "dekker: options for wavelength range definition" 

procedure t_dekker()
#------------------------------------------------------------------------------
#
#   -----------
#   D E K K E R
#   -----------
#
#   Description
#   -----------
#   Define the dekker limits and draw them on a display
#
#   
#   Scope of program
#   ----------------
#   - Only 2-d files acceptable
#   - Data type irrelevant
#   - Variance arrays irrelevant
#   - Batch execution not supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   nobj       Number of objects to be used (int)(prompted for)
#
#   axis       Dispersion axis (1=X, 2=Y) (int) (prompted for)
#
#   track      TRUE if a trace file exists for the displayed file (bool)
#              (prompted for)
#
#   trace      Name of file with trace coefficients (string)(prompted for)
#
#   dekker     Name of output file for dekker limits (string)(prompted for)
#
#   frame      Display frame where the image has been displayed (int) 
#              (hidden)
#
#   colour     Colour index to be used for overplot (int) (hidden)
#
#   imagecur   Parameter for use with cursor. (*imcur)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   - A one dimensional output file of type REAL is written with just the two 
#     limits of the dekker.
#
#   
#   Method
#   ------
#   - The user must first display the frame on the image display
#   - The user is then asked to indicate where on the image the dekker limits lie.
#   - The order of the limits is unimportant as they will be put into the correct
#     order anyway.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgetb
#      clgeti
#      clgstr
#
#   Library DS:
#      imd_mapframe
#      iw_close
#      iw_im2fb
#      iw_open
#
#   Library ETC:
#      erract
#      error
#      pagefile
#
#   Library FMTIO:
#      eprintf
#      printf
#   
#   Library IMIO:
#      imflush
#      immap
#      impl2r
#      imunmap
#
#   Library MEMIO:
#      salloc
#      sfree
#      smark
#
#   Library RTOOLS:
#      rt_crfile
#
#   Library SAMSUB:
#      sam_ctrace
#      sam_gtrace
#      sam_otrace
#
#   Library VOPS:
#      amovkr
#
#   Internal subroutines called
#   ---------------------------
#      sub_dekker
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <error.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   22-FEB-93     - Original program based on VAX program SAM_DEKKER
#   12-NOV-93     - Modified to allow for multiple spectra on a frame (JRL)
#
#
#------------------------------------------------------------------------------

# Local variables

int        axis        # Dispersion axis
int        colour      # Colour index for overplot
int        dims[2]     # Dimensions of displayed file
pointer    dp          # Display frame descriptor
pointer    dptr        # Pointer for dekker header
pointer    dname       # Pointer for output dekker file name
int        frame       # Frame number for display
pointer    iname       # Pointer for input file name
int        iobj        # Counter for number of object
pointer    iptr        # Pointer for displayed file
int        iw          # Display frame WCS descriptor
int        ndim        # Number of dimensions for output file
int        nobj        # Number of objects in trace file
int        nord        # Order of polynomial for trace
int        odims[2]    # Dimensions for output file
pointer    ptr1        # Trace pointer
pointer    ptr_de      # Dekker output pointer
bool       sametrace   # TRUE if all spectra use the same trace
pointer    sp          # Stack marker
pointer    tname       # Pointer for trace name
pointer    tptr        # Pointer for trace header
bool       track       # TRUE if a trace is being used
int        wcs_status  # Status variable

# Functions

bool clgetb()
int clgeti()
pointer imd_mapframe(),iw_open(),immap(),rt_crfile(),sam_gtrace(),impl2r()

# Define error exit point

define errexit_ 99

begin

      # Get stack workspace

      call smark(sp)
      call salloc(tname,SZ_FNAME,TY_CHAR)
      call salloc(dname,SZ_FNAME,TY_CHAR)
      call salloc(iname,SZ_FNAME,TY_CHAR)

      # Get the image display descriptor

      frame = clgeti("frame")
      colour = clgeti("colour")
      iferr {
         dp = imd_mapframe(frame,READ_WRITE,YES)
         iw = iw_open(dp,frame,Memc[iname],SZ_FNAME,wcs_status)
      } then {
         call eprintf("Error opening display window\n")
         call erract(EA_ERROR)
      }

      # Open the displayed file and get its dimensions

      iferr (iptr = immap(Memc[iname],READ_ONLY,0)) call erract(EA_ERROR)
      dims[1] = IM_LEN(iptr,1)
      dims[2] = IM_LEN(iptr,2)
      call imunmap(iptr)

      # Get some parameters

      axis = clgeti("axis")
      track = clgetb("track")
      nobj = clgeti("nobj")

      # If using a trace file, then open the trace file now.

      if (track) call clgstr("trace",Memc[tname],SZ_FNAME)
      iferr (call sam_otrace(track,tname,nobj,NULL,nord,tptr,sametrace)) {
         call eprintf("Error opening trace file\n")
         call erract(EA_ERROR)
      }

      # Open output file for dekker info...

      call clgstr("dekker",Memc[dname],SZ_FNAME)
      ndim = 2
      odims[1] = 4
      odims[2] = nobj
      iferr(dptr = rt_crfile(Memc[dname],NEW_FILE,0,ndim,odims,TY_REAL))
         call erract(EA_ERROR)

      # Now do the work. Loop for each object...

      call printf("Dekker limits for %d objects will be required\n")
      call pargi(nobj)
      do iobj = 1,nobj {

         # First a message...
     
         call printf("Mark limits for object #%d\n")
         call pargi(iobj)

         # Now get the trace for that object

         ptr1 = sam_gtrace(track,dims[axis],tptr,sametrace,iobj)

         # Get the appropriate row for output file

         ptr_de = impl2r(dptr,iobj)

         # Mark the limits

         iferr (call sub_dekker(dp,iw,colour,dims,axis,nord,Memr[ptr1],
            Memr[ptr_de])) {
            call eprintf("Error defining dekker positions\n")
            call erract(EA_ERROR)
         }
      }

      # Unmap files

      call imunmap(dptr)
      call sam_ctrace(track,tptr,ptr1)
      call imunmap(dp)
      call iw_close(iw)

      # Free workspace

      call sfree(sp)
end

procedure sub_dekker(dp,iw,colour,dims,axis,nord,trace,output)
#------------------------------------------------------------------------------
#      SUB_DEKKER -- Locate the limits of the data for subsequent analysis
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) dp:         (pointer) Pointer to the display frame
#      (>) iw:         (pointer) Pointer to the WCS of the display frame
#      (>) colour:     (int) Colour index for overplot
#      (>) dims:       (int[2]) Dimensions of displayed file
#      (>) axis:       (int) Dispersion axis
#      (>) nord:       (int) Order of polynomial for fit
#      (>) trace:      (real[nord+4]) Trace polynomial coefficients
#      (<) output:     (real[4]) Output dekker limits (spatial and spectral)
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        22-FEB-1993
#      UPDATE:      16-AUG-1994 -- Add in help in cursor mode. (JRL)
#
#------------------------------------------------------------------------------

# Parameters

pointer dp,iw
int colour,dims[2],axis,nord
real trace[nord+4],output[4]

# Local variables

pointer sp,cmd,ptr1
int nst,i,wcs,key,iy1,iy2,j,ii,jj,nx,ny
real x1[2],dum[2,2],y1[2],temp,yscale,zero,x2[2],xx,yy,fx,fy,dum1,dum2
bool quit,whole,tlim

# Functions

pointer imps2s()
int clgcur()
real apolr()

errchk imps2s,clgcur

begin

      # Get stack workspace

      call smark(sp)
      call salloc(cmd,SZ_LINE,TY_CHAR)

      # Indicate extremes of dekker

      quit = false
      call printf("Mark spatial limits of dekker with cursor\n")
      i = 0
      while (i < 2) {
         nst = clgcur("imagecur",dum1,dum2,wcs,key,Memc[cmd],SZ_LINE)
         switch (key) {
            case 'q':        # end point selection prematurely
               quit = true
               break
            case '?':        # help...
               call pagefile(HELPFIL,PROMPT)
            default:
               i = i + 1
               dum[1,i] = dum1
               dum[2,i] = dum2
         }
      }
      if (quit) call error(1,"Point selection ended prematurely")
      x1[1] = dum[3-axis,1]
      x1[2] = dum[3-axis,2]
      y1[1] = dum[axis,1]
      y1[2] = dum[axis,2]

      # Make sure they are in the right order

      if (x1[1] > x1[2]) {
         temp = x1[2]
         x1[2] = x1[1]
         x1[1] = temp
         temp = y1[2]
         y1[2] = y1[1]
         y1[1] = temp }
      else if (x1[1] == x1[2]) {
         call error(0,"Dekker of zero width??? Get outta here!!")
      }

      # Now indicate the spectral limits...

      whole = false
      tlim = false
      call printf("Indicate the spectral limits of the spectrum\n")
      call printf("(typing 'w' will give the whole spectral range)\n")
      call printf("(typing 't' will give the trace limits)\n")
      i = 0
      while (i < 2) {
         nst = clgcur("imagecur",dum1,dum2,wcs,key,Memc[cmd],SZ_LINE)
         switch (key) {
            case 'q':      # quit, end point selection prematurely
               quit = true
               break
            case 'w':      # use whole spectral range
               whole = true
               break
            case 't':      # use spectral range as defined by trace
               tlim = true
               break
            case '?':      # HELP!
               call pagefile(HELPFIL2,PROMPT2)
            default:
               i = i + 1
               dum[1,i] = dum1
               dum[2,i] = dum2
         }
      }
      if (quit) 
         call error(1,"Point selection ended prematurely") 
      else if (whole) {
         output[3] = 1.0
         output[4] = real(dims[axis]) }
      else if (tlim) {
         output[3] = trace[nord+2]
         output[4] = trace[nord+3] }
      else {
         output[3] = real(nint(dum[axis,1]))
         output[4] = real(nint(dum[axis,2]))
         call sam_band(output[3],output[4])
         output[3] = max(output[3],1.0)
         output[4] = min(real(dims[axis]),output[4])
      }

      # Get dekker limits at lowest value of the spectral axis to be considered.
      # Write them to the output file

      yscale = trace[nord+4]
      zero = apolr(output[3]*yscale,trace,nord+1)
      x2[1] = x1[1] - (apolr(y1[1]*yscale,trace,nord+1) - zero)
      x2[2] = x1[2] - (apolr(y1[2]*yscale,trace,nord+1) - zero)
      output[1] = max(real(nint(x2[1])),1.0)
      output[2] = min(real(nint(x2[2])),real(dims[3-axis]))

      # Plot a line on either side of the dekker...

      iy1 = nint(output[3])
      iy2 = nint(output[4])
      zero = apolr(trace[nord+2]*yscale,trace,nord+1)
      nx = IM_LEN(dp,1)
      ny = IM_LEN(dp,2)
      do j = 1,2 {
         do i = iy1,iy2 {
            yy = real[i]
            xx = apolr(yy*yscale,trace,nord+1) - zero + x2[j]
            if (axis == 1) {
               call iw_im2fb(iw,yy,xx,fy,fx)
               ii = nint(fy)
               jj = nint(fx) }
            else {
               call iw_im2fb(iw,xx,yy,fx,fy)
               ii = nint(fx)
               jj = nint(fy)
            }
            if (ii >= 1 && ii <= nx && jj >= 1 && jj <= ny) {
               ptr1 = imps2s(dp,ii,ii,jj,jj)
               Mems[ptr1] = colour
            }
         }
      }
      call imflush(dp)

      # Get rid of workspace

      call sfree(sp)
end
