include <imhdr.h>
include <fset.h>
include <error.h>

procedure t_extract()
#------------------------------------------------------------------------------
#
#   -------------
#   E X T R A C T
#   -------------
#
#   Description
#   -----------
#   Extract spectra from a 2-d frame.
#
#   
#   Scope of program
#   ----------------
#   - Handles 2-D images only
#   - All data types converted to REAL
#   - Variance arrays supported
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   input      Image with spectrum to be extracted (string)(prompted for)
#
#   output     Name of file for output spectrum (string)(prompted for)
#
#   optimal    TRUE if an optimal extraction is to be done. (bool)(hidden)
#
#   ext_thr    Threshold in sigma for bad pixel detection (real)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   - Output header is inherited from the input file
#   - Output data array is type REAL.  Each row has the same number of pixels
#     as there are in the requested spectral range, e.g. if the spectral range
#     requested earlier on was 100,901, then the output spectrum will have
#     800 pixels.
#   - Each object in the input data will be stored as a separate row in the
#     output file
#
#   
#   Method
#   ------
#   - A section of the input file is read in and organised so that it is
#     straightish and has the dispersion axis in the y direction
#   - If an optimal extraction has been requested, then a profile file is opened
#     and a similar section is read out of it.
#   - The optimal extraction is then iterated so that bad pixels are removed
#     one at a time, until no more are found.
#   - For a non-optimal extraction a simple extraction between two tramlines is
#     performed.
#   - If there is more than one object on the slit then this whole procedure is
#     repeated.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgetb
#      clgetr
#      clgstr
#
#   Library ERRPACK:
#      err_close
#      err_open
#
#   Library ETC:
#      erract
#      error
#
#   Library FIO:
#      fseti
#      fstati
#
#   Library FMTIO:
#      eprintf
#      pargi
#      pargstr
#      printf
#      sprintf
#   
#   Library IMIO:
#      imflush
#      imgetb
#      imgeti
#      imgetr
#      imgstr
#      immap
#      impl2r
#      imtclose
#      imtgetim
#      imtlen
#      imtopen
#      imunmap
#
#   Library MEMIO:
#      calloc
#      malloc
#      mfree
#      salloc
#      sfree
#      smark
#
#   Library RTOOLS:
#      rt_crfile
#
#   Library SAMSUB:
#      sam_aroff
#      sam_ctrace
#      sam_deltemp
#      sam_filex
#      sam_gdekker
#      sam_gtrace
#      sam_ifile
#      sam_odekker
#      sam_otrace
#      sam_rsect
#      sub_optx
#      sub_tramx
#
#   Library VOPS:
#      amovkr
#
#   Internal subroutines called
#   ---------------------------
#   None
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <fset.h>
#   include <error.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   22-FEB-93     - Original program based on VAX program SAM_EXTRACT
#   03-MAR-93     - Now processes an input file list (JRL)
#   12-NOV-93     - Minor modifications for new trace strategy.  Each
#                   object is now put in a separate file (JRL)
#
#
#------------------------------------------------------------------------------


# Local variables

int         axis        # Dispersion axis
pointer     cp1         # Dummy character pointer
pointer     cp2         # Dummy character pointer
pointer     dname       # Pointer for dekker file name
pointer     dptr        # Poineter for dekker file header
pointer     ename       # Pointer for error file name
pointer     enptr       # Pointer for input error section file
pointer     eptr        # Pointer for input error file header
bool        er_ex       # TRUE if error file exists
real        ext_thr     # Threshold for rejection of bad pixels
pointer     ilist       # Pointer for input file list
pointer     iname       # Pointer for input file name
pointer     inptr       # Pointer for input section file
int         iobj        # Counter for spectrum number
pointer     iptr        # Pointer for input file header
int         j           # Dummy integer counter
int         listin      # Input file list descriptor
int         listout     # Output file list descriptor
int         nnx         # Number of spacial pixels in section file
int         nny         # Number of spectral pixels in section file
int         nobj        # Number of objects on the slit
int         nord        # Order of polynomial for trace
int         nt          # Number of elements in offset array
int         nx          # X dimension of input file
int         nx1         # Lower limit of seeing disc
int         nx2         # Upper limit of seeing disc
int         ny          # Y dimension of input file
int         ny1         # Lower limit of requested spectral range
int         ny2         # Upper limit of requested spectral range
int         odims       # Dimensions of output file
pointer     oeptr       # Pointer for output error file header
pointer     olist       # Pointer for output file list
pointer     oname       # Pointer for output file name
bool        opt         # TRUE if extraction is requested to be optimal
bool        optimal     # TRUE if extraction actually is optimal
pointer     optr        # Pointer for output file header
pointer     pname       # Pointer for profile file name
pointer     pnptr       # Pointer for profile section file
pointer     pptr        # Pointer for profile file header
pointer     ptr_de      # Pointer for dekker info
pointer     ptr_out     # Pointer for output data array
pointer     ptr_outerr  # Pointer for output variance array
pointer     ptr_tr      # Pointer for trace
bool        sametrace   # TRUE if all spectra use the same trace
pointer     sp          # Stack marker
pointer     tname       # Pointer for trace file name
pointer     toff        # Pointer for offset array
pointer     tptr        # Pointer for trace file header
bool        track       # TRUE if a trace file exists
real        width[2]    # Endpoints of seeing disc
pointer     wpt1        # Pointer for workspace

# Functions

int fstati(),imgeti(),imtopen(),imtgetim(),imtlen(),sam_filex()
pointer immap(),sam_rsect(),impl1r(),err_open(),rt_crfile(),sam_gtrace(),
   sam_gdekker()
bool clgetb(),imgetb()
real imgetr(),clgetr()

begin

      # Mark stack and get some workspace...

      call smark(sp)
      call salloc(iname,SZ_FNAME,TY_CHAR)
      call salloc(ename,SZ_FNAME,TY_CHAR)
      call salloc(pname,SZ_FNAME,TY_CHAR)
      call salloc(oname,SZ_FNAME,TY_CHAR)
      call salloc(dname,SZ_FNAME,TY_CHAR)
      call salloc(tname,SZ_FNAME,TY_CHAR)
      call salloc(cp1,SZ_LINE,TY_CHAR)
      call salloc(cp2,SZ_FNAME,TY_CHAR)
      call salloc(ilist,SZ_LINE,TY_CHAR)
      call salloc(olist,SZ_LINE,TY_CHAR)

      # Flush buffered output immediately

      if (fstati(STDOUT,F_REDIR) == NO) call fseti(STDOUT,F_FLUSHNL,YES)

      # Get the input file name list.

      call clgstr("input",Memc[ilist],SZ_LINE)

      # Do you want optimal extraction? If so, get a threshold for zapping bad
      # pixels.

      opt = clgetb("optimal")
      if (opt) ext_thr = clgetr("ext_thr")

      # Get output file name list

      call clgstr("output",Memc[olist],SZ_LINE)

      # Open the input and output file lists

      listin = imtopen(Memc[ilist])
      listout = imtopen(Memc[olist])
      if (imtlen(listin) != imtlen(listout)) {
         call imtclose(listin)
         call imtclose(listout)
         call error(0,"Number of input/output images not the same")
      }

      # Check that none of the output files already exists...

      if (sam_filex(listout) > 0) call error(0,
         "These output files already exist and cannot be clobbered")         

      # Now loop for the input images...

      while (imtgetim(listin,Memc[iname],SZ_FNAME) != EOF && 
         imtgetim(listout,Memc[oname],SZ_FNAME) != EOF) {

         # Send a friendly message to the user...

         call printf("Beginning work on file %s \n")
            call pargstr(Memc[iname])

         # Open the input file name check its dimensions

         iferr (call sam_ifile(Memc[iname],iptr,nx,ny,Memc[ename],eptr,er_ex)) {
            call eprintf("Error opening input file\n")
            call erract(EA_WARN)
            next
         }

         # Do you want an optimal extraction? If there are no errors in the input
         # file, then an optimal extraction cannot be performed... 

         if (!er_ex && opt) {
            call printf("No error file exists -- a tramline extraction will be done\n") 
            optimal = false }
         else {
            optimal = opt
         }

         # If you are doing an optimal extraction, then get the name of the profile
         # file and check that its dimensions are the same as the input file's

         if (optimal) {
            call imgstr(iptr,"pr_file",Memc[pname],SZ_FNAME)
            iferr (pptr = immap(Memc[pname],READ_ONLY,0)) {
               call eprintf("Error opening profile file\n")
               call erract(EA_WARN)
               next
            }
            if (IM_LEN(iptr,1) != IM_LEN(pptr,1) || IM_LEN(iptr,2) != IM_LEN(pptr,2)) {
               call eprintf("Input and profile files have different sizes!\n") 
               next  
            }     
         }

         # Get some information which you'll need from the header

         axis = imgeti(iptr,"axis")
         nobj = imgeti(iptr,"nobj")

         # Open dekker file...

         call imgstr(iptr,"dekker",Memc[dname],SZ_FNAME)
         iferr(call sam_odekker(dname,nobj,dptr)) {
            call eprintf("Error opening dekker file\n")
            call erract(EA_WARN)
            next
         }

         # If using a trace file, then open it now.

         track = imgetb(iptr,"track")
         if (track) call imgstr(iptr,"trace",Memc[tname],SZ_FNAME)
         iferr (call sam_otrace(track,tname,nobj,dptr,nord,tptr,sametrace)) {
            call eprintf("Error opening trace file\n")
            call erract(EA_WARN)
            next
         }

         # Loop for the objects now...

         do iobj = 1,nobj {

            # Friendly message to the user...

            call printf("Starting work on object %d\n")
               call pargi(iobj)

            # First read the position of each object...

            do j = 1,2 {
               call sprintf(Memc[cp1],SZ_LINE,"swid%d_%d")
                  call pargi(iobj)
                  call pargi(j)
               width[j] = imgetr(iptr,Memc[cp1])
            }
            nx1 = nint(width[1])
            nx2 = nint(width[2])

            # Get the dekker info for this object

            ptr_de = sam_gdekker(dptr,iobj)
            ny1 = nint(Memr[ptr_de+2])
            ny2 = nint(Memr[ptr_de+3])

            # Get the trace for this object

            ptr_tr = sam_gtrace(track,ny2,tptr,sametrace,iobj)

            # Get offset array workspace

            call sam_aroff(track,ny1,ny2,nt,toff)

            # Open desired section of input file and its error file

            nnx = nx2 - nx1 + 1
            nny = ny2 - ny1 + 1
            iferr {
               inptr = sam_rsect(Memc[iname],nx1,nx2,ny1,ny2,track,Memr[ptr_tr],nord,
                  axis,nt,Mems[toff])
               if (er_ex) enptr = sam_rsect(Memc[ename],nx1,nx2,ny1,ny2,track,
                  Memr[ptr_tr],nord,axis,nt,Mems[toff])
            } then {
               call eprintf("Error opening input sections file\n")
               call erract(EA_ERROR)
            }

            # Open output file...

            if (nobj > 1) {
               call sprintf(Memc[cp2],SZ_LINE,"%s_%03d")
                  call pargstr(Memc[oname])
                  call pargi(iobj) }
            else {
                call strcpy(Memc[oname],Memc[cp2],SZ_FNAME)
            }
            odims = nny
            iferr(optr = rt_crfile(Memc[cp2],NEW_COPY,iptr,1,odims,TY_REAL)) {
               call eprintf("Error opening output file\n")
               call erract(EA_WARN)
               next
            }

           # Open an output error file, if we really want one...

            if (er_ex) {
               iferr (oeptr = err_open(optr,NEW_FILE,false)) {
                  call eprintf("Error opening output error file\n")
                  call erract(EA_WARN)
                  next
               }
            }

            # Right, if this is an optimal extraction then prepare for that now...

            if (optimal) {

               # Open the desired section of the profile file...

               iferr (pnptr = sam_rsect(Memc[pname],nx1,nx2,ny1,ny2,track,
                  Memr[ptr_tr],nord,axis,nt,Mems[toff])) {
                  call eprintf("Error opening profile sections file\n")
                  call erract(EA_ERROR)
               }
 
               # First get a bit of workspace

               call calloc(wpt1,nnx*nny,TY_SHORT)

               # Open the line of the output file...

               ptr_out = impl1r(optr)
               ptr_outerr = impl1r(oeptr)

               # Now do the extraction...

               iferr (call sub_optx(inptr,enptr,nnx,nny,axis,nt,Mems[toff],
                  ext_thr,pnptr,Mems[wpt1],Memr[ptr_out],Memr[ptr_outerr])) {
                  call eprintf("Error extracting object %d \n")
                     call pargi(iobj)
                  call erract(EA_WARN)
               }

               # Delete the workspace...

               call mfree(wpt1,TY_SHORT)

               # Unmap and delete the temporary profile file...

               iferr (call sam_deltemp(pnptr)) {
                  call eprintf("Error deleting temporary file\n")
                  call erract(EA_WARN)
               }
            }

            # If this isn't an optimal extraction, then prepare for the tramline

            else {

               # Open rows of output files. If an error array isn't desired, 
               # then get a bit of workspace to fill in the gaps...

               ptr_out = impl1r(optr)
               if (er_ex) {
                  ptr_outerr = impl1r(oeptr) }
               else {
                  call malloc(ptr_outerr,nny,TY_REAL)
               }

               # Do the extraction

               iferr (call sub_tramx(inptr,enptr,er_ex,nnx,nny,Memr[ptr_out],
                  Memr[ptr_outerr])) {
                  call eprintf("Error extracting object %d \n")
                     call pargi(iobj)
                  call erract(EA_WARN)
               }

               # If you opened a workspace, delete it now...

               if (!er_ex) call mfree(ptr_outerr,TY_REAL)
            }

            # Delete temporary files...

            iferr {
               call sam_deltemp(inptr)
               if (er_ex) call sam_deltemp(enptr)
            } then {
               call eprintf("Error deleting temporary file\n")
               call erract(EA_WARN)
            }

            # Ditch offset array 

            call mfree(toff,TY_SHORT)

         } # End of object loop

         # Close up the remaining files...

         if (er_ex) call err_close(oeptr)
         call imunmap(optr)
         if (er_ex) call err_close(eptr)
         call imunmap(iptr)
         if (optimal) call imunmap(pptr)
         call sam_ctrace(track,tptr,ptr_tr)
         call mfree(toff,TY_SHORT)
      
      } # End of template file loop

      # Close up file lists

      call imtclose(listin)
      call imtclose(listout)

      # Ditch remaining workspace

      call sfree(sp)
end
