include <imhdr.h>
include <error.h>
include <fset.h>

procedure t_find()
#------------------------------------------------------------------------------
#
#   -------
#   F I N D
#   -------
#
#   Description
#   -----------
#   Locate spectra on a 2-d frame.
#
#   
#   Scope of program
#   ----------------
#   - Handles 2-D images only
#   - All data types converted to REAL
#   - Variance arrays irrelevant
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   input      Image with spectra to be located (string)(prompted for)
#
#   autofind   TRUE if the spectra are to be located automatically rather than
#              interactively with a cursor. (bool)(prompted for)
#
#   axis       Axis to which spectra are parallel -- i.e. dispersion axis.
#              1 = X, 2 = Y. (int)(hidden)
#
#   track      TRUE if a trace file is being used. (bool)(hidden)
#
#   trace      Name of the trace file if required. (string)(hidden)
#
#   dekker     Name of the dekker file. (string)(hidden)
#
#   nom_width  Nominal width of the seeing disc.  This is just a first guess
#              to give the auto-finding routine something to go on. (int)
#              (hidden)
#
#   ansig      Once the sigma of the profile has been found, this is the
#              number of sigma from the centre out to which the spectrum
#              will be spatially delimited. In other words, 
#              limit = x0 +/- ansig*sigma
#
#   max_sep    Maximum separation between the last spectrum pixel and the
#              last pixel in a sky band.  This is to guard against sky pixels
#              which are too far away from the spectrum to be of any value.
#              Required by autofinding routine. (int)(hidden)
#
#   graphcur   Parameter for use in cursor input in the interactive finding
#              routine. (*gcur)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   Not relevant
#
#   
#   Method
#   ------
#   - The data between the dekker endpoints if summed in the spectral direction
#     following the trace if necessary
#   - If the user chooses the interactive algorithm, then the summed data is 
#     displayed as a line plot.  The user is then asked to indicate the 
#     location of the spectrum, the limits of the spectrum (the limits of 
#     the seeing disc) and two sky bands -- one on either side.
#   - If the user chooses the auto-finding algorithm, then the summed data is
#     first gaussian smoothed and searched for the greatest local maximum.  
#     The algorithm then searches either side of the maximum for locations 
#     where the second derivatives are zero.  This defines 1 sigma.
#     Sky bands are chosen so that they are local to the spectrum in question.
#   - Information on the location of the spectrum and sky bands are written 
#     to the input file header.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgetb
#      clgeti
#      clgetr
#      clgstr
#
#   Library ERRPACK
#      err_close
#
#   Library ETC:
#      erract
#
#   Library FIO:
#      fseti
#      fstati
#
#   Library FMTIO:
#      eprintf
#      pargi
#      pargstr
#      printf
#      sprintf
#   
#   Library IMIO:
#      imaddb
#      imaddi
#      imaddr
#      imastr
#      imtclose
#      imtgetim
#      imtopen
#      imunmap
#
#   Library MEMIO:
#      calloc
#      malloc
#      mfree
#      salloc
#      sfree
#      smark
#
#   Library SAMSUB:
#      sam_aroff
#      sam_ctrace
#      sam_deltemp
#      sam_gdekker
#      sam_gtrace
#      sam_ifile
#      sam_odekker
#      sam_otrace
#      sam_rsect
#      sub_finda
#      sub_findi
#
#   Library VOPS:
#      amovr
#
#
#   Internal subroutines called
#   ---------------------------
#   None
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <error.h>
#   include <fset.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   23-FEB-93     - Original program based on VAX program SAM_FIND
#   02-MAR-93     - Now processes an input file list (JRL)
#   12-NOV-93     - Now processes objects in a loop.  This was added in order
#                   to allow for multi-object spectroscopy (JRL)
#   15-NOV-93     - Modified slightly to include new autofinding routine (JRL)
#
#
#------------------------------------------------------------------------------

# Local variables

real         ansig       # Number of sigma to delimit spectrum
bool         auto        # TRUE if using auto-finding routine
int          axis        # Dispersion axis
pointer      cp1         # Pointer for dummy string
real         dek[2]      # Array for holding current dekker limits
pointer      dname       # Pointer for dekker file
pointer      dptr        # Pointer for dekker file header
pointer      ename       # Pointer for error file name
pointer      eptr        # Pointer for error file header
bool         er_ex       # TRUE if an error file exists for input file
pointer      ilist       # Pointer to input file list
pointer      iname       # Pointer for input file name
pointer      inptr       # Pointer for input section file
int          iobj        # Number of objects actually found
pointer      iptr        # Pointer for input file header
int          j           # Integer counter
int          listin      # Input file list descriptor
int          max_sep     # Maximum separation between object pixels and sky band
int          nnx         # X dimension of section file
int          nny         # Y dimension of section file
int          nobj        # Maximum number of objects on a slit
int          nom_width   # Nominal width of the seeing disc
int          nord        # Order of trace polynomial
int          nprof       # Number of elements for workspace allocation
int          nt          # Number of elements for offset array
int          nx          # X dimension of input data
int          nx1         # Lower spacial limit for section file
int          nx2         # Upper spacial limit for section file
int          ny          # Y dimension of input data
int          ny1         # Lower limit of requested spectral range
int          ny2         # Upper limit of requested spectral range
real         pos         # Position of the spectrum
pointer      ptr_de      # Pointer for dekker file data
pointer      ptr_tr      # Pointer for trace data
bool         sametrace   # TRUE if all the spectra use the same trace
int          sky1[2]     # Limits of first sky band
int          sky2[2]     # Limits of second sky band
pointer      sp          # Stack marker    
real         spec_rng[2] # Range of spectral pixels to be considered
pointer      tname       # Pointer for trace file
pointer      toff        # Pointer for offset array
pointer      tptr        # Pointer for trace file header
bool         track       # TRUE if using a trace file
int          width[2]    # Location of limits of seeing disc
pointer      wpt1        # Workspace pointer
pointer      wpt2        # Workspace pointer
pointer      wpt3        # Workspace pointer

# Functions

pointer sam_rsect(),sam_gtrace(),sam_gdekker()
int clgeti(),fstati(),imtopen(),imtgetim()
real clgetr()
bool clgetb()

begin

      # Get stack workspace

      call smark(sp)
      call salloc(iname,SZ_FNAME,TY_CHAR)
      call salloc(ename,SZ_FNAME,TY_CHAR)
      call salloc(tname,SZ_FNAME,TY_CHAR)
      call salloc(dname,SZ_FNAME,TY_CHAR)
      call salloc(cp1,SZ_LINE,TY_CHAR)
      call salloc(ilist,SZ_LINE,TY_CHAR)

      # Flush buffered output immediately

      if (fstati(STDOUT,F_REDIR) == NO) call fseti(STDOUT,F_FLUSHNL,YES)

      # Get the input file name.

      call clgstr("input",Memc[ilist],SZ_LINE)

      # Get a few more parameters

      axis = clgeti("axis")

      # Parameters for use by automatic finding routine

      auto = clgetb("autofind")
      if (auto) {
         nom_width = clgeti("nom_width")
         ansig = clgetr("ansig")
         max_sep = clgeti("max_sep")
      }

      # Open a dekker file

      call clgstr("dekker",Memc[dname],SZ_FNAME)
      iferr(call sam_odekker(dname,nobj,dptr)) {
         call eprintf("Error opening dekker file\n")
         call erract(EA_ERROR)
      }

      # If using a trace file, then open it now.

      track = clgetb("track")
      if (track) call clgstr("trace",Memc[tname],SZ_FNAME)
      iferr (call sam_otrace(track,tname,nobj,dptr,nord,tptr,sametrace)) {
         call eprintf("Error opening trace file\n")
         call erract(EA_ERROR)
      }

      # Open input file list

      listin = imtopen(Memc[ilist])

      # Now loop for the input images...

      while (imtgetim(listin,Memc[iname],SZ_FNAME) != EOF) {

         # Send a friendly message to the user...

         call printf("Beginning work on file %s \n")
            call pargstr(Memc[iname])

         # Open the input file name check its dimensions

         iferr (call sam_ifile(Memc[iname],iptr,nx,ny,Memc[ename],eptr,er_ex)){
            call eprintf("Error opening input file\n")
            call erract(EA_WARN)
            next
         }

         # Loop for all the objects now...

         do iobj = 1,nobj {

            # Message to the user...

            if (nobj > 1) {
               call printf("Working on object %d\n")
               call pargi(iobj)
            }

            # Get the limits for this object from the dekker file...

            ptr_de = sam_gdekker(dptr,iobj)
            call amovr(Memr[ptr_de],dek,2)
            nx1 = nint(dek[1])
            nx2 = nint(dek[2])
            call amovr(Memr[ptr_de+2],spec_rng,2)
            ny1 = nint(spec_rng[1])
            ny2 = nint(spec_rng[2])

            # Check that the requested spectral range is not too big for the file
  
            if (nint(spec_rng[2]) > IM_LEN(iptr,axis)) {
               call eprintf("Requested spectral range too big for: %s\n")
                  call pargstr(Memc[iname])
               next
            }

            # Get the trace you want

            ptr_tr = sam_gtrace(track,ny2,tptr,sametrace,iobj)

            # Get the requisite workspace for offset array

            call sam_aroff(track,ny1,ny2,nt,toff)

            # Open the section files...

            nnx = nx2 - nx1 + 1
            nny = ny2 - ny1 + 1
            iferr (inptr = sam_rsect(Memc[iname],nx1,nx2,ny1,ny2,track,
               Memr[ptr_tr],nord,axis,nt,Mems[toff])) {
               call eprintf("Error opening section input file\n")
               call erract(EA_ERROR)
            }

            # Grab a bit of workspace

            iferr {
               nprof = nnx
               call calloc(wpt1,nprof,TY_REAL)
               if (auto) {
                  call malloc(wpt3,nprof,TY_SHORT) }
               else {
                  call calloc(wpt2,nprof,TY_REAL)
               }
            } then {
               call eprintf("Could not get the workspace needed\n")
               call erract(EA_ERROR)
            }

            # If not doing autofind, then run the interactive finder...

            if (! auto) {
               iferr (call sub_findi(inptr,dek,Memr[ptr_tr],nord,spec_rng,
               nprof,Memr[wpt1],Memr[wpt2],pos,width,sky1,sky2))
               call erract(EA_WARN) }

            # Otherwise run the auto finder...

            else {
               iferr (call sub_finda(inptr,dek,Memr[ptr_tr],nord,
               spec_rng,nom_width,ansig,max_sep,nprof,Memr[wpt1],
               Mems[wpt3],pos,width,sky1,sky2)) call erract(EA_WARN)
            }
      
            # Ditch temporary file now...
      
            iferr (call sam_deltemp(inptr)) call erract(EA_WARN)

            # Write some results into the header of the file

            iferr {

               # First the position

               call sprintf(Memc[cp1],SZ_LINE,"pos%d")
                  call pargi(iobj)
               call imaddr(iptr,Memc[cp1],pos)

               # The width of the seeing disc

               do j = 1,2 {
                  call sprintf(Memc[cp1],SZ_LINE,"swid%d_%d")
                     call pargi(iobj)
                     call pargi(j)
                  call imaddi(iptr,Memc[cp1],width[j])
               }

               # The first sky band

               do j = 1,2 {
                  call sprintf(Memc[cp1],SZ_LINE,"skya%d_%d")
                     call pargi(iobj)
                     call pargi(j)
                  call imaddi(iptr,Memc[cp1],sky1[j])
               }

               # The second sky band

               do j = 1,2 {
                  call sprintf(Memc[cp1],SZ_LINE,"skyb%d_%d")
                     call pargi(iobj)
                     call pargi(j)
                  call imaddi(iptr,Memc[cp1],sky2[j])
               }
            } then {
               call eprintf("Problems encountered writing results to input file")
               call eprintf(" header.  Better check this before proceding\n")
               call erract(EA_WARN)
            }

            # Ditch some workspace

            call mfree(wpt1,TY_REAL)
            if (auto) {
               call mfree(wpt3,TY_SHORT) }
            else {
               call mfree(wpt2,TY_REAL)
            }
            call mfree(toff,TY_SHORT)


         # End of object loop

         }

         # Add a few more things into the header...
         
         iferr {
            call imaddi(iptr,"nobj",nobj)
            call imaddi(iptr,"axis",axis)
            call imaddb(iptr,"track",track)
            if (track) {
              call imastr(iptr,"trace",Memc[tname])
            }
            call imastr(iptr,"dekker",Memc[dname])
         } then {
            call eprintf("Problems encountered writing results to input file")
            call eprintf(" header.  Better check this before proceding\n")
            call erract(EA_WARN)
         }
         
         # Close the file

         if (er_ex) call err_close(eptr)
         call imunmap(iptr)
      }

      # Close the remaining files

      call imunmap(dptr)
      iferr (call sam_ctrace(track,tptr,ptr_tr)) call erract(EA_WARN)
      call imtclose(listin)

      # Free workspace

      call sfree(sp)
end
