include <imhdr.h>
include <error.h>
include <fset.h>

procedure t_mkerr()
#------------------------------------------------------------------------------
#
#   ---------
#   M K E R R
#   ---------
#
#   Description
#   -----------
#   Make an error file for a file which needs one.
#
#   
#   Scope of program
#   ----------------
#   - Handles 2-D images only
#   - All data types supported, but REAL arithmetic used
#   - Variance arrays irrelevant
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   input      List of images to be processed (string)(prompted for)
#
#   readout    Readout noise of the chip in electrons (real)(prompted for)
#
#   gain       Gain of the chip in electrons per adu (real)(prompted for)
#
#   Propogation of data structure
#   -----------------------------
#   Not relevant
#
#   
#   Method
#   ------
#   - If the error file already exists, then the input is skipped.
#   - If an error file is required then the simple 
#        variance[i] = (readout/gain)**2 + data[i]/gain
#     model is assumed, where the varaince is expressed in ADUs.
#   - The output file has the same name as the input file, with the suffix
#     ERR appended.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgetr
#      clgstr
#
#   Library ERRPACK
#      err_open
#      err_close
#
#   Library ETC:
#      erract
#
#   Library FIO:
#      fseti
#      fstati
#
#   Library FMTIO:
#      eprintf
#      pargstr
#      printf
#   
#   Library IMIO:
#      imgnlr
#      impnlr
#      imtclose
#      imtgetim
#      imtopen
#      imunmap
#
#   Library MEMIO:
#      salloc
#      sfree
#      smark
#
#   Library SAMSUB:
#      sam_ifile
#
#   Library VOPS:
#      altmr
#      amovkl
#
#
#   Internal subroutines called
#   ---------------------------
#   None
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <error.h>
#   include <fset.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   15-AUG-94     - Original program 
#   04-MAR-96     - Fixed bug which allowed negative variances (JRL)
#
#------------------------------------------------------------------------------

# Local variables

pointer    ename              # Name for output error file
pointer    eptr               # Pointer for error file header
bool       er_ex              # TRUE if an error file exists
real       gain               # Gain of the chip used for observations
real       gain2              # Inverse of gain of the chip 
pointer    ilist              # Pointer for input file list
pointer    iname              # Pointer for individual input file
pointer    iptr               # Pointer for input file header
int        junk               # Dummy integer variable
int        listin             # Descriptor for input file list
int        nx                 # X dimension of data frame
int        ny                 # Y dimension of data frame
long       one_l              # Long integer value of 1
pointer    ptr1               # Pointer for input data
pointer    ptr2               # Pointer for output data
real       read               # Readout noise of the chip
real       read2              # Readout variance of the chip
pointer    sp                 # Pointer for stack memory
long       v1[IM_MAXDIM]      # Indices for input data lines
long       v2[IM_MAXDIM]      # Indices for output data lines
pointer    wptr               # Workspace pointer

# Functions

real clgetr()
int imtopen(),imtgetim(),imgnlr(),impnlr(),fstati()
pointer err_open()

begin

      # Get stack workspace

      call smark(sp)
      call salloc(iname,SZ_FNAME,TY_CHAR)
      call salloc(ename,SZ_FNAME,TY_CHAR)
      call salloc(ilist,SZ_LINE,TY_CHAR)

      # Flush buffered output immediately

      if (fstati(STDOUT,F_REDIR) == NO) call fseti(STDOUT,F_FLUSHNL,YES)

      # Get the input file name.

      call clgstr("input",Memc[ilist],SZ_LINE)

      # Get readout and gain

      read = clgetr("readout")
      gain = clgetr("gain")
      gain2 = 1.0/gain
      read2 = read*read*gain2*gain2

      # Initialise a variable

      one_l = 1

      # Open input file list

      listin = imtopen(Memc[ilist])

      # Now loop for the input images...

      while (imtgetim(listin,Memc[iname],SZ_FNAME) != EOF) {

         # Send a friendly message to the user...

         call printf("Beginning work on file %s \n")
            call pargstr(Memc[iname])

         # Open the input file name check its dimensions

         iferr (call sam_ifile(Memc[iname],iptr,nx,ny,Memc[ename],eptr,er_ex)){
            call eprintf("Error opening input file\n")
            call erract(EA_WARN)
            next
         }

         # If an error file already exists for this frame, then issue a message
         # and go on to the next frame

         if (er_ex) {
            call eprintf("Error file already exists for this file -- skipping")
            call erract(EA_WARN)
            next
         }

         # If it doesn't exist, then create it

         eptr = err_open(iptr,NEW_FILE,false)

         # Get some workspace

         call malloc(wptr,nx,TY_REAL)

         # Now fill it with the correct values...

         call amovkl(one_l,v1,IM_MAXDIM)
         call amovkl(one_l,v2,IM_MAXDIM)
         while (imgnlr(iptr,ptr1,v1) != EOF) {
            junk = impnlr(eptr,ptr2,v2)
            call aabsr(Memr[ptr1],Memr[wptr],nx)
            call altmr(Memr[wptr],Memr[ptr2],nx,gain2,read2)
         }

         # Close things up...

         call mfree(wptr,TY_REAL)
         call err_close(eptr)
         call imunmap(iptr)
      }

      # Tidy and exit...

      call imtclose(listin)
      call sfree(sp)
end
