include <imhdr.h>
include <fset.h>
include <error.h>

procedure t_profile()
#------------------------------------------------------------------------------
#
#   -------------
#   P R O F I L E
#   -------------
#
#   Description
#   -----------
#   Fit an optimal extraction profile to seeing discs
#
#   
#   Scope of program
#   ----------------
#   - Handles 2-D images only
#   - All data types converted to REAL
#   - Variance arrays supported
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   input      Image with spectra to be fit (string)(prompted for)
#
#   output     Name of file for output profile (string)(prompted for)
#
#   block      Blocking factor in the spectral direction for profile fitting.
#              (int)(hidden)
#
#   pol_dist   Distance between polynomials for Marsh's profile fitting algorithm.
#              (real)(hidden)
#
#   pord       Order of polynomials used for profile fitting (int)(hidden)
#
#   pro_thr    Threshold for deleting bad blocks in profile fitting stage
#              (real)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   - Optimal extract profiles will be stored in a real type file of the
#     same dimensionality as the input file.
#
#   
#   Method
#   ------
#   - If a trace is being used then the routine automatically selects Marsh's 
#     algorithm for fitting the profiles.  Otherwise Horne's algorithm is used.
#   - The section of the input file between the seeing disc limits and the
#     spectral limits is copied to a temporary file.  The trace is followed if
#     necessary.
#   - The relevant model is fitted to the data.  If so requested the data is
#     blocked in the spectral direction before fitting takes place. Whether blocked
#     or not the output profile is evaluated in an unblocked way.
#   - The output section is written to a file which has the same dimensionality as
#     the input file.
#   - If there are more than one spectrum on the slit, then this whole procedure
#     is repeated.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgeti
#      clgetr
#      clgstr
#
#   Library ERRPACK:
#      err_close
#
#   Library ETC:
#      erract
#      error
#
#   Library FIO:
#      fseti
#      fstati
#
#   Library FMTIO:
#      eprintf
#      pargi
#      pargstr
#      printf
#      sprintf
#   
#   Library IMIO:
#      imastr
#      imgetb
#      imgeti
#      imgetr
#      imgstr
#      immap
#      impnlr
#      imps2r
#      imtclose
#      imtgetim
#      imtlen
#      imtopen
#      imunmap
#
#   Library MEMIO:
#      malloc
#      mfree
#      salloc
#      sfree
#      smark
#
#   Library SAMSUB:
#      sam_aroff
#      sam_ctrace
#      sam_deltemp
#      sam_filex
#      sam_gdekker
#      sam_gtrace
#      sam_ifile
#      sam_maket
#      sam_odekker
#      sam_otrace
#      sam_rsect
#      sam_wsect
#      sub_profh
#      sub_profm
#
#   Library VOPS:
#      amovkl
#      amovkr
#
#
#   Internal subroutines called
#   ---------------------------
#   None
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <error.h>
#   include <fset.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   23-FEB-93     - Original program based on VAX program SAM_PROFILE
#   02-MAR-93     - Now processes an input file list (JRL)
#   12-NOV-93     - Minor modifications (JRL)
#   16-AUG-94     - Now an error not to have error files (JRL)
#
#
#------------------------------------------------------------------------------

# Local variables

int        axis               # Dispersion axis
int        block_size         # Blocking factor
pointer    cp1                # Dummy character pointer
pointer    dname              # Pointer for name of dekker file
pointer    dptr               # Pointer for dekker file header
pointer    ename              # Pointer for name of error file
pointer    enptr              # Pointer for error section file header
pointer    eptr               # Pointer for error file header
bool       er_ex              # TRUE if error file exists
pointer    ilist              # Pointer for input file list
pointer    iname              # Pointer for input file name
pointer    inptr              # Pointer for input section file header
int        iobj               # Counter for object numbere
pointer    iptr               # Pointer for input file header
int        j                  # Integer counter
int        listin             # Input list descriptor
int        listout            # Output list descriptor
int        nchunks            # Number of pixels in blocked data
int        nnx                # Spacial dimension of section file
int        nny                # Spectral dimension of section file
int        nobj               # Number of objects on the slit
int        nord               # Order of polynomial for trace
int        npoly              # Number of polynomials needed for Marsh algorithm
int        nt                 # Size of offset array
int        nx                 # X dimension of input file
int        nx1                # Lower limit of seeing disc
int        nx2                # Upper limit of seeing disc
int        ny                 # Y dimension of input file
int        ny1                # Lower limit of spectral range
int        ny2                # Upper limit of spectral range
pointer    olist              # Pointer for output file list
pointer    oname              # Pointer for output file name
long       one_l              # Dummy long variable
pointer    onptr              # Pointer for output section file header
pointer    optr               # Pointer for output file header
real       pol_dist           # Distance between polynomials for Marsh algorithm
int        pord               # Order of polynomials for profile fit
real       pro_thr            # Threshold for rejection of bad blocks in fitting
pointer    ptr                # Dummy pointer
pointer    ptr_de             # Pointer for dekker info
pointer    ptr_out            # Pointer for output file
pointer    ptr_tr             # Pointer for trace
bool       sametrace          # TRUE if all spectra use the same trace
pointer    sp                 # Stack marker
pointer    tempname           # Dummy pointer
pointer    tname              # Pointer for name of trace file
pointer    toff               # Pointer for offset array
pointer    tptr               # Pointer for trace file header
bool       track              # TRUE if trace file is being used
long       v1[IM_MAXDIM]      # Dummy long array
real       width[2]           # Limits of seeing disc
pointer    wpt1               # Pointer for workspace
pointer    wpt10              # Pointer for workspace
pointer    wpt11              # Pointer for workspace
pointer    wpt12              # Pointer for workspace
pointer    wpt13              # Pointer for workspace
pointer    wpt2               # Pointer for workspace
pointer    wpt3               # Pointer for workspace
pointer    wpt4               # Pointer for workspace
pointer    wpt5               # Pointer for workspace
pointer    wpt6               # Pointer for workspace
pointer    wpt7               # Pointer for workspace
pointer    wpt8               # Pointer for workspace
pointer    wpt9               # Pointer for workspace

# Functions

bool imgetb()
int imgeti(),clgeti(),fstati(),imtopen(),imtgetim(),imtlen(),sam_filex()
pointer immap(),impnlr(),sam_maket(),imps2r(),sam_rsect(),sam_gtrace(),
   sam_gdekker()
real clgetr(),imgetr()

begin

      # Mark stack and get some workspace...

      call smark(sp)
      call salloc(iname,SZ_FNAME,TY_CHAR)
      call salloc(ename,SZ_FNAME,TY_CHAR)
      call salloc(oname,SZ_FNAME,TY_CHAR)
      call salloc(dname,SZ_FNAME,TY_CHAR)
      call salloc(tname,SZ_FNAME,TY_CHAR)
      call salloc(tempname,SZ_FNAME,TY_CHAR)
      call salloc(cp1,SZ_LINE,TY_CHAR)
      call salloc(ilist,SZ_LINE,TY_CHAR)
      call salloc(olist,SZ_LINE,TY_CHAR)

      # Flush buffered output immediately

      if (fstati(STDOUT,F_REDIR) == NO) call fseti(STDOUT,F_FLUSHNL,YES)

      # Get the input file name.

      call clgstr("input",Memc[ilist],SZ_LINE)

      # Get a few more parameters

      block_size = clgeti("block")
      pord = clgeti("pord")
      pro_thr = clgetr("pro_thr")

      # Get output file name.

      call clgstr("output",Memc[olist],SZ_LINE)

      # Open the input and output file lists

      listin = imtopen(Memc[ilist])
      listout = imtopen(Memc[olist])
      if (imtlen(listin) != imtlen(listout)) {
         call imtclose(listin)
         call imtclose(listout)
         call error(0,"Number of input/output images not the same")
      }

      # Check that none of the output files already exists...

      if (sam_filex(listout) > 0) call error(0,
         "These output files already exist and cannot be clobbered")         

      # Now loop for the input images...

      while (imtgetim(listin,Memc[iname],SZ_FNAME) != EOF && 
         imtgetim(listout,Memc[oname],SZ_FNAME) != EOF) {

         # Send a friendly message to the user...

         call printf("Beginning work on file %s \n")
            call pargstr(Memc[iname])

         # Open the input file name check its dimensions

         iferr (call sam_ifile(Memc[iname],iptr,nx,ny,Memc[ename],eptr,er_ex)) {
            call eprintf("Error opening input file\n")
            call erract(EA_WARN)
            next
         }

         # Check that error info is available

         if (!er_ex) {
            call eprintf("No error file for %s. Skip to next file...\n")
               call pargstr(Memc[iname])
            next
         }

         # Get some information which you'll need from the header

         axis = imgeti(iptr,"axis")
         nobj = imgeti(iptr,"nobj")

         # Open dekker file...

         call imgstr(iptr,"dekker",Memc[dname],SZ_FNAME)
         iferr(call sam_odekker(dname,nobj,dptr)) {
            call eprintf("Error opening dekker file\n")
            call erract(EA_WARN)
            next
         }

         # If using a trace file, then open it now.

         track = imgetb(iptr,"track")
         if (track) {
            pol_dist = clgetr("pol_dist")
         }
         if (track) call imgstr(iptr,"trace",Memc[tname],SZ_FNAME)
         iferr (call sam_otrace(track,tname,nobj,dptr,nord,tptr,sametrace)) {
            call eprintf("Error opening trace file\n")
            call erract(EA_WARN)
            next
         }

         # Open the output file.  Zero it all then close the file.
         # Force the data type to be real

         iferr (optr = immap(Memc[oname],NEW_COPY,iptr)) {
            call eprintf("Error opening output file\n")
            call erract(EA_WARN)
            next
         }
         IM_PIXTYPE(optr) = TY_REAL
         one_l = 1
         call amovkl(one_l,v1,IM_MAXDIM)
         while (impnlr(optr,ptr,v1) != EOF) {
            call amovkr(0.0,Memr[ptr],nx)
         }
         call imunmap(optr)
         optr = immap(Memc[oname],READ_ONLY,0)

         # Loop for each object now...

         do iobj = 1,nobj {
      
            # Friendly message to the user...

            call printf("Starting work on object %d\n")
               call pargi(iobj)

            # Get the dekker info for this object

            ptr_de = sam_gdekker(dptr,iobj)
            ny1 = nint(Memr[ptr_de+2])
            ny2 = nint(Memr[ptr_de+3])

            # Get the trace for this object

            ptr_tr = sam_gtrace(track,ny2,tptr,sametrace,iobj)

            # First read the position of each object...

            do j = 1,2 {
               call sprintf(Memc[cp1],SZ_LINE,"swid%d_%d")
                  call pargi(iobj)
                  call pargi(j)
               width[j] = imgetr(iptr,Memc[cp1])
            }
            nx1 = nint(width[1])
            nx2 = nint(width[2])
            if (track) {
               nx1 = max(1,nx1-1)
               nx2 = min(IM_LEN(iptr,3-axis),nx2+1)
            }

            # Get offset array workspace

            call sam_aroff(track,ny1,ny2,nt,toff)

            # Open desired section of input file and its error file

            nnx = nx2 - nx1 + 1
            nny = ny2 - ny1 + 1
            iferr {
               inptr = sam_rsect(Memc[iname],nx1,nx2,ny1,ny2,track,Memr[ptr_tr],nord,
                  axis,nt,Mems[toff])
               enptr = sam_rsect(Memc[ename],nx1,nx2,ny1,ny2,track,Memr[ptr_tr],nord,
                  axis,nt,Mems[toff])
            } then {
               call eprintf("Error opening input section files\n")
               call erract(EA_ERROR)
            }

            # Open an output file with the correct dimensions...

            iferr {
               onptr = sam_maket(nnx,nny,TY_REAL,optr)
               ptr_out = imps2r(onptr,1,nnx,1,nny)
            } then {
               call eprintf("Error opening output section file\n")
               call erract(EA_ERROR)
            }

            # If you don't have a TRACE file...

            if (!track) {

               # Get the workspace you need

               iferr {
                  call malloc(wpt1,nnx,TY_REAL)
                  call malloc(wpt2,nnx,TY_REAL)
                  call malloc(wpt3,nnx,TY_REAL)
                  call malloc(wpt4,nnx*nny,TY_REAL)
                  call malloc(wpt5,nnx*nny,TY_REAL)
                  call malloc(wpt6,nnx*nny,TY_REAL)
                  call malloc(wpt7,nnx*nny,TY_REAL)
                  call malloc(wpt8,nnx*nny,TY_REAL)
                  call malloc(wpt9,nny,TY_REAL)
                  call malloc(wpt10,nnx*(pord+1),TY_REAL)
                  call malloc(wpt11,nnx*nny,TY_BOOL)
                  call malloc(wpt12,nny,TY_BOOL)
               } then {
                  call eprintf("Could not get the necessary workspace\n")
                  call erract(EA_ERROR)
               }
         
               # Do the work...If there is an error just move on to the next object

               iferr (call sub_profh(inptr,enptr,nnx,nny,pord,block_size,pro_thr,
                  Memr[wpt1],Memr[wpt2],Memr[wpt3],Memr[wpt4],Memr[wpt5],Memr[wpt6],
                  Memr[wpt7],Memr[wpt8],Memr[wpt9],Memr[wpt10],Memb[wpt11],Memb[wpt12],
                  Memr[ptr_out])) {
                  call eprintf("Error in profile fit on object %d\n")
                     call pargi(j)
                  call erract(EA_WARN)
               }

               # Now get rid of any workspaces...

               call mfree(wpt1,TY_REAL)
               call mfree(wpt2,TY_REAL)
               call mfree(wpt3,TY_REAL)
               call mfree(wpt4,TY_REAL)
               call mfree(wpt5,TY_REAL)
               call mfree(wpt6,TY_REAL)
               call mfree(wpt7,TY_REAL)
               call mfree(wpt8,TY_REAL)
               call mfree(wpt9,TY_REAL)
               call mfree(wpt10,TY_REAL)
               call mfree(wpt11,TY_BOOL)
               call mfree(wpt12,TY_BOOL) }

            # If you do have a TRACE file...

            else if (track) {

               # Set up a few variables

               nchunks = (ny2 - ny1 + 1)/block_size
               if (mod(ny2-ny1+1,block_size) != 0) nchunks = nchunks + 1
               npoly = int(real(nnx)/pol_dist) + 1

               # Get the workspace you need...

               iferr {
                  call malloc(wpt1,nchunks,TY_REAL)
                  call malloc(wpt2,nchunks,TY_REAL)
                  call malloc(wpt3,nnx,TY_REAL)
                  call malloc(wpt4,nnx,TY_REAL)
                  call malloc(wpt5,nnx,TY_REAL)
                  call malloc(wpt6,nnx*nchunks,TY_REAL)
                  call malloc(wpt7,nnx*nchunks,TY_REAL)
                  call malloc(wpt8,((pord+1)*npoly)**2,TY_REAL)
                  call malloc(wpt9,(pord+1)*npoly,TY_REAL)
                  call malloc(wpt10,nchunks*npoly,TY_REAL)
                  call malloc(wpt11,nnx*nchunks,TY_BOOL)
                  call malloc(wpt12,nchunks,TY_BOOL)
                  call malloc(wpt13,nchunks*(2*pord+1),TY_REAL)
                } then {
                  call eprintf("Could not get the necessary workspace\n")
                  call erract(EA_ERROR)
               }

               # Do the work...If there is an error, go on to the next object

               iferr (call sub_profm(inptr,enptr,nnx,nny,ny1,ny2,nchunks,
                  block_size,nord,Memr[ptr_tr],pol_dist,npoly,pord,pro_thr,
                  Memr[wpt1],Memr[wpt2],Memr[wpt3],Memr[wpt4],Memr[wpt5],
                  Memr[wpt6],Memr[wpt7],Memr[wpt8],Memr[wpt9],Memr[wpt10],
                  Memb[wpt11],Memb[wpt12],Memr[wpt13],Memr[ptr_out])) {
                  call eprintf("Error in profile fit on object %d\n")
                     call pargi(j)
                  call erract(EA_WARN)
               }

               # Now get rid of any workspaces...

               call mfree(wpt1,TY_REAL)
               call mfree(wpt2,TY_REAL)
               call mfree(wpt3,TY_REAL)
               call mfree(wpt4,TY_REAL)
               call mfree(wpt5,TY_REAL)
               call mfree(wpt6,TY_REAL)
               call mfree(wpt7,TY_REAL)
               call mfree(wpt8,TY_REAL)
               call mfree(wpt9,TY_REAL)
               call mfree(wpt10,TY_REAL)
               call mfree(wpt11,TY_BOOL)
               call mfree(wpt12,TY_BOOL) 
               call mfree(wpt13,TY_REAL)
            }

            # Now write the results to the output file...

            iferr (call sam_wsect(onptr,Memc[oname],nx1,nx2,ny1,ny2,track,axis,
               nt,Mems[toff])) {
               call eprintf("Error writing from section file on object %d \n")
                  call pargi(j)
               call erract(EA_WARN)
            }

            # Ditch offset array 

            call mfree(toff,TY_SHORT)

            # Deal with the temporary files which were set up by SAM_RSECT

            iferr {
               call sam_deltemp(onptr)
               call sam_deltemp(enptr)
               call sam_deltemp(inptr)
            } then {
               call eprintf("Problem deleting some/all temporary section files\n")
               call erract(EA_WARN)
            }

         } # End of object loop

         # Write name of profile file into header

         call imastr(iptr,"pr_file",Memc[oname])

         # Close up remaining files...

         call err_close(eptr)
         call imunmap(iptr)
         call sam_ctrace(track,tptr,ptr_tr)
         call imunmap(optr)
         call imunmap(dptr)

      } # End of input template loop

      # Close the template files

      call imtclose(listin)
      call imtclose(listout)

      # Ditch workspace

      call sfree(sp)
end
