include <imhdr.h>
include <error.h>
include <fset.h>

procedure t_showme()
#------------------------------------------------------------------------------
#
#   -----------
#   S H O W M E 
#   -----------
#
#   Description
#   -----------
#   Show the locations of the seeing disc and the sky bands on a 
#   displayed image
#
#   
#   Scope of program
#   ----------------
#   - Handles 2-D images only
#   - Data type irrelevant
#   - Variance arrays irrelevant
#   - Batch execution not supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   frame      Number of display frame onto which to draw (int)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   Not relevant
#
#   
#   Method
#   ------
#   - The routine first checks that there is an image display to which it can write
#   - The appropriate information about the location of each spectrum is read
#     from the header of the input file
#   - The dekker limits are drawn in red, the sky bands in green and the limits of
#     the seeing disc in blue.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgeti
#
#   Library DS:
#      imd_mapframe
#      iw_close
#      iw_im2fb
#      iw_open
#
#   Library ETC:
#      erract
#      error
#
#   Library FMTIO:
#      eprintf
#      pargi
#      sprintf
#   
#   Library IMIO:
#      imgetb
#      imgeti
#      imgetr
#      imgl1r
#      imgstr
#      immap
#      imps2s
#      imunmap
#
#   Library MEMIO:
#      malloc
#      mfree
#      salloc
#      sfree
#      smark
#
#   Library SAMSUB:
#      sam_ctrace
#      sam_gdekker
#      sam_gtrace
#      sam_odekker
#      sam_otrace
#
#   Library VOPS:
#      aaddkr
#      amovkr
#      amovr
#      apolr
#
#
#   Internal subroutines called
#   ---------------------------
#      drawline
#      trace_off
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <error.h>
#   include <fset.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   23-FEB-93     - Original program based on VAX program SAM_SHOW
#   11-NOV-93     - Modified to do multiple slit frames properly (JRL)
#
#
#------------------------------------------------------------------------------

# Local variables

int        axis               # Dispersion axis
short      blue               # Colour index for blue
pointer    cp1                # Dummy character pointer
pointer    dname              # Pointer for dekker file name
pointer    dp                 # Display frame descriptor
pointer    dptr               # Pointer for dekker file header
int        frame              # Frame in which to draw
short      green              # Colour index for green
int        i                  # Integer counter
pointer    iname              # Pointer for input file name
pointer    iptr               # Pointer for input file header
pointer    iw                 # Pointer for WCS of plotted frame
int        j                  # Integer counter
int        nobj               # Number of objects on frame
int        nord               # Order of polynomial for trace
int        nspec              # Number of pixels for workspace
int        ny1                # Lower limit of spectral range
int        ny2                # Upper limit of spectral range
pointer    ptr_de             # Pointer for dekker file
pointer    ptr_tr             # Pointer for trace file
short      red                # Colour index for red
bool       sametrace          # TRUE of all spectra use the same trace
real       sky1[2]            # First sky band
real       sky2[2]            # Second sky band
pointer    sp                 # Stack marker
pointer    tname              # Pointer for name of trace
pointer    tptr               # Pointer for trace file header
bool       track              # TRUE if a trace file is being used
int        wcs_status         # WCS status variable
real       width[2]           # Limits of seeing disc
pointer    wpt1               # Pointer for workspace
pointer    wpt2               # Pointer for workspace

# Functions

pointer immap(),imd_mapframe(),iw_open(),sam_gtrace(),sam_gdekker()
int imgeti(),clgeti()
bool imgetb()
real imgetr()

begin

      # Get stack workspace

      call smark(sp)
      call salloc(iname,SZ_FNAME,TY_CHAR)
      call salloc(tname,SZ_FNAME,TY_CHAR)
      call salloc(dname,SZ_FNAME,TY_CHAR)
      call salloc(cp1,SZ_FNAME,TY_CHAR)

      # Open the frame and the WCS

      frame = clgeti("frame")
      iferr {
         dp = imd_mapframe(frame,READ_WRITE,YES)
         iw = iw_open(dp,frame,Memc[iname],SZ_FNAME,wcs_status)
      } then {
         call eprintf("Error opening display window\n")
         call erract(EA_ERROR)
      }

      # Get the input file name.  Open it and check its dimensions...

      iptr = immap(Memc[iname],READ_WRITE,0)
      if (IM_NDIM(iptr) != 2) 
         call error(0,"Only 2-d arrays are considered!")

      # Get a few more parameters

      axis = imgeti(iptr,"axis")

      # Open a dekker file

      call imgstr(iptr,"dekker",Memc[dname],SZ_FNAME)
      iferr(call sam_odekker(dname,nobj,dptr)) {
         call eprintf("Error opening dekker file\n")
         call erract(EA_ERROR)
      }

      # If using a trace file, then open it now.

      track = imgetb(iptr,"track")
      if (track) call imgstr(iptr,"trace",Memc[tname],SZ_FNAME)
      iferr (call sam_otrace(track,tname,nobj,dptr,nord,tptr,sametrace)) {
         call eprintf("Error opening trace file\n")
         call erract(EA_ERROR)
      }

      # Set up colour indices

      red = 204
      green = 205
      blue = 206

      # Now loop for each object

      do i = 1,nobj {

         # Get dekker info for this object

         ptr_de = sam_gdekker(dptr,i)
         ny1 = nint(Memr[ptr_de+2])
         ny2 = nint(Memr[ptr_de+3])

         # Get the trace for the object and create the offset array

         ptr_tr = sam_gtrace(track,ny2,tptr,sametrace,i)

         # If you are using a trace, then get an array for the trace offsets.

         nspec = ny2 - ny1 + 1
         if (track) {
            call malloc(wpt1,nspec,TY_REAL)

            # Create trace offsets

            call trace_off(nspec,ny1,nord,Memr[ptr_tr],Memr[wpt1])
         }

         # Get some workspace for the plotting routine
   
         call malloc(wpt2,nspec,TY_REAL)

         # Do the dekker limits in red

         iferr {
            call drawline(dp,iw,nspec,track,axis,wpt1,wpt2,ny1,
               Memr[ptr_de],red,frame)
            call drawline(dp,iw,nspec,track,axis,wpt1,wpt2,ny1,
               Memr[ptr_de+1],red,frame)
         } then {
            call eprintf("Error drawing dekker limits\n")
            call erract(EA_ERROR)
         }

         # Read the location of the sky bands and the objects from the header

         do j = 1,2 {
            call sprintf(Memc[cp1],SZ_LINE,"skya%d_%d")
               call pargi(i)
               call pargi(j)
            sky1[j] = imgetr(iptr,Memc[cp1])
         }
         do j = 1,2 {
            call sprintf(Memc[cp1],SZ_LINE,"skyb%d_%d")
               call pargi(i)
               call pargi(j)
            sky2[j] = imgetr(iptr,Memc[cp1])
         }
         do j = 1,2 {
            call sprintf(Memc[cp1],SZ_LINE,"swid%d_%d")
               call pargi(i)
               call pargi(j)
            width[j] = imgetr(iptr,Memc[cp1])
         }

         # Do the sky bands in green

         iferr {
            call drawline(dp,iw,nspec,track,axis,wpt1,wpt2,ny1,sky1[1],
               green,frame)
            call drawline(dp,iw,nspec,track,axis,wpt1,wpt2,ny1,sky1[2],
               green,frame)
            call drawline(dp,iw,nspec,track,axis,wpt1,wpt2,ny1,sky2[1],
               green,frame)
            call drawline(dp,iw,nspec,track,axis,wpt1,wpt2,ny1,sky2[2],
               green,frame)
         } then {
            call eprintf("Error drawing sky bands \n")
            call erract(EA_ERROR)
         }

         # Do the spacial limits in blue

         iferr {
            call drawline(dp,iw,nspec,track,axis,wpt1,wpt2,ny1,width[1],
               blue,frame)
            call drawline(dp,iw,nspec,track,axis,wpt1,wpt2,ny1,width[2],
               blue,frame)
         } then {
            call eprintf("Error drawing spectrum band \n")
            call erract(EA_ERROR)
         }

         # Free some workspace
   
         if (track) call mfree(wpt1,TY_REAL)
         call mfree(wpt2,TY_REAL)

      } # End of object loop

      # Close up frame and WCS

      call iw_close(iw)
      call imunmap(dp) 

      # Close up files...

      call imunmap(iptr)
      call imunmap(dptr)
      call sam_ctrace(track,tptr,ptr_tr)

      # Free workspace
   
      call sfree(sp)
end

procedure drawline(dp,iw,n,track,axis,wpt1,wpt2,ny1,xoff,colour,frame)
#------------------------------------------------------------------------------
#      DRAWLINE -- Draw a line on the screen while following a trace
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) dp:         (pointer) Pointer to image display
#      (>) iw:         (pointer) Pointer to image WCS
#      (>) n:          (int) Number of elements in array
#      (>) track:      (bool) TRUE if a trace is being used
#      (>) axis:       (int) 1: horizontal spectra, 2: vertical spectra
#      (>) wpt1:       (pointer) Pointer to offset array
#      (>) wpt2:       (pointer) Pointer to work space array
#      (>) ny1:        (int) Lower limit in spectral direction for plot
#      (>) xoff:       (real) Offset in spatial direction
#      (>) colour:     (short) Colour index for line
#      (>) frame:      (int) Frame into which to plot line
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        23-FEB-1993
#      UPDATE:
#
#------------------------------------------------------------------------------

# Parameters

int n,ny1,frame,axis
real xoff
short colour
bool track
pointer wpt1,wpt2,dp,iw

# Local variables

real x,y,fx,fy
int ii,jj,i
pointer ptr1

# Functions

pointer imps2s()

errchk imps2r

begin

      # Orgainise the offset array

      if (track) {
         call amovr(Memr[wpt1],Memr[wpt2],n) }
      else {
         call amovkr(0.0,Memr[wpt2],n)
      }
      call aaddkr(Memr[wpt2],xoff,Memr[wpt2],n)

      # Now loop for the individual pixels

      do i = 1,n {
         x = real(ny1+i-1)
         y = Memr[wpt2+i-1]
         if (axis == 1) {
            call iw_im2fb(iw,x,y,fx,fy)
            ii = nint(fx)
            jj = nint(fy) }
         else {
            call iw_im2fb(iw,y,x,fy,fx)
            ii = nint(fy)
            jj = nint(fx)
         }
         ptr1 = imps2s(dp,ii,ii,jj,jj) 
         Mems[ptr1] = colour
      }
end


procedure trace_off(n,ny1,nord,trace,array)
#------------------------------------------------------------------------------
#      TRACE_OFF -- Set trace offset into an array
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) n:          (int) Number of elements in array
#      (>) ny1:        (int) Lower limit for plotting in spectral direction
#      (>) nord:       (int) Order of polynomial for trace
#      (>) trace:      (real[nord+4]) Trace coefficients
#      (<) array:      (real[n]) Output offset array
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        23-FEB-1993
#      UPDATE:
#
#------------------------------------------------------------------------------

# Parameters

int n,nord,ny1
real trace[nord+4],array[n]

# Local variables

real zero,yscale,x
int i

# Functions

real apolr()

begin

      # Set up a few variables

      yscale = trace[nord+4]
      zero = apolr(trace[nord+2]*yscale,trace,nord+1)

      # Now loop for each element...

      do i = 1,n {
         x = real(ny1+i-1)*yscale
         array[i] = apolr(x,trace,nord+1) - zero
      }
end
