include <imhdr.h>
include <error.h>
include <fset.h>

procedure t_sky()
#------------------------------------------------------------------------------
#
#   -----
#   S K Y
#   -----
#
#   Description
#   -----------
#   Subtract sky estimate from each spectrum
#
#   
#   Scope of program
#   ----------------
#   - Handles 2-D images only
#   - All data types converted to REAL
#   - Variance arrays supported
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   input      Image with spectra to be processed (string)(prompted for)
#
#   output     Name of file for output frame (string)(prompted for)
#
#   sky_ord    Order of polynomial to fit to sky rows. (int)(hidden)
#
#   nreject    Maximum number of pixel to reject during sky row fitting. 
#              (int)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   - Output file is exactly the same as input except that a sky estimate has
#     been subtracted between the limits of the sky bands.
#
#   
#   Method
#   ------
#   - The section of the input file between the two sky bands and the
#     spectral limits is copied to a temporary file.  The trace is followed if
#     necessary.
#   - A polynomial is fit two the sky info for each row.
#   - For each row a maximum of 'nreject' pixels will be rejected from the fit
#   - The final polynomial fit is subtracted from the data between the seeing
#     disc.
#   - If there are more than one spectrum on the slit, then this whole procedure
#     is repeated.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgeti
#      clgstr
#
#   Library ERRPACK:
#      err_open
#      err_fname
#      err_close
#
#   Library ETC:
#      erract
#      error
#
#   Library FIO:
#      fseti
#      fstati
#
#   Library FMTIO:
#      eprintf
#      pargi
#      pargstr
#      printf
#      sprintf
#   
#   Library IMIO:
#      imflush
#      imgetb
#      imgeti
#      imgetr
#      imgnlr
#      imgstr
#      immap
#      impnlr
#      imtclose
#      imtgetim
#      imtlen
#      imtopen
#      imunmap
#
#   Library MEMIO:
#      malloc
#      mfree
#      salloc
#      sfree
#      smark
#
#   Library RTOOLS:
#      rt_crfile
#
#   Library SAMSUB:
#      sam_aroff
#      sam_ctrace
#      sam_deltemp
#      sam_filex
#      sam_gdekker
#      sam_gtrace
#      sam_ifile
#      sam_odekker
#      sam_otrace
#      sam_rsect
#      sam_wsect
#      sub_sky
#
#   Library VOPS:
#      amovkl
#      amovr
#
#
#   Internal subroutines called
#   ---------------------------
#   None
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <error.h>
#   include <fset.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   23-FEB-93     - Original program based on VAX program SAM_SKY
#   02-MAR-93     - Now processes an input file list (JRL)
#   12-NOV-93     - Minor modifications (JRL)
#
#
#------------------------------------------------------------------------------

# Local variables

int        axis               # Dispersion axis
pointer    cp1                # Dummy character pointer
pointer    dname              # Pointer for dekker file name
pointer    dptr               # Pointer to dekker file
pointer    ename              # Pointer for error file name
pointer    eptr               # Pointer for error file header
bool       er_ex              # TRUE if an error file exists
pointer    ilist              # Pointer for input file list
pointer    iname              # Pointer for input file name
int        iobj               # Counter for the number of objects
pointer    iptr               # Pointer for input file header
int        j                  # Integer counter
int        listin             # Input list descriptor
int        listout            # Output list descriptor
int        nobj               # Number of objects on the slit
int        nord               # Order of trace
int        npix               # Dummy integer
int        nreject            # Number of sky pixels to reject in fit
int        nt                 # Size of offset array
int        nw                 # Size of workspace allocation
int        nx                 # X dimension of input file
int        nx1                # Lower spacial limit of the input section
int        nx2                # Upper spacial limit of the input section
int        ny                 # Y dimension of input file
int        ny1                # Lower spectral limit of the input section
int        ny2                # Upper spectral limit of the input section
int        odims[2]           # Dimensions of output file
pointer    oename             # Pointer for output error file name
pointer    oenptr             # Pointer for output error section file header
pointer    oeptr              # Pointer for output error file header
pointer    olist              # Pointer for output file list
pointer    oname              # Pointer for output file name
long       one_l              # Dummy long variable
pointer    onptr              # Pointer for output section file header
pointer    optr               # Pointer for output file header
pointer    ptr1               # Dummy pointer
pointer    ptr2               # Dummy pointer
pointer    ptr3               # Dummy pointer
pointer    ptr4               # Dummy pointer
pointer    ptr_de             # Pointer for dekker info
pointer    ptr_tr             # Pointer for trace info
bool       sametrace          # TRUE if all spectra use the same trace
real       sky1[2]            # First sky band
real       sky2[2]            # Second sky band
int        sky_ord            # Order of polynomial to fit to sky pixels
pointer    sp                 # Stack marker
pointer    tempname           # Pointer for temporary file name
pointer    tname              # Pointer for trace file name
pointer    toff               # Pointer for offset array
pointer    tptr               # Pointer for trace file header
bool       track              # TRUE if a trace file is being used
long       v1[IM_MAXDIM]      # Dummy long array
long       v2[IM_MAXDIM]      # Dummy long array
long       v3[IM_MAXDIM]      # Dummy long array
long       v4[IM_MAXDIM]      # Dummy long array
pointer    wpt1               # Pointer for workspace
pointer    wpt2               # Pointer for workspace
pointer    wpt3               # Pointer for workspace
pointer    wpt4               # Pointer for workspace

# Functions

pointer immap(),err_open(),sam_rsect(),rt_crfile(),sam_gtrace(),sam_gdekker()
int fstati(),imgeti(),clgeti(),imgnlr(),impnlr(),imtopen(),imtgetim(),imtlen(),
   sam_filex()
real imgetr()
bool imgetb()

begin

      # Get stack workspace

      call smark(sp)
      call salloc(iname,SZ_FNAME,TY_CHAR)
      call salloc(ename,SZ_FNAME,TY_CHAR)
      call salloc(tname,SZ_FNAME,TY_CHAR)
      call salloc(dname,SZ_FNAME,TY_CHAR)
      call salloc(oname,SZ_FNAME,TY_CHAR)
      call salloc(oename,SZ_FNAME,TY_CHAR)
      call salloc(tempname,SZ_FNAME,TY_CHAR)
      call salloc(cp1,SZ_LINE,TY_CHAR)
      call salloc(ilist,SZ_LINE,TY_CHAR)
      call salloc(olist,SZ_LINE,TY_CHAR)

      # Flush buffered output immediately

      if (fstati(STDOUT,F_REDIR) == NO) call fseti(STDOUT,F_FLUSHNL,YES)

      # Get the input file name.

      call clgstr("input",Memc[ilist],SZ_LINE)

      # Get order of fit for sky polynomial and the number of points to be
      # rejected from each fit. 

      sky_ord = clgeti("sky_ord")
      nreject = clgeti("nreject")

      # Get output file name...

      call clgstr("output",Memc[olist],SZ_LINE)

      # Open the input and output file lists

      listin = imtopen(Memc[ilist])
      listout = imtopen(Memc[olist])
      if (imtlen(listin) != imtlen(listout)) {
         call imtclose(listin)
         call imtclose(listout)
         call error(0,"Number of input/output images not the same")
      }

      # Check that none of the output files already exists...

      if (sam_filex(listout) > 0) call error(0,
         "These output files already exist and cannot be clobbered")         

      # Now loop for the input images...

      while (imtgetim(listin,Memc[iname],SZ_FNAME) != EOF && 
         imtgetim(listout,Memc[oname],SZ_FNAME) != EOF) {

         # Send a friendly message to the user...

         call printf("Beginning work on file %s \n")
            call pargstr(Memc[iname])

         # Open the input file name check its dimensions

         iferr (call sam_ifile(Memc[iname],iptr,nx,ny,Memc[ename],eptr,er_ex)) {
            call eprintf("Error opening input file\n")
            call erract(EA_WARN)
            next
         }

         # Get a few more parameters

         axis = imgeti(iptr,"axis")
         nobj = imgeti(iptr,"nobj")

         # Open dekker file...

         call imgstr(iptr,"dekker",Memc[dname],SZ_FNAME)
         iferr(call sam_odekker(dname,nobj,dptr)) {
            call eprintf("Error opening dekker file\n")
            call erract(EA_WARN)
            next
         }

         # If using a trace file, then open it now.

         track = imgetb(iptr,"track")
         if (track) call imgstr(iptr,"trace",Memc[tname],SZ_FNAME)
         iferr (call sam_otrace(track,tname,nobj,dptr,nord,tptr,sametrace)) {
            call eprintf("Error opening trace file\n")
            call erract(EA_WARN)
            next
         }

         # Open the output file and copy the input file to it...

         iferr {
            odims[1] = IM_LEN(iptr,1)
            odims[2] = IM_LEN(iptr,2)
            optr = rt_crfile(Memc[oname],NEW_COPY,iptr,2,odims,
               IM_PIXTYPE(iptr))
            if (er_ex) {
               oeptr = err_open(optr,NEW_COPY,false)
               call err_fname(optr,Memc[oename])
            }
         } then {
            call eprintf("Error opening output file\n")
            call erract(EA_WARN)
            next
         }
         one_l = 1.0
         call amovkl(one_l,v1,IM_MAXDIM)
         call amovkl(one_l,v2,IM_MAXDIM)
         call amovkl(one_l,v3,IM_MAXDIM)
         call amovkl(one_l,v4,IM_MAXDIM)
         while (imgnlr(iptr,ptr1,v1) != EOF) {
            npix = impnlr(optr,ptr3,v3)
            call amovr(Memr[ptr1],Memr[ptr3],npix)
            if (er_ex) {
               npix = imgnlr(eptr,ptr2,v2)
               npix = impnlr(oeptr,ptr4,v4)
               call amovr(Memr[ptr2],Memr[ptr4],npix)
            }
         }
         if (er_ex) call err_close(oeptr)
         call imunmap(optr)
         if (er_ex) call err_close(eptr)
      
         # Unmap the input files to get rid of the workspace associated with them,
         # but reopen the input file so that the header can be read

         call imunmap(iptr)
         iptr = immap(Memc[iname],READ_ONLY,0)

         # Now loop for each object

         do iobj = 1,nobj {

            # Friendly message to the user...

            call printf("Starting work on object %d\n")
               call pargi(iobj)

            # Get dekker info for this object

            ptr_de = sam_gdekker(dptr,iobj)
            ny1 = nint(Memr[ptr_de+2])
            ny2 = nint(Memr[ptr_de+3])

            # Get the trace for this object

            ptr_tr = sam_gtrace(track,IM_LEN(iptr,axis),tptr,sametrace,iobj)

            # Read the location of the sky bands from the header

            do j = 1,2 {
               call sprintf(Memc[cp1],SZ_LINE,"skya%d_%d")
                  call pargi(iobj)
                  call pargi(j)
               sky1[j] = imgetr(iptr,Memc[cp1])
            }
            do j = 1,2 {
               call sprintf(Memc[cp1],SZ_LINE,"skyb%d_%d")
                  call pargi(iobj)
                  call pargi(j)
               sky2[j] = imgetr(iptr,Memc[cp1])
            }

            # Get workspace for offset array

            call sam_aroff(track,ny1,ny2,nt,toff)

            # Open the section files...
          
            nx1 = nint(sky1[1])
            nx2 = nint(sky2[2])
            iferr {
               onptr = sam_rsect(Memc[oname],nx1,nx2,ny1,ny2,track,Memr[ptr_tr],nord,
                  axis,nt,Mems[toff])
               if (er_ex) oenptr = sam_rsect(Memc[oename],nx1,nx2,ny1,ny2,track,
                  Memr[ptr_tr],nord,axis,nt,Mems[toff])
            } then {
               call eprintf("Error opening input section files\n")
               call erract(EA_ERROR)
            }

            # Get some workspace for the main processing routine

            iferr {
               nw = nx2 - nx1 + 1
               call malloc(wpt1,nw,TY_REAL)
               call malloc(wpt2,nw,TY_REAL)
               call malloc(wpt3,nw,TY_REAL)
               call malloc(wpt4,sky_ord+1,TY_REAL)
            } then {
               call eprintf("Could not get the needed workspace \n")
               call erract(EA_ERROR)
            }

            # Now do the work...

            iferr(call sub_sky(er_ex,Mems[toff],sky_ord,nreject,axis,sky1,sky2,
               ny1,ny2,Memr[wpt1],Memr[wpt2],Memr[wpt3],nw,Memr[wpt4],onptr,
               oenptr)) 
               call erract(EA_WARN)

            # Flush output buffers...

            call imflush(onptr)
            if (er_ex) call imflush(oenptr)

            # Now transfer these results over...

            iferr {
               call sam_wsect(onptr,Memc[oname],nx1,nx2,ny1,ny2,track,axis,
                  nt,Mems[toff])
               if (er_ex) call sam_wsect(oenptr,Memc[oename],nx1,nx2,ny1,ny2,track,
                  axis,nt,Mems[toff])
            } then {
               call eprintf("Error writing output from section files\n")
               call erract(EA_ERROR)
            }

            # Free some workspace

            call mfree(wpt1,TY_REAL)
            call mfree(wpt2,TY_REAL)
            call mfree(wpt3,TY_REAL)
            call mfree(wpt4,TY_REAL)
            call mfree(toff,TY_SHORT)

            # Deal with temporary files created by SAM_RSECT

            call sam_deltemp(onptr)
            if (er_ex) call sam_deltemp(oenptr)

         } # End of object loop

         # Close up some files and ditch some workspace

         call imunmap(iptr)
         call sam_ctrace(track,tptr,ptr_tr)
         call imunmap(dptr)

      } # End of input template loop

      # Close remaining files

      call imtclose(listin)
      call imtclose(listout)

      # Free some workspace

      call sfree(sp)
end
                                     
