include <imhdr.h>
include <error.h>
include <fset.h>

procedure t_strt()
#-------------------------------------------------------------------------------
#
#   -------
#   S T R T
#   -------
#
#   Description
#   -----------
#   Straighten a spectrum which has an associated trace file.  Note that the
#   results from this operation should not then be used for optimal extraction
#   as the independence of error values will have been violated
#
#   
#   Scope of program
#   ----------------
#   - Handles 2-D images only
#   - All data types converted to REAL
#   - Variance arrays supported
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   input      List of images with spectrum to be straightened 
#              (string)(prompted for)
#
#   output     List of output file names (string)(prompted for)
#
#   axis       Axis to which spectra are parallel -- i.e. dispersion axis.
#              1 = X, 2 = Y. (int)(hidden)
#
#   trace      Name of the trace file if required. (string)(hidden)
#
#   dekker     Name of the dekker file. (string)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   - Output file will have all the same attributes as the input file
#     including all its header information.
#
#   
#   Method
#   ------
#   - The appropriate section of the input file is copied to a temporary 
#     file, following the trace if necessary.
#   - Pixels are shifted by the integral amount in the section algorithm and
#     by the fractional amount by a linear interpolation scheme.
#   - The output section is written to an output file which is the same size,
#     etc as the input file.  The trace is followed if necessary.
#   - The pixels which are not covered by the output spectrum will be filled
#     with zeros.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgeti
#      clgstr
#
#   Library ERRPACK:
#      err_close
#      err_fname
#      err_open
#
#   Library ETC:
#      erract
#      error
#
#   Library FIO
#      fseti
#      fstati
#
#   Library FMTIO:
#      eprintf
#      pargstr
#      pargi
#      printf
#   
#   Library IMIO:
#      imflush
#      imgl1r
#      imgl2r
#      immap
#      impl2r
#      impnlr
#      imtclose
#      imtgetim
#      imtlen
#      imtopen
#      imunmap
#
#   Library MEMIO:
#      malloc
#      mfree
#      salloc
#      sfree
#      smark
#
#   Library SAMSUB:
#      sam_aroff
#      sam_ctrace
#      sam_deltemp
#      sam_filex
#      sam_gdekker
#      sam_gtrace
#      sam_ifile
#      sam_odekker
#      sam_otrace
#      sam_rsect
#      sam_wsect
#
#   Library VOPS:
#      amovkl
#      amovkr
#      apolr
#
#
#   Internal subroutines called
#   ---------------------------
#      sub_strt
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <error.h>
#   include <fset.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   26-FEB-93     - Original program based on VAX program SAM_STRT
#   02-MAR-93     - Now processes an input file list (JRL)
#   12-NOV-93     - Minor modifications (JRL)
#
#
#-------------------------------------------------------------------------------

# Local variables

int        axis               # Dispersion axis
pointer    dname              # Pointer for dekker file name
pointer    dptr               # Pointer for dekker file header
pointer    ename              # Pointer for error file name
pointer    enptr              # Pointer for error section file name
pointer    eptr               # Pointer for error file header
bool       er_ex              # TRUE if error file exists
pointer    ilist              # Pointer for input file list
pointer    iname              # Pointer for input file name
pointer    inptr              # Pointer for input section file name
int        iobj               # Counter for number of objects on frame
pointer    iptr               # Pointer for input file header
int        listin             # Input file list descriptor
int        listout            # Output file list descriptor
int        nnx                # Spacial size of input section file
int        nobj               # Total number of objects on frame
int        nord               # Order of polynomial for trace
int        npix               # Dummy integer
int        nt                 # Size of offset array
int        nx                 # X dimension of input data
int        nx1                # Lower spacial limit of section
int        nx2                # Upper spacial limit of section
int        ny                 # Y dimension of input data
int        ny1                # Lower spectral limit of section
int        ny2                # Upper spectral limit of section
pointer    oename             # Pointer for output error file name
pointer    oenptr             # Pointer for output error section file header
pointer    oeptr              # Pointer for output error file header
pointer    olist              # Pointer for output file list
pointer    oname              # Pointer for output file name
long       one_l              # Dummy long variable
pointer    onptr              # Pointer for output section file header
pointer    optr               # Pointer for output file header
pointer    ptr1               # Dummy pointer
pointer    ptr2               # Dummy pointer
pointer    ptr_de             # Pointer for dekker file
pointer    ptr_tr             # Pointer for trace file
bool       sametrace          # TRUE if all spectra use the same trace
pointer    sp                 # Stack marker
pointer    tname              # Pointer for trace file name
pointer    toff               # Pointer for offset array
pointer    tptr               # Pointer for trace file header
bool       track              # Dummy boolean variable (always TRUE)
long       v1[IM_MAXDIM]      # Dummy long array
long       v2[IM_MAXDIM]      # Dummy long array

# Functions

pointer immap(),err_open(),sam_rsect(),sam_gtrace(),sam_gdekker()
int clgeti(),impnlr(),imtopen(),imtlen(),imtgetim(),fstati(),sam_filex()

begin

      # Get stack workspace

      call smark(sp)
      call salloc(iname,SZ_FNAME,TY_CHAR)
      call salloc(ename,SZ_FNAME,TY_CHAR)
      call salloc(tname,SZ_FNAME,TY_CHAR)
      call salloc(dname,SZ_FNAME,TY_CHAR)
      call salloc(oname,SZ_FNAME,TY_CHAR)
      call salloc(oename,SZ_FNAME,TY_CHAR)
      call salloc(ilist,SZ_LINE,TY_CHAR)
      call salloc(olist,SZ_LINE,TY_CHAR)

      # Flush buffered output immediately

      if (fstati(STDOUT,F_REDIR) == NO) call fseti(STDOUT,F_FLUSHNL,YES)

      # Get the input file name.

      call clgstr("input",Memc[ilist],SZ_LINE)

      # Get a few more parameters

      axis = clgeti("axis")

      # Open a dekker file

      call clgstr("dekker",Memc[dname],SZ_FNAME)
      iferr (call sam_odekker(dname,nobj,dptr)) {
         call eprintf("Error opening dekker file\n")
         call erract(EA_ERROR)
      }

      # Open trace file. (You must have one, otherwise there's no 
      # point in being here)

      call clgstr("trace",Memc[tname],SZ_FNAME)
      track = true
      iferr (call sam_otrace(track,tname,nobj,dptr,nord,tptr,sametrace)) {
         call eprintf("Error opening trace file\n")
         call erract(EA_ERROR)
      }
 
      # Get output file name...

      call clgstr("output",Memc[olist],SZ_LINE)

      # Open the input and output file lists

      listin = imtopen(Memc[ilist])
      listout = imtopen(Memc[olist])
      if (imtlen(listin) != imtlen(listout)) {
         call imtclose(listin)
         call imtclose(listout)
         call error(0,"Number of input/output images not the same")
      }

      # Check that none of the output files already exists...

      if (sam_filex(listout) > 0) call error(0,
         "These output files already exist and cannot be clobbered")         

      # Now loop for the input images...

      while (imtgetim(listin,Memc[iname],SZ_FNAME) != EOF && 
         imtgetim(listout,Memc[oname],SZ_FNAME) != EOF) {

         # Send a friendly message to the user...

         call printf("Beginning work on file %s \n")
            call pargstr(Memc[iname])

         # Open the input file name check its dimensions

         iferr (call sam_ifile(Memc[iname],iptr,nx,ny,Memc[ename],eptr,er_ex)) {
            call eprintf("Error opening input file\n")
            call erract(EA_WARN)
            next
         }

         # Open the output file and fill it with zeros...

         iferr (optr = immap(Memc[oname],NEW_COPY,iptr)) {
            call eprintf("Error opening output file \n")
            call erract(EA_WARN)
            next
         }
         if (er_ex) {
            call err_fname(optr,Memc[oename])
            iferr (oeptr = err_open(optr,NEW_COPY,false)) {
               call eprintf("Error opening output error file\n")
               call erract(EA_WARN)
               next
            }
         }
         one_l = 1.0
         call amovkl(one_l,v1,IM_MAXDIM)
         call amovkl(one_l,v2,IM_MAXDIM)
         while (impnlr(optr,ptr1,v1) != EOF) {
            call amovkr(0.0,Memr[ptr1],nx)
            if (er_ex) {
               npix = impnlr(oeptr,ptr2,v2)
               call amovkr(0.0,Memr[ptr2],nx)
            }
         }
         if (er_ex) call err_close(oeptr)
         call imunmap(optr)
         if (er_ex) call err_close(eptr)

         # Loop for number of objects...

         do iobj = 1,nobj {

            # Message to the user...
 
            call printf("Working on object # %d\n")
            call pargi(iobj)

            # Get dekker info for this object
 
            ptr_de = sam_gdekker(dptr,iobj)
            nx1 = nint(Memr[ptr_de])
            nx2 = nint(Memr[ptr_de+1])
            ny1 = nint(Memr[ptr_de+2])
            ny2 = nint(Memr[ptr_de+3])

            # Get the trace for this object

            ptr_tr = sam_gtrace(track,ny2,tptr,sametrace,iobj)

            # Check that the requested spectral range is not too big for the file

            if (ny2  > IM_LEN(iptr,axis)) {
               call eprintf("Requested spectral range too big for: %s\n")
                  call pargstr(Memc[iname])
               next
            }

            # Get workspace for offset array

            call sam_aroff(track,ny1,ny2,nt,toff)

            # Section the input file so that it has one extra column on 
            # either side than the output section file...This may need 
            # to be checked to be sure that the dekker isn't say on column 
            # 1 or column nx...

            nx1 = nx1 - 1
            nx2 = nx2 + 1
            iferr {
               inptr = sam_rsect(Memc[iname],nx1,nx2,ny1,ny2,track,
                  Memr[ptr_tr],nord,axis,nt,Mems[toff])
               if (er_ex) enptr = sam_rsect(Memc[ename],nx1,nx2,ny1,ny2,track,
                  Memr[ptr_tr],nord,axis,nt,Mems[toff])
            } then {
               call eprintf("Error opening input section files\n")
               call erract(EA_ERROR)
            }

            # Now section output files...

            nx1 = nx1 + 1
            nx2 = nx2 - 1
            nnx = nx2 - nx1 + 1
            iferr {
               onptr = sam_rsect(Memc[oname],nx1,nx2,ny1,ny2,track,
                  Memr[ptr_tr],nord,axis,nt,Mems[toff])
               if (er_ex) oenptr = sam_rsect(Memc[oename],nx1,nx2,ny1,ny2,
                  track,Memr[ptr_tr],nord,axis,nt,Mems[toff])
            } then {
               call eprintf("Error opening output section files\n")
               call erract(EA_ERROR)
            }

            # Now do the work

            iferr (call sub_strt(inptr,enptr,er_ex,nnx,ny1,ny2,Memr[ptr_tr],
               nord,onptr,oenptr)) {
               call eprintf("Error straightening spectrum \n")
               call erract(EA_WARN)
               next
            }

            # Flush the output buffers

            call imflush(onptr)
            if (er_ex) call imflush(oenptr)

            # Now transfer these results over...
    
            iferr {
               call sam_wsect(onptr,Memc[oname],nx1,nx2,ny1,ny2,false,axis,
                  nt,Mems[toff])
               if (er_ex) call sam_wsect(oenptr,Memc[oename],nx1,nx2,ny1,ny2,
                  false,axis,nt,Mems[toff])
            } then {
               call eprintf("Error writing section to output file\n")
               call erract(EA_ERROR)
            }

            # Ditch offset array workspace

            call mfree(toff,TY_SHORT)

            # Deal with temporary files...

            iferr {
               call sam_deltemp(onptr)
               if (er_ex) call sam_deltemp(oenptr)
               call sam_deltemp(inptr)
               if (er_ex) call sam_deltemp(enptr)
            } then {
               call eprintf("Error deleting temporary files\n")
               call erract(EA_WARN)
            }
         } # End object loop

         # Close input file

         call imunmap(iptr)
        
      } # End of template loop

      # Close up any remaining files...

      call imunmap(dptr)
      call sam_ctrace(track,tptr,ptr_tr)
      call imtclose(listin)
      call imtclose(listout)

      # Free some workspace

      call sfree(sp)
end



procedure sub_strt(iptr,eptr,er_ex,nnx,ny1,ny2,trace,nord,optr,oeptr)
#-------------------------------------------------------------------------------
#      SUB_STRT -- Straigten spectra with a linear interpolation
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) iptr:       (pointer) Pointer to the input data file header
#      (>) eptr:       (pointer) Pointer to the input error file header
#      (>) er_ex:      (bool) TRUE if an error file exits
#      (>) nnx:        (int) Number of pixels in a row to interpolate
#      (>) ny1:        (int) Lower limit of requested spectral range
#      (>) ny2:        (int) Upper limit of requested spectral range
#      (>) trace:      (real[nord+4]) Trace information
#      (>) nord:       (int) Order of polynomial for trace fit
#      (>) optr:       (pointer) Pointer to output data file header
#      (>) oeptr:      (pointer) Pointer to output error file header
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        26-FEB-1993
#      UPDATE:
#
#-------------------------------------------------------------------------------

# Parameters

pointer iptr,eptr,optr,oeptr
int ny1,ny2,nord,nnx
real trace[nord+4]
bool er_ex

# Local variables

real yscale,zero,x,y,diff,dx,w1,w2,dat1,dat2
int irow,j,ioff,i
pointer ptr1,ptr2,ptr3,ptr4

# Functions

real apolr()
pointer imgl2r(),impl2r()

errchk imgl2r,impl2r

begin

      # Set up a few constants first...

      yscale = trace[nord+4]
      zero = apolr(trace[nord+2]*yscale,trace,nord+1)

      # Loop for each pixel in the spectral direction...

      irow = 0
      do j = ny1,ny2 {
         irow = irow + 1

         # Find the offset at this wavelength

         y = real(j)*yscale
         x = apolr(y,trace,nord+1) - zero

         # "diff" tells whether the SAM sectioning routines have moved forward
         # by an integer pixel or not.  If diff < 0 then the interpolation is
         # between the pixel in question and it's neighbour on the lhs.  If
         # diff > 0 then the neighbour on the rhs is used.

         diff = x - nint(x)
         if (diff <= 0) {
            ioff = -1 }
         else {
            ioff = 0
         }

         # Sort out weights...

         x = x + 1               # Deal with negative shifts
         dx = x - real(int(x))
         w1 = 1.0 - dx
         w2 = dx

         # Read in rows of data

         ptr1 = imgl2r(iptr,irow)
         ptr2 = impl2r(optr,irow)
         if (er_ex) {  
            ptr3 = imgl2r(eptr,irow)
            ptr4 = impl2r(oeptr,irow)
         }

         # Now loop for each pixel in the row.  Note that the input file has an
         # extra row on either side...

         do i = 1,nnx {
            dat1 = Memr[ptr1+i+ioff]
            dat2 = Memr[ptr1+i+1+ioff]
            Memr[ptr2+i-1] = dat1*w1 + dat2*w2
            if (er_ex) {
               dat1 = Memr[ptr3+i+ioff]
               dat2 = Memr[ptr3+i+1+ioff]
               Memr[ptr4+i-1] = dat1*w1 + dat2*w2
            }
         }
      } # end of row loop
end
