include <imhdr.h>
include <error.h>
include <mach.h>
include <math.h>

define HELPFIL    "rgo$sam/curfile/trhlp1.key"
define HELPFIL2   "rgo$sam/curfile/trhlp2.key"
define PROMPT     "trace: options for locating spectra"
define PROMPT2    "trace: options for wavelength range definition"

procedure t_trace()
#------------------------------------------------------------------------------
#
#   ---------
#   T R A C E
#   ---------
#
#   Description
#   -----------
#   Trace the curvature of a spectrum and fit a polynomial to it.
#
#   
#   Scope of program
#   ----------------
#   - Handles 2-D images only
#   - All data types converted to REAL
#   - Variance arrays supported, but not used
#   - Batch execution not supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   input      Image with spectrum to be traced(string)(prompted for)
#
#   output     Output file for the trace coefficients (string)(prompted for)
#
#   axis       Axis to which spectra are parallel -- i.e. dispersion axis.
#              1 = X, 2 = Y. (int)(prompted for)
#
#   nord       Order of polynomial to be fit to spectrum trace (int)
#              (prompted for)
#
#   nobj       Number of objects to which to fit a trace (int)(prompted for)
#
#   nom_width  Nominal width of the seeing disc (FWHM).  This will 
#              also be used as the FWHM of a smoothing kernel.
#              (int)(hidden)
#
#   block      Blocking factor in the spectral direction for centring.
#              (int)(hidden)
#
#   clip       Threshold in sigma at which to clip blocks from the fit.
#              (real)(hidden)
#
#   ncycle     Number of centring/fitting cycles to be done. (int)(hidden)
#
#   frame      Display frame onto which to draw the trace (int)(hidden)
#
#   colour     Colour index to using in drawing trace (int)(hidden)
#
#   imagecur   Parameter for use in cursor input (*imcur)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   - The output file will have a 2-d data array with nord+4 elements
#     in each row.  (One row per object to be traced). The first nord+1 
#     elements will be used for the trace coefficients; the next two 
#     will be used for the spectral range over which the fit was 
#     performed; the last one will be a factor which is used to scale 
#     the independent variable.
#
#   
#   Method
#   ------
#   - The user should first display the frame in question.
#   - Each spectrum is dealt with in turn.
#   - The user is requested to indicate the spectrum and the spectral
#     range over which to fit the trace
#   - From this point, the program procedes to locate the centre of the 
#     spectrum at each wavelength, blocking if requested.  
#   - The position of the centres as a function of wavelength (pixel space)
#     is fitted by a polynomial.  Bad blocks are clipped out and the program
#     redoes the centring the fitting until the specified number of cycles 
#     has be performed
#   - The fit is drawn on the display and the coefficients are written to the 
#     output file
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgcur
#      clgeti
#      clgetr
#      clgstr
#
#   Library DS:
#      imd_mapframe
#      iw_close
#      iw_im2fb
#      iw_open
#
#   Library ERRPACK:
#      err_close
#
#   Library ETC:
#      erract
#      error
#      pagefile
#
#   Library FMTIO:
#      eprintf
#      pargi
#      pargr
#      printf
#   
#   Library IMIO:
#      imgs2s
#      impl2r
#      imps2s
#      imunmap
#
#   Library MEMIO:
#      malloc
#      mfree
#      salloc
#      sfree
#      smark
#
#   Library RTOOLS:
#      rt_crfile
#
#   Library SAMSUB:
#      sam_band
#      sam_ifile
#      sam_peaks
#      sam_poly
#
#   Library VOPS:
#      aaddr
#      acnvrr
#      adivkr
#      amovr
#      amovkr
#      apolr
#
#   Library XTOOLS:
#      cogetr
#      comap
#      counmap
#
#
#   Internal subroutines called
#   ---------------------------
#      sub_trace
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <error.h>
#   include <mach.h>
#   include <math.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   26-FEB-93     - Original program based on VAX program SAM_TRACE
#   12-NOV-93     - Modified to do multiple traces on a single frame (JRL)
#   23-NOV-93     - Major revamp to include new and simpler centring
#                   algorithm (JRL)
#
#
#------------------------------------------------------------------------------

# Local variables

int        axis               # Dispersion axis
int        block_size         # Blocking factor in spectral direction
real       clip               # Clipping threshold for fitting cycle
int        colour             # Colour index used for over plot
pointer    ename              # Pointer for error file name
pointer    eptr               # Pointer for error file header
bool       er_ex              # TRUE if error file exists
int        frame              # Frame onto which over plot will be done
pointer    iname              # Pointer for input file name
int        iobj               # Number of the spectrum
pointer    iptr               # Pointer for input file header
int        ncycle             # Number of centring/fitting cycles to do
int        ndim               # Number of dimensions in output file
int        nobj               # Total number of objects
int        nom_width          # Nominal width of seeing disc
int        nord               # Order of polynomial to fit to trace
int        nt                 # Dummy integer variable
int        nx                 # X dimension of input data
int        ny                 # Y dimension of input data
int        odims[2]           # Output file dimensions
pointer    oname              # Pointer for output file name
pointer    optr               # Pointer for output file
pointer    ptr                # Dummy pointer
pointer    sp                 # Stack marker

# Functions

pointer impl2r(),rt_crfile()
int clgeti()
real clgetr()

begin

      # Get stack workspace

      call smark(sp)
      call salloc(iname,SZ_FNAME,TY_CHAR)
      call salloc(ename,SZ_FNAME,TY_CHAR)
      call salloc(oname,SZ_FNAME,TY_CHAR)

      # Get the input file name.  Open it and check its dimensions

      call clgstr("input",Memc[iname],SZ_FNAME)
      iferr (call sam_ifile(Memc[iname],iptr,nx,ny,Memc[ename],eptr,er_ex)) {
         call eprintf("Error opening input file\n")
         call erract(EA_ERROR)
      }

      # Get a few more parameters

      axis = clgeti("axis")
      nord = clgeti("nord")
      nom_width = clgeti("nom_width")
      clip = clgetr("clip")
      block_size = clgeti("block")
      ncycle = clgeti("ncycle")
      frame = clgeti("frame")
      colour = clgeti("colour")
      nobj = clgeti("nobj")

      # Get output file name
 
      call clgstr("output",Memc[oname],SZ_FNAME)

      # Open the output file

      ndim = 2
      odims[1] = nord + 4
      odims[2] = nobj
      iferr(optr = rt_crfile(Memc[oname],NEW_FILE,0,ndim,odims,TY_REAL))
         call erract(EA_ERROR)

      # Switch the dimensions of the maps if the dispersion axis is the x axis, as
      # the main processing routine assumes that the dispersion axis is the y axis.

      if (axis == 1) {
         nt = ny
         ny = nx
         nx = nt
      }

      # Now call the main processing routine for each object

      do iobj = 1,nobj {

         # Map output data array...

         ptr = impl2r(optr,iobj)

         # Now fit the trace ...

         iferr(call sub_trace(nx,ny,iptr,axis,nom_width,clip,block_size,nord,
            ncycle,frame,colour,iobj,Memr[ptr])) {
            call eprintf("Trace fitting routine failed\n")
            call erract(EA_WARN)
            next
         }
      }

      # Unmap the input and output files

      call imunmap(iptr)
      call imunmap(optr)

      # Free some workspace

      call sfree(sp)
end



procedure sub_trace(nx,ny,iptr,axis,spec_width,clip,block_size,nord,
   ncycle,frame,colour,iobj,coefs)
#------------------------------------------------------------------------------
#      SUB_TRACE -- Do the centring and fitting for the trace.  Plot up the
#                   the resulting polynomial
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) nx:         (int) X dimension of data
#      (>) ny:         (int) Y dimension of data
#      (>) iptr:       (pointer) Input data file pointer
#      (>) axis:       (int) Dispersion axis
#      (>) spec_width: (int) Nominal FWHM of the seeing disc
#      (>) clip:       (real) Threshold for clipping during fit cycle
#      (>) block_size: (int) Number of rows to block together to form one block
#      (>) nord:       (int) Order of polynomial for fit
#      (>) ncycle:     (int) Number of centring/fit cycles to do
#      (>) frame:      (int) Frame onto which overplot is to be done
#      (>) colour:     (int) Colour index for overplot
#      (>) iobj:       (int) Number of the spectrum
#      (<) coefs:      (real[nord+4]) Output trace coefs
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        26-FEB-1993
#      UPDATE:      11-NOV-1993 -- Minor modifications plus a few bug 
#                                  fixes.  (JRL)
#                   23-NOV-1993 -- Major changes to use a new, simpler and
#                                  (hopefully) more stable centring algorithm
#                                  (JRL)
#                                  
#
#------------------------------------------------------------------------------

# Parameters

int nx,ny,block_size,nord,ncycle,axis,frame,colour,iobj,spec_width
real clip,coefs[nord+4]
pointer iptr

# Local variables

pointer sp,cmd,imname,pptr,wpptr,cnvptr,kptr,xptr,xfit,yfit,wfit,co1,ptr1,
   wpt1,wpt2,wpt3,dp,iw
bool quit,whole,ok
int nst,wcs,key,i,nnx,iext,knpix,nk1,isuccess,iblock,k,jstart,jend,jinc,
   j1,j2,i1,i2,nix,jj,ntot,status,ifirst,nfit,nterms,icycle,jtot,
   nblock,remain,nclip,nused,irem,n,ii,j,wcs_status,ndx,ndy
real dum1[2],x0,y0,spec_rng[2],sig_gau,anorm,denom,ymax,xstart,xlast,ylast,
   ysol,wsol,xsol,sigma,slope,aint,ynew,dev_max,sum,sumsq,xx,yy,dev,rms,
   fx,fy,dum2[2,2],chisqr

# Functions

pointer imgs2r(),imps2s(),imd_mapframe(),iw_open(),comap(),cogetr()
real    apolr()
int     clgcur(),errcode()

errchk imps2s,imgs2s,imd_mapframe,iw_open,comap,cogetr,clgcur,sam_poly

begin

      # Get some stack workspace
 
      call smark(sp)
      call salloc(cmd,SZ_LINE,TY_CHAR)
      call salloc(imname,SZ_FNAME,TY_CHAR)

      # Start by indicating location of spectrum.  This defines where on the
      # image to begin finding the position of the spectrum

      quit = false
      call printf("Indicate location of spectrum #%d with cursor\n")
      call pargi(iobj)
      ok = false
      while (!ok) {
         nst = clgcur("imagecur",dum1[1],dum1[2],wcs,key,Memc[cmd],SZ_LINE)
         switch (key) {
            case 'q':
               quit = true
               break
            case '?':
               call pagefile(HELPFIL,PROMPT)
            default:
               ok = true
         }
      }
      if (quit) call error(1,"Point selection ended prematurely")
      x0 = dum1[3-axis]
      y0 = dum1[axis]

      # Indicate the spectral range you want to fit with two
      # cursor points.  If the user hits 'w', then the full spectral range
      # will be used.

      call printf("Indicate limits in spectral direction for fitting\n")
      call printf("(typing 'w' will give the whole spectral range)\n")
      whole = false
      i = 0
      while (i < 2) {
         nst = clgcur("imagecur",dum1[1],dum1[2],wcs,key,Memc[cmd],SZ_LINE)
         switch (key) {
            case 'q':
               quit = true 
               break
            case 'w':
               whole = true
               break
            case '?':
               call pagefile(HELPFIL2,PROMPT2)
            default:
               i = i + 1
               dum2[1,i] = dum1[1]
               dum2[2,i] = dum1[2]
         }
      }
      if (quit) {
         call error(1,"Point selection ended prematurely") }
      else if (whole) {
         spec_rng[1] = 1.0
         spec_rng[2] = real(ny) }
      else {
         spec_rng[1] = real(nint(dum2[axis,1]))
         spec_rng[2] = real(nint(dum2[axis,2]))
         call sam_band(spec_rng[1],spec_rng[2])
         if (spec_rng[1] < 1.0) spec_rng[1] = 1.0
         if (spec_rng[2] > real(ny)) spec_rng[2] = real(ny)
      }
      
      # Get some workspace for centring routine

      iferr {
         nnx = 4*spec_width + 1  # 4*FWHM of seeing disc is extracted
         iext = spec_width       # Extend profile by 1 FWHM
         knpix = 2*iext + 1      # Size of smoothing kernel is 2*FWHM
         nk1 = iext + 1
         call malloc(pptr,nnx,TY_REAL)
         call malloc(wpptr,nnx+knpix-1,TY_REAL)
         call malloc(cnvptr,nnx+knpix-1,TY_REAL)
         call malloc(kptr,knpix,TY_REAL)
         call malloc(xptr,nnx,TY_REAL)

         # Get some workspace for storing centroids

         call malloc(xfit,ny,TY_REAL)
         call malloc(yfit,ny,TY_REAL)
         call malloc(wfit,ny,TY_REAL)
      } then {
         call error(3,"Could not get necessary workspace")
      }

      # Set up smoothing kernel

      sig_gau = real(spec_width)/2.355
      anorm = 1.0/(sqrt(TWOPI)*sig_gau)
      denom = -0.5/sig_gau**2
      do i = 1,knpix
         Memr[kptr+i-1] = anorm*exp(denom*(real(i-nk1)**2))

      # Set for column access if it's needed

      if (axis == 1) co1 = comap(iptr,nx)

      # Set a few useful variables

      isuccess = 0
      iblock = 0
      ymax = real(ny)

      # Loop twice here.  The first time procede upwards from indicated point.
      # The second time, downwards...

      do k = 1,2 {

         # Set up a few useful things depending on whether the loop is up or down

         if (k == 1) {
            jstart = nint(y0)
            jend = nint(spec_rng[2])
            jinc = block_size
            xstart = x0 }
         else {
            jstart = nint(y0) - block_size
            jend = nint(spec_rng[1])
            jinc = -block_size
            xstart = x0
            if (isuccess .ne. 0) {
               xlast = Memr[xfit+ifirst-1]
               ylast = Memr[yfit+ifirst-1]
            }
         }

         # Start the main centroiding loop

         do j = jstart,jend,jinc {

            # Get limits of the block

            j1 = max(nint(spec_rng[1]),j - block_size/2)
            j2 = min(j + block_size/2,nint(spec_rng[2]))
            i1 = max(1,min(nx,nint(xstart)-2*spec_width))
            i2 = max(1,min(nx,nint(xstart)+2*spec_width))
            nix = i2 - i1 + 1

            # Zero profile array

            call amovkr(0.0,Memr[pptr],nix)

            # Get the mean seeing profile in the block

            do jj = j1,j2 {
               if (axis  == 2) {
                  ptr1 = imgs2r(iptr,i1,i2,jj,jj)}
               else {
                  ptr1 = cogetr(co1,jj,i1,i2)
               }
               call aaddr(Memr[pptr],Memr[ptr1],Memr[pptr],nix)
            }
            ok = false
            do jj = 1,nix {
               if (Memr[pptr+jj-1] > 0.0) {
                  ok = true
                  break
               }
            }

            # If the profile is ok,  then try to get the median position

            if (ok) {
               iblock = iblock + 1

               # Shift the profile over to temporary extended storage
               # and convolve with Gaussian

               ntot = nix + 2*iext 
               do jj = 1,iext
                  Memr[wpptr+jj-1] = Memr[pptr]
               do jj = nix+iext+1,ntot
                  Memr[wpptr+jj-1] = Memr[pptr+nix-1]
               call amovr(Memr[pptr],Memr[wpptr+iext],nix)
               call amovkr(0.0,Memr[cnvptr],ntot)
               call acnvrr(Memr[wpptr],Memr[cnvptr],nix,Memr[kptr],knpix)

               # Now shift smoothed profile back

               call amovr(Memr[cnvptr],Memr[pptr],nix)

               # Set up X array

               do jj = 1,nix
                  Memr[xptr+jj-1] = real(jj) + i1 - 1

               # Now do centring...

               status = 0
               iferr (call sam_peaks(Memr[xptr],Memr[pptr],nix,xsol,sigma)) {
                  if (errcode() == 0) status = 1
               }
               if (status == 0) {
                  ysol = 0.5*real(j1+j2)
                  isuccess = isuccess + 1
                  wsol = 1.0
                  if (isuccess != 1) {
                     slope = (xsol - xlast)/(ysol - ylast)
                     aint = xlast - slope*ylast
                     ynew = real(j+jinc)
                     xstart = ynew*slope + aint }
                  else {
                     xstart = xsol
                     ifirst = iblock
                  }
                  Memr[yfit+iblock-1] = ysol
                  Memr[xfit+iblock-1] = xsol
                  Memr[wfit+iblock-1] = wsol
                  xlast = xsol
                  ylast = ysol }
               else {
                  Memr[wfit+iblock-1] = 0.0
               }
            }
         } # End of main centroiding loop
      }

      # Normalise independent variable so they are all between 0 and 1

      nfit = iblock
      if (nfit == 0) call error(0,"No points to fit polynomial")
      call adivkr(Memr[yfit],ymax,Memr[yfit],nfit)

      # Get some workspace for the fitting routine
  
      nterms = nord + 1
      call malloc(wpt1,2*nterms-1,TY_DOUBLE)
      call malloc(wpt2,nterms,TY_DOUBLE)
      call malloc(wpt3,nterms*nterms,TY_DOUBLE)

      # Loop for the number of iteration cycles requested

      jstart = nint(spec_rng[1])
      icycle = 0
      jtot = (nint(spec_rng[2]) - nint(spec_rng[1]) + 1)
      nblock = jtot/block_size
      remain = nint((real(jtot)/real(block_size) -
         real(int(real(jtot)/real(block_size))))*block_size)
      if (remain != 0) nblock = nblock + 1
      while (icycle .lt. ncycle) {
         icycle = icycle + 1

         # Do the fit and reject points one at a time until no more
         # can be rejected

         nclip = 1
         while (nclip != 0) {
            call sam_poly(Memr[yfit],Memr[xfit],Memr[wfit],nfit,nterms,
               Memd[wpt1],Memd[wpt2],Memd[wpt3],coefs,chisqr)

            # Find out whether to clip a point, and if so, which one...

            nclip = 0
            nused = 0
            dev_max = -MAX_REAL
            sum = 0.0
            sumsq = 0.0
            do i = 1,nfit {
               if (Memr[wfit+i-1] != 0) {
                  nused = nused + 1
                  yy = Memr[yfit+i-1]
                  xx = apolr(yy,coefs,nterms)
                  dev = xx - Memr[xfit+i-1]
                  sum = sum + dev
                  sumsq = sumsq + dev*dev
                  if (abs(dev) .gt. dev_max) {
                     dev_max = abs(dev)
                     irem = i
                  }
               }
            }
            rms = sumsq/real(nused) - (sum/real(nused))**2
            if (rms >= 0.0) {
               rms = sqrt(rms)
               if (dev_max > clip*rms) {
                  nclip = 1
                  Memr[wfit+irem-1] = 0.0
               }
            }
         }

         # Right, having refined the fit, we now use the polynomial evaluation
         # as a first guess for a re-centring of the spectrum.  Don't bother
         # doing this for the last cycle, though...

         if (icycle != ncycle) {
            iblock = 0
            do n = 1,nblock {
               j1 = (n - 1)*block_size + jstart
               j2 = j1 + block_size - 1
               j2 = min(j2,nint(spec_rng[2]))
               

               # Get the value of Y at this point and evaluate the polynomial
               # to find X.  Don't forget that the coefs are relative to scaled
               # values of Y...

               yy = 0.5*real(j1 + j2)/ymax
               xx = apolr(yy,coefs,nterms)
               i1 = max(1,min(nx,nint(xx)-2*spec_width))
               i2 = max(1,min(nx,nint(xx)+2*spec_width))
               nix = i2 - i1 + 1
  
               # Zero profile array

               call amovkr(0.0,Memr[pptr],nix)

               # Sum up profiles...

               do jj = j1,j2 {
                  if (axis  == 2) {
                     ptr1 = imgs2r(iptr,i1,i2,jj,jj) }
                  else {
                     ptr1 = cogetr(co1,jj,i1,i2)
                  }
                  call aaddr(Memr[pptr],Memr[ptr1],Memr[pptr],nix)
               }
               ok = false
               do jj = 1,nix {
                  if (Memr[pptr] > 0.0) {
                     ok = true
                     break
                  }
               }

               # Now get the new centres
     
               if (ok) {
                  iblock = iblock + 1

                  # Shift the profile over to temporary extended storage
                  # and convolve with Gaussian

                  ntot = nix + 2*iext + 1
                  do jj = 1,iext
                     Memr[wpptr+jj-1] = Memr[pptr]
                  do jj = nix+iext+1,ntot
                     Memr[wpptr+jj-1] = Memr[pptr+nix-1]
                  call amovr(Memr[pptr],Memr[wpptr+iext],nix)
                  call amovkr(0.0,Memr[cnvptr],ntot)
                  call acnvrr(Memr[wpptr],Memr[cnvptr],nix,Memr[kptr],knpix)
               
                  # Now shift smoothed profile back

                  call amovr(Memr[cnvptr],Memr[pptr],nix)

                  # Set up X array

                  do jj = 1,nix
                     Memr[xptr+jj-1] = real(jj) + i1 - 1

                  # Now do centring...

                  status = 0
                  iferr (call sam_peaks(Memr[xptr],Memr[pptr],nix,xsol,sigma)){
                     if (errcode() == 0) status = 1
                  }
                  if (status == 0) {
                     ysol = yy
                     Memr[wfit+iblock-1] = 1.0
                     Memr[xfit+iblock-1] = xsol
                     Memr[yfit+iblock-1] = ysol
                  }
                  else {
                     Memr[wfit+iblock-1] = 0.0
                  }
               }
            } # End of block loop
            nfit = iblock
            if (nfit == 0) call error(0,"No points for fit")
         } 
      } # End cycle loop

      # Send a little message about the fit to the user

      call printf("Successful fit with reduced chi-squared = %9.4f \n")
         call pargr(chisqr)

      # Add some extra info into the coefs file

      coefs[nord+2] = spec_rng[1]
      coefs[nord+3] = spec_rng[2]
      coefs[nord+4] = 1.0/ymax

      # Now do the plot...

      if (frame != 0) {
         dp = imd_mapframe(frame,READ_WRITE,YES)
         iw = iw_open(dp,frame,Memc[imname],SZ_FNAME,wcs_status)
         ndx = IM_LEN(dp,1)
         ndy = IM_LEN(dp,2)
         do i = nint(spec_rng[1]),nint(spec_rng[2]) {
            yy = real[i]
            xx = apolr(yy/ymax,coefs,nterms)
            if (axis == 1) {
               call iw_im2fb(iw,yy,xx,fy,fx)
               ii = nint(fy)
               jj = nint(fx) }
            else {
               call iw_im2fb(iw,xx,yy,fx,fy)
               ii = nint(fx)
               jj = nint(fy)
            }
            if (ii >= 1 && ii <= ndx && jj >= 1 && jj <= ndy) {
               ptr1 = imps2s(dp,ii,ii,jj,jj) 
               Mems[ptr1] = colour
            }
         }
         call iw_close(iw)
         call imunmap(dp) }
      else {
         call eprintf("No frame specified -- line plot skipped\n")
      }  

      # Free workspace

      call sfree(sp)
      call mfree(wpptr,TY_REAL)
      call mfree(pptr,TY_REAL)
      call mfree(cnvptr,TY_REAL)
      call mfree(kptr,TY_REAL)
      call mfree(xptr,TY_REAL)
      call mfree(xfit,TY_REAL)
      call mfree(yfit,TY_REAL)
      call mfree(wfit,TY_REAL)
      call mfree(wpt1,TY_DOUBLE)
      call mfree(wpt2,TY_DOUBLE)
      call mfree(wpt3,TY_DOUBLE)

      # Unmap column buffers if they were needed
 
      if (axis == 1) call counmap(co1)
end

