include <error.h>

procedure check_proc(bias,sub_frame,bias_frame,flat,flat_div,flat_field,dark,
   dark_cor,dark_frame,prog,suffix,listb,listd,listf,listp)

#--------------------------------------------------------------------------------------
#      CHECK_PROC -- Check that processing instructions make sense
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) bias:       (bool) True if bias frames are being reduced
#      (>) sub_frame:  (bool) True if bias frame is being subtracted
#      (>) bias_frame: (char) Name of mean bias frame
#      (>) flat:       (bool) True if flat frames are being reduced
#      (>) flat_div:   (bool) True if dividing by flat field
#      (>) flat_field: (char) Name of mean flat frame
#      (>) dark:       (bool) True if dark frames are being reduced
#      (>) dark_cor:   (bool) True if dark correction is being done
#      (>) dark_frame: (char) Name of mean dark frame
#      (>) prog:       (bool) True if program frames are being reduced
#      (>) suffix:     (char) Suffix used to create output file names
#      (>) listb:      (int) Input bias file list descriptor
#      (>) listd:      (int) Input dark file list descriptor
#      (>) listf:      (int) Input flat file list descriptor
#      (>) listp:      (int) Input program file list descriptor
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        18-NOV-1992
#      UPDATE:      21-OCT-1993 -- Now checks that output files don't
#                                  already exist. (JRL)
#
#--------------------------------------------------------------------------------------

# Parameters

bool sub_frame,flat_div,dark_cor,bias,flat,dark,prog
char bias_frame[ARB],flat_field[ARB],dark_frame[ARB],suffix[ARB]
int listb,listd,listf,listp

# Local variables

pointer sp,filename
int nerr
bool exist,clobber

# Functions

int strlen(),imtgetim(),imtlen()

begin

      # Get workspace

      call smark(sp)
      call salloc(filename,SZ_FNAME,TY_CHAR)

      # Check for bias frame errors...

      nerr = 0
      if (bias || sub_frame) {

         # Check that a name has be given for the mean bias frame

         if (strlen(bias_frame) == 0) {
            call eprintf("***No mean bias frame specified***\n")
            nerr = nerr + 1
         }

         # If bias frames are being processed check that the output file
         # doesn't already exist.

         if (bias) {
            iferr(call rt_fstat(bias_frame,NEW_FILE,false,exist,clobber)) {
               call erract(EA_WARN)
               nerr = nerr + 1 
            }

            # Check that input bias frames have been specified

            if (imtlen(listb) == 0) {
               call eprintf("***No input bias frames specified***\n")
               nerr = nerr + 1
            } 

            # Check that the input bias frames exist

            else {
               while (imtgetim(listb,Memc[filename],SZ_FNAME) != EOF) {
                  iferr(call rt_fstat(Memc[filename],READ_ONLY,true,
                  exist,clobber)) {
                     call erract(EA_WARN)
                     nerr = nerr + 1
                  }
               }
               call imtrew(listb)
            }
         }

         # If bias frames are not being processed, but you still wish to 
         # subtract a bias frame, check that the specified mean frame exists

         else if (sub_frame) {
            iferr(call rt_fstat(bias_frame,READ_ONLY,true,exist,clobber)) {
               call erract(EA_WARN)
               nerr = nerr + 1
            }
         }
      }

      # Check flat field frame errors

      if (flat || flat_div) {
         if (strlen(flat_field) == 0) {
            call eprintf("***No mean flat field frame specified***\n")
            nerr = nerr + 1
         }

         # If flat frames are being processed check that the output file
         # doesn't already exist.

         if (flat) {
            iferr(call rt_fstat(flat_field,NEW_FILE,false,exist,clobber)) {
               call erract(EA_WARN)
               nerr = nerr + 1 
            } 

            # Check that input flat frames have been specified

            if (imtlen(listf) == 0) {
               call eprintf("***No input flat frames specified***\n")
               nerr = nerr + 1
            } 

            # Check that the input flat frames exist

            else {
               while (imtgetim(listf,Memc[filename],SZ_FNAME) != EOF) {
                  iferr(call rt_fstat(Memc[filename],READ_ONLY,true,
                  exist,clobber)) {
                     call erract(EA_WARN)
                     nerr = nerr + 1
                  }
               }
               call imtrew(listf)
            }
         }

         # If flat frames are not being processed but you still want to
         # divide by a flat, then check that that flat exists

         else if (flat_div) {
            iferr(call rt_fstat(flat_field,READ_ONLY,true,exist,clobber)) {
               call erract(EA_WARN)
               nerr = nerr + 1
            }
         }
      }

      # Check dark frame existence

      if (dark_cor || dark) {
         if (strlen(dark_frame) == 0) {
            call eprintf("***No mean dark frame specified***\n")
            nerr = nerr + 1
         }

         # If dark frames are being processed check that the output file
         # doesn't already exist.

         if (dark) {
            iferr(call rt_fstat(dark_frame,NEW_FILE,false,exist,clobber)) {
               call erract(EA_WARN)
               nerr = nerr + 1 
            } 

            # Check that input dark frames have been specified

            if (imtlen(listd) == 0) {
               call eprintf("***No input dark frames specified***\n")
               nerr = nerr + 1
            } 

            # Check that the input dark frames exist

            else {
               while (imtgetim(listd,Memc[filename],SZ_FNAME) != EOF) {
                  iferr(call rt_fstat(Memc[filename],READ_ONLY,true,
                  exist,clobber)) {
                     call erract(EA_WARN)
                     nerr = nerr + 1
                  }
               }
               call imtrew(listd)
            }
         }

         # If you are not processing dark frames, but still want to take
         # out the dark counts, then see if mean dark frame exists

         else if (dark_cor) {
            iferr(call rt_fstat(dark_frame,READ_ONLY,true,exist,clobber)) {
               call erract(EA_WARN)
               nerr = nerr + 1
            }
         }
      }

      # Check that suffix isn't a null string

      if (prog && strlen(suffix) == 0) {
         call eprintf("***No suffix specified***\n")
         nerr = nerr + 1
      }

      # Check output file names. If they exist and can't be clobbered
      # then signal an error

      if (prog) {
         while (imtgetim(listp,Memc[filename],SZ_FNAME) != EOF) {
            call strcat(suffix,Memc[filename],SZ_FNAME)
            iferr(call rt_fstat(Memc[filename],NEW_FILE,false,exist,clobber)) {
               call erract(EA_WARN)
               nerr = nerr + 1
            }
         }
         call imtrew(listp)
      }

      # Free workspace

      call sfree(sp)

      # If there have been errors, then signal send a final message here

      if (nerr > 0) call error(1,
         "Please check your parameters and start again")
end
