include <imhdr.h>
include <math/curfit.h>
include <error.h>

procedure onedflat(iptr,eptr,ex_err,npix,axis,nord,output)
#--------------------------------------------------------------------------------------
#      ONEDFLAT -- Do a fit to a 1d flat field frame, display it and
#                  create an output array with values to use for future
#                  division
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) iptr:       (pointer) Input data pointer
#      (>) eptr:       (pointer) Input variance pointer
#      (>) ex_err:     (bool) Flag for the existence of error information
#      (>) npix:       (int) Number of pixels in the spectral direction
#      (>) axis:       (int) The axis parallel to which the fit should be done.
#      (>) nord:       (int) Order of the Chebyshev polynomial to be fit
#      (<) output:     (real[npix]) Array of flat values to be used for division
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        11-FEB-1993 
#      UPDATE:
#
#--------------------------------------------------------------------------------------

# Parameters

pointer iptr,eptr
int npix,axis,nord
real output[npix]
bool ex_err

# Local variables

int maxbuf,nw,ni,nrows,i,ier
pointer co,coe,wpt1,wpt2,wpt3,ptr1,ptr2,gp,cv
real ave,ymin,ymax

# Functions

pointer comap(),imgl2r(),cogetr(),gopen()
real cveval(),asumr(),ahivr(),alovr()

extern fix_zerdiv

begin

      # If the axis of interest is the y axis, then prepare for column access

      if (axis == 2) {
         nrows = IM_LEN(iptr,2)
         maxbuf = 10*nrows
         co = comap(iptr,maxbuf)
         if (ex_err) coe = comap(eptr,maxbuf)
      }

      # Allocate some workspace for summations

      nw = npix
      call calloc(wpt1,nw,TY_REAL)
      call calloc(wpt2,nw,TY_REAL)
      call calloc(wpt3,nw,TY_REAL)

      # Initialise a few things...

      ni = IM_LEN(iptr,3-axis)
      do i = 1,npix {
         Memr[wpt3+i-1] = real(i)
      }

      # Now loop for each row/column and add all this up into the buffers

      do i = 1,ni {
         if (axis == 1) {
            ptr1 = imgl2r(iptr,i)
            if (ex_err) ptr2 = imgl2r(eptr,i) 
         }
         else {
            ptr1 = cogetr(co,i,1,nrows)
            if (ex_err) ptr2 = cogetr(coe,i,1,nrows)
         }
         if (ex_err) {
            call w_sum_frames(npix,Memr[ptr1],Memr[ptr2],Memr[wpt1],Memr[wpt2],
               Memr[wpt1],Memr[wpt2]) 
         }
         else {
            call sum_frames(npix,Memr[ptr1],Memr[wpt1],Memr[wpt2],Memr[wpt1],
               Memr[wpt2])
         }
      }

      # Get rid of column access workspace

      if (axis == 2) {
         call counmap(co)
         if (ex_err) call counmap(coe)
      }

      # Now take the average

      call advzr(Memr[wpt1],Memr[wpt2],Memr[wpt1],npix,fix_zerdiv)

      # Find the average of this array and divide it through so that the
      # average value in the array is one.

      ave = asumr(Memr[wpt1],npix)/real(npix)
      call adivkr(Memr[wpt1],ave,Memr[wpt1],npix)

      # Initialise the the fitting routine...

      call cvinit(cv,CHEBYSHEV,nord,1.0,real(npix))

      # Now do the fit.

      call cvfit(cv,Memr[wpt3],Memr[wpt1],Memr[wpt2],npix,WTS_UNIFORM,ier)

      # If the fit worked, then evaluate it over each point and plot the result

      switch (ier) {
      case OK:
         do i = 1,npix {
            output[i] = cveval(cv,Memr[wpt3+i-1])
         }
         ifnoerr (gp = gopen("stdgraph",NEW_FILE,STDGRAPH)) {
            ymax = ahivr(Memr[wpt1],npix) + 0.1
            ymin = alovr(Memr[wpt1],npix) - 0.1
            call gswind(gp,1.0,real(npix),ymin,ymax)
            call glabax(gp,"1d Flat field and fit","X","")
            call gpline(gp,Memr[wpt3],Memr[wpt1],npix)
            call gpline(gp,Memr[wpt3],output,npix)
            call gclose(gp) }
         else {
            call erract(EA_WARN)
         }
         call cvfree(cv)
      case SINGULAR:
         call cvfree(cv)
         call error(0,"Fit to 1d flat failed with a singular matrix") 
      case NO_DEG_FREEDOM:
         call cvfree(cv)
         call error(0,"Fit to 1d flat failed with too few data points for fit")
      }

      # Free some workspaces

      call mfree(wpt1,TY_REAL)
      call mfree(wpt2,TY_REAL)
      call mfree(wpt3,TY_REAL)
end
