include <error.h>

procedure read_chip(fd,log_descr,nx,ny,dsect,bsect,readout,gain)

#--------------------------------------------------------------------------------------
#      READ_CHIP -- Read the chip file and extract info
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) fd:         (int) Descriptor of chip file
#      (>) log_descr:  (int) Descriptor for log file
#      (<) nx:         (int) X dimension of chip 
#      (<) ny:         (int) Y dimension of chip
#      (<) dsect:      (char) Data section
#      (<) bsect:      (char) Bias section
#      (<) readout:    (real) Readout noise for chip in e-
#      (<) gain:       (real) Chip gain in e-/adu
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        18-NOV-1992
#      UPDATE:      11-FEB-1993 -- Traps errors in a better way (JRL)
#                   26-OCT-1993 -- Better trapping of image section errors (JRL)
#
#--------------------------------------------------------------------------------------

# Parameters

int fd,nx,ny,log_descr
real readout,gain
char dsect[ARB],bsect[ARB]

# Local variables

pointer sp,lp1
int iline,index,ilen,ncomma

# Functions

int stridxs(),strlen(),ctoi(),ctor()

define errexit_ 99

begin

      # Get some workspace

      call smark(sp)
      call salloc(lp1,SZ_LINE,TY_CHAR)

      # Skip the first line (comment line)

      iline = 0
      iferr (call chip_data(iline,fd,Memc[lp1])) {
         call erract(EA_WARN)
         goto errexit_
      }

      # Read chip name

      iferr (call chip_data(iline,fd,Memc[lp1])) {
         call erract(EA_WARN)
         goto errexit_
      }
      ilen = strlen(Memc[lp1])
      for (index = stridxs("!",Memc[lp1]); index <= ilen && Memc[lp1+index-1] != ' ';
         index = index + 1)                              # Find first non-white space
      call fprintf(log_descr,"Chip name is: ")
      if (index <= ilen) {
         call fprintf(log_descr,"%s \n")
            call pargstr(Memc[lp1+index-1])
      }

      # Get data array size

      iferr (call chip_data(iline,fd,Memc[lp1])) {
         call erract(EA_WARN)
         goto errexit_
      }
      index = stridxs("[",Memc[lp1]) + 1
      if (ctoi(Memc[lp1],index,nx) == 0) go to errexit_
      index = index + 1
      if (ctoi(Memc[lp1],index,ny) == 0) go to errexit_
      call fprintf(log_descr,"Chip dimensions are: %d %d \n")
         call pargi(nx)
         call pargi(ny)

      # Get data section

      iferr (call chip_data(iline,fd,Memc[lp1])) {
         call erract(EA_WARN)
         goto errexit_
      }
      call imgsection(Memc[lp1],dsect,SZ_LINE)
      ilen = strlen(dsect)
      ncomma = 0
      do index = 1,ilen 
         if (dsect[index] == ',') ncomma = ncomma + 1
      if (ilen == 0 || ncomma > 1) {
         call eprintf("Illegal image section for data region\n")
         goto errexit_
      }
      call fprintf(log_descr,"Chip data section is: %s\n")
         call pargstr(dsect)

      # Get bias section

      iferr (call chip_data(iline,fd,Memc[lp1])) {
         call erract(EA_WARN)
         goto errexit_
      }
      call imgsection(Memc[lp1],bsect,SZ_LINE) 
      ilen = strlen(bsect)
      ncomma = 0
      do index = 1,ilen
         if (bsect[index] == ',' ) ncomma = ncomma + 1
      if (ilen == 0 || ncomma > 1) {
         call eprintf("Illegal image section for bias strip\n")
         goto errexit_
      }
      call fprintf(log_descr,"Chip bias section is: %s\n")
         call pargstr(bsect)

      # Get readout noise

      iferr (call chip_data(iline,fd,Memc[lp1])) {
         call erract(EA_WARN)
         goto errexit_
      }
      index = stridxs("!",Memc[lp1]) + 1
      if (ctor(Memc[lp1],index,readout) == 0) goto errexit_
      call fprintf(log_descr,"Chip readout noise is: %8.2f \n")
         call pargr(readout)

      # Get gain

      iferr (call chip_data(iline,fd,Memc[lp1])) {
         call erract(EA_WARN)
         goto errexit_
      }
      index = stridxs("!",Memc[lp1]) + 1
      if (ctor(Memc[lp1],index,gain) == 0) goto errexit_
      call fprintf(log_descr,"Chip gain is: %8.2f \n")
         call pargr(gain)
      if (gain <= 0.0) {
         call eprintf("Chip gain is not allowed to be zero\n")
         goto errexit_
      }

      # Normal exit

      call sfree(sp)
      return

      # Error exit

errexit_
      call sprintf(Memc[lp1],SZ_LINE,
         "There is an error on line %d in your chip file")
         call pargi(iline)
      call error(1,Memc[lp1])
end

procedure chip_data(iline,fd,data)

#--------------------------------------------------------------------------------------
#      CHIP_DATA -- Read a line of data from the chip file
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (!) iline:      (int) Line of file you're currently on
#      (>) fd:         (int) Descriptor of chip file
#      (<) data:       (char) Output line of data
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        18-NOV-1992
#      UPDATE:      11-FEB-1993 -- Better error trapping used (JRL)
#
#--------------------------------------------------------------------------------------

# Parameters

int iline,fd
char data[SZ_LINE]

# Functions

int fscan()

begin

      # Increment the line number, try and read the file and copy the info over

      iline = iline + 1
      if (fscan(fd) == EOF) call error(0,"Premature end of file")
      call gargstr(data,SZ_LINE)
end
