procedure specflat(input,output,chipfile,speaxis,space1,space2)
#------------------------------------------------------------------------------
#
#   ---------------
#   S P E C F L A T
#   ---------------
#
#   Description
#   -----------
#   Take a number of spectral flat fields and create a mean
#
#   
#   Scope of program
#   ----------------
#   - Only 2d frames supported
#   - Data type irrlevant, output is REAL
#   - Error files not supported
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   CL script
#
#
#   Parameters
#   ----------
#   input      List of images with flats to be processed. Note that these
#              should be debiassed _before_ using them here. (character)
#              (prompted for)
#
#   output     The output file name. (character)(prompted for)
#
#   chipfile   This is the name of the SCP chip file used in debiassing the
#              input files. (character)(prompted for)
#
#   speaxis    This is the axis parallel to the dispersion direction. 
#              1 = x, 2 = y (int)(prompted for)
#
#   space1     This is the lower limit of the spatial domain to include in
#              the calculation.  If this is 0, then the lower limit is 
#              set to 1. (int)(prompted for)
#
#   space2     This is the upper limit of the spatial domain to include in
#              the calculation.  If this is 0, then the uperr limit is 
#              set to the maximum for the input frames. (int)(prompted for)
#
#   sig        This is the dispersion of the Gaussian kernel used for
#              smoothing the initial flat field (real)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   - Output is one file of type REAL, with dimensions the same as the
#     input files.
#
#   
#   Method
#   ------
#   - The routine combines the input flat frames into one averaged frame.
#   - The section of the averaged frame which excludes the overscan regions
#     (as defined from the chip file) and the spatial limits defined by the
#     user is then block averaged to form a 1d spectrum.
#   - The 1d spectrum is smoothed with a very mild gaussian kernel.
#   - The smoothed 1d spectrum is block replicated to create a 2d image with
#     the same spatial dimension as the input files and the spectral 
#     dimension which excludes any overscan regions.
#   - A dummy array of 1s is created with the same dimensions as the input
#     file.
#   - The 2d block replicated smoothed flat is copied into the dummy array
#     so that it is located in the data region defined in the chip file.
#   - The original combined file is divided by this 2d flat file to form
#     the final flat field.
#
#
#   External applications called
#   ----------------------------
#   Package artdata:
#      mkpattern
#
#   Package images:
#      blkavg
#      blkrep
#      gauss
#      imarith
#      imcombine
#      imcopy
#      imdelete
#
#
#   Possible future upgrades
#   ------------------------
#     
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   17-AUG-94     - Original program 
#
#
#------------------------------------------------------------------------------

# Query Parameters

string    input          {prompt="List of input flat frames"}
string    output         {prompt="Output mean flat frame"}
string    chipfile       {prompt="Name of chip file"}
int       speaxis        {1,min=1,max=2,prompt="Which is the spectral axis?"}
int       space1         {0,prompt="Lower spatial endpoint for analysis"}
int       space2         {0,prompt="Upper spatial endpoint for analysis"}
real      sig            {1.0,min=0.0,prompt="Sigma for Gaussian smooth"}
struct    *listf

begin

      # Local variables
 
      string infil,sumfil,final,data_reg,outfil
      int axis,ns,i1,i2,xx[2,2],yy[2,2],ii,jj,spa1,spa2,bb1,bb2,nx,ny
      struct lined
      real sigma
  
      # First combine all of the input frames into a temporary output frame

      sumfil = mktemp("SCP")
      infil = input
      final = output
      imcombine(infil,sumfil,plfile="",sigma="",
        logfile="STDOUT",combine="average",reject="none",project-,
        outtype="real",offsets="none",masktype="none",maskvalue=0.0,blank=0.0,
        scale="none",zero="none",weight="none",statsec="",expname="",
        lthreshold=INDEF,hthreshold=INDEF,nlow=1,nhigh=1,nkeep=1,mclip+,
        lsigma=3.0,hsigma=3.0,rdnoise="0.0",gain="1.0",snoise="0.0",
        sigscale=0.1,pclip=-0.5,grow=0,>& "dev$null")
      imgets(sumfil,"i_naxis1")
      nx = int(imgets.value)
      imgets(sumfil,"i_naxis2")
      ny = int(imgets.value)

      # Now get the limits of the data region for the block average

      listf = chipfile
      axis = speaxis
      while (fscan(listf,lined) != EOF) {
         if (substr(lined,1,11) == "Data Region") {
            ns = strlen(lined)
            i1 = stridx("[",lined) + 1
            i2 = stridx(":",lined) - 1
            xx[1,1] = int(substr(lined,i1,i2))
            i1 = i2 + 2
            i2 = stridx(",",lined) - 1
            xx[2,1] = int(substr(lined,i1,i2))
            i1 = i2 + 2
            i2 = stridx(":",substr(lined,i1,ns)) + i1 - 2
            xx[1,2] = int(substr(lined,i1,i2))
            i1 = i2 + 2
            i2 = stridx("]",lined) - 1
            xx[2,2] = int(substr(lined,i1,i2))
         }
      } 
      for (j = 1; j <= 2; j += 1)
         for (i = 1; i <= 2; i += 1)
            yy[i,j] = xx[i,j]
      spa1 = space1
      if (spa1 != 0) yy[1,3-axis] = spa1
      spa2 = space2
      if (spa2 != 0) yy[2,3-axis] = spa2
      data_reg = "["//yy[1,1]//":"//yy[2,1]//","//yy[1,2]//":"//yy[2,2]//"]"

      # Do the block average

      infil = sumfil
      outfil = mktemp("SCP")
      if (axis == 1) {
         bb1 = 1
         bb2 = yy[2,2] - yy[1,2] + 1
      } else {
         bb1 = yy[2,1] - yy[1,1] + 1
         bb2 = 1
      }
      blkavg(infil//data_reg,outfil,bb1,bb2,1,1,1,1,1,option="average")

      # Smooth the block averaged file

      infil = outfil
      outfil = mktemp("SCP")
      sigma = sig
      gauss(infil,outfil,sigma,ratio=1.0,theta=0.0,nsigma=4.0,bilinear+,
         boundary="nearest",constant=0.0)
      imdelete(infil,go+,ver-,def+)

      # Block replicate this smoothed file now.  Replicate it to the original
      # spatial dimensions...

      infil = outfil
      outfil = mktemp("SCP")
      if (axis == 1) {
         bb1 = 1
         bb2 = xx[2,2] - xx[1,2] + 1
      } else {
         bb1 = xx[2,1] - xx[1,1] + 1
         bb2 = 1
      }
      blkrep(infil,outfil,bb1,bb2,1,1,1,1,1)
      imdelete(infil,go+,ver-,def+)

      # Now create a dummy file with all ones. Make it the same size
      # as the original frames.

      infil = mktemp("SCP")
      mkpattern(infil,output=infil,pattern="constant",option="replace",
         v1=1.0,v2=1.0,size=1,title="",pixtype="real",ndim=2,
         ncols=nx,nlines=ny,n3=1,n4=1,n5=1,n6=1,n7=1,header="")

      # Now copy the block replicated, smoothed flat into the dummy file     

      data_reg = "["//xx[1,1]//":"//xx[2,1]//","//xx[1,2]//":"//xx[2,2]//"]"
      imcopy(outfil,infil//data_reg,ver-)
      imdelete(outfil,go+,ver-,def+)

     # Finally divide this flat into the original averaged array.  This
     # is the final flat field.

     imarith(sumfil,"/",infil,final,title="",divzero=0.0,hparams="",
        pixtype="",calctype="",ver-,noact-)
     imdelete(sumfil,go+,ver-,def+)
     imdelete(infil,go+,ver-,def+)
end
