include <imhdr.h>
include <error.h>

procedure t_chip()
#------------------------------------------------------------------------------
#
#   -------
#   C H I P
#   -------
#
#   Description
#   -----------
#   Write out the chip info for later use.
#
#   
#   Scope of program
#   ----------------
#   - Data dimensionality irrelevant
#   - Data type irrelevant
#   - Batch execution not supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   chipfile   Name of output text file where chip info will be stored. (string)
#              (prompted for)
#
#   comment    A line of comments for the chip file. (string)(prompted for)
#
#   chipname   A name for the chip. (string)(prompted for)
#
#   nx,ny      The physical dimensions of the chip. (integer)(prompted for)
#
#   data_reg   Region on the chip where there is valid data.  This should be
#              expressed as an iraf subsection. (string)(prompted for)
#
#   bias_reg   Region on the chip where the bias should be calculated.  This should
#              be expressed as an iraf subsection. (string)(prompted for)
#
#   readout    Readout noise of the chip in electrons. (double)(prompted for)
#
#   gain       Chip gain in electrons per adu. (double)(prompted for)
#
#   bcfile     File with a list of bad columns.  These should be listed one entry
#              per line. (string)(prompted for)
#
#   brfile     File with a list of bad rows.  These should be listed one entry per 
#              line. (string)(prompted for)
#
#   bpfile     File with a list of bad pixels (not part of a bad row or column).
#              These should be listed one entry per line. (string)(prompted for)
#
#
#   Propogation of data structure
#   -----------------------------
#   Not relevant
#
#   
#   Method
#   ------
#   - The user is prompted for various bits of information about the chip.  These
#     are written to the chip file in a standard format.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgeti
#      clgetr
#      clgstr
#
#   Library ETC:
#      erract
#      error
#
#   Library FIO:
#      close
#      open
#
#   Library FMTIO:
#      fprintf
#      fscan
#      gargi
#      pargi
#      pargr
#      pargstr
#      reset_scan
#      strcpy
#      streq
#      strncmp
#   
#   Library IMIO:
#      imgsection
#
#   Library MEMIO:
#      salloc
#      sfree
#      smark
#
#
#   Internal subroutines called
#   ---------------------------
#      set_bad
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <error.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   08-OCT-92     - Original program based on VAX program SCP_NOISE
#
#
#--------------------------------------------------------------------------------------

# Local variables

pointer   cp1     # Stack workspace pointer
pointer   cp2     # Stack workspace pointer
pointer   cp3     # Stack workspace pointer
int       fd      # File descriptor for output text file
real      gain    # Chip gain
int       nx      # X dimension of chip
int       ny      # Y dimension of chip
real      read    # Chip readout noise
pointer   sp      # Stack memory marker

# Functions

int       clgeti(),open(),strncmp()
real      clgetr()

begin

      # Mark the memory stack and allocate some space

      call smark(sp)
      call salloc(cp1,SZ_FNAME,TY_CHAR)
      call salloc(cp2,SZ_LINE,TY_CHAR)
      call salloc(cp3,SZ_LINE,TY_CHAR)

      # Get name of chip file and open it

      call clgstr("chipfile",Memc[cp1],SZ_FNAME)
      fd = open(Memc[cp1],NEW_FILE,TEXT_FILE)

      # Get a comment line and write it to the file
     
      call clgstr("comment",Memc[cp2],SZ_LINE)
      call fprintf(fd,"***%s***\n")
         call pargstr(Memc[cp2])
      
      # Get the chip name and write it to the file
      
      call clgstr("chipname",Memc[cp2],SZ_LINE)
      call fprintf(fd,"Chip Name! %s\n")
         call pargstr(Memc[cp2])

      # Get the dimensions of the chip and write them to the file

      nx = clgeti("nx")
      ny = clgeti("ny")
      call fprintf(fd,"Dimensions! [%d,%d]\n")
         call pargi(nx)
         call pargi(ny)
 
      # Get location of data region

      call clgstr("data_reg",Memc[cp2],SZ_LINE)
      call strcpy(Memc[cp2],Memc[cp3],SZ_LINE)
      iferr (call imgsection(Memc[cp3],Memc[cp2],SZ_LINE)) call erract(EA_ERROR)
      if (strncmp("[",Memc[cp2],1) != 0) {
         call error(0,"This is not a proper image section") }
      else {
         call fprintf(fd,"Data Region! %s\n")
            call pargstr(Memc[cp2])
      }

      # Get location of bias region

      call clgstr("bias_reg",Memc[cp2],SZ_LINE)
      call strcpy(Memc[cp2],Memc[cp3],SZ_LINE)
      iferr (call imgsection(Memc[cp3],Memc[cp2],SZ_LINE)) call erract(EA_ERROR)
      if (strncmp("[",Memc[cp2],1) != 0) {
         call error(0,"This is not a proper image section") }
      else {
         call fprintf(fd,"Bias Strip! %s\n")
            call pargstr(Memc[cp2])
      }

      # Get readout noise in electrons

      read = clgetr("readout")
      call fprintf(fd,"Readout noise! %-8.2f \n")
         call pargr(read)

      # Get gain in electrons/adu

      gain = clgetr("gain")
      call fprintf(fd,"Electrons per ADU! %-8.2f \n")
         call pargr(gain)

      # Get the bad columns

      call fprintf(fd,"Bad pixels!\n")
      iferr (call set_bad(nx,ny,"bcfile",Memc[cp1],fd,"C")) call erract(EA_WARN)

      # Get the bad rows

      iferr (call set_bad(nx,ny,"brfile",Memc[cp1],fd,"R")) call erract(EA_WARN)

      # Get bad individual pixels
      
      iferr (call set_bad(nx,ny,"bpfile",Memc[cp1],fd," ")) call erract(EA_WARN)

      # Close the chip file
    
      call close(fd)

      # Release the workspace

      call sfree(sp) 

end



procedure set_bad(nx,ny,parname,filename,fd,pre)
#--------------------------------------------------------------------------------------
#      SET_BAD -- Write the coords of bad columns, rows and
#                pixels to output buffer
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) nx:         (int) X dimension of chip
#      (>) ny:         (int) Y dimension of chip
#      (>) parname:    (char) Parameter name for the file to be opened
#      (W) filename:   (char) Character workspace for input file name
#      (>) fd:         (int) Output file descriptor
#      (>) pre:        (char) Prefix to indicate either row, column or pixel
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        08-OCT-1992
#      UPDATE:
#
#--------------------------------------------------------------------------------------

# Parameters

char filename[SZ_FNAME],parname[ARB],pre
int fd,nx,ny

# Local variables

int fd2,ix,iy,ix1,ix2,iy1,iy2
char message[SZ_LINE]

# Functions

int fscan(),open(),strlen()
bool streq()

define exit_ 99

begin

      # Start by getting the file name 

      call clgstr(parname,filename,SZ_FNAME)
      if (strlen(filename) == 0) goto exit_

      # If it exists, then open it

      ifnoerr (fd2 = open(filename,READ_ONLY,TEXT_FILE)) {

         # Read for as long as you can.  If this is a bad column or row, then
         # read only one pixel per line. If it's a bad individual pixel then read
         # two pixels per line

         while (fscan(fd2) != EOF) {
            call reset_scan()
            call gargi(ix)
            if (streq(pre," ")) call gargi(iy) 

            # Set limits now

            if (streq(pre,"R")) {
               ix1 = 1
               ix2 = nx
               iy1 = ix
               iy2 = ix }
            else if (streq(pre,"C")) {
               ix1 = ix
               ix2 = ix
               iy1 = 1
               iy2 = ny }
            else {
               ix1 = ix
               ix2 = ix
               iy1 = iy
               iy2 = iy
            }

            # Write this out now

            call fprintf(fd,"%d %d %d %d\n")
               call pargi(ix1)
               call pargi(ix2)
               call pargi(iy1)
               call pargi(iy2)
         }
        
         # Close the file when done

         call close(fd2) }

      # If there is an error opening the file, then send out a warning...

      else {
         call sprintf(message,SZ_LINE,"Could not open file %s")
            call pargstr(filename)
         call error(0,message)
      }
exit_
      return
end         
