include <error.h>
include <imhdr.h>

procedure t_cleanup()
#------------------------------------------------------------------------------
#
#   -------------
#   C L E A N U P
#   -------------
#
#   Description
#   -----------
#   Clean dud pixels out of a frame and do 16 bit correction
#
#   
#   Scope of program
#   ----------------
#   - Handles 2-D images only
#   - All data types converted to REAL
#   - Variance arrays NOT supported
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   input      List of input file names (string) (prompted for)
#
#   output     List of output file names (string) (prompted for)
#
#   sixteen    TRUE if you want to do the 16 bit correction (bool)(prompted for)
#
#   dud        TRUE if you want to remove dud pixels from a frame. (bool)
#              (prompted for)
#
#   chipfile   Name of file which chip information (string)(prompted for)
#
#
#   Propogation of data structure
#   -----------------------------
#   - Output file is identical in size, shape and data type to input file
#
#   
#   Method
#   ------
#   - If the 16 bit correction is being done, then the appropriate value is
#     is added to the negative values in the data array.
#   - If dud pixels are being removed, then the chip file is read to determine
#     where the bad pixels are.  
#   - If there are no bad pixels in the chip file, then the program exits
#   - For each row with a bad pixel, the routine takes an average of adjacent
#     good pixels for the replacement value.
#   - If there are no adjacent good pixels, then that pixel goes unrepaired.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgetb
#      clgstr
#
#   Library COSMETICS:
#      dud_mask_set
#      dud_pixels
#      sixteen_bit
#
#   Library ETC:
#      erract
#      error
#
#   Library FIO:
#      close
#      open
#
#   Library FMTIO:
#      fprintf
#      pargi
#      pargstr
#   
#   Library IMIO:
#      imflush
#      imgnlr
#      imgs2r
#      immap
#      impnlr
#      imtclose
#      imtgetim
#      imtlen
#      imtopen
#      imunmap
#
#   Library MEMIO:
#      calloc
#      malloc
#      mfree
#      salloc
#      sfree
#      smark
#
#   Library PREP:
#      check_size
#      read_chip
#      scplog
#
#   Library VOPS:
#      amovkl
#      amovr
#      amovs
#      asums
#
#   Internal subroutines called
#   ---------------------------
#      sub_cleanup
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <error.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   11-MAR-93     - Original program
#
#
#--------------------------------------------------------------------------------------


# Local variables

pointer    bsect              # Pointer for bias strip section
int        chip_descr         # Descriptor for chip file
pointer    chipfile           # Pointer for chip file name
pointer    dsect              # Pointer for data section
bool       dud                # TRUE if dud pixels are being replaced
pointer    dudm               # Pointer for dud pixel mask
real       gain               # CCD gain
pointer    ilist              # Pointer for input file list
int        listin             # Descriptor for input file list
int        listout            # Descriptor for output file list
int        log_descr          # Descriptor for log file
int        ndud               # Number of dud pixels in dud pixel mask
int        nx                 # X dimension of chip
int        ny                 # Y dimension of chip
pointer    olist              # Pointer for output file list
real       readout            # Chip readout noise
bool       sixteen            # TRUE if sixteen bit correction is being done
pointer    sp                 # Stack marker

# Functions

int imtopen(),imtlen(),scplog(),open()
bool clgetb()

begin

      # Get some stack workspace

      call smark(sp)
      call salloc(ilist,SZ_LINE,TY_CHAR)
      call salloc(olist,SZ_LINE,TY_CHAR)
      call salloc(bsect,SZ_LINE,TY_CHAR)
      call salloc(dsect,SZ_LINE,TY_CHAR)
      call salloc(chipfile,SZ_FNAME,TY_CHAR)

      # Get the input file list...

      call clgstr("input",Memc[ilist],SZ_LINE)
      listin = imtopen(Memc[ilist])

      # Get the output file list...
  
      call clgstr("output",Memc[olist],SZ_LINE)
      listout = imtopen(Memc[olist])

      # If the input/output lists aren't the same size, then get out of here

      if (imtlen(listin) != imtlen(listout)) call error(0,
         "Input and output lists have different sizes")

      # Open a logfile

      log_descr = scplog()

      # What do you want done?

      sixteen = clgetb("sixteen")
      dud = clgetb("dud")
      if (!dud && !sixteen) call error(0,"You don't appear to want to do anything")

      # Get chip file

      call clgstr("chipfile",Memc[chipfile],SZ_FNAME)
      iferr (chip_descr = open(Memc[chipfile],READ_ONLY,TEXT_FILE)) {
         call erract(EA_ERROR)
      }

      # Read the chip file...

      iferr (call read_chip(chip_descr,log_descr,nx,ny,Memc[dsect],Memc[bsect],
         readout,gain)) {
         call erract(EA_ERROR)
      }

      # Get the workspace for the dud pixel mask and set it

      call calloc(dudm,nx*ny,TY_SHORT)
      if (dud) call dud_mask_set(chip_descr,nx,ny,Mems[dudm],ndud)
      call close(chip_descr)
      if (ndud == 0 && !sixteen) call error(0,
         "No dud pixels were found in the chip file")

      # Do the work now...

      call fprintf(log_descr," \n")
      call fprintf(log_descr,"********************\n")
      call sub_cleanup(nx,ny,listin,listout,sixteen,dud,Mems[dudm],ndud,log_descr)

      # Close log file

      call close(log_descr)

      # Free workspace

      call imtclose(listin)
      call imtclose(listout)
      call mfree(dudm,TY_SHORT)
      call sfree(sp)
end


procedure sub_cleanup(nx,ny,listin,listout,sixteen,dud,dud_mask,ndud,log_descr)
#--------------------------------------------------------------------------------------
#      SUB_CLEANUP -- Do 16 bit correction and fix dud pixels
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) nx:         (int) X dimension of chip 
#      (>) ny:         (int) Y dimension of chip
#      (>) listin:     (int) Input list descriptor
#      (>) listout:    (int) Output list descriptor
#      (>) sixteen:    (bool) TRUE if the 16 bit correction is being done
#      (>) dud:        (bool) TRUE if dud pixels are to be fixed
#      (>) dud_mask:   (short[nx,ny]) Mask with location of dud pixel (1=bad, 0=good)
#      (>) ndud:       (int) Number of dud pixels in the mask
#      (>) log_descr:  (int) Descriptor for log file
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        11-MAR-1993
#      UPDATE:
#
#--------------------------------------------------------------------------------------

# Parameters

int nx,ny,ndud,log_descr,listin,listout
bool sixteen,dud
short dud_mask[nx,ny]

# Local variables

pointer sp,infile,outfile,iptr,optr,ptr1,ptr2,wpt1
int junk,nunable,nfixed,i,drow,iy1,iy2,nny,nf,nu,j
long one_l,v1[IM_MAXDIM],v2[IM_MAXDIM]

# Functions

pointer immap(),imgs2r()
int imtgetim(),imgnlr(),impnlr()
real asums()

begin

      # Get some workspace

      call smark(sp)
      call salloc(infile,SZ_FNAME,TY_CHAR)
      call salloc(outfile,SZ_FNAME,TY_CHAR)

      # Loop for the files in the list...

      one_l = 1
      while (imtgetim(listin,Memc[infile],SZ_FNAME) != EOF &&
         imtgetim(listout,Memc[outfile],SZ_FNAME) != EOF) {

         # Open the input and output files. Check the size of the input file

         iptr = immap(Memc[infile],READ_ONLY,0)
         iferr(call check_size(iptr,nx,ny)) {
            call erract(EA_WARN)
            call imunmap(iptr)
            next
         }
         optr = immap(Memc[outfile],NEW_COPY,iptr)

         # Now send some messages to the log file...

         call fprintf(log_descr,"Frame:  %s\n")
            call pargstr(Memc[infile])

         # First do the 16 bit correction, if that's been requested...

         if (sixteen) {
            call amovkl(one_l,v1,IM_MAXDIM)
            call amovkl(one_l,v2,IM_MAXDIM)
            while (imgnlr(iptr,ptr1,v1) != EOF) {
               junk = impnlr(optr,ptr2,v2)
               call sixteen_bit(nx,Memr[ptr1],Memr[ptr2])
            }
            call fprintf(log_descr,"    Sixteen bit correction applied\n")
            call imflush(optr)
            call imunmap(iptr) }

         # Otherwise, just copy the file straight over

         else {
            call amovkl(one_l,v1,IM_MAXDIM)
            call amovkl(one_l,v2,IM_MAXDIM)
            while (imgnlr(iptr,ptr1,v1) != EOF) {
               junk = impnlr(optr,ptr2,v2)
               call amovr(Memr[ptr1],Memr[ptr2],nx)
            }
            call imflush(optr)
            call imunmap(iptr)
         }

         # Do dud pixel correction if it's required...

         if (dud) {

            # Get some workspace and initialise some counters

            call malloc(wpt1,3*nx,TY_SHORT)
            nunable = 0
            nfixed = 0

            # Now loop for each row. Set up some more section offsets

            do i = 1,ny {

               # First check and see if there are any dud pixels in that row... If
               # there aren't any then just skip over that row

               if (nint(asums(dud_mask[1,i],nx)) == 0) next

               # Set up some variables which tell the subroutine where to look
               # in the subsection for the dud pixel

               if (i == 1) {
                  drow = 1
                  iy1 = 1
                  iy2 = 2
                  nny = 2 }
               else if (i == ny) {
                  drow = 2
                  iy1 = ny - 1
                  iy2 = ny
                  nny = 2 }
               else {
                  drow = 2
                  iy1 = i - 1
                  iy2 = i + 1
                  nny = 3
               }

               # Get the data sections and move the relevant part of the mask over

               ptr1 = imgs2r(optr,1,nx,iy1,iy2)
               do j = iy1,iy2 {
                  call amovs(dud_mask[1,j],Mems[wpt1+nx*(j-iy1)],nx)
               }
               call dud_pixels(nx,nny,Memr[ptr1],Mems[wpt1],drow,iy1,nf,
                  nu,optr)
               nfixed = nfixed + nf
               nunable = nunable + nu
            }

            # Ditch workspace

            call mfree(wpt1,TY_SHORT)
          
            # Friendly message...

            call fprintf(log_descr,"    Number of dud pixels removed: %d\n")
               call pargi(nfixed)
            if (nunable != 0) {
               call fprintf(log_descr,"    Unable to fix %d dud pixels\n")
                  call pargi(nunable)
            }
            call fprintf(log_descr,"    New file is called: %s\n")
               call pargstr(Memc[outfile])
         }

         # Close up the output file
       
         call imunmap(optr)

      # End of list loop

      }

      # Ditch workspace

      call sfree(sp)
end
