include <imhdr.h>
include <gset.h>
include <error.h>
include <math/nlfit.h>

# Macro definitions for starting values of parameters

define    STA_READ     1.0
define    STA_GAIN     1.0
define    STA_GRAIN    0.01
define    MAX_PAR      3

procedure t_noise()
#------------------------------------------------------------------------------
#
#   ---------
#   N O I S E
#   ---------
#
#   Description
#   -----------
#   Analyse a CCD frame in order to calculate the readout noise,
#   a/d conversion and grain noise
#
#   
#   Scope of program
#   ----------------
#   - Handles 2-D images only
#   - All data types converted to DOUBLE for calculations
#   - Batch execution not supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   image      Name of the CCD image (string)(prompted for)
#
#   xbox,ybox  Size of the analysis boxes in the x and y directions
#              (integer)(prompted for)
#
#   key        Flag to vary a particular parameter.  If you wish to vary a
#              parameter, then the value of key for that parameter should be
#              set to 1.  If you wish it held constant, then the value of
#              key for that parameter should be 0.  The order in which the
#              parameters go is: readout, gain, grain. This is read
#              as a string which is parsed a short array (string)(prompted for)
#
#   readout    Default readout noise in electrons (real)(prompted for)
#
#   gain       Default value for the CCD gain in electrons/adu. (real)(prompted for)
#
#   grain      Default value for the grain noise expressed as a percentage.
#              (real)(prompted for)
#
#   tol        Tolerance for fit (real) (hidden)
#
#   itmax      Maximum number of iterations for the fitting routine (int)
#              (hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   Not relevant
#
#   
#   Method
#   ------
#   - The user specifies the image and subset desired.
#   - The subset is divided in to rectangular boxes with dimensions specified also
#     by the user.  For each of these boxes a mean value of the noise and average 
#     signal is calculated
#   - The user can allow any of the three parameters to be varied and by 
#     specifying the values of those to be held constant.
#   - The "signal" for each pixel is defined as the average signal in a 3x3 square
#     around that pixel.
#   - The deviation for each pixel is defined as the absolute difference between
#     the central pixel value and the average defined above.  The noise for a given
#     box is then the average of the absolute deviations for all the pixels in that
#     box.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library CLIO:
#      clgeti
#      clgetr
#      clgstr
#
#   Library ETC:
#      erract
#      error
#      locpr
#
#   Library FMTIO:
#      eprintf
#      pargr
#      printf
#   
#   Library GIO:
#      gclose
#      gflush
#      glabax
#      gopen
#      gpline
#      gpmark
#      gswind
#
#   Library IMIO:
#      imgs2d
#      immap
#      imunmap
#
#   Library MEMIO:
#      malloc
#      mfree
#      salloc
#      sfree
#      smark
#
#   Library NLFIT:
#      nlerrorsr
#      nlfitr
#      nlfreer
#      nlinitr
#      nlpgetr
#      nlvectorr
#
#   Library RTOOLS:
#      rt_par_arrs
#
#   Library VOPS:
#      asums
#
#
#   Internal subroutines called
#   ---------------------------
#      dnmod
#      do_sums
#      nmod
#      sub_fit
#      sub_noise
#
#
#   INCLUDE statements
#   ------------------
#   include <imhdr.h>
#   include <gset.h>
#   include <error.h>
#   include <math/nlfit.h>
#   include <fset.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   05-OCT-92     - Original program based on VAX program SCP_NOISE
#   03-FEB-93     - Modified to use NLFIT and for general tidy. (JRL)
#   25-MAR-93     - Parameter KEY now read in as a string and parsed (JRL)
#
#
#--------------------------------------------------------------------------------------

# Local variables

real      coefs[3]            # Array with estimates of parameters
real      errs[3]             # Array with errors in parameters 
int       i                   # Integer counter
pointer   image               # File name pointer
pointer   imp                 # Input file name pointer
int       itmax               # Maximum number of iterations for fit
short     key[3]              # Integer flags for parameters to be varied
int       nbox                # Total number of boxes
int       nboxx               # Number of boxes in X direction
int       nboxy               # Number of boxes in Y direction
int       nread               # Number of values parsed from string for KEY parameter
int       ntot                # Total number of boxes with valid data
int       nx                  # X dimension of input frame
int       ny                  # Y dimension of input frame
pointer   sp                  # Stack marker
real      tol                 # Tolerance for fit
pointer   wpt1a               # Workspace pointer
pointer   wpt1b               # Workspace pointer
pointer   wpt2a               # Workspace pointer
pointer   wpt2b               # Workspace pointer
pointer   wpt3a               # Workspace pointer
pointer   wpt3b               # Workspace pointer
int       xbox                # Size of boxes in X direction
int       ybox                # Size of boxes in Y direction

# Functions

pointer   immap()
int       clgeti()
real      clgetr(),asums()

begin

      # Mark stack and get some workspace

      call smark(sp)
      call salloc(image,SZ_FNAME,TY_CHAR)

      # Open the file and get dimensional information

      call clgstr("image",Memc[image],SZ_FNAME)
      iferr (imp = immap(Memc[image],READ_ONLY,0)) call erract(EA_ERROR)
      if (IM_NDIM(imp) != 2) call error(0,"Only 2d images, please")
      nx = IM_LEN(imp,1)
      ny = IM_LEN(imp,2)

      # How big a box to use?

      xbox = clgeti("xbox")
      ybox = clgeti("ybox")

      # How many boxes in each direction does that make?
 
      nboxx = nx/xbox
      nboxy = ny/ybox
      nbox = nboxx*nboxy
      if (nbox <= 0) call error(0,"Box size is bigger than data array")

      # Get some workspace for calculating stats and for plotting

      call malloc(wpt2a,nbox,TY_REAL)
      call malloc(wpt2b,nbox,TY_REAL)

      # Specify which parameters are to be varied and give a default
      # value for the ones to be held constant

      iferr(call rt_par_arrs("key",MAX_PAR,nread,key)) call erract(EA_ERROR)
      do i = 1,3 {
         if (key[i] == 0) {
            switch (i) {
               case 1:
                  coefs[i] = clgetr("readout")
               case 2:
                  coefs[i] = clgetr("gain")
               case 3:
                  coefs[i] = 0.01*clgetr("grain")
            } }
         else {
            key[i] = 1   # Make sure everything is either 0 or 1
         }
      }
      
      # Give each of the parameters a starting value if it is being varied

      if (key[1] == 1) coefs[1] = STA_READ
      if (key[2] == 1) coefs[2] = STA_GAIN
      if (key[3] == 1) coefs[3] = STA_GRAIN

      # Check to make sure you're not doing something stupid

      if (coefs[2] == 0.0) call error(0,"Gain = 0.0? Get outta here!!")
 
      # If there are any variables, get the tolerance for the fit and maximum
      # number of iterations

      if (nint(asums(key,3)) > 0) {
         tol = clgetr("tol")
         itmax = clgeti("itmax")
      }

      # Do the noise stats now...

      iferr (call sub_noise(nx,ny,imp,xbox,ybox,nboxx,nboxy,nbox,ntot,Memr[wpt2a],
         Memr[wpt2b])) {
         call eprintf("Error calculating statistics\n")
         call erract(EA_ERROR)
      }

      # Check that there are boxes with data...
      
      if (ntot == 0) call error(0,"No boxes with good data")

      # Close the file and allocate some more workspace

      call imunmap(imp)
      call malloc(wpt1a,500,TY_REAL)
      call malloc(wpt1b,500,TY_REAL)
      call malloc(wpt3a,ntot,TY_REAL)
      call malloc(wpt3b,ntot,TY_REAL)

      # Do the fit to the noise model

      iferr (call sub_fit(ntot,Memr[wpt2a],Memr[wpt2b],key,tol,itmax,
         Memr[wpt1a],Memr[wpt1b],Memr[wpt3a],Memr[wpt3b],coefs,errs)) {
         call eprintf("Error in fitting routine")
         call erract(EA_ERROR)
      }

      # Write out the results 

      else {
         call printf("Readout noise: %.6g +/- %.6g electrons\n")
            call pargr(abs(coefs[1]))
            call pargr(errs[1])
         call printf("Gain: %.6g +/- %.6g electrons/data number \n")
            call pargr(coefs[2])
            call pargr(errs[2])
         call printf("Graininess: %.6g +/- %.6g percent\n")
            call pargr(coefs[3]*100.0)
            call pargr(100.0*errs[3])
      }

      # Free workspace

      call sfree(sp)
      call mfree(wpt1a,TY_REAL)
      call mfree(wpt1b,TY_REAL)
      call mfree(wpt2a,TY_REAL)
      call mfree(wpt2b,TY_REAL)
      call mfree(wpt3a,TY_REAL)
      call mfree(wpt3b,TY_REAL)
end


procedure sub_noise(nx,ny,imp,xbox,ybox,nboxx,nboxy,nbox,ntot,ave,sig)
#--------------------------------------------------------------------------------------
#      SUB_NOISE -- Compute the noise estimates
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) nx:         (int) X dimension of data 
#      (>) ny:         (int) Y dimension of data
#      (>) imp:        (pointer) Pointer for input data frame
#      (>) xbox:       (int) X dimension of box for computation
#      (>) ybox:       (int) Y dimension of box for computation
#      (>) nboxx:      (int) Number of boxes on a frame in X direction
#      (>) nboxy:      (int) Number of boxes on a frame in Y direction
#      (>) nbox:       (int) Total number of boxes on a frame 
#      (<) ntot:       (int) Total number of boxes with good data
#      (<) ave:        (real[nbox]) Array of average estimates
#      (<) sig:        (real[nbox]) Array of noise estimates
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        05-OCT-1992
#      UPDATE:
#
#--------------------------------------------------------------------------------------

# Parameters

int       nx,ny,xbox,ybox,nboxx,nboxy,nbox,ntot
real      ave[nbox],sig[nbox]
pointer   imp

# Local variables

int       i,j,yst,yfn,xst,xfn,ncont,nx2,ny2
pointer   iptr
double    dn,suma,sumr,sumr2

# Functions

pointer   imgs2d()

errchk imgs2d

begin
      # Initialise the box counter

      ntot = 0

      # Loop for each box in the area of interest

      do j = 1,nboxy {
         yst = real(j-1)*ybox + 2 
         yfn = yst + ybox - 1
         if (j == nboxy) yfn = ny - 1
         do i = 1,nboxx {
            xst = real(i-1)*xbox + 2 
            xfn = xst + xbox - 1
            if (i == nboxx) xfn = nx - 1

            # Get the image section you need

            iptr = imgs2d(imp,xst-1,xfn+1,yst-1,yfn+1)
            nx2 = xfn - xst + 2
            ny2 = yfn - yst + 2

            # Now do the necessary sums

            call do_sums(Memd[iptr],nx2,ny2,suma,sumr,sumr2,ncont)

            # Calculate the mean and rms for the box

            if (ncont != 0) {
               ntot = ntot + 1
               dn = double(ncont)
               ave[ntot] = real(suma/dn)
               sig[ntot] = real(sumr2/dn - sumr*sumr/(dn*dn))
            }
         }
      }
end

procedure do_sums(input,nx,ny,suma,sumr,sumr2,ncont)
#--------------------------------------------------------------------------------------
#      DO_SUMS -- Do necessary sums to for noise estimate computation
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) input:      (double[nx,ny]) Input data subset
#      (>) nx:         (int) X dimension of data subset
#      (>) ny:         (int) Y dimension of data subset
#      (<) suma:       (double) Sum of the data averages
#      (<) sumr:       (double) Sum of absolute residuals
#      (<) sumr2:      (double) Sum of squares of absolute residuals
#      (<) ncont:      (int) Number of pixels included in sums
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        05-OCT-1992
#      UPDATE:
#
#-------------------------------------------------------------------------------------

# Parameters

double    input[nx,ny],suma,sumr,sumr2
int       nx,ny,ncont

# Local variables

int       i1,j1,i2,j2,i,j,ii,jj,ninc
double    sumd,average,resid

begin

      # Get the array of indices to be used in the looping

      i1 = 2
      j1 = 2
      i2 = nx - 1 
      j2 = ny - 1 

      # Initialise a few things

      suma = 0.0; sumr = 0.0; sumr2 = 0.0
      ncont = 0 

      # Loop over each of the pixels within the main cell

      do j = j1,j2 {
         do i = i1,i2 {
            sumd = 0.0
            ninc = 0
            
            # Get the average around the pixel in question

            do jj = j-1,j+1 {
               do ii = i-1,i+1 {
                  if ((jj != j) || (ii != i)) {
                     sumd = sumd + input[ii,jj]
                     ninc = ninc + 1
                  }
               }  
            }

            # So long as there is at least one good pixel, then
            # finish doing the stats

            if (ninc != 0) {
               average = sumd/double(ninc)
               suma = suma + average
               resid = input[i,j] - average
               sumr = sumr + resid
               sumr2 = sumr2 + resid**2
               ncont = ncont + 1
            }
         }
      }
end

procedure sub_fit(ntot,ave,sig,key,tol,itmax,work1,work2,work3,work4,coefs,errs)
#--------------------------------------------------------------------------------------
#      SUB_FIT -- Do a fit to the noise estimate data
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) ntot:       (int) Total number of boxes with good data
#      (>) ave:        (real[nbox]) Array of average estimates
#      (>) sig:        (real[nbox]) Array of noise estimates
#      (>) key:        (short[3]) Array of keys to tell which parameters are varied
#      (>) tol:        (real) Tolerance for fit
#      (>) itmax:      (int) Maximum number of iterations
#      (W) work1:      (real[500]) Workspace for plotting 
#      (W) work2:      (real[500]) Workspace for plotting
#      (W) work3:      (real[ntot]) Workspace for both fitting and plotting 
#      (W) work4:      (real[ntot]) Workspace for both fitting and plotting
#      (!) coefs:      (real[3]) Input and output values for solution vector
#      (<) errs:       (real[3]) Values for solution vector errors
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        05-OCT-1992
#      UPDATE:      03-FEB-1993 -- Modified to use NLFIT routines. (JRL)
#
#--------------------------------------------------------------------------------------

# Parameters

int      ntot,itmax
short    key[3]
real     ave[ntot],sig[ntot],work1[500],work2[500],work3[ntot],work4[ntot],
         coefs[3],tol,errs[3]

# Local variables

int      m,i,ier,fnc,dfnc,nfpar,list[3],npar
pointer  gp,nl,wpt1,wpt2
real     xmin,xmax,ymin,ymax,av,si,dx,x,dcoefs[3],var,chisqr,temperrs[3]
bool     do_fit

# Functions

pointer  gopen()
int locpr()
real asums()

# External routines

extern   nmod,dnmod

errchk gopen,nlinitr

begin
      
      # Begin by plotting the results of the statistical analysis.
      # First take the log of everything...

      m = 0
      xmin = 0.0
      xmax = -1.0
      ymin = 0.0
      ymax = -1.0
      do i = 1,ntot {
         av = ave[i]
         si = sig[i]
         if ((av > 0.0) && (si > 0.0)) {
            m = m + 1
            av = log10(av)
            si = 0.5*log10(si)
            work3[m] = av
            work4[m] = si
            xmax = max(xmax,av)
            ymax = max(ymax,si)
         }
      }
      xmax = real(int(xmax) + 1)
      ymax = real(int(ymax) + 1)
      
      # Open a graphics device

      gp = gopen("stdgraph",NEW_FILE,STDGRAPH)
 
      # Set the window

      call gswind(gp,xmin,xmax,ymin,ymax)

      # Set up the axes

      call glabax(gp,"Noise Profile","<X>","Sigma")

      # Now plot the damn thing

      call gpmark(gp,work3,work4,m,GM_POINT,3.0,3.0)
      call gflush(gp)

      # Only do this next section if you have any parameters to vary!

      call amovkr(0.0,errs,3)
      do_fit = false
      if (nint(asums(key,3)) > 0) do_fit = true
      if (do_fit) {

         # Convert the values back to non-logarithmic

         do i = 1,m {
            work3[i] = 10.0**work3[i]
            work4[i] = 10.0**(2.0*work4[i])
         }

         # Set up some things for nlfit

         nfpar = 0
         do i = 1,3 {
            if (key[i] == 1) {
               nfpar = nfpar + 1
               list[nfpar] = i
            }
            dcoefs[i] = 0.0
         }
         fnc = locpr(nmod)
         dfnc = locpr(dnmod)
         npar = 3

         # Get a bit of workspace for fitting routines

         call malloc(wpt1,m,TY_REAL)
         call malloc(wpt2,m,TY_REAL)

         # Do the fit now...

         call nlinitr(nl,fnc,dfnc,coefs,dcoefs,npar,list,nfpar,tol,itmax)
         call nlfitr(nl,work3,work4,Memr[wpt1],m,1,WTS_UNIFORM,ier)

         # Did it converge? If it did, then extract the coefs and their errors.
         # If it didn't then report back with the proper error message

         switch (ier) {
         case DONE:
            call nlpgetr(nl,coefs,npar)
            call nlvectorr(nl,work3,Memr[wpt2],m,1)
            call nlerrorsr(nl,work4,Memr[wpt2],Memr[wpt1],m,var,chisqr,temperrs)
            call printf("Fit converged with error: %.4g\n")
               call pargr(var**0.25) # 0.25 because its a variance of a variance
            nfpar = 0
            do i = 1,3 {
               if (key[i] == 1) {
                  nfpar = nfpar + 1
                  errs[i] = temperrs[nfpar] }
               else {
                  errs[i] = 0.0
               } 
            }                 
         case SINGULAR:
            call nlfreer(nl)
            call error(0,"Fit failed -- singular matrix")
         case NO_DEG_FREEDOM:
            call nlfreer(nl)
            call error(0,"Fit failed -- Too few data points")
         case NOT_DONE:
            call nlfreer(nl)
            call error(0,"Fit failed -- maximum number of iterations reached")
         }
      }

      # Plot the solution...
      
      xmin = 10.0**xmin
      xmax = 10.0**xmax
      dx = (xmax - xmin)/499.0
      do i = 1,500 {
         x = xmin + real(i-1)*dx
         work1[i] = log10(x)
         si  = (coefs[1]/coefs[2])**2 + x/coefs[2] + (coefs[3]*x)**2
         work2[i] = 0.5*log10(si)
      }
      call gpline(gp,work1,work2,500)

      # Close it down

      call gclose(gp)

      # Release nlfit workspace

      if (do_fit) call nlfreer(nl)

end

procedure nmod(x,nvars,params,nparams,zfit)
#--------------------------------------------------------------------------------------
#      NMOD -- Calcuate the value of the noise model at a point
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) x:          (real[nvars]) Point at which to calculate model
#      (>) nvars:      (int) Number of independent variables
#      (>) params:     (real[nparams]) Coefficients for the fit
#      (>) nparams:    (int) Number of coefficients for the fit
#      (<) zfit:       (real) Output value
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        03-FEB-1993
#      UPDATE:
#
#--------------------------------------------------------------------------------------

# Parameters

int nvars,nparams
real x[nvars],params[nparams],zfit

# Local variables

real p12,p2,p3

begin

      # Calculate the model

      if (params[2] == 0.0) call error(0,"Gain has become zero...")
      p12 = (params[1]/params[2])**2
      p2 = 1.0/params[2]
      p3 = params[3]**2
      zfit = p12 + p2*x[1] + p3*x[1]**2
end

procedure dnmod(x,nvars,params,dparams,nparams,zfit,derivs)
#--------------------------------------------------------------------------------------
#      NMOD -- Calcuate the value of the noise model at a point
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) x:          (real[nvars]) Point at which to calculate model
#      (>) nvars:      (int) Number of independent variables
#      (>) params:     (real[nparams]) Coefficients for the fit
#      (>) dparams:    (real[nparams]) Step size for each coef (not used)
#      (>) nparams:    (int) Number of coefficients for the fit
#      (<) zfit:       (real) Output value
#      (<) derivs:     (real[nparams]) First derivatives of the function
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        03-FEB-1993
#      UPDATE:
#
#--------------------------------------------------------------------------------------

# Parameters

int nvars,nparams
real x[nvars],params[nparams],dparams[nparams],zfit,derivs[nparams]

# Local variables

real p12,p2,p3,p22

begin

      # Calculate the model

      if (params[2] == 0.0) call error(0,"Gain has become zero...")
      p12 = (params[1]/params[2])**2
      p2 = 1.0/params[2]
      p3 = params[3]**2
      zfit = p12 + p2*x[1] + p3*x[1]**2

      # Now the derivatives...

      p22 = params[2]**2
      p12 = 2.0*params[1]/p22
      p2 = -2.0*params[1]**2/(p22*params[2])
      p3 = 2.0*params[3]
      derivs[1] = p12
      derivs[2] = p2 - x[1]/p22
      derivs[3] = p3*x[1]*x[1]
end
