include <imhdr.h>
include <error.h>
include <imio.h>

procedure t_proc()
#------------------------------------------------------------------------------
#
#   -------
#   P R O C
#   -------
#
#   Description
#   -----------
#   Main routine used to process CCD data
#
#   
#   Scope of program
#   ----------------
#   - Handles 2-D images only
#   - Output data type will always be REAL
#   - Variance arrays supported
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   bias        TRUE if bias frames are being reduced (bool)(prompted for)
#
#   dark        TRUE if dark frames are being reduced (bool)(prompted for)
#
#   flat        TRUE if flat frames are being reduced (bool)(prompted for)
#
#   prog        TRUE if program frames are being reduced (bool)(prompted for)
#
#   chipfile    Name of file with chip info as created by scp.chip. (char)
#               (not prompted for)
#
#   bias_frames List of bias frames to be reduced. (char)(prompted for)
#
#   dark_frames List of dark frames to be reduced. (char)(prompted for)
#
#   flat_frames List of flat frames to be reduced. (char)(prompted for)
#
#   prog_frames List of program frames to be reduced. (char)(prompted for)
#
#   exphead     Location in header for exposure time.  If blank, then this
#               defaults to "exptime". (char)(hidden)
#
#   meanbias    File name for mean bias frame. (char)(hidden)
#
#   sub_bias    TRUE if a bias estimate is to be subtracted. (bool)(hidden)
#
#   sub_frame   TRUE if the bias estimate is from a bias frame (bool)(hidden)
#
#   overscan    TRUE if the bias frame subtracted is to be scaled by the mean
#               bias value in the overscan regions. (bool)(hidden)
#
#   meandark    File name for mean dark frame. (char)(hidden)
#
#   dark_cor    TRUE if correction for dark counts is being done. (bool)
#               (hidden)
#
#   meanflat    File name for mean flat frame. (char)(hidden)
#
#   flat_div    TRUE if program frames are divided by flat field (bool)
#               (hidden)
#
#   trim        TRUE if overscan region is to be trimmed from program frames.
#               (bool)(hidden)
#
#   want_err    TRUE if you want error files for output data files. (bool)
#               (hidden)
#
#   suffix      Suffix to tack onto input file name to form output file name for
#               program frames (character)(hidden)
#
#
#   sixteen     TRUE if the 16 bit correction is to be applied. (bool) 
#               (hidden)
#
#   repval      Replacement value used for divisor in case of an attempted
#               division by zero. (real)(hidden)
#
#
#   Propogation of data structure
#   -----------------------------
#   - For mean bias, dark or flat frames, the mean output file is defined by
#     the size of the chip as defined in the chip file.  The output type is
#     always real.
#   - For output program files, the size is based on the chip size and whether
#     or not the user has decided to trim the bias areas or not. Again the output
#     type is always real.
#
#   
#   Method
#   ------
#   - Bias frames are averaged and a mean bias frame is written out.  If 
#     desired, the mean value of the overscan region of this file is calculated.
#   - Dark, flat and program frames can all be bias subtracted.
#   - Flat and program frames can be dark corrected
#   - Only program frames can be flat fielded.
#   - Mean dark frames are normalised to be in units of dark counts/pix/sec.
#   - Mean flat frames are normalised so that the average in the data section is 1
#   - Only program frames can be trimmed.
#   - A log file called scp_proc.lis is written out with all sorts of useful info.
#
#
#   External functions & subroutines called
#   ---------------------------------------
#   Library ARITH:
#      fix_zerdiv
#      stats
#      sum_frames
#      var_div
#      w_sum_frames
#
#   Library CLIO:
#      clgetb
#      clgetr
#      clgstr
#
#   Library COSMETIC:
#      sixteen_bit
#
#   Library ERRPACK:
#      err_close
#      err_exist
#      err_fname
#      err_imdel
#      err_open
#
#   Library ETC:
#      erract
#      errcode
#
#   Library FIO:
#      close
#      flush
#      open
#
#   Library FMTIO:
#      eprintf
#      fprintf
#      pargi
#      pargr
#      pargstr
#      printf
#      strcat
#      strcpy
#      streq
#      strlen
#   
#   Library IMIO:
#      imflush
#      imgnlr
#      imgs2r
#      immap
#      impl2r
#      impnlr
#      imtclose
#      imtgetim
#      imtopen
#      imunmap
#
#   Library MEMIO:
#      calloc
#      malloc
#      mfree
#      salloc
#      sfree
#      smark
#
#   Library PREP:
#      check_proc
#      check_size
#      cre_var
#      get_exp
#      open_mean
#      prep_os
#      read_chip
#      scplog
#
#   Library RTOOLS:
#      rt_crfile
#
#   Library VOPS:
#      adivkr
#      advzr
#      amovkl
#      amovkr
#      amovr
#      arczr
#      asubkr
#      awsur
#
#
#   Internal subroutines called
#   ---------------------------
#      proc_bias
#      proc_frame
#
#
#   INCLUDE statements
#   ------------------
#   include <error.h>
#   include <imhdr.h>
#   include <imio.h>
#
#
#   Possible future upgrades
#   ------------------------
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   18-NOV-92     - Original program based on VAX FIGARO program SCP_PROC
#   11-FEB-93     - Modifications mainly to subroutines. (JRL)
#   10-MAR-93     - Removed cosmic ray and dud pixel fixing.  Made changes
#                   to internal routines to speed things up a bit (JRL)
#   09-DEC-93     - Removed 1d flat fielding routines. (JRL)
#
#
#------------------------------------------------------------------------------

# Local variables

bool       bias               # TRUE if processing bias frames
pointer    bsect              # Bias strip data section
int        chip_descr         # Descriptor for chip file
pointer    chipfile           # Pointer for chip file name
bool       dark               # TRUE if processing dark frames
bool       dark_cor           # TRUE if frames are to be dark corrected
pointer    dsect              # Data section in input frame
char       exphead[SZ_FNAME]  # Location in header of exposure time
real       fbias              # Bias strip value of mean bias frame
bool       flat               # TRUE if processing flat field frames
bool       flat_div           # TRUE if frames will be flat field corrected
real       gain               # CCD chip gain
pointer    imtlist            # Input file list
int        listb              # List descriptor for input bias frames
int        listd              # List descriptor for input dark frames
int        listf              # List descriptor for input flat frames
int        listp              # List descriptor for input prog frames
int        log_descr          # Descriptor for log file
int        nx                 # X dimension of chip
int        ny                 # Y dimension of chip
int        nzero              # Number of zero divides which have occurred
pointer    out_bias           # Pointer for output mean bias name
pointer    out_dark           # Pointer for output mean dark name
pointer    out_flat           # Pointer for output mean flat name
bool       overscan           # TRUE if overscan correction is applied to bias
bool       prog               # TRUE if processing program frames
real       readout            # Chip readout noise
real       repval             # Replacement value in case of zero divide
bool       sixteen            # TRUE if 16 bit correction is being done
pointer    sp                 # Stack memory marker
bool       sub_bias           # TRUE if a bias estimate is being subtracted
bool       sub_frame          # TRUE if bias estimate is in form of a 2d frame
pointer    suffix             # Suffix to attach to input name to form output name
bool       trim               # TRUE if overscan regions are to be trimmed away
bool       want_err           # TRUE if you want an error array

common /replace/ repval,nzero

# Functions

int open(),scplog(),strlen(),imtopen()
real clgetr()
bool clgetb()

begin

      # Get some stack memory

      call smark(sp)
      call salloc(out_bias,SZ_FNAME,TY_CHAR)
      call salloc(out_flat,SZ_FNAME,TY_CHAR)
      call salloc(out_dark,SZ_FNAME,TY_CHAR)
      call salloc(suffix,SZ_FNAME,TY_CHAR)
      call salloc(chipfile,SZ_FNAME,TY_CHAR)
      call salloc(dsect,SZ_LINE,TY_CHAR)
      call salloc(bsect,SZ_LINE,TY_CHAR)
      call salloc(imtlist,SZ_LINE,TY_CHAR)

      # Open up the log file

      log_descr = scplog()

      # Set some defaults...

      sub_frame = false

      # Get all the parameters you want... Start by asking what type of files
      # you want to reduce on this particular run

      bias = clgetb("bias")
      dark = clgetb("dark")
      flat = clgetb("flat")
      prog = clgetb("prog")

      # Open input file lists...

      if (bias) {
         call clgstr("bias_frames",Memc[imtlist],SZ_LINE)
         listb = imtopen(Memc[imtlist])
      }
      if (dark) {
         call clgstr("dark_frames",Memc[imtlist],SZ_LINE)
         listd = imtopen(Memc[imtlist])
      }
      if (flat) {
         call clgstr("flat_frames",Memc[imtlist],SZ_LINE)
         listf = imtopen(Memc[imtlist])
      }
      if (prog) {
         call clgstr("prog_frames",Memc[imtlist],SZ_LINE)
         listp = imtopen(Memc[imtlist])
      }

      # Where is the exposure time information?

      call clgstr("exphead",exphead,SZ_FNAME)
      if (strlen(exphead) == 0) call strcat("exptime",exphead,SZ_FNAME)

      # Read the bias info

      sub_bias = clgetb("sub_bias")
      if (sub_bias) {
         sub_frame = clgetb("sub_frame")
         overscan = clgetb("overscan")
      }
      if (bias || sub_frame) {
         call clgstr("meanbias",Memc[out_bias],SZ_FNAME)
      }

      # Read in value of divisor to be used in case of zero divides...

      repval = clgetr("repval")

      # Read dark frame info

      dark_cor = clgetb("dark_cor")
      if (dark || dark_cor) {
         call clgstr("meandark",Memc[out_dark],SZ_FNAME)
      }

      # Read flat field info

      flat_div = clgetb("flat_div")
      if (flat || flat_div) call clgstr("meanflat",Memc[out_flat],SZ_FNAME)

      # Miscellaneous questions about program frames

      sixteen = clgetb("sixteen")
      trim = clgetb("trim")
      want_err = clgetb("want_err")
      call clgstr("suffix",Memc[suffix],SZ_FNAME)

      # Check to make sure all these things make sense

      iferr (call check_proc(bias,sub_frame,Memc[out_bias],flat,flat_div,
         Memc[out_flat],dark,dark_cor,Memc[out_dark],prog,Memc[suffix],
         listb,listd,listf,listp)) call erract(EA_ERROR)

      # Get the chip file

      call clgstr("chipfile",Memc[chipfile],SZ_FNAME)
      iferr (chip_descr = open(Memc[chipfile],READ_ONLY,TEXT_FILE)) {
         call eprintf("***Error opening chip file***\n")
         call erract(EA_ERROR)
      }

      # Read the chip file...

      call fprintf(log_descr,"The chip file was: %s \n")
         call pargstr(Memc[chipfile])
      iferr (call read_chip(chip_descr,log_descr,nx,ny,Memc[dsect],Memc[bsect],
         readout,gain)) {
         call erract(EA_ERROR)
      }
      call close(chip_descr)

      # Ok, now start with the bias frames, if you have any...

      if (bias) {
         call proc_bias(listb,nx,ny,Memc[bsect],log_descr,Memc[out_bias],
            fbias)
      }

      # Do a bit of preliminary analysis on mean bias frames if is really needed

      if (dark || flat || prog) {
         call prep_os(nx,ny,bias,overscan,Memc[out_bias],Memc[bsect],
            log_descr,fbias)
      }

      # Next process your dark frames

      if (dark) {
         call proc_frame(nx,ny,listd,"d",exphead,sub_bias,
            sub_frame,overscan,Memc[out_bias],fbias,dark_cor,Memc[out_dark],
            flat_div,Memc[out_flat],Memc[dsect],Memc[bsect],
            sixteen,trim,want_err,Memc[suffix],log_descr,readout,gain)
      }

      # Process flat frames

      if (flat) {
         call proc_frame(nx,ny,listf,"f",exphead,sub_bias,
            sub_frame,overscan,Memc[out_bias],fbias,dark_cor,Memc[out_dark],
            flat_div,Memc[out_flat],Memc[dsect],Memc[bsect],
            sixteen,trim,want_err,Memc[suffix],log_descr,readout,gain)
      }

      # Process program frames

      if (prog) {
         call proc_frame(nx,ny,listp,"p",exphead,sub_bias,
            sub_frame,overscan,Memc[out_bias],fbias,dark_cor,Memc[out_dark],
            flat_div,Memc[out_flat],Memc[dsect],Memc[bsect],
            sixteen,trim,want_err,Memc[suffix],log_descr,readout,gain)
      }

      # Tidy and exit

      call close(log_descr)
      call sfree(sp)
end


procedure proc_bias(inlist,nx,ny,bsect,log_descr,bias_frame,fbias)

#------------------------------------------------------------------------------
#      PROC_BIAS -- Process the bias frames
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) inlist:     (int) Input file list
#      (>) nx:         (int) X dimension of chip 
#      (>) ny:         (int) Y dimension of chip
#      (>) bsect:      (char) Bias section
#      (>) log_descr:  (int) Descriptor for log file
#      (>) bias_frame: (char) Name of output mean bias frame
#      (<) fbias:      (real) Value of average bias in overscan region of output frame
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        18-NOV-1992
#      UPDATE:      09-MAR-1993 -- Slightly modified to speed things up a little (JRL)
#
#------------------------------------------------------------------------------

# Parameters

char bsect[ARB],bias_frame[ARB]
int nx,ny,log_descr,inlist
real fbias

# Local variables

pointer sp,lp1,out,ptr1,optr,wptr,inp,ptr2,ptr3
long one_l,v1[IM_MAXDIM],v2[IM_MAXDIM],v3[IM_MAXDIM]
int length,length2,npix,nrow,junk,nzero,nfile
real sig,repval

common /replace/ repval,nzero

# Functions

pointer immap(),imgs2r()
int imtgetim(),strlen(),errcode(),impnlr(),imgnlr()

extern fix_zerdiv

begin

      # Get some stack workspace

      call smark(sp)
      call salloc(lp1,SZ_FNAME,TY_CHAR)

      # Open an output file for the mean bias frame

      out = immap(bias_frame,NEW_IMAGE,0)
      IM_NDIM(out) = 2
      IM_LEN(out,1) = nx
      IM_LEN(out,2) = ny
      IM_PIXTYPE(out) = TY_REAL

      # Get a bit of zeroed workspace

      call calloc(wptr,nx*ny,TY_REAL)

      # Write some stuff out to the log file

      call fprintf(log_descr," \n")
      call fprintf(log_descr,"*****  Bias Frame Processing  *****\n")
      call fprintf(log_descr," \n")
      call fprintf(log_descr,"The following files were processed: \n")
      length = 0
      nzero = 0
      nfile = 0

      # Loop now for each frame

      while (imtgetim(inlist,Memc[lp1],SZ_FNAME) != EOF) {

         # Open the file

         inp = immap(Memc[lp1],READ_ONLY,0)
         iferr(call check_size(inp,nx,ny)) {
            call erract(EA_WARN)
            next
         }
         nfile = nfile + 1

         # Add the data into a buffer, zeroing initial buffer as you go along

         one_l = 1
         call amovkl(one_l,v1,IM_MAXDIM)
         call amovkl(one_l,v2,IM_MAXDIM)
         call amovkl(one_l,v3,IM_MAXDIM)
         while (imgnlr(inp,ptr1,v1) != EOF) {
            junk = impnlr(out,ptr3,v3)
            if (nfile == 1) {
               ptr2 = ptr3
               call amovkr(0.0,Memr[ptr2],nx) }
            else {
               junk = imgnlr(out,ptr2,v2) 
            }
            call sum_frames(nx,Memr[ptr1],Memr[ptr2],Memr[wptr+(v1[2]-2)*nx],
               Memr[ptr3],Memr[wptr+(v1[2]-2)*nx])
         }
         call imflush(out)

         # Close the input file...

         call imunmap(inp)

         # Add this file's name to the buffer for logging

         length2 = strlen(Memc[lp1]) + 1
         if (length2+length > SZ_LINE-2) {
            call fprintf(log_descr,"\n")
            length = 0
         }
         call fprintf(log_descr,"%s ")
            call pargstr(Memc[lp1])
         length = length + length2


         # Friendly message to user...

         call printf("Processing complete on frame %s \n")
            call pargstr(Memc[lp1])
            call flush(STDOUT)
      }

      # Flush out log buffer

      call fprintf(log_descr,"\n")

      # Flush the output buffer

      call imflush(out)

      # Close input file list

      call imtclose(inlist)

      # Now divide the summed frame by the workspace frame

      one_l = 1
      call amovkl(one_l,v1,IM_MAXDIM)
      call amovkl(one_l,v2,IM_MAXDIM)
      while (imgnlr(out,ptr1,v1) != EOF) {
         junk = impnlr(out,ptr2,v2)
         call advzr(Memr[ptr1],Memr[wptr+(v1[2]-2)*nx],Memr[ptr2],nx,fix_zerdiv)
      }

      # Ditch the workspace

      call mfree(wptr,TY_REAL)

      # Close the output file

      call imunmap(out)

      # Write a message to the log file

      call fprintf(log_descr,"Mean output bias frame was %s \n")
         call pargstr(bias_frame)
      if (nzero != 0) {
         call fprintf(log_descr,"%d zero divides trapped in making mean bias\n")
            call pargi(nzero)
      }

      # Reopen it with the bias section

      call strcpy(bias_frame,Memc[lp1],SZ_FNAME)
      call strcat(bsect,Memc[lp1],SZ_FNAME)
      out = immap(Memc[lp1],READ_ONLY,0)
      npix = IM_LEN(out,1)
      nrow = IM_LEN(out,2)
      optr = imgs2r(out,1,npix,1,nrow)

      # Find the overscan bias of this mean frame...

      iferr (call stats(npix*nrow,Memr[optr],fbias,sig,3.0)) {
         if (errcode() == 1) {
            call fprintf(log_descr,"Mean value in bias strip was zero\n")
            call erract(EA_ERROR) }
         else if (errcode() == 2) {
            call fprintf(log_descr,"Bias strip sig**2 < 0\n")
            call erract(EA_WARN)
         }
      }
      else {
         call fprintf(log_descr,
            "    Bias strip average of mean bias frame was %0.7g +/- %0.7g \n")
            call pargr(fbias)
            call pargr(sig)
      }

      # Free stack memory

      call sfree(sp)
end


procedure proc_frame(nx,ny,inlist,type,exphead,sub_bias,sub_frame,
   overscan,bias_frame,fbias,dark_cor,dark_frame,flat_div,flat_frame,
   dsect,bsect,sixteen,trim,want_err,suffix,log_descr,readout,gain)
#------------------------------------------------------------------------------
#      PROC_FRAME -- Main routine for processing frames
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) nx:         (int) X dimension of chip 
#      (>) ny:         (int) Y dimension of chip
#      (>) inlist:     (int) Input file list
#      (>) type:       (char) Type of file being reduced
#      (>) exphead:    (char[ARB]) Location in header for exposure time info
#      (>) sub_bias:   (bool) True if bias is being subtracted
#      (>) sub_frame:  (bool) True if bias frame is being subtracted
#      (>) overscan:   (bool) True if overscan correction is being done
#      (>) bias_frame: (char) Name of mean bias frame
#      (>) fbias:      (real) Average value in overscan region of mean bias frame
#      (>) dark_cor:   (bool) True if dark correction is being done
#      (>) dark_frame: (char) Name of mean dark frame
#      (>) flat_div:   (bool) True if dividing by flat field
#      (>) flat_frame: (char) Name of mean flat frame
#      (>) dsect:      (char) Data section
#      (>) bsect:      (char) Bias section
#      (>) sixteen:    (bool) True if 16 bit correction is being done
#      (>) trim:       (bool) True if overscan regions are to be trimmed
#      (>) want_err:   (bool) True if error files are wanted
#      (>) suffix:     (char) Suffix used to create output file names
#      (>) log_descr:  (int) Descriptor for log file
#      (>) readout:    (real) Readout noise for chip in e-
#      (>) gain:       (real) Chip gain in e-/adu
#
#
#      AUTHOR:      Jim Lewis -- RGO
#      DATE:        18-NOV-1992
#      UPDATE:      11-FEB-1993 -- Several modifications including addition of
#                                  1d flat fielding. Error reporting tightened. (JRL)
#                   10-MAR-1993 -- Cosmic ray and dud pixel fixing removed. Several
#                                  changes to main processing loop for speed (JRL)
#                   09-DEC-1993 -- 1d flat fielding removed (JRL)
#
#------------------------------------------------------------------------------

# Parameters

int nx,ny,log_descr,inlist
char type[ARB],bias_frame[ARB],dark_frame[ARB],flat_frame[ARB],
   dsect[ARB],bsect[ARB],suffix[ARB],exphead[ARB]
bool sub_bias,sub_frame,overscan,dark_cor,flat_div,sixteen,trim,want_err
real fbias,readout,gain

# Local variables

pointer sp,lp1,lp2,ind,outd,indv,outdv,ptr1,ptr2,
   ptr3,ptr4,ptr5,inb,ibias,iflat,idark,outfr,outfrv,wpt1,wpt2,
   edark,eflat,ptr_go,ptr_po,ptr_gb,ptr_goe,ptr_poe,ptr_gd,ptr_gde,
   ptr_gf,ptr_gfe,ptr_gm,ptr_gme,ptr_pm,ptr_pme
real phot,read2,exp_time,fbiaso,sigo,sig,ave,ofactor,dfactor,dfactor2,divc,repval
bool ex_err,bias_open,dark_open,flat_open,ex_errf,ex_errd,should_map,ff,
   use_sect
long one_l,v1[IM_MAXDIM],v2[IM_MAXDIM],v3[IM_MAXDIM],v4[IM_MAXDIM],
   v_go[IM_MAXDIM],v_po[IM_MAXDIM],v_gb[IM_MAXDIM],v_goe[IM_MAXDIM],
   v_poe[IM_MAXDIM],v_gd[IM_MAXDIM],v_gde[IM_MAXDIM],v_gf[IM_MAXDIM],
   v_gfe[IM_MAXDIM],v_gm[IM_MAXDIM],v_gme[IM_MAXDIM],v_pm[IM_MAXDIM],
   v_pme[IM_MAXDIM]
int npix,nrow,junk,irow,nzero,nfile,odims[2]

common /replace/ repval,nzero

# Functions

bool streq(),err_exist()
pointer immap(),imgs2r(),err_open(),impl2r(),rt_crfile()
int imtgetim(),errcode(),imgnlr(),impnlr()

extern fix_zerdiv

begin

      # Get some stack memory

      call smark(sp)
      call salloc(lp1,SZ_FNAME,TY_CHAR)
      call salloc(lp2,SZ_FNAME,TY_CHAR)

      # Initialise a few things...

      phot = 1.0/gain
      read2 = (readout*phot)**2
      one_l = 1
      bias_open = false
      dark_open = false
      flat_open = false
      ex_errf = false
      ex_errd = false
      nfile = 0

      # Write some stuff to the log file...

      call fprintf(log_descr," \n")
      call fprintf(log_descr," \n")
      if (streq("d",type)) {
         call fprintf(log_descr,"*****  Dark Frame Processing  *****\n") }
      else if (streq("f",type)) {
         call fprintf(log_descr,"*****  Flat Frame Processing  *****\n") }
      else if (streq("p",type)) {
         call fprintf(log_descr,"*****  Program Frame Processing  *****\n")
      }
      call fprintf(log_descr," \n")

      # Write out a few messages to the log file...

      if ((sub_bias) && (sub_frame)) {
         call fprintf(log_descr,"    Bias frame was: %s \n")
            call pargstr(bias_frame)
      }
      if (dark_cor) {
         call fprintf(log_descr,"    Dark frame was: %s \n")
            call pargstr(dark_frame)
      }
      if ((streq("p",type)) && (flat_div)) {
         call fprintf(log_descr,"    Flat frame was: %s \n")
            call pargstr(flat_frame)
      }

      # If this is a run of flat images or dark images, then open the output
      # file here.  This will initially be used to do the sums of the frames, but
      # will eventually store the mean dark or flat frame.

      if (streq(type,"f") || streq(type,"d")) {
         if (streq(type,"f")) {
            call strcpy(flat_frame,Memc[lp2],SZ_FNAME) }
         else {
            call strcpy(dark_frame,Memc[lp2],SZ_FNAME)
         }
         odims[1] = nx
         odims[2] = ny
         outfr = rt_crfile(Memc[lp2],NEW_FILE,0,2,odims,TY_REAL)

         # Get an output error array even if you don't really want it in the end.

         iferr(outfrv = err_open(outfr,NEW_COPY,false)) call erract(EA_ERROR)
      }

      # Open mean bias frame if it is needed...

      if (sub_bias && sub_frame) {
         should_map = false
         ff = false
         use_sect = false
         iferr(call open_mean(bias_frame,nx,ny,dsect,type,want_err,should_map,
            ex_err,ff,use_sect,read2,phot,ibias,ptr1)) call erract(EA_ERROR)
         bias_open = true
      }

      # Open mean dark frame if it is needed...

      if (dark_cor && !streq(type,"d")) {
         should_map = true
         ff = false
         use_sect = false
         iferr(call open_mean(dark_frame,nx,ny,dsect,type,want_err,should_map,
            ex_errd,ff,use_sect,read2,phot,idark,edark)) call erract(EA_ERROR)
         dark_open = true
      }

      # Open mean flat frame if it is needed...

      if (flat_div && streq(type,"p")) {
         should_map = true
         ff = true
         use_sect = trim
         iferr(call open_mean(flat_frame,nx,ny,dsect,type,want_err,should_map,
            ex_errf,ff,use_sect,read2,phot,iflat,eflat)) call erract(EA_ERROR)
         flat_open = true
      }

      # Loop now for each frame

      while (imtgetim(inlist,Memc[lp1],SZ_FNAME) != EOF) {

         # Open up the file and check its credentials

         nzero = 0
         iferr {
            ind = immap(Memc[lp1],READ_ONLY,0)
            call check_size(ind,nx,ny)
            call get_exp(ind,exphead,exp_time)
         } then {
            call erract(EA_WARN)
            call imunmap(ind)
            next 
         }
         if (exp_time <= 0.0) {
            call eprintf("%s has exposure time of zero. No further work done on it\n")
               call pargstr(Memc[lp1])
            call fprintf(log_descr,"Frame %s has exposure time of zero.\n")
               call pargstr(Memc[lp1])
            call fprintf(log_descr,"    No further work done on it\n")
            call imunmap(ind)
            next 
         }

         # Did you want to include errors?  If so, then see if an error array exists

         if (want_err) {
            call err_fname(ind,Memc[lp2])
            ex_err = err_exist(ind,READ_ONLY,Memc[lp2],false)
         }

         # Now send some messages to the log file...

         call fprintf(log_descr,"Frame:  %s\n")
            call pargstr(Memc[lp1])
         call fprintf(log_descr,"    Exposure time: %10.0f seconds\n")
            call pargr(exp_time)

         # Now close and reopen the file if you want to trim it

         if (trim && streq(type,"p")) {
            call imunmap(ind)
            call strcpy(Memc[lp1],Memc[lp2],SZ_FNAME)
            call strcat(dsect,Memc[lp2],SZ_FNAME)
            ind = immap(Memc[lp2],READ_ONLY,0)
         }

         # If you wanted errors and an error array exists, then open it...

         if (want_err && ex_err) {
            iferr(indv = err_open(ind,READ_ONLY,false)) {
               call erract(EA_WARN)
               call imunmap(indv)
               next
            }  
         }

         # Open the output file for program frames...Force REAL data arrays

         if (streq(type,"p")) {
            iferr {
               call strcpy(Memc[lp1],Memc[lp2],SZ_FNAME)
               call strcat(suffix,Memc[lp2],SZ_FNAME)
               outd = immap(Memc[lp2],NEW_COPY,ind)
               IM_PIXTYPE(outd) = TY_REAL
               ptr1 = impl2r(outd,1)
               call imunmap(outd)
               outd = immap(Memc[lp2],READ_WRITE,ind)

               # Open an output variance array if you want one...

               if (want_err) outdv = err_open(outd,NEW_FILE,false)
            } then {
               call erract(EA_WARN)
               next
            }
         }

         if (sub_bias) {

            # If you are doing the overscan correction or if your bias estimate
            # is coming soley from the overscan strip then first open the
            # bias section of the original file. Then get an estimate of the
            # mean bias in that strip

            if ((overscan) || (! sub_frame)) {
               call strcpy(Memc[lp1],Memc[lp2],SZ_FNAME)
               call strcat(bsect,Memc[lp2],SZ_FNAME)
               inb = immap(Memc[lp2],READ_ONLY,0)
               npix = IM_LEN(inb,1)
               nrow = IM_LEN(inb,2)
               ptr1 = imgs2r(inb,1,npix,1,nrow)
               iferr (call stats(npix*nrow,Memr[ptr1],fbiaso,sigo,3.0)) {
                  call erract(EA_WARN)
                  if (errcode() == 1) {
                     call fprintf(log_descr,"Bias strip has all bad values\n")
                     next }
                  else if (errcode() == 2) {
                     call fprintf(log_descr,"Bias strip sig**2 < 0\n")
                  }
               }
               else {
                  call fprintf(log_descr,
                     "    Bias strip average was %0.7g +/- %0.7g \n")
                     call pargr(fbiaso)
                     call pargr(sigo)
               }
               call imunmap(inb)
            }
         }

         # Define overscan correction factor

         if (sub_frame && overscan) {
            ofactor = fbiaso/fbias
            call fprintf(log_descr,"    Overscan correction factor was %0.7g\n")
               call pargr(ofactor) }
         else {
            ofactor = 1.0
         }

         # Define dark count scaling factor

         dfactor = exp_time
         dfactor2 = dfactor**2

         # Set up the indices to the various arrays...

         npix = IM_LEN(ind,1)
         call amovkl(one_l,v_go,IM_MAXDIM)
         call amovkl(one_l,v_po,IM_MAXDIM)
         call amovkl(one_l,v_gb,IM_MAXDIM)
         call amovkl(one_l,v_gd,IM_MAXDIM)
         call amovkl(one_l,v_gf,IM_MAXDIM)
         call amovkl(one_l,v_gm,IM_MAXDIM)
         call amovkl(one_l,v_pm,IM_MAXDIM)
         call amovkl(one_l,v_gme,IM_MAXDIM)
         call amovkl(one_l,v_pme,IM_MAXDIM)
         if (want_err) {
            call amovkl(one_l,v_goe,IM_MAXDIM)
            call amovkl(one_l,v_poe,IM_MAXDIM)
            call amovkl(one_l,v_gde,IM_MAXDIM)
            call amovkl(one_l,v_gfe,IM_MAXDIM)
         }

         # Get workspace to store results for each row...

         call malloc(wpt1,npix,TY_REAL)
         call malloc(wpt2,npix,TY_REAL)

         # Main Processing Loop

         irow = 0
         nfile = nfile + 1
         while (imgnlr(ind,ptr_go,v_go) != EOF) {
            ptr_po = wpt1
            irow = irow + 1

            # Take out sixteen bit correction

            if (sixteen) {
               call sixteen_bit(npix,Memr[ptr_go],Memr[ptr_po])
               ptr_go = ptr_po
            }

            # Bias subtraction section...

            if (sub_bias) {

               # Section for subtracting a bias frame

               if (sub_frame) {
                  junk = imgnlr(ibias,ptr_gb,v_gb)
                  call awsur(Memr[ptr_go],Memr[ptr_gb],Memr[ptr_po],npix,1.0,
                  -ofactor) }

               # Section for subtracting a bias value

               else {
                  call asubkr(Memr[ptr_go],fbiaso,Memr[ptr_po],npix)
               }

               # Make the input the same as the output now...

               ptr_go = ptr_po
            }

            # Section to create an error array if it's needed

            if (want_err && !ex_err) {
               ptr_poe = wpt2
               call cre_var(npix,Memr[ptr_go],read2,phot,Memr[ptr_poe])
               ptr_goe = ptr_poe }
            else if (want_err && ex_err) {
               junk = imgnlr(indv,ptr_goe,v_goe)
               ptr_poe = wpt2
            }

            # Dark count subtraction section...

            if (dark_cor && !streq(type,"d")) {
               junk = imgnlr(idark,ptr_gd,v_gd)
               call awsur(Memr[ptr_go],Memr[ptr_gd],Memr[ptr_po],npix,1.0,-dfactor)
               ptr_go = ptr_po

               # Deal with error arrays if you wish...
   
               if (want_err) {
                  junk = imgnlr(edark,ptr_gde,v_gde)
                  call awsur(Memr[ptr_goe],Memr[ptr_gde],Memr[ptr_poe],npix,1.0,
                  dfactor2)
                  ptr_goe = ptr_poe
               }
            }

            # Flat fielding section...

            if (flat_div && streq(type,"p")) {

               # Division by image flat...

               junk = imgnlr(iflat,ptr_gf,v_gf)
               if (want_err) {
                  junk = imgnlr(eflat,ptr_gfe,v_gfe)
                  call var_div(npix,Memr[ptr_go],Memr[ptr_goe],Memr[ptr_gf],
                  Memr[ptr_gfe],Memr[ptr_poe])
                  ptr_goe = ptr_poe
               }
               call advzr(Memr[ptr_go],Memr[ptr_gf],Memr[ptr_po],npix,fix_zerdiv)
               ptr_go = ptr_po 
            }

            # If these are dark frames, then divide by exposure time in order that
            # the mean dark frame will be in units of counts/sec/pix

            if (streq(type,"d")) {
               call adivkr(Memr[ptr_go],dfactor,Memr[ptr_po],npix)
               ptr_go = ptr_po
               if (want_err) {
                  call adivkr(Memr[ptr_goe],dfactor2,Memr[ptr_poe],npix)
                  ptr_goe = ptr_poe
               }
            }

            # If these are either dark frames or flat frames, then add 
            # them into the mean dark/flat file output buffer.

            if (streq(type,"d") || streq(type,"f")) {
               junk = impnlr(outfr,ptr_pm,v_pm)
               junk = impnlr(outfrv,ptr_pme,v_pme)
               if (nfile != 1) {
                  junk = imgnlr(outfr,ptr_gm,v_gm)
                  junk = imgnlr(outfrv,ptr_gme,v_gme) }
               else {
                  ptr_gm = ptr_pm
                  ptr_gme = ptr_pme
                  call amovkr(0.0,Memr[ptr_gm],npix)
                  call amovkr(0.0,Memr[ptr_gme],npix)
               }
               if (want_err) {
                  call w_sum_frames(npix,Memr[ptr_go],Memr[ptr_goe],
                  Memr[ptr_gm],Memr[ptr_gme],Memr[ptr_pm],Memr[ptr_pme]) }
               else {
                  call sum_frames(npix,Memr[ptr_go],Memr[ptr_gm],
                  Memr[ptr_gme],Memr[ptr_pm],Memr[ptr_pme])
               } }

            # Otherwise, this is a program frame, and the workspace is then
            # written to the appropriate row of the output file

            else {
               junk = impnlr(outd,ptr1,v_po)
               call amovr(Memr[ptr_po],Memr[ptr1],npix)
               if (want_err) {
                  junk = impnlr(outdv,ptr1,v_poe)
                  call amovr(Memr[ptr_poe],Memr[ptr1],npix)
               }
            }
         } # End of row loop

         # Send a few messages to the log file...

         if (sixteen) {
            call fprintf(log_descr,"    Sixteen bit correction done\n")
         }         
         if (sub_bias) {
            if (sub_frame) {
               call fprintf(log_descr,"    Bias frame was subtracted\n") }
            else {
               call fprintf(log_descr,"    Bias value was subtracted\n")
            }
         }
         if (dark_cor && !streq(type,"d")) 
            call fprintf(log_descr,"    Dark frame was subtracted\n")
         if (flat_div && streq(type,"p"))
            call fprintf(log_descr,"    Divided by image flat field\n")

         # If this is a program frame, then close the output file...

         if (streq(type,"p")) {
            npix = IM_LEN(outd,1)
            nrow = IM_LEN(outd,2)
            call fprintf(log_descr,
               "    Output file has dimensions %d %d and is named %s \n")
               call pargi(npix)
               call pargi(nrow)
               call pargstr(IM_HDRFILE(outd))
            call imunmap(outd)
            if (want_err) call err_close(outdv) }

         # Flush the output buffer for the mean arrays...

         else {
            call imflush(outfr)
            call imflush(outfrv) 
         }

         # Close input files...

         call imunmap(ind)
         if (want_err && ex_err) call err_close(indv) 

         # Ditch the workspace

         call mfree(wpt1,TY_REAL)
         call mfree(wpt2,TY_REAL)

         # Friendly message to the user

         call printf("Processing complete on frame %s\n")
            call pargstr(Memc[lp1])
            call flush(STDOUT)

         # Tell the user about any zero divides which have been trapped

         if (nzero != 0) {
            call fprintf(log_descr,"%d zero divides trapped in processing this frame\n")
               call pargi(nzero)
         }

      # End of Main Loop

      }

      # Close up the template list

      call imtclose(inlist)

      # If this was flat field or dark frame processing, then finish averaging
      # the data

      nzero = 0
      if (streq("f",type) || streq("d",type)) {
         npix = IM_LEN(outfr,1)
         call amovkl(one_l,v1,IM_MAXDIM)
         call amovkl(one_l,v2,IM_MAXDIM)
         call amovkl(one_l,v3,IM_MAXDIM)
         call amovkl(one_l,v4,IM_MAXDIM)
         while (imgnlr(outfr,ptr1,v1) != EOF) {
            junk = imgnlr(outfrv,ptr2,v2)
            junk = impnlr(outfr,ptr3,v3)
            call advzr(Memr[ptr1],Memr[ptr2],Memr[ptr3],npix,fix_zerdiv)
            if (want_err) {
               junk = impnlr(outfrv,ptr4,v4)
               call arczr(1.0,Memr[ptr2],Memr[ptr4],npix,fix_zerdiv)
            }
         }
         if (streq(type,"d")) {
            call imflush(outfr)
            if (want_err) call imflush(outfrv)
         }

         # Send another message to the log file...

         call fprintf(log_descr,"    Output file is named: %s\n")
            if (streq("f",type)) {
               call pargstr(flat_frame) }
            else {
               call pargstr(dark_frame)
            }

         # If this is a flat field, then normalise it so that the average is 1

         if (streq("f",type)) {
            call imunmap(outfr)
            call err_close(outfrv)
            call strcpy(flat_frame,Memc[lp2],SZ_FNAME)
            call strcat(dsect,Memc[lp2],SZ_FNAME)
            outfr = immap(Memc[lp2],READ_WRITE,0)
            if (want_err) outfrv = err_open(outfr,READ_WRITE,true)
            npix = IM_LEN(outfr,1)
            nrow = IM_LEN(outfr,2)
            ptr1 = imgs2r(outfr,1,npix,1,nrow)
            ifnoerr (call stats(npix*nrow,Memr[ptr1],ave,sig,3.0)) {
               if (ave > 0.0) {
                  divc = ave*ave
                  call amovkl(one_l,v1,IM_MAXDIM)
                  call amovkl(one_l,v2,IM_MAXDIM)
                  call amovkl(one_l,v3,IM_MAXDIM)
                  call amovkl(one_l,v4,IM_MAXDIM)
                  while (imgnlr(outfr,ptr2,v1) != EOF) {
                     junk = impnlr(outfr,ptr3,v2)
                     call adivkr(Memr[ptr2],ave,Memr[ptr3],npix)
                     if (want_err) {
                        junk = imgnlr(outfrv,ptr4,v3)
                        junk = impnlr(outfrv,ptr5,v4)
                        call adivkr(Memr[ptr4],divc,Memr[ptr5],npix)
                     }
                  }
                  call fprintf(log_descr,
                     "    Flat field had an average of %0.7g +/- %0.7g\n")
                     call pargr(ave)
                     call pargr(sig) }
               else  {
                  call eprintf("Mean of flat field was zero -- check results carefully\n")
                  call fprintf(log_descr,
                     "    Mean of flat field was zero -- check results carefully\n")
               } 
            }
            else {
               call eprintf(
                  "Couldn't do stats on mean flat field -- check results carefully\n")
               call fprintf(log_descr,
                 "    Couldn't do stats on mean flat field -- check results carefully\n")
            }
         }

         # Now close the file

         if (want_err) call err_close(outfrv)
         if (!want_err) call err_imdel(outfr)
         call imunmap(outfr)

         # Signal for trapped zero divides in mean frame

         if (nzero != 0) {
            call fprintf(log_descr,"%d zero divides trapped in creating mean frame\n")
               call pargi(nzero)
         }

      # End of extra processing for dark and flat frames

      }

      # Close up the mean bias frame

      if (bias_open) call imunmap(ibias)

      # Close up the mean dark frame

      if (dark_open) {
         if (want_err) {
            call err_close(edark)
            if (!ex_errd) call err_imdel(idark)
         }
         call imunmap(idark)
      }

      # Close up mean flat frame

      if (flat_open) {
        if (want_err) {
            call err_close(eflat)
            if (!ex_errf) call err_imdel(iflat)
         }
         call imunmap(iflat)
      }
            
      # Free stack workspace

      call sfree(sp)
end
