# include <stdlib.h>
# include <stdio.h>

# include <c_iraf.h>
# include <xclio.h>
# include <ximio.h>
# include <xtables.h>

# include "apm.h"
# include "ap.h"

void iraferr() {
	printf("IRAF error code %d\nIRAF message: %s\n",
		c_iraferr(),c_iraferrmsg());
	exit(0);
}

int main(int argc, char **argv) {
	IRAFPointer in, out, tab, list;
	long *in_data;
	float *out_data;
	int i, j, k;
	int no_dims, imfile;
	long dim1, dim2;
	IRAFType in_type;
	char name[BUFSIZ], backname[80], tabname[80], inlist[BUFSIZ];
	ap_t ap;
	int maxib;

	int *bak;
	int sky;
	int *start, *bstart, *smoothed;
	int *prev, *next;

	/* must always remember to call c_irafinit */
	c_irafinit(argc, argv);

	/* install the error handler */
	c_pusherr(iraferr);

	printf("IRAF APM program\n");

        /* Get the input file from the command line... */

        if (argc > 1) {
             strcpy(inlist,argv[1]);
        } else {
             printf("Syntax: apm images\n");
             exit(1);
        }

        /* Open the list and test it */

        if ((list = c_imtopen(inlist)) == (IRAFPointer)NULL) {
            printf("Error opening file list\n");
            exit(1);
        }

        /* Loop for each file in the list */

        for (imfile = 0; imfile < c_imtlen(list); imfile++) {
            c_imtgetim(list,name,BUFSIZ);
   	    printf("Reading image file %s\n",name);

	    /* open the image */
 	    in = c_immap(name,IRAF_READ_ONLY,0);
	    printf("Error code is %d\n",c_iraferr());

	    /* get the number of dimensions */ 
	    no_dims = c_imgndim(in);
	    if (no_dims != 2) {
		printf("Sorry! This example only works for two dimensions\n");
		continue;
	    }

  	    /* and the size */
	    dim1 = c_imglen(in,1);
	    dim2 = c_imglen(in,2);
	    /* get the type of pixel */
	    in_type = (IRAFType)c_imgtypepix(in);
	    printf("Input:  %d dimensions -- dim1 = %d  dim2 = %d of type %d\n",
		no_dims, dim1, dim2, in_type);
	    printf("Error code is %d\n",c_iraferr());

	    /* Now, we'll open the output image */
	    (void)strcpy(backname, name);
	    (void)strcat(backname, ".bg.fits");
	    out = c_immap(backname,IRAF_NEW_COPY,in);
	    /* and the output table: */
	    (void)strcpy(tabname, name);
	    (void)strcat(tabname, ".tab");
	    tab = c_tbtopn(tabname, IRAF_NEW_FILE, 0);
	    tabInit(tab);

	    printf("Generating background image to %s\n",backname);

	    c_imptypepix(out, (int)IRAF_INT);

	    bak = malloc(dim1*dim2*sizeof(*bak));

	    /* Calculate background map for sky subtraction: */
	    background(in, out, bak, &ap);
	    sky = ap.background;
	    printf("Sky background = %d, sigma = %f\n", sky, ap.sigma);

	    /* Set up image-detecton analysis: */
	    ap.lsiz = dim1;
	    smoothed = malloc(dim1 * sizeof(*smoothed));
	    apinit(&ap);
	    /* Stick in some arbitrary values: */
	    ap.thresh = 2.0*ap.sigma;
	    ap.multiply = 1;
	    ap.ipnop = 5;
	    tabAddHead(&ap);

	    /* Extract images: */
	    bstart = bak;
	    maxib = 0;
	    printf("Extract images...\n");
	    for (j=1; j<=dim2; j++)
	    {
		/* Hanning filter; need this line, previous and next,
		 * but use the same one over on first and last lines: */
		if(j == 1)
		{
			start = c_imgl2i(in, 1);
			prev = start;
			next = c_imgl2i(in, 2);
		}	
		else if (j < dim2)
		{
			prev = start;
			start = next;
			next = c_imgl2i(in, j+1);
		}
		else
		{
			prev = next;
			start = next;
		}
		smoothed[0] = (prev[0] + 2*start[0] + next[0] + 2)/4 -
				bstart[0] + sky;
		for (i=1; i<dim1-1; i++)
		{
			smoothed[i] = (prev[i-1]+2*prev[i]+prev[i+1]
				+ 2*start[i-1]+4*start[i]+2*start[i+1]
				+ next[i-1]+2*next[i]+next[i+1] + 8)/16;
			smoothed[i] = sky + smoothed[i] - bstart[i];
		}
		smoothed[dim1-1] = (prev[dim1-1]+2*start[dim1-1]+next[dim1-1]
					+ 2)/4 - bstart[dim1-1] + sky;
		for(i=0; i<dim1; i++)
			start[i] += (sky - bstart[i]);
		apline(&ap, start, smoothed, j);
		if(ap.ibstack > maxib) maxib = ap.ibstack;
		if(ap.ibstack > (10000 - ap.lsiz) )
		{
			printf("APFU!: ibstack = %d\n",ap.ibstack);
			apfu(&ap);
		}
		if(ap.ipstack > (ap.maxpa*3/4) )
		{
			printf("APFU!: ipstack = %d\n",ap.ipstack);
			for(i=0; i < 200*3/8; i++) apfu(&ap);
		}
		if(ap.ipstack > 1) terminate(&ap);
		bstart += dim1;
	    }

   	    /* close both images */
	    c_imunmap(in);
	    c_imunmap(out);
            c_imdelete(backname);

	    /* And the output table */
	    c_tbtclo(tab);

            /* Free workspace */

            free(bak);
            free(smoothed);
            apclose(&ap);
	    printf("Error code is %d\n",c_iraferr());
        }
        c_imtclose(list);

	return 0;
}
