include <mach.h>
include <math.h>
include <error.h>
include <imhdr.h>
include <smw.h>
include <pkg/center1d.h>

define    MAXBINS     2048
define    FTYPES      |emission|absorption|

procedure t_autoarc()
#------------------------------------------------------------------------------
#
#   -------------
#   A U T O A R C
#   -------------
#
#   Description
#   -----------
#   Do automatic arc line identification 
#
#   
#   Scope of program
#   ----------------
#   - Handles up to 3-D 
#   - All data types converted to REAL
#   - Batch execution supported
#
#
#   Environment
#   -----------
#   IRAF
#
#
#   Language
#   --------
#   SPP
#
#
#   Parameters
#   ----------
#   arcfiles   List of input files with arc data. (string)(prompted for)
#
#   line       Aperture/line number corresponding to the spectrum 
#              (int)(prompted for)
#
#   band       Band number of the spectrum (int)(prompted for)
#
#   apshift    Number of pixels to shift spectrum in order for the central
#              wavelength requested later will in fact be located near the
#              centre of the spectrum.  This is in case you know that the
#              central wavelength in the header is miscalibrated by a 
#              certain amount. (real)(hidden)
#
#   coordlist  Name of file with wavelengths (string)(hidden)
#
#   dflip      Do you want to consider negative or positive dispersions
#              or both?  Answer -1 for negative only, 1 for positive only
#              or zero for both.  Note that the sign refers to the input
#              dispersion, e.g. "negative" means the inverse of the input
#              dispersion. (integer)(hidden)
#
#   database   Name of subdirectory where id database should be dumped
#              (string)(hidden)
#
#   arclog     Name of optional logfile (string)(hidden)
#
#   idpars     Pset defining identification algorithm parameters (pset)
#
#   cenpars    Pset with parameters specific to the peak finding and centring
#              algorithms (pset)
#
#
#   Method
#   ------
#   - The input file header is read to get a rough wavelength centre and disp.
#   - A peak finding algorithm finds all the major peaks and assigns these
#     a rough wavelength 
#   - The region of wavelength space covered by the rough calibration is read
#     from a wavelength file
#   - The projection method is employed to then identify a better dispersion
#     and offset.  Note that this method works best for spectra where the
#     dispersion is nearly linear.  See the routine aa_proj for more notes
#     on this algorithm.
#   - The ids inferred from the best solution are written to a database file
#     of a similar format to that used by the onedspec routines 'identify'
#     and 'reidentify'
#
#
#   External libraries needed for linkage
#   -------------------------------------
#      libiminterp.a
#      libsmw.a
#      libxtools.a
#
#   Internal libraries required for linkage
#   ---------------------------------------
#      libpkg.a
#
#
#   INCLUDE statements
#   ------------------
#   include <error.h>
#   include <imhdr.h>
#   include <mach.h>
#   include <math.h>
#   include <smw.h>
#   include <pkg/center1d.h>
#
#
#   Possible future upgrades
#   ------------------------
#   -- Allow for a range of apertures in a file, all with different central
#      wavelengths and dispersions as with echelle data.
#
#
#   Author/s
#   --------
#   Jim Lewis      RGO
#
#
#   History
#   -------
#   16-FEB-95     - Original program 
#
#
#------------------------------------------------------------------------------

# Local variables

pointer sp,arcname,database,wavefile,logfile,iptr,pptr,wptr,
   nptr,optr,cw,dp,images,sh,mw
int nx,fd,sep,edge,nmax,type,npeaks,np,i,index,nalloc,nw,wd,nd,
   noo,nmatch,nt,nn,ip,apnum,dflip,list,nline,nband
real cenwave,disp,wave1,wave2,contrast,pkthresh,width,crad,cthresh,
   dwave,dfrac,dd[2],control[7],wfact,wave,apshift,slop
bool debug

# Functions

int clgeti(),access(),open(),find_peaks(),fscan(),aa_best(),ctor(),imtopen(),
   imtgetim(),imaccess(),clgwrd()
real imgetr(),center1d(),clgetr()

errchk rgetimage

begin

      # Get stack memory

      call smark(sp)
      call salloc(images,SZ_FNAME,TY_CHAR)
      call salloc(arcname,SZ_FNAME,TY_CHAR)
      call salloc(database,SZ_FNAME,TY_CHAR)
      call salloc(wavefile,SZ_FNAME,TY_CHAR)
      call salloc(logfile,SZ_FNAME,TY_CHAR)
      call salloc(cw,SZ_LINE,TY_CHAR)
      call salloc(dp,SZ_LINE,TY_CHAR)

      # Open arc file list

      call clgstr("arcfiles",Memc[images],SZ_FNAME)
      list = imtopen(Memc[images])

      # Initialise a few things

      iptr = NULL
      sh = NULL
      nline = 0
      nband = 0

      # Get database file name

      call clgstr("database",Memc[database],SZ_FNAME)

      # Get name of wavelength file and see if it's accessible

      call clgstr("coordlist",Memc[wavefile],SZ_FNAME)
      if (access(Memc[wavefile],READ_ONLY,TEXT_FILE) == NO) {
         call eprintf("Cannot access wavelength file %s\n")
            call pargstr(Memc[wavefile])
         call error(0,"Check your input parameters...")
      }

      # Open a log file.  If it is already there, then append the data
      # from this run to it.  If it doesn't exist, then open a new file.
      # If the file name is blank, then don't bother logging anything.

      call clgstr("arclog",Memc[logfile],SZ_FNAME)
      if (Memc[logfile] != EOS) {
         if (access(Memc[logfile],APPEND,TEXT_FILE) == YES) 
            fd = open(Memc[logfile],APPEND,TEXT_FILE)
         else 
            iferr (fd = open(Memc[logfile],NEW_FILE,TEXT_FILE)) 
               call error(0,"Can't open your log file")
      } else {
         fd = NULL
      }

      # Set up some parameters for peak finding algorithm. These are 
      # similar to parameters found for 'identify'

      sep = clgeti("sep")
      contrast = clgetr("contrast")
      edge = clgeti("edge")
      nmax = clgeti("maxfeatures")
      pkthresh = clgetr("pkthresh")
      debug = false
      width = clgetr("fwidth")
      type = clgwrd("ftype",Memc[cw],SZ_LINE,FTYPES)
      crad = clgetr("cradius")
      cthresh = clgetr("cthresh")

      # Set up parameters for automatic line identifier.

      dfrac = clgetr("dfrac")
      nd = clgeti("ndisp")
      dd[1] = 1.0/(1.0 + dfrac)   # In search algorithm, dispersion is defined
      dd[2] = 1.0/(1.0 - dfrac)   # by dp/dw.
      control[1] = clgetr("limit_angle")
      control[2] = real(clgeti("ns_select"))
      control[3] = real(clgeti("ng_select"))
      control[4] = 1.0  # bin_factor = 1.0
      control[5] = 1.0  # nb_covers = 1
      control[6] = 0.0  # trial_angles = false
      control[7] = real(clgeti("test_stat"))
      wfact = clgetr("wfact")
      slop = clgetr("padding")

      # Get some workspace for autoarc routines

      nn = MAXBINS
      call malloc(nptr,nn,TY_INT)
      noo = nint(control[2] + control[3]) + 1
      call malloc(optr,noo*5,TY_REAL)

      # Now loop for each file in the input list

      while (imtgetim(list,Memc[arcname],SZ_FNAME) != EOF) {

         # See if the current input file is available

         if (imaccess (Memc[arcname], READ_ONLY) == NO) {
            call eprintf ("Can't get image %s\n")
               call pargstr (Memc[arcname])
            next
         }

         # Get the spectrum you want.  If it has already been scrunched
         # then skip over it.

         call rgetimage(Memc[arcname],nline,nband,apnum,iptr,mw,sh)
         nx = SN(sh)
         if (DC(sh) != DCNO) {
            call eprintf(
               "Image: %s, Ap: %d, band: %d is already dispersion corrected\n")
               call pargstr(Memc[arcname])
               call pargi(apnum)
               call pargi(nband)
            call shdr_close(sh)
            call smw_close(mw)
            call imunmap(iptr)
            next
         }

         # Get the aperture shift and the flip parameter

         apshift = clgeti("apshift")
         dflip = clgeti("dflip")

         # Read the central wavelength and dispersion parameters. If you can't
         # read the parameters into a real number, we figure that the user
         # has given us a header keyword name. Calculate an approximate 
         # wavelength region 

         call clgstr("cenwave",Memc[cw],SZ_LINE)
         ip = 1
         if (ctor(Memc[cw],ip,cenwave) == 0) 
            iferr (cenwave = imgetr(iptr,Memc[cw])) call erract(EA_ERROR)
         call clgstr("disp",Memc[dp],SZ_LINE)
         ip = 1
         if (ctor(Memc[dp],ip,disp) == 0)
            iferr (disp = imgetr(iptr,Memc[dp])) call erract(EA_ERROR)
         cenwave = cenwave + apshift*disp
         wave1 = cenwave - 0.5*disp*real(nx)
         wave2 = cenwave + 0.5*disp*real(nx)

         # Get an initial allocation for the wavelengths in the wavelength file

         nalloc = 64
         call malloc(wptr,nalloc,TY_REAL)

         # Now read wavelengths from the wavelength file.  Pad out the end
         # points to allow for slop in the definition of the central wavelength
         # in the header... Close the wavelength file when finished

         dwave = (wave2 - wave1)*slop
         wave1 = wave1 - dwave
         wave2 = wave2 + dwave
         nw = 0
         wd = open(Memc[wavefile],READ_ONLY,TEXT_FILE)
         while (fscan(wd) != EOF) {
            call gargr(wave)
            if (wave >= wave1 && wave <= wave2) {
               nw = nw + 1
               if (nw > nalloc) {
                  nalloc = nalloc + 64
                  call realloc(wptr,nalloc,TY_REAL)
               }
               Memr[wptr+nw-1] = wave
            }
         }
         call realloc(wptr,nw,TY_REAL)
         call close(wd)
      
         # Find all the peaks and centre them.  Remove the ones which don't
         # centre well.

         call malloc(pptr,nx,TY_REAL)
         npeaks = find_peaks(Memr[SY(sh)],Memr[pptr],nx,contrast,sep,edge,nmax,
            pkthresh,debug)
         np = 0
         do i = 1,npeaks {
            index = pptr + i - 1
            Memr[index] = center1d(Memr[index],Memr[SY(sh)],nx,width,type,crad,
               cthresh)
            if (!IS_INDEF(Memr[index])) {
               np = np + 1
               Memr[pptr+np-1] = cenwave + (Memr[index] - real(nx)*0.5)*disp
            }
         }
         call realloc(pptr,np,TY_REAL)

         # Now run the automatic identification scheme.  If there was an error
         # encountered then signal it.

         iferr (call aa_proj(Memr[pptr],np,Memr[wptr],nw,dd,nd,control,
            fd,dflip,Memi[nptr],nn,Memr[optr],noo,nt)) {
               call eprintf("Error in finding solution for %s, Ap %d, band %d\n")
                  call pargstr(Memc[arcname])
                  call pargi(apnum)
                  call pargi(nband)
               call erract(EA_WARN)

         # If the algorithm completed normally but failed to find a solution,
         # then signal an error here.

         } else if (nt == 0) {
            call eprintf("No solutions found for %s, Ap %d, band %d\n")
               call pargstr(Memc[arcname])
               call pargi(apnum)
               call pargi(nband)

         # Otherwise, choose the best solution and write the results to
         # the database file.

         } else {
            nmatch = aa_best(fd,Memr[pptr],np,Memr[wptr],nw,Memr[optr],noo,
               nt,wfact,Memi[nptr])
            call aa_wdb(Memc[arcname],apnum,nband,apshift,nx,Memr[pptr],np,
               Memr[wptr],nw,Memi[nptr],nmatch,cenwave,disp,width,type,
               Memc[database],sh)
         }

         # Close things up and free some workspace

         call mfree(wptr,TY_REAL)
         call mfree(pptr,TY_REAL)
         call shdr_close(sh)
         call smw_close(mw)
         call imunmap(iptr)
      }
         
      # Close up log file and release remaining workspace

      if (fd != NULL) call close(fd)
      call mfree(optr,TY_REAL)
      call mfree(nptr,TY_INT)
      call imtclose(list)
      call sfree(sp)
end


procedure aa_wdb(arcfile,apnum,nband,apshift,nx,pt,np,wt,nw,ids,nids,cenwave,
   disp,width,ftype,database,sh)
#------------------------------------------------------------------------------
#      AA_WDB -- Write the list of ids to a database file which can be
#                read by onedspec.
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) arcfile:    (char[ARB]) Input arc data file name
#      (>) apnum:      (int) Aperture number of the spectrum
#      (>) nband:      (int) Band number of the spectrum
#      (>) apshift:    (real) Aperture shift between physical to logical coord
#      (>) nx:         (int) Number of pixels in arc data
#      (>) pt:         (real[np]) Measured pixel positions of lines
#      (>) np:         (int) Number of pixel positions
#      (>) wt:         (real[nw]) Wavelength positions of known arc lines
#      (>) nw:         (int) Number of wavelengths
#      (>) ids:        (int[nw]) Indices such that wt[i] => pt[ids[i]]
#      (>) nids:       (int) Number of ids made
#      (>) cenwave:    (real) Original central wavelength from header
#      (>) disp:       (real) Original dispersion from header
#      (>) width:      (real) Feature width
#      (>) ftype:      (int) Feature type (emission or absorption)
#      (>) database:   (char[ARB]) Subdirectory for depositing database file
#      (>) sh:         (pointer) Spectral header data pointer
#
#
#      AUTHORS:     Jim Lewis -- RGO
#      DATE:        16-FEB-1995
#      UPDATE:
#
#------------------------------------------------------------------------------

# Parameters

char arcfile[ARB],database[ARB]
int np,nw,ids[nw],nids,nx,apnum,ftype,nband
real pt[np],wt[nw],cenwave,disp,width,apshift
pointer sh

# Local variables

pointer sp,line,rootname,file,dt,pixptr,wtptr,line2
int i,index,ntot,junk[2]
real pixpos,lcor

begin

      # Get some stack workspace

      call smark(sp)
      call salloc(rootname,SZ_FNAME,TY_CHAR)
      call salloc(file,SZ_FNAME,TY_CHAR)
      call salloc(line,SZ_LINE,TY_CHAR)
      call salloc(pixptr,np,TY_REAL)
      call salloc(wtptr,np,TY_REAL)
      call salloc(line2,SZ_LINE,TY_CHAR)

      # Set up the database file name.  Strip of the input file name
      # extension.  The open the database file.

      call xt_imroot(arcfile,Memc[rootname],SZ_FNAME)
      call sprintf(Memc[file],SZ_LINE,"id%s")
         call pargstr(Memc[rootname])
      call dtremap(dt,database,Memc[file],APPEND)

      # Get ID section

      junk[1] = apnum
      junk[2] = nband
      call rid_dbsection(sh,Memc[rootname],junk,Memc[line2],SZ_LINE)

      # Write the necessary header information...

      call dtptime(dt)
      call sprintf(Memc[line],SZ_LINE,"begin\tidentify %s%s\n")
         call pargstr(Memc[rootname])
         call pargstr(Memc[line2])
      call dtput(dt,Memc[line])
      call sprintf(Memc[line],SZ_LINE,"\tid\t%s\n")
         call pargstr(Memc[rootname])
      call dtput(dt,Memc[line])
      call dtput(dt,"\ttask\tautoarc\n")
      call sprintf(Memc[line],SZ_LINE,"\timage\t%s%s\n")
         call pargstr(Memc[rootname])
         call pargstr(Memc[line2])
      call dtput(dt,Memc[line])
      call sprintf(Memc[line],SZ_LINE,"\taperture\t%d\n")
         call pargi(apnum)
      call dtput(dt,Memc[line])
      call sprintf(Memc[line],SZ_LINE,"\taplow\t%g\n")
         call pargr(APLOW(sh,1))
      call dtput(dt,Memc[line])
      call sprintf(Memc[line],SZ_LINE,"\taphigh\t%g\n")
         call pargr(APHIGH(sh,1))
      call dtput(dt,Memc[line])

      # Calculate the pixel positions of the lines that matched

      ntot = 0
      do i = 1,nw {
         index = ids[i]
         if (index != 0) {
            ntot = ntot + 1
            pixpos = (pt[index] - cenwave)/disp + 0.5*real(nx)
            Memr[pixptr+ntot-1] = pixpos
            Memr[wtptr+ntot-1] = wt[i]
         }
      }

      # Now calculate the pixel positions of the features which didn't
      # match and give them an indefinite wavelength

      do index = 1,np {
         do i = 1,nw
            if (index == ids[i])
               break
         if (i > nw) {
            pixpos = (pt[index] - cenwave)/disp + 0.5*real(nx)
            ntot = ntot + 1
            Memr[pixptr+ntot-1] = pixpos
            Memr[wtptr+ntot-1] = INDEF
         }
      }

      # Sort the results into increasing order of pixel position

      call xt_sort2(Memr[pixptr],Memr[wtptr],ntot)

      # Now write the info about the features.  The first item is the 
      # total number of features (matched or not). The rest is the
      # info on the lines.  This includes its pixel position, its wcs
      # position (presumably rough wavelength calibration) and the 
      # wavelength assigned to it.

      call sprintf(Memc[line],SZ_LINE,"\tfeatures\t%d\n")
         call pargi(ntot)
      call dtput(dt,Memc[line])
      do i = 1,ntot {
         call sprintf(Memc[line],SZ_LINE,
            "\t    %10.2f %10.8g %10.8g %5.1f %d 1\n")
            call pargr(Memr[pixptr+i-1])
            lcor = Memr[pixptr+i-1] + apshift
            call pargr(lcor)
            call pargr(Memr[wtptr+i-1])
            call pargr(width)
            call pargi(ftype)
         call dtput(dt,Memc[line])
      }
      call dtput(dt,"\n")

      # Close up the database file, release the workspace and
      # get out of here.

      call dtunmap(dt)
      call sfree(sp)
end


int procedure aa_best(fd,pt,np,wt,nw,table,ntab,nt,wfact,ids)
#------------------------------------------------------------------------------
#      AA_BEST -- Decide which is the best solution and make line assignments
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) fd:         (int) Descriptor for log file
#      (>) pt:         (real[np]) Measured pixel positions of lines
#      (>) np:         (int) Number of pixel positions
#      (>) wt:         (real[nw]) Wavelength positions of known arc lines
#      (>) nw:         (int) Number of wavelengths
#      (>) table:      (real[ntab,5]) Table with possible solutions
#      (>) ntab:       (int) First dimension of input table
#      (>) nt:         (int) Number of solutions found.
#      (>) wfact:      (real) Factor for use in widening the id window.
#      (<) ids:        (int[nw]) List of ids so that wt[i] => pt[ids[i]]
#
#
#      AUTHORS:     Jim Lewis -- RGO
#      DATE:        16-FEB-1995
#      UPDATE:
#
#------------------------------------------------------------------------------

# Parameters

int fd,np,nw,ntab,ids[nw],idents,nt
real pt[np],wt[nw],table[ntab,5],wfact

# Local variables

int i,nsol,nidents
real linear[2],range[2]

# Functions

int aa_load()

begin

      # Start by initialising

      idents = 0
      nsol = 0

      # Now loop through each solution and see which one gives the most
      # identifications...

      do i = 1,nt {
         linear[1] = table[i,2]
         linear[2] = table[i,3]
         iferr (nidents = aa_load(fd,pt,np,wt,nw,linear,wfact,ids,range)) {
            call erract(EA_WARN)
            next
         }
         if (nidents > idents) {
            idents = nidents
            nsol = i
         }
      }

      # Now get the solution for the best fit...

      linear[1] = table[nsol,2]
      linear[2] = table[nsol,3]
      iferr (idents = aa_load(fd,pt,np,wt,nw,linear,wfact,ids,range))
         call erract(EA_WARN)

      # Return the number of ids

      return (idents)
end


procedure aa_proj(pt,np,wt,nw,dd,nd,control,fd,dflip,number,nn,output,no,nt)
#------------------------------------------------------------------------------
#      AA_PROJ -- Identify arclines using projection method.
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) pt:         (real[np]) Measured pixel positions of lines
#      (>) np:         (int) Number of pixel positions
#      (>) wt:         (real[nw]) Wavelength positions of known arc lines
#      (>) nw:         (int) Number of wavelengths
#      (>) dd:         (real[2]) Range of dispersions to be considered
#      (>) nd:         (int) Number of angles/dispersions to try
#      (>) control:    (real[7]) Control variables. See below
#      (>) fd:         (int) Descriptor for log file
#      (>) dflip:      (int) Flag for whether dispersions will be flipped
#      (W) number:     (int[nn]) Workspace
#      (>) nn:         (int) Size of workspace
#      (<) output:     (real[no,5]) Table with possible solutions
#      (>) no:         (int) First dimension of output table
#      (<) nt:         (int) Number of solutions found.
#
#
#      AUTHORS:     E.J. Zuiderwijk -- RGO and Kapteyn Lab:  Original FORTRAN
#                   Jim Lewis -- RGO: SPP version
#      DATE:        16-FEB-1995
#      UPDATE:
#
# Comments from EJZ's original routine. (Modified for this SPP version)
#
# method : Arc line positions are identified with reference wavelength using
#          the projection method described in detail in :
#
#          "Algorithms for Automated Identification of Arc lines", by
#           E J Zuiderwijk and D Mills
#
#          This routine will find the best value of dispersion by searching
#          for the maximum density of points in the projection space.
#
#        - Position and wavelength data are transformed to dimensionless
#          coordinates  x = (W-<W>)/WW and y = (P-<P>)/PP, where the scaling
#          parameters <W> and WW are the centre of the wavelength interval,
#          and the wavelength range covered by the input table, respectively,
#          and similarly for <P> and PP.
#        - For a trial dispersion (chosen based on input data), the direction
#          of a projection line is determined. Subsequently, all data points
#          x,y are projected on this line, and sampled in equidistant bins,
#          whose width is also calculated from the input data.
#        - The bin with the largest number of projection points, found after
#          using a number of trial dispersions, represents the solution found.
#          The corresponding angle and distance to the origin in the x,y
#          space, translate directly into the coefficients of a linear
#          dispersion relation P=C+D*W, where D is the dispersion in pixel/A.
#        - The significance of the solution(s) found, is determined by
#          comparing the observed maximum density, with the average density
#          expected on the assumption of a random background distribution of
#          projected points.
#
#          The indices representing the identifications W(i) <==> P(index(i))
#          will have to be determined subsequently with routine aa_load.
#
# usage  : call aa_proj (pt,np,wt,nw,dd,nd,control,fd,number,nn,output,no,nt)
#
# input  : pt = real*4 array of length np, with observed positions of arc lines
#          wt = real*4 array of length nw, with reference wavelengths
#          dd = real*4 array of length 2,  with dispersion data (see further)
#          nd = integer*4 constant to define search method  (see further)
#          control = real*4 array of length 7 with control parameters (see further)
#          fd = file descriptor for log file.
#          number = integer*4 array of length nn for scratch storage
#
#        - control contains other input which must be defined outside this routine:
#
#          limit_angle   (real*4) : boundary angle for default search
#          ns_select  (integer*4) : number of solutions considered per scan
#          ng_select  (integer*4) : number of solutions considered globally
#          bin_factor    (real*4) : scaling factor applied to default bin width
#          nb_covers  (integer*4) : number of covers per bin (sliding mean)
#          trial_angles (logical) : .true.  for equidistant test angles
#                                   .false. for equidistant trial dispersions
#          test_statistic (int*4) : 1 for raw counts; 2 for chi ; else invalid
#
#
#        - ND determines the number of trial dispersions, as follows:
#
#               nd=1 : uses only one dispersion, to be provided in d(1);
#                      if no default dispersion is defined, exit with fail=0
#               nd=2 or nd>2 : use nd trial dispersion in range d(1) - d(2).
#               nd=0 : use default dispersion range, defined by default range
#                      of angles from : limit_angle to pi/2-limit_angle, on a
#                      default number of equidistant points equal nw+np.
#               nd<0 : as for nd=0, but using iabs(nd) points
#
#                         
# output : output = real*4 array of dimension no,5 for output data
#          nt     = integer*4 variable for output of number of solutions
#
#        - TABLE contains the following parameters per solution nr i:
#
#          output(i,1) = value of test statistic used
#          output(i,2) = constant term (C) of linear relation (P=C+D*W)
#          output(i,3) = dispersion (D) defined as dP/dW (NOT dw/dp!)
#          output(i,4) = raw counts in corresponding bin
#          output(i,5) = significance of solution
#
#
#------------------------------------------------------------------------------

# Parameters

real pt[np],wt[nw],dd[2],control[7],output[no,5]
int np,nw,nd,fd,number[nn],nn,no,nt

# Local variables

real limit_angle,bin_factor,lnsq2pi,pp[2],ww[2],window,angles[2],angle,
   disper[2],dang,xline,yline,dmin,dmax,smax,pmaximum,wmaximum,tangent,
   predicted,drange,bin_zero,spurious,wfactor,pfactor,py,px,offset,temp,
   average,fraction,raw_count,statistic,poisson,signif,ranges[4],pc[2],wc[2]
int ns_select,ng_select,nb_covers,test_statistic,nad,nangle,i,na,nbin,
   sign,j,index,k,l,threshold,nc_half,dflip,nsign,jj
bool trial_angles

# Functions

real aa_bin()

errchk aa_bin

begin

      # Initialise some useful variables

      limit_angle = control[1]
      ns_select = nint(control[2])
      ng_select = nint(control[3])
      bin_factor = control[4]
      nb_covers = nint(control[5])
      trial_angles = (nint(control[6]) != 0)
      test_statistic = nint(control[7])

      # Test the number of covers. Make sure it is odd

      if (nb_covers <= 0) nb_covers = 1
      if (mod(nb_covers,2) == 0) nb_covers = nb_covers - 1
      nc_half = nb_covers/2

      # Get the spacing information

      if (trial_angles)
         nad = 0
      else
         nad = 2

      # Useful constants

      lnsq2pi = 0.5*log(TWOPI)

      # Get scaling info for x,y transformations

      call aa_scale(pt,np,true,pp)
      call aa_scale(wt,nw,true,ww)

      # Test length of output buffer

      nt = ns_select + ng_select + 1
      if (nt > no) call error(0,"AA_PROJ -- Output buffer too small")

      # Zero the output table

      call aclrr(output,no*5)

      # Get the bin width for beta = 0.5 and gamma = 0.368.  Adjust the window
      # for the scale factor and for the multi-cover (running mean) option

      window = aa_bin(np,nw)*bin_factor/real(nb_covers)

      # Get number of trail angles/dispersions

      nangle = abs(nd)

      # Set initial angles for standard global search

      if (nd <= 0) {
         angles[1] = limit_angle
         angles[2] = HALFPI - limit_angle

         # If nd = 0 get the default number of trial angles. Use angle 
         # stepsize of 1.4*window size

         if (nd == 0) {
            angle = angles[2] - angles[1]
            angle = abs(angle)/(window*real(nb_covers)*1.4)
            nd = nint(angle)
            nangle = abs(nd)
         }

         # Get the range of dispersions. If either the angles are zero, the exit

         do i = 1,2 {
            if (abs(angles[i]) <= EPSILONR) 
               call error(0,"AA_PROJ -- Input angle of zero")
            disper[i] = -tan(angles[i])*pp[2]/ww[2]
         }
   
      # Set initial angles for search of specified number of dispersions

      } else {
         do i = 1,min(nd,2) {
            disper[i] = dd[i]
            if (abs(disper[i]) <= EPSILONR) 
               call error(0,"AA_PROJ -- Input dispersion of zero")
            angles[i] = -atan(disper[i]*ww[2]/pp[2])
            if (nd == 1) {
               disper[2] = disper[1]
               angles[2] = angles[1]
            }
         }
      }
      do i = 1,2 {
         ranges[i] = angles[i]
         ranges[i+2] = disper[i]
      }

      # Sort out ranges depending on whether both positive and negative
      # dispersions will be tried...

      switch (dflip) {
         case 0:
            nsign = 2
         case 1:
            nsign = 1
         case -1:
            nsign = 1
            do i = 1,3,2 {
               temp = -ranges[i+1]
               ranges[i+1] = -ranges[i]
               ranges[i] = temp
            }
      }
            
      # Start the search procedure by printing out some info to the log file

      if (fd != NULL) {
         call fprintf(fd,"Global scan for %d position angles\n")
            call pargi(nangle)
         call fprintf(fd,"   angle = %10.3g for dispersion %10.3g \n")
            call pargr(RADTODEG(angles[1]))
            call pargr(disper[1])
         if (nangle > 1) {
            call fprintf(fd,"to angle = %10.3g for dispersion %10.3g \n")
               call pargr(RADTODEG(angles[2]))
               call pargr(disper[2])
         }
         call fprintf(fd,"Nominal bin size = %10.3g")
            call pargr(window*real(nb_covers))
         call fprintf(fd,"; bin scale factor = %10.3g; Covers = %d\n")
            call pargr(bin_factor)
            call pargi(nb_covers)
      }

      # Begin the main search loop.  Loop for either angles or dispersions.

      if (nangle > 1)
         dang = (ranges[nad+2] - ranges[nad+1])/real(nangle-1)
      else 
         dang = 0.0
      do na = 1,nangle {

         # Choose angle or choose dispersion and convert to an angle

         angle = ranges(nad+1) + real(na-1)*dang
         if (nad == 2) angle = -atan(angle*ww[2]/pp[2])

         # Get projection line direction fo rthis angle

         xline = sin(angle)
         yline = cos(angle)

         # Get some quantities to be used for segment length estimates.
         # Project two adjacent corners (0.5,0.5 and 0.5,-0.5) onto the
         # projection line;  the projected points are critical points for 
         # calculation of segment length
  
         dmin = min(0.5*abs(xline+yline),0.5*abs(xline-yline))
         dmax = max(0.5*abs(xline+yline),0.5*abs(xline-yline))

         # Now the larger distance corresponds to length = 0, while the smaller
         # one corresponds to a (maximum) segment length = 1.0/max(xline,yline)

         smax = 1.0/max(abs(xline),abs(yline))

         # Predicted number of entries for largest segment

         pmaximum = real(np)
         wmaximum = real(nw)
         tangent = abs(yline/xline)
         if (tangent < 1.0) 
            wmaximum = wmaximum*tangent
         else 
            pmaximum = pmaximum/tangent

         # The expected number of ids is at most the smaller of these two

         predicted = min(wmaximum,pmaximum)

         # Get number of bins from (bin) window size.  The average number of
         # random entries in a bin equals: smax*np*nw*window where smax is
         # a correction factor due to the orientation of the projection line.

         drange = 2.0*dmax
         
         # Put an upper limit to the range in order to skip very small 
         # segments at ends

         if (drange > 1.0) drange = drange - 0.25*(drange-1.0)
         nbin = nint(drange/window)
         if (nbin >= nn) nbin = nn - 1
         if (mod(nbin,2) == 0) nbin = nbin - 1
         if (nbin <= 3) nbin = 3

         # Get zeropoint for offsets and define the expected number of
         # random hits per bin
  
         bin_zero = -0.5*real(nbin)*window
         spurious = smax*real(np*nw)*window

         # Do a scan first for both positive and negative angles if requested

         do sign = 1,nsign {

            # Zero counters for this scan.  Clear storage bins.

            do jj = 1,5
               call aclrr(output[1,jj],ns_select+1)
            call aclrr(number,nn)

            # Collect projected positions of all combinations P,W

            wfactor = xline/ww[2]
            pfactor = yline/pp[2]
            do j = 1,np {
               py = (pt[j] - pp[1])*pfactor - bin_zero
               do i = 1,nw {
                  px = (wt[i] - ww[1])*wfactor
                  index = int((px+py)/window) - nc_half
                  do k = 1,nb_covers {
                     l = index + k
                     if (l >= 1 && l <= nbin) number[l] = number[l] + 1
                  }
               }
            }

            # Collect NS_SELECT points of maximum statistic in scan

            threshold = ns_select + 2
            if (abs(output[threshold,1]) < EPSILON) threshold = 2
            do i = 1,nbin {
 
               # At least 3 points define a line

               if (number[i] >= 3) { 

                  # Get central position of search window

                  offset = bin_zero + (real(i) - 0.5)*window
               
                  # Get average frequency in bin from segment length

                  if (abs(offset) >= dmax) {
                     average = 0.0
                     fraction = 0.0
                  } else if (abs(offset) <= dmin) {
                     average = spurious
                     fraction = 1.0
                  } else {
                     fraction = (dmax - abs(offset))/(dmax - dmin)
                     average = spurious*fraction
                  }

                  # We have the expected frequency -- get the real one.

                  raw_count = real(number[i])/real(nb_covers)

                  # Do further analysis if more than average number is found

                  if (raw_count > average) {
                  
                     # Get test statistic
                     
                     if (test_statistic == 1)
                        statistic = raw_count
                     else if (test_statistic == 2)
                        if (abs(average) > EPSILON) 
                           statistic =((raw_count - average)**2)/average
                        else
                           statistic = 0.0

                     # Register the largest densities and corresponding
                     # quantities
  
                     if (statistic > output[threshold,1]) {

                        # Get the significance of this number[i] in bin i
                        # assuming Poisson statistics;  the significance 
                        # is -log(P) with P equal to the probability of finding
                        # at least one bin with this number in nbin bins. 
                        # Use Sterling's formula to get the probability
                        # where N! = N**N*exp(-N)*sqrt(2*pi*N)
                        
                        poisson = raw_count*log(average) -
                           (raw_count + 0.5)*log(raw_count) + 
                           (raw_count - average) - lnsq2pi

                        # Correct probability to get approx. P(n>=N,a) 
                        # P somewhat overestimated

                        poisson = exp(poisson)/(1.0 - average/(raw_count+1.0))
   
                        # Get the expected (avereage) number of bins in a scan
                        # with a number of counts >= number[i]

                        poisson = real(nbin)*poisson

                        # Get Poisson probability of finding at least one
                        # such bin from avereage

                        if (poisson > 1.0e-4) 
                           poisson = 1.0 - exp(-poisson)
                     
                        # And the significance = -log10(prob)
                       
                        if (poisson > 0.0) 
                           signif = -log10(poisson)
                        else
                           signif = INDEF
                     
                        # Put local variables into output buffer in 1st pos.

                        output[1,1] = statistic
                        output[1,2] = angle
                        output[1,3] = offset
                        output[1,4] = raw_count
                        output[1,5] = signif

                        # Now sort the buffer.  If you sort ns_select+1 
                        # then the worst solution will always be put into
                        # index #2 leaving index #1 free for the next attempt

                        call aa_sort5(output[1,1],output[1,2],output[1,3],
                           output[1,4],output[1,5],ns_select+1)
                     } # end statistic if
                  } # end raw_count if
               } # end number[i]>3 if
            } # end i=1,nbin do

            # Print out to logfile

            if (fd != NULL) {
               call fprintf(fd,"\n")
               call fprintf(fd,"scan %3d (%d): angle (degrees) = %10.3f")
                  call pargi(na)
                  call pargi(sign)
                  call pargr(RADTODEG(angle))
               call fprintf(fd,"   dispersion = %10.5g\n")
                  call pargr(-tan(angle)*pp[2]/ww[2])
               call fprintf(fd,"              sampling window = %10.5f")
                  call pargr(window)
               call fprintf(fd,"   number of bins = %d\n")
                  call pargi(nbin)
               call fprintf(fd,"              largest segment = %10.3f")
                  call pargr(smax)
               call fprintf(fd,"   expected identifications = %d\n")
                  call pargi(nint(predicted))

               # Print out ns_select best solutions

               call fprintf(fd,"\n")
               call fprintf(fd,"The best %d solutions in this scan :\n\n")
                  call pargi(ns_select)
               call fprintf(fd,"offset  position range  wavelength range")
               call fprintf(fd,"  raw count   statistic  significance\n\n")
               do i = 1,ns_select {
                  statistic = output[ns_select-i+2,1]
                  offset = output[ns_select-i+2,3]
                  raw_count = output[ns_select-i+2,4]
                  signif = output[ns_select-i+2,5]
                  call aa_cover(angle,offset,pc,wc)
                  if (test_statistic == 2) statistic = sqrt(statistic)
                  call fprintf(fd,"%6.3f %7.3f%7.3f   %7.3f%7.3f")
                     call pargr(offset)
                     call pargr(pc[1])
                     call pargr(pc[2])
                     call pargr(wc[1])
                     call pargr(wc[2])
                  call fprintf(fd,"       %3d    %10.4e     %7.3f\n")
                     call pargr(raw_count)
                     call pargr(statistic)
                     call pargr(signif)
               }
            } # end of logfile listing

            # Now sort the output table so that the ng_select best solutions
            # are in locations 2+ns_select to ns_select+ng_select+1
                  
            call aa_sort5(output[2,1],output[2,2],output[2,3],
               output[2,4],output[2,5],ns_select+ng_select)

            # reversion the sign of the angle by flipping the y component
 
            angle = -angle
            xline = -xline
         } # end of sign loop
      } # end of na angles loop

      # Get ng_select best solutions.  Make the significance negative to fool
      # the sorting routine into sorting into descending order.

      do i = 1,nt
         output[i,5] = -output[i,5]
      call aa_sort5(output[1,5],output[1,2],output[1,3],output[1,4],
         output[1,1],ns_select+ng_select+1)

      # Get output table; transform angle/offset into linear terms

      nt = 0
      do i = 1,ng_select {
         output[i,5] = -output[i,5]
         if (test_statistic == 2) output[i,1] = sqrt(output[i,1])
         
         # Transform angle into linear relation

         xline = sin(output[i,2])
         yline = cos(output[i,2])
         if (abs(xline) > EPSILON && abs(yline) > EPSILON) {
            nt = nt + 1
            offset = output[i,3]

            # Transform angle and offset into linear terms

            output[i,3] = -(pp[2]*xline)/(ww[2]*yline)
            output[i,2] = pp[1] - output[i,3]*ww[1] + pp[2]*offset/yline
         }
      }

      # Finally list the possible solutions found

      call aa_list(fd,pt,np,wt,nw,output,no,nt)
end


procedure aa_sort5 (a1,a2,a3,a4,a5,npts)
#------------------------------------------------------------------------------
#      AA_SORT5 -- Sort an array and cosort four others.
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#           
#      (!) a1:         (real[npts]) Array to be sorted.
#      (!) a2:         (real[npts]) Array to be cosorted.
#      (!) a3:         (real[npts]) Array to be cosorted.
#      (!) a4:         (real[npts]) Array to be cosorted.
#      (!) a5:         (real[npts]) Array to be cosorted.
#      (>) npts:       (int) Size of each array
#
#        
#      AUTHORS:     Jim Lewis -- RGO Based on xtools routine.
#      DATE:        16-FEB-1995
#      UPDATE:
#           
#------------------------------------------------------------------------------

real    a1[npts],a2[npts],a3[npts],a4[npts],a5[npts]  # Arrays to be sorted
int     npts                                          # Number of points

int     i, j
pointer sp, index, ptr

int     xts_compare()
extern  xts_compare

begin
        call smark (sp)
        call salloc (index, npts, TY_INT)
        call salloc (ptr, npts, TY_REAL)

        do i = 1, npts
            Memi[index+i-1] = ptr + i - 1

        call amovr (a1, Memr[ptr], npts)

        call qsort (Memi[index], npts, xts_compare)

        do i = 1, npts {
            j = Memi[index+i-1]
            a1[i] = Memr[j]
        }

        call amovr (a2, Memr[ptr], npts)
        do i = 1, npts {
            j = Memi[index+i-1]
            a2[i] = Memr[j]
        }

        call amovr (a3, Memr[ptr], npts)
        do i = 1, npts {
            j = Memi[index+i-1]
            a3[i] = Memr[j]
        }

        call amovr (a4, Memr[ptr], npts)
        do i = 1, npts {
            j = Memi[index+i-1]
            a4[i] = Memr[j]
        }

        call amovr (a5, Memr[ptr], npts)
        do i = 1, npts {
            j = Memi[index+i-1]
            a5[i] = Memr[j]
        }

        call sfree (sp)
end


real procedure aa_bin(npos,nwave)
#------------------------------------------------------------------------------
#      AA_BIN -- Approximate the bin size. Calculated for beta = 0.5
#                and gamma = 0.368.  See A&A article.
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#           
#      (>) npos:       (int) Number of positions
#      (>) nwave:      (int) Number of wavelengths
#
#      Returns:
#          width:      (real) Output width of bins
#
#        
#      AUTHORS:     E.J. Zuiderwijk -- RGO and Kapteyn Lab:  Original FORTRAN
#                   Jim Lewis -- RGO: SPP version
#      DATE:        09-FEB-1995
#      UPDATE:
#           
#------------------------------------------------------------------------------

# Parameters

int npos,nwave
real width

# Local variables

int np,nw
real xx,err,cc[4],dd[4],c,d

# Data statements

data cc /-0.31148651435,-0.05075997465, -0.5157731794, 1.313592024/
data dd /-0.463938906, -0.3072819211, -0.8744600561, 1.21460075/

begin

      # Check for silly situations where the number of positions or lines <= 1

      width = INDEF
      if (npos <= 1 || nwave <= 1) 
         call error(0,"AA_BIN -- Either only one peak or only one wavelength")

      # Now let np be the larger of npos and nwave

      np = max(npos,nwave)
      nw = min(npos,nwave)

      # Get coefficients of linear fit as function of nw

      xx = log10(real(nw))
      err = -1.692*xx - 0.878
      err = max(err,-3.80)
      err = exp(LN_10*err)
      xx = log(xx)

      # Calculate slope and intercept

      c = cc[1] + cc[2]*xx + cc[3]*exp(-cc[4]*xx)
      d = dd[1] + dd[2]*xx + dd[3]*exp(-dd[4]*xx)

      # Get log10(width) as a function of NP.  Correct it for error level

      xx = c + d*log10(real(np))
      xx = xx - err

      # Finally calculate width

      width = exp(LN_10*xx)

      # Exit point

      return(width)
end


int procedure aa_load(fd,pt,np,wt,nw,linear,lw_factor,iw,range)
#------------------------------------------------------------------------------
#      AA_LOAD -- Load the indentifications from a particular solution
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#           
#      (>) fd:         (int) Log file descriptor
#      (>) pt:         (real[np]) Input line positions
#      (>) np:         (int) Number of input line positions
#      (>) wt:         (real[nw]) Input line wavelengths
#      (>) nw:         (int) Number of input wavelengths
#      (>) linear:     (real[2]) Coeffs of linear dispersion relation
#                                [1] = offset, [2] = dispersion
#      (>) lw_factor:  (real) Expansion factor for uncertainty window.
#      (<) iw:         (int[nw]) Indices associating wavelength with position
#      (<) range:      (real[2]) Output wavelength range.
#
#      returns:
#          status:     (int) Status variable.  Can assume the following values:
#                            status >= 0  -- Number of matches found
#                            status < 0 -- Error condition
#
#        
#      AUTHORS:     E.J. Zuiderwijk -- RGO and Kapteyn Lab:  Original FORTRAN
#                   Jim Lewis -- RGO: SPP version
#      DATE:        09-FEB-1995
#      UPDATE:
#           
#------------------------------------------------------------------------------

# Parameters

int fd,np,nw,iw[nw],status
real pt[np],wt[nw],linear[2],lw_factor,range[2]

# Local variables

pointer sp,line,hits
real pp[2],ww[2],ao[2],angle,offset,pc[2],wc[2],xline,yline,dmin,dist,px,py,
   expected_p,expected_w,expected,spurious,section,window
int i,j,jrem,zero_i

# Functions

real asumi()

begin

      # Get some stack workspace

      call smark(sp)
      call salloc(line,SZ_LINE,TY_CHAR)
      call salloc(hits,np,TY_INT)

      # Initialise status variable

      status = 0

      # Get current range of position and wavelength

      call aa_scale(pt,np,true,pp)
      call aa_scale(wt,nw,true,ww)

      # Test input position and wavelength data.  If the range in either
      # is zero then signal and error

      if (ww[2] <= EPSILONR || pp[2] <= EPSILONR) 
         call error(0,"AA_LOAD -- Wavelength or position range of zero?")

      # Test that the input dispersion is not zero.

      if (abs(linear[2]) <= EPSILONR) 
         call error(0,"AA_LOAD -- Input dispersion of zero")

      # Get the wavelength range covered by the position data.

      do i = 1,2 {
         range[i] = (real(i) - 1.5)*pp[2]
         range[i] = (pp[1] - linear[1])/linear[2] + range[i]/abs(linear[2])
      }

      # Get orientation angle and projection offset from linear relation

      iferr (call aa_map(pp,ww,linear,ao)) call erract(EA_WARN)
      angle = ao[1]
      offset = ao[2]

      # Check that an error hasn't occured with the angle...

      if (IS_INDEF(angle)) call error(0,"AA_LOAD -- Undefined angle")
      
      # Get the length of the line segment defined by angle and offset
      # within the box defined by normalised position and wavelength space.

      call aa_cover(angle,offset,pc,wc)

      # Check to make sure the segment is physically inside the box.
      # Signal an error if it isn't

      yline = (pc[1] + pc[2])*0.5
      xline = (wc[1] + wc[2])*0.5
      if (abs(xline) >= 0.5 || abs(yline) >= 0.5) {
         if (fd != NULL) {
            call strcpy("\n ! input solution outside range:\n",Memc[line],
               SZ_LINE)
            call strcat("  wavelength range from %0.6g to %0.6g\n",Memc[line],
               SZ_LINE)
            call strcat("  position range from %0.6g to %0.6g\n",Memc[line],
               SZ_LINE)
            call fprintf(fd,Memc[line])
               call pargr(wc[1])
               call pargr(wc[2])
               call pargr(pc[1])
               call pargr(pc[2])
         }
         call error(0,"AA_LOAD -- Solution is outside allowed range")
      }

      # Get the expected number of matches from the fraction of space covered.

      expected_w = real(nw)*abs(wc[2] - wc[1])
      expected_p = real(np)*abs(pc[2] - pc[1])
      expected = min(expected_w,expected_p)

      # The average number of spurious (random) id's = section*nw*np*window
      # given the dispersion relation, get the window to be used by requiring
      # that the number of spurious id's is less than the uncertainty in 
      # the expected number of true id's

      spurious = max(1.0,sqrt(expected))
      section = sqrt((pc[2] - pc[1])**2 + (wc[2] - wc[1])**2)
      window = spurious/(section*real(np*nw))
      
      # Scale this window by the factor given by the calling routine

      window = lw_factor*window

      # Define line direction components

      xline = sin(angle)
      yline = cos(angle)

      # Now get the indices by finding the distance of each projected
      # [position,wavelength] combination and compare it with the offset

      zero_i = 0
      call amovki(zero_i,Memi[hits],np)
      do j = 1,nw {
         iw[j] = 0
         dmin = MAX_REAL
         px = ((wt[j] - ww[1])/ww[2])*xline
         do i = 1,np {
            py = ((pt[i] - pp[1])/pp[2])*yline
            dist = abs(px + py - offset)
            if (dist < dmin) {
               dmin = dist
               iw[j] = i
            }
         } # end of i loop 

         # If the deviation is too large, ignore this...

         if (dmin > 0.5*window) iw[j] = 0
         if (iw[j] != 0) Memi[hits+iw[j]-1] = Memi[hits+iw[j]-1] + 1
      }

      # Now loop to remove double hits, i.e. positions which have been assigned
      # more than one wavelength.

      do i = 1,np {
         if (Memi[hits+i-1] > 1) {
            dmin = MAX_REAL
            py = ((pt[i] - pp[1])/pp[2])*yline
            do j = 1,nw {
               if (iw[j] == i) {
                  px = ((wt[j] - ww[1])/ww[2])*xline
                  dist = abs(px + py - offset)
                  if (dist < dmin) {
                     dmin = dist
                     jrem = j
                  }
               }
            }
            do j = 1,nw 
               if (iw[j] == i && j != jrem) iw[j] = 0
            Memi[hits+i-1] = 1
         }
      }
      status = nint(asumi(Memi[hits],np))

      # Send some useful info the the log file

      if (fd != NULL) {
         call fprintf(fd,"\n %d lines found for dispersion %0.6g\n")
            call pargi(status)
            call pargr(linear[2])
         call fprintf(fd,"\n\n  wavelength range covered by input position")
         call fprintf(fd," table %0.6g to %0.6g\n")
            call pargr(range[1])
            call pargr(range[2])
         call fprintf(fd,"\n  angle = %0.6g,  offset = %0.6g")
            call pargr(RADTODEG(angle))
            call pargr(offset)
         call fprintf(fd," ; expected number of identifications = %d \n")
            call pargi(nint(expected))
      }

      # Exit point

      call sfree(sp)
      return(status)
end


procedure aa_cover(angle,offset,pc,wc)
#------------------------------------------------------------------------------
#      AA_COVER -- Find the coverage of the line defined by 'angle' and 
#                  'offset' within the normalised (position,wavelength)
#                  square.  This is similar to locating the points of
#                  intersection of a line and a square.
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) angle:      (real) Input angle of projection line
#      (>) offset:     (real) Input offset of projection line
#      (<) pc:         (real[2]) Position space coordinates of intersection
#      (<) wc:         (real[2]) Wavelength space coordinates of intersection
#
#
#      AUTHORS:     E.J. Zuiderwijk -- RGO and Kapteyn Lab:  Original FORTRAN
#                   Jim Lewis -- RGO: SPP version
#      DATE:        09-FEB-1995
#      UPDATE:
#
#------------------------------------------------------------------------------

# Parameters

real angle,offset,pc[2],wc[2]

# Local variables

real sina,cosa,x[4],y[4],d[4]
int i,j

define  MINSC  1.0e-6

begin

      # Get normal values.  Make sure they aren't zero...

      sina = sin(angle)
      if (abs(sina) < EPSILON) sina = MINSC
      cosa = cos(angle)
      if (abs(cosa) < EPSILON) cosa = MINSC

      # Get the four possible points of intersection with the edge of the box.
      # i.e. The line must intersect on two boundries.  The boundries are
      # defined by x = +/- 0.5 and y = +/- 0.5.  Entries 1 and 2 are for 
      # x = +/- 0.5 and entries 3 and 4 are for y = +/- 0.5.  Calculate the
      # distance of these four points to the origin...

      do i = 1,2 {
         j = i + 2
         x[i] = real(i) - 1.5
         y[i] = (offset - x[i]*sina)/cosa
         y[j] = x[i]
         x[j] = (offset - y[j]*cosa)/sina
         d[i] = x[i]**2 + y[i]**2
         d[j] = x[j]**2 + y[j]**2
      }

      # Sort these entries by increasing distance 

      call xt_sort3(d,x,y,4)

      # The two which are closest to the origin will be the intersection points

      do i = 1,2 {
         pc[i] = x[i]
         wc[i] = y[i]
      }
end

procedure aa_scale(x,nx,extend,xx)
#------------------------------------------------------------------------------
#      AA_SCALE -- Find the mean value and range of an array.  Extend the
#                  range if appropriate.
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) x:          (real[nx]) Input array
#      (>) nx:         (int) Dimension of input array
#      (>) extend:     (bool) TRUE if the range is to be extended
#      (<) xx:         (real[2]) Output mean value and range
#
#
#      AUTHORS:     E.J. Zuiderwijk -- RGO and Kapteyn Lab:  Original FORTRAN
#                   Jim Lewis -- RGO: SPP version
#      DATE:        09-FEB-1995
#      UPDATE:
#
#------------------------------------------------------------------------------

# Parameters

real x[nx],xx[2]
int nx
bool extend

# Local variables

real xlow,xhigh

begin
 
      # Start by findind the min and max of the array.

      call alimr(x,nx,xlow,xhigh)

      # Store the mean and the range

      xx[1] = (xhigh + xlow)*0.5
      xx[2] = xhigh - xlow

      # Extend the range by the average value between two x-values if 
      # desired (and possible).

      if (extend && nx > 1) 
         xx[2] = xx[2]*(1.0 + 1.0/real(nx-1))
end


procedure aa_map(pp,ww,xx,yy)
#------------------------------------------------------------------------------
#      AA_MAP -- Map projection space - dispersion relation. Transform the
#                coefficients of the linear relation: P = C + D*W into a
#                line defined by angle A and a distance Od from the origin:
#
#                   X SIN(A) + Y COS(A) = Od
#
#                in the plane of the collapsed coordinates:
#
#                   X = (W-<W>)/range_W
#                   Y = (P-<P>)/range_P
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) pp:         (real[2]) <P> and range_P
#      (>) ww:         (real[2]) <W> and range_W
#      (>) xx:         (real[2]) C and D
#      (<) yy:         (real[2]) A and Od
#
#
#      AUTHORS:     E.J. Zuiderwijk -- RGO and Kapteyn Lab:  Original FORTRAN
#                   Jim Lewis -- RGO: SPP version
#      DATE:        09-FEB-1995
#      UPDATE:
#
#------------------------------------------------------------------------------

# Parameters

real pp[2],ww[2],xx[2],yy[2]

begin

      # Preset output to be INDEF

      call amovkr(INDEF,yy,2)

      # Protect against bad input (e.g. a positional range of 0)

      if (abs(pp[2]) <= EPSILON) 
         call error(0,"AA_MAP -- Positional range of zero?")

      # Get the angle and the distance to the origin

      yy[1] = -atan(xx[2]*ww[2]/pp[2])
      yy[2] = (xx[1] - pp[1] + xx[2]*ww[1])*cos(yy[1])/pp[2]
      return
end

procedure aa_list(fd,pt,np,wt,nw,ids_table,ni,nt)
#------------------------------------------------------------------------------
#      AA_LIST -- List all possible solutions to the log file.
#
#      Parameters (">" Input, "<" Output, "W" Workspace, "!" Changed)
#
#      (>) fd:         (int) Descriptor for log file
#      (>) pt:         (real[np]) Measured pixel positions of lines
#      (>) np:         (int) Number of pixel positions
#      (>) wt:         (real[nw]) Wavelength positions of known arc lines
#      (>) nw:         (int) Number of wavelengths
#      (>) ids_table:  (real[ni,5]) Table with possible solutions
#      (>) ni:         (int) Input first dimension of ids_table
#      (>) nt:         (int) Number of solutions found.
#
#
#      AUTHORS:     E.J. Zuiderwijk -- RGO and Kapteyn Lab:  Original FORTRAN
#                   Jim Lewis -- RGO: SPP version
#      DATE:        08-FEB-1995
#      UPDATE:
#
#------------------------------------------------------------------------------

# Parameters

real pt[np],wt[nw],ids_table[ni,5]
int fd,np,nw,ni,nt

# Local variables

real pp[2],ww[2],data[5],wsf[2],ao[2]
int i,j
pointer sp,line

# Error exit

define exit_ 10

begin

      # Get some stack workspace

      call smark(sp)
      call salloc(line,SZ_LINE,TY_CHAR)

      # Tell the user how many solutions have been found

      call printf("%d possible solutions found\n")
         call pargi(nt)

      # If logging hasn't been requested, then get out.

      if (fd == NULL) goto exit_

      # Get scaling and print banner

      call aa_scale(pt,np,true,pp)
      call aa_scale(wt,nw,true,ww)
      call strcpy("\n%3d possible solutions found :\n\n",Memc[line],SZ_LINE)
      call strcat(" Angle  Offset  Dispersion     Wavelength Cover",
         Memc[line],SZ_LINE)
      call strcat("   Count  Statistic  Significance\n\n",Memc[line],SZ_LINE)
      call fprintf(fd,Memc[line])
        call pargi(nt)
    
      # Get out if there were no solutions

      if (nt <= 0) goto exit_

      # Now loop for each solution

      do i = 1,nt {
         if (i > ni) break

         # Pass the info over 

         do j = 1,5
            data[j] = ids_table[i,j]

         # Trap for zero dispersion...

         ao[1] = 0.0
         if (! IS_INDEF(data[3])) {
            do j = 1,2 {
               wsf[j] = (real(j) - 1.5)*pp[2]
               wsf[j] = (pp[1] - data[2])/data[3] + wsf[j]/abs(data[3])
            }
            iferr (call aa_map(pp,ww,data[2],ao)) call erract(EA_WARN)
            if (!IS_INDEF(ao[1])) ao[1] = RADTODEG(ao[1])
         }
    
         # Now write it out

         if (! IS_INDEF(ao[1])) {
            call strcpy("%6.2f  %6.3f  %8.3f    %10.3f%10.3f%6d",Memc[line],
               SZ_LINE)
            call strcat(" %10.4e     %5.1f\n",Memc[line],SZ_LINE)
            call fprintf(fd,Memc[line])
               do j = 1,2
                  call pargr(ao[j])
               call pargr(data[3])
               do j = 1,2
                  call pargr(wsf[j])
               call pargi(nint(data[4]))
               call pargr(data[1])
               call pargr(data[5])
         }
      } # end of do loop

exit_ 
      call sfree(sp)
end
