# FIBRESPONSE -- Make a aperture response spectrum using a flat field
# and a throughput file or image.

procedure w_fibresponse (flat, throughput, apreference, response)

string  flat                    {prompt="Flat field spectrum"}
string  throughput              {prompt="Throughput file or image"}
string  apreference             {prompt="Aperture reference spectrum"}
string  response                {prompt="Response spectrum"}
file    apidtable               {prompt="Aperture identifications"}

bool    clean = no              {prompt="Detect and replace bad pixels?"}
bool    fitflat = no            {prompt="Fit and ratio flat field spectrum?"}
bool    interactive = yes       {prompt="Fit flat field interactively?"}
string  function = "spline3"    {prompt="Fitting function",
                                 enum="spline3|legendre|chebyshev|spline1"}
int     order = 20              {prompt="Fitting function order", min=1}
bool    batch                   {prompt="Do processing in batch?"}

struct  *fd

begin
      string  imtype,mstype,aps,sample,flat2d,skyflat2d,apref,resp
      file    temp1,log1,log2,log3
      int     wyflim1,wyflim2,sam1,sam2,maxshift
      real    mean,ishift
      bool    dsaw

      # Get image and ms type

      imtype = "." // envget ("imtype")
      mstype = ".ms" // imtype

      # Get the input frame names and create a temporary file name

      flat2d = flat
      skyflat2d = throughput
      apref = apreference
      resp = response
      temp1 = mktemp ("tmp$iraf")
      dsaw = wyfpars.dosaw

      # Check required input and output.

      if (resp == "" || resp == flat2d || resp == skyflat2d)
         error (1, "Bad response image name")
      if (flat2d == "" && skyflat2d == "")
         error (1, "No flat field or throughput specified")
      if (apref == "")
         error (1, "No aperture reference specified")
      if (flat2d != "") {
         if (!access (flat2d // imtype))
            error (1, "Flat field spectrum not found - " // flat2d)
      }
      if (!access (skyflat2d // imtype))
         error (1,"Throughput file or image not found - " // skyflat2d)
      if (!access (apref // imtype))
         error (1, "Aperture reference image required")

      # Set logging

      tee.append = yes
      if (logfile == "")
         log1 = "dev$null"
      else
         log1 = logfile
      if (verbose && !batch)
         log2 = "STDOUT"
      else
         log2 = "dev$null"
      if (batch)
         log3 = "dev$null"
      else
         log3 = "STDOUT"

      # Initialize APSCRIPT

      apscript.references = apref
      apscript.ansrecenter = "YES"
      apscript.ansresize = "NO"
      apscript.ansedit = "NO"
      apscript.anstrace = "NO"
      apscript.ansextract = "YES"

      # Find the limits of the spectra from the aperture reference...

      hselect(apref,"WYFLIM*",yes) | scan(wyflim1,wyflim2)
      if (nscan() < 2)
         error(0,"Must have spectral limits in aperture reference header")

      # Find the sample limits for stats.  This is necessary so that when
      # you are normalising the flats you are only comparing the spectra
      # in the regions where the overlap.

      if (dsaw) {
         maxshift = 0
         fd = wyfpars.sfile
         while (fscan(fd,ishift,aps) != EOF) {
            if (nint(abs(ishift)) > maxshift)
               maxshift = nint(abs(ishift))
         }
         fd = ""
         sam1 = wyflim1 + maxshift
         sam2 = wyflim2 - maxshift
         sample = sam1//":"//sam2
      } else {
         sample = "*"
         sam1 = INDEF
         sam2 = INDEF
      }

      # If using a flat field extract it if necessary and possibly fit it
      # and ratio the individual apertures by an overall smooth function

      if (flat2d != "") {
         if (!access (flat2d // mstype)) {
            print ("Extract flat field ", flat2d) | tee (log1, > log3)
            if (flat2d != apref)
               apscript (flat2d, output=resp, background="none",
                  clean=clean, extras=no)
            else
               apscript (flat2d, output=resp, ansrecenter="NO",
                  ansrecenter="NO", ansresize="NO", ansedit="NO",
                  anstrace="NO", background="none",
                  clean=clean, extras=no)
            imcopy(resp, flat2d//mstype, verbose=no)
            sapertures (resp, apertures="", apidtable=apidtable,
               wcsreset=no, verbose=no, beam=INDEF, dtype=INDEF, w1=INDEF,
               dw=INDEF, z=INDEF, aplow=INDEF, aphigh=INDEF, title=INDEF)
            if (dsaw) {
               fd = wyfpars.sfile
               while (fscan(fd,ishift,aps) != EOF)
                  if (ishift != 0.0) 
                     specshift(resp,ishift,apertures=aps,verbose=no)
               fd = ""
            }
            w_mksafe(resp,wyflim1,wyflim2)
         } else
            imcopy (flat2d//mstype,resp,verbose=no)
         if (fitflat) {
            print ("Fit and ratio flat field ", flat2d) | tee (log1, > log3)
            scombine.log = no
            scombine(resp,temp1,noutput="",logfile="STDOUT",
               apertures="",group="all",combine="average",reject="none",
               first=no,w1=INDEF,w2=INDEF,dw=INDEF,nw=INDEF,
               scale="none",zero="none",weight="none",sample="",
               lthreshold=INDEF,hthreshold=INDEF,nlow=1,nhigh=1,
               nkeep=1,mclip=yes,lsigma=3.0,hsigma=3.0,rdnoise=0.0,
               gain=1.0,snoise=0.0,pclip=-0.5,grow=0,blank=0.0) | 
               tee (log1, >log2)
            fit1d (temp1,temp1,"fit",axis=1,interactive=interactive,
               sample="*",naverage=1,function=function,order=order,
               low_reject=3.0,high_reject=3.0,niterate=0,grow=0.,
               graphics="stdgraph")
            sarith (resp,"/",temp1,resp,w1=INDEF,w2=INDEF,
               apertures="",bands="",beams="",apmodulus=0,reverse=no,
               ignoreaps=yes,format="multispec",renumber=no,
               offset=0,clobber=yes,merge=yes,errval=0,verbose=no)
            imdelete (temp1, verify=no)
         }
      }

      # Extract sky flat image and apply it to the flat field.
      # Compute the throughput through each aperture.

      print ("Extract throughput image ", skyflat2d) | 
         tee (log1, > log3)
      apscript (skyflat2d, output=temp1, background="none",
         clean=clean, extras=no)
      sapertures (temp1, apertures="", apidtable=apidtable,
         wcsreset=no, verbose=no, beam=INDEF, dtype=INDEF, w1=INDEF,
         dw=INDEF, z=INDEF, aplow=INDEF, aphigh=INDEF, title=INDEF)
      if (dsaw) {
         fd = wyfpars.sfile
         while (fscan(fd,ishift,aps) != EOF) 
            if (ishift != 0.0)
               specshift(temp1,ishift,apertures=aps,verbose=no)
         fd = ""
      }
      w_mksafe(temp1,wyflim1,wyflim2)

      # Attempt to get rid of CR hits and night sky emission lines.

      sfit(temp1,temp1,lines="*",bands="1",type="data",
         replace=yes,wavescale=yes,logscale=no,override=no,
         listonly=no,logfile="",interac=no,
         sample=sample,naverage=1,function="legendre",
         order=10,low_rej=4.0,high_rej=4.0,niterate=5,grow=0)

      if (flat2d != "") {
         print ("Correct flat field to throughput image") | 
            tee (log1, > log3)
         sarith (temp1, "/", resp, temp1, w1=INDEF, w2=INDEF,
            apertures="", bands="", beams="", apmodulus=0,
            reverse=no, ignoreaps=no, format="multispec",
            renumber=no, offset=0, clobber=yes, merge=no,
            errval=0, verbose=no)
         hedit(temp1,"SFIT","",add=no,delete=yes,verify=no,show=no,
            update=yes)
         hedit(temp1,"SFITB","",add=no,delete=yes,verify=no,show=no,
            update=yes)
         sfit(temp1,temp1,lines="*",bands="1",type="fit",
            replace=no,wavescale=yes,logscale=no,override=no,
            listonly=no,logfile=logfile,interac=no,
            sample=sample,naverage=1,function="legendre",
            order=1,low_rej=3.0,high_rej=3.0,niterate=0,grow=0)
         sarith (resp, "*", temp1, resp, w1=INDEF, w2=INDEF,
            apertures="", bands="", beams="", apmodulus=0,
            reverse=no, ignoreaps=no, format="multispec",
            renumber=no, offset=0, clobber=yes, merge=no,
            errval=0, verbose=no)
         imdelete (temp1, verify=no)
      } else {
         print ("Compute aperture throughput from image") | 
            tee (log1, > log3)
         hedit(temp1,"SFIT","",add=no,delete=yes,verify=no,show=no,
            update=yes)
         hedit(temp1,"SFITB","",add=no,delete=yes,verify=no,show=no,
            update=yes)
         sfit(temp1,resp,lines="*",bands="1",type="fit",
            replace=no,wavescale=yes,logscale=no,override=no,
            listonly=no,logfile=logfile,interac=no,
            sample=sample,naverage=1,function="legendre",
            order=1,low_rej=3.0,high_rej=3.0,niterate=0,grow=0)
      }

      # The final response is normalized to overall unit mean and the
      # average aperture response is printed.

      print ("Create the normalized response ", resp) | tee (log1, > log3)

      scopy(resp,temp1,w1=sam1,w2=sam2,apertures="",bands="",
         beams="",apmodulus=0,format="multispec",renumber=no,
         offset=0,clobber=yes,merge=no,rebin=no,verbose=no)
      imstat(temp1,fields="mean",lower=INDEF,upper=INDEF,
         binwidth=0.1,format=no) | scan(mean)
      imdelete(temp1,verify=no)

      imarith(resp,"/",mean,resp,title="",divzer=0.0,hparams="",
         pixtype="",calctype="",verb=no,noact=no)
      print(resp," -> ",resp," using bzero: 0.  and bscale: ",
         mean,"\n") | tee (log1, >log2)
      hedit(resp,"SFIT","",add=no,delete=yes,verify=no,show=no,
         update=yes)
      hedit(resp,"SFITB","",add=no,delete=yes,verify=no,show=no,
         update=yes)
      sfit(resp,temp1,lines="*",bands="1",type="fit",replace=no,
         wavescale=yes,logscale=no,listonly=no,logfiles=logfile,
         interactive=no,sample=sample,naverage=1,function="chebyshev",
         order=1,low_reject=3.0,high_reject=3.0,niterate=0,grow=0)    
      imdelete (temp1, verify=no)
end
